// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package content.assessment;

/*$Id: Jury.java,v 1.9 2008/05/06 12:47:23 lhakulin Exp $*/

import content.interfaces.SimulationExerciseModel;
import matrix.animation.AnimatedStructures;
import matrix.animation.Animator;
import matrix.structures.FDT.FDT;
import matrix.structures.FDT.probe.UndirectedGraphImpl;
import matrix.util.Note;

public class Jury extends AssessmentStrategy{

    private static Jury singleton;
    
    public static Jury getInstance(){
    	if (singleton == null)
            singleton = new Jury();
        
        return singleton;
    }
    
    
    
    
    /**
     * This method creates the judgement for a simulation exercise when we want
     * to grade the different user structures as a group
     * 
     * @param user          The structures used by the user
     * @param userAnimator  The Animator used by the user
     * @param exercise      The exercise being solved
     * @return              The result of the judging
     */
    
    protected Judgement getJudgementJoinedFDTs(FDT[] user, Animator userAnimator, SimulationExerciseModel exercise) {
        //Otto removed the synchronized keyword 15.3.2005

        // sanity check
        if(exercise == null || userAnimator == null || user == null) {
            Note.err(null, "judge: null argument in judge, cannot judge exercise");
            return new Judgement(0);
        }
        
        Animator org = Animator.getActiveAnimator();
        Animator solverAnimator = new Animator();
        Animator.setActiveAnimator(solverAnimator);
        
        FDT[] modelSolution = exercise.solve();
        Animator.setActiveAnimator(org);
        return getJudgementJoinedFDTs(user, userAnimator, modelSolution, solverAnimator);
    }
    
    public Judgement getJudgementJoinedFDTs(FDT[] user, Animator userAnimator,
                                            FDT[] modelSolution, Animator solverAnimator){    

        int modelCount   = 0; //number of correct states is modelSolution
        int userCount    = 0;  //number of states in user's sulution
        int correctCount = 0; // number of correct states in user solution
        //int userState = 0; //offset from userAnimator. Null state is currents state
        
        int userSolutionState;
        FDT[] userSolution  = user; // userSolution is a class variable
                        
        if(modelSolution.length != userSolution.length) {
            Note.warning(null, "User solution and model solution have different number of structures, stopping judgement.");
            return new Judgement(1);
        }
            
        //Note.out(null, "rewinding model solution");
        modelCount = solverAnimator.rewind();
        //Note.out(null, "rewinding user solution");
        // GUI handles placing the beginning state of the user animator to the right place 
            
        // the states of the user animator are evaluated only until the current state.
        // If there are any more states in the user animator, they are removed
            
        // Panu commented this out 2003-01-21
        //Stack redoStack = userAnimator.disposeRedo(); 
            
        userCount = userAnimator.rewind();
        //Note.out(null, "starting judgement user: "+userCount+" model: "+modelCount);
        int oldUserState = 0; //state of user solution in last iteration
            
        //rewindAnimator(solverAnimator);
        //rewindAnimator(userAnimator);
        //userState = -userCount;

        userSolutionState = 0;
        correctCount = 0;
        
        //Searching for the beginning state from the model states
        while(!(allEqual(userSolution,modelSolution)) && solverAnimator.hasNextOperation()) {
            solverAnimator.redo();
            modelCount--;
        }

        boolean isEmpty = true;
        
        // Check whether user answer is empty.
        // The check is done by skipping all states that match the beginning state
        // until a new state is found or the sequence ends.
        // if the last state still matches the model, there isn't but a single state
        // in the user solution.

        while(allEqual(userSolution,modelSolution) && userAnimator.hasNextOperation())
            userAnimator.redo();
        
        if (allEqual(userSolution, modelSolution) == false)
            isEmpty = false;
        
        userAnimator.rewind();
        // Emptiness check ends 
    
        
        
        Judgement result = new Judgement(new String[]{"NEVER USED!"},1,modelCount);

        
        
        if(!(allEqual(userSolution,modelSolution)) || modelCount == 0) {
            Note.warning(null, "No state matching the starting state of the user animator found.");
            result.setPoints(0, 0);                
        }
                        
        for (int i=0; i<modelCount; i++) {
            solverAnimator.redo();
            //Note.out(null, "Comparing states "+solverAnimator);
            oldUserState = userSolutionState;
                
            while(!(allEqual(userSolution,modelSolution)) && userAnimator.hasNextOperation() ) {
                userSolutionState++;
                userAnimator.redo();
            }
            if (allEqual(userSolution,modelSolution)){
                //Note.out(null,"state "+i+" correct");
                // Added by Harri 11-08-2003
                // all states must in correct order
                if (i==correctCount){
                    correctCount++;
                    result.setFinalUserState(userSolutionState);
                }
            }
            else {
                //Note.out(null,"state " + i + " wrong");
                if (!result.errorStateIndexSet()) {
                    result.setErrorStateIndex(i);
                }
                    
				// Mark up that erroneous states were found                    
                result.setErrorStates();
                for(; userSolutionState > oldUserState; userSolutionState--)
                    userAnimator.undo();
                
                // no need to go any further, because following states would
                // be all incorrect after this incorrect state
                if (user[0] instanceof UndirectedGraphImpl){
                	break;
                }
                
            }
                
        }
        while(userAnimator.hasNextOperation()) {
            userAnimator.redo();
            //Check and mark if erroneous states are found
            if (!allEqual(userSolution, modelSolution))
                result.setErrorStates();
        }
            
        result.setPoints(correctCount, 0);        
            
        if(isFinished(userAnimator, solverAnimator, userSolution, modelSolution)) {
            result.setComplete();
            //Note.out(null, "exercise is complete");
        }
            
        userAnimator.end();
            
        //If the result was found to be empty, mark it in the judgement    
        if (isEmpty)
            result.setEmpty();
        return result;
    }           

    /**
     * This method creates the judgement for a simulation exercise when we want
     * to grade the different user structures separately of each other.
     * 
     * @param user          The structures used by the user
     * @param userAnimator  The Animator used by the user
     * @param exercise      The exercise being solved
     * @return              The result of the judging
     */
  
    public Judgement getJudgementSeparateFDTs(FDT[] user, Animator a, SimulationExerciseModel exercise) {

        Animator solverAnimator = new Animator();
        Animator.setActiveAnimator(solverAnimator);
        
        FDT[] modelSolution= exercise.solve();

        return getJudgementSeparateFDTs(user, a, modelSolution, solverAnimator);
    }
    
    
    public Judgement getJudgementSeparateFDTs(  FDT[] userSolution, Animator userAnimator,
    		                                    FDT[] modelSolution, Animator solverAnimator){    

        int modelCount   = 0; //number of correct states is modelSolution
        int userCount    = 0; //number of states in user's sulution
        int correctCount = 0; // number of correct states in user solution
        int userSolutionState;
        
        //This variable is used to find the last user state
        //matching a model state. As the structures are checked
        //separate, we must find the structure that has the first
        //error and return the value for that structure. Hence
        //the dummy starting value from which we go down.
        int minimalUserSolutionState = 100000; 
        
        if(modelSolution.length != userSolution.length) {
            Note.warning(null, "The user solution has a different number of structures with "+
                               "modelSolution, stopping judgement.");
            return new Judgement(modelSolution.length);
        }
        //Note.out(null, "rewinding model solution");
        modelCount = solverAnimator.rewind();
        Judgement result = new Judgement(
                new String[]{"NEVER USED!"}/*exercise.getStructureNames()*/,
				modelSolution.length, modelCount); 

        //Note.out(null, "rewinding user solution");
        // GUI handles placing the beginning state of the user animator to the right place 
        
        // the states of the user animator are evaluated only until the current state.
        // If there are any more states in the user animator, they are removed
            
        // Panu commented this out 2003-01-21
        //Stack redoStack = userAnimator.disposeRedo(); 
            
        userCount = userAnimator.rewind();
        //Note.out(null, "starting judgement");
        int oldUserState = 0; //state of user solution in last iteration
            
        boolean isEmpty = true;        
        
        // For-loop executed for each FDT in modelSolution
        for(int j = 0; j < modelSolution.length; j++) {
            int maxUserSolutionState = 0; //
            solverAnimator.rewind();
            userAnimator.rewind();
            //userState = -userCount;
            userSolutionState = 0;
            correctCount = 0;
            //Note.out(null, "COMPARE " + userSolution[j] + " with " + modelSolution[j]);
            boolean cmpResult = userSolution[j].equals(modelSolution[j]);
            while((cmpResult == false) && solverAnimator.hasNextOperation()) {
                solverAnimator.redo();
                cmpResult = userSolution[j].equals(modelSolution[j]);
            }
            if (cmpResult == false) {
                Note.out(null, "No state matching the starting state of the user animator found *");
                solverAnimator.rewind();
                userAnimator.rewind();
                Note.out(null, userSolution[j] + " - " + modelSolution[j]);
                
                
                result.setPoints(0, j);
                continue;
            }
            
            // --- Emptiness check begins
            // Check whether user answer is empty. This is done
            // by skipping user states until a state that differs from the beginning model
            // state is found or the sequence ends. If the latter
            // happens, the student sequence only contains the beginning state.
            while(userSolution[j].equals(modelSolution[j]) && userAnimator.hasNextOperation())
                userAnimator.redo();
            
            if (userSolution[j].equals(modelSolution[j]) == false)
                isEmpty=false;
            
            // restoring the user sequence to where it was before
            // the emptiness check
            userAnimator.rewind();
            // --- Emptiness check ends 
            
            for (int i=0; i<modelCount; i++) {
                solverAnimator.redo();
                //Note.out(null, "Comparing states");
                oldUserState = userSolutionState;
                                
                /** 
                 * Search model solution's current state from user's solution. Search starts
                 * from position defined by Animator
                 */
                while(!(userSolution[j].equals(modelSolution[j])) && userAnimator.hasNextOperation() ) {
                    userSolutionState++;
                    userAnimator.redo();
                }
                
                if (userSolution[j].equals(modelSolution[j])) {
                    //Note.out(null,"state "+i+" correct");
                    // Added by Harri 11-08-2003
                    // all states must in correct order
                	if (i==correctCount)
                        correctCount++;
                    
                    // Each passed correct state is marked, to know
                    // the index of the final correct state
                    maxUserSolutionState = userSolutionState;
                    result.setFinalUserState(maxUserSolutionState);
                }
                else {
                    //Note.out(null,"state " + i + " wrong");
                    if (!result.errorStateIndexSet()) {
                        result.setErrorStateIndex(i);
                    }
                    
                    // The fact that there were incorrect states is marked
                    result.setErrorStates();
                                     
                    for(; userSolutionState > oldUserState; userSolutionState--)
                        userAnimator.undo();
                }
                
            }

			// We are trying to find the sequence that had the least amount of
			// matched states
            if (minimalUserSolutionState > maxUserSolutionState)
                minimalUserSolutionState = maxUserSolutionState;

            
            while(userAnimator.hasNextOperation()) {
                userAnimator.redo();
                
                // If the states do not match, mark it up.
                if (!userSolution[j].equals(modelSolution[j]))
                    result.setErrorStates();
            }
            result.setPoints(correctCount, j);
            
        }
        if(isFinished(userAnimator, solverAnimator, userSolution, modelSolution)) {
            result.setComplete();
            //Note.out(null, "exercise is complete");
        }
        //Note.out(null, "judge end");
        userAnimator.end();
        
        //Finally mark if the result was completely empty
        if (isEmpty)
            result.setEmpty();

        return result;
    }
    
    /**
     * Rewinds the model animator inside the modelStructures until a state
     * matching the beginning state of the user structures is found.
     * 
     * @param userStructures user structures and the corresponding animator
     * @param modelStructures model structures and the corresponding animator
     * @return
     */

    public static boolean findCommonJoinedStartingState(
            AnimatedStructures userStructures,
            AnimatedStructures modelStructures){    

        FDT[] userSolution     = userStructures.getStructures();
        FDT[] modelSolution    = modelStructures.getStructures();
        
        Animator userAnimator  = userStructures.getAnimator();
        Animator modelAnimator = modelStructures.getAnimator();
        
        if(modelSolution.length != userSolution.length) {
            Note.warning(null, "User solution and model solution have different number of structures, stopping judgement.");
            return false;
        }   

        modelAnimator.rewind();
        userAnimator.rewind();

        //Searching for the beginning state from the model states
        while(!(allEqual(userSolution,modelSolution)) && modelAnimator.hasNextOperation()) {
            modelAnimator.redo();
        }

        // Check whether user answer is empty.
        // The check is done by skipping all states that match the beginning state
        // until a new state is found or the sequence ends.
        // if the last state still matches the model, there isn't but a single state
        // in the user solution.

        boolean isEmpty = true;
        
        while(allEqual(userSolution,modelSolution) && userAnimator.hasNextOperation())
            userAnimator.redo();

        if (allEqual(userSolution, modelSolution) == false)
            isEmpty = false;

        userAnimator.rewind();
        // Emptiness check ends 


        if(!(allEqual(userSolution,modelSolution))) {
            Note.warning(null, "No state matching the starting state of the user animator found.");
            return false;                
        }
    
        return true;
    }
}