// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package content.assessment;

/*$Id: SkippingJury.java,v 1.2 2007/09/07 13:24:52 vkaravir Exp $*/

import content.interfaces.SimulationExerciseModel;
import matrix.animation.Animator;
import matrix.structures.FDT.FDT;
import matrix.util.Note;

public class SkippingJury extends AssessmentStrategy{
    
    /**
     * This method creates the judgement for a simulation exercise when we want
     * to grade the different user structures as a group
     * 
     * IMPORTANT NOTE: This method uses the skipping strategy
     * 
     * @param user          The structures used by the user
     * @param userAnimator  The Animator used by the user
     * @param exercise      The exercise being solved
     * @return              The result of the judging
     */
    
    protected Judgement getJudgementJoinedFDTs(FDT[] user, Animator userAnimator, SimulationExerciseModel exercise) {
        //Otto removed the synchronized keyword 15.3.2005

        // sanity check
        if(exercise == null || userAnimator == null || user == null) {
            Note.err(null, "judge: null argument in judge, cannot judge exercise");
            return new Judgement(0);
        }
        
        int modelCount   = 0; //number of correct states is modelSolution
        int userCount    = 0;  //number of states in user's sulution
        int correctCount = 0; // number of correct states in user solution
        //int userState = 0; //offset from userAnimator. Null state is currents state
        
        Animator solverAnimator = new Animator();
        Animator.setActiveAnimator(solverAnimator);
        
        FDT[] modelSolution = exercise.solve();
        FDT[] userSolution  = user; // userSolution is a class variable
        int userSolutionState;
                        
        if(modelSolution.length != userSolution.length) {
            Note.warning(null, "User solution and model solution have different number of structures, stopping judgement.");
            return new Judgement(1);
        }
            
        //Note.out(null, "rewinding model solution");
        modelCount = solverAnimator.rewind();
        //Note.out(null, "rewinding user solution");
        // GUI handles placing the beginning state of the user animator to the right place 
            
        // the states of the user animator are evaluated only until the current state.
        // If there are any more states in the user animator, they are removed
            
        // Panu commented this out 2003-01-21
        //Stack redoStack = userAnimator.disposeRedo(); 
            
        userCount = userAnimator.rewind();
        //Note.out(null, "starting judgement user: "+userCount+" model: "+modelCount);
        int oldUserState = 0; //state of user solution in last iteration
            
        //rewindAnimator(solverAnimator);
        //rewindAnimator(userAnimator);
        //userState = -userCount;

        userSolutionState = 0;
        correctCount = 0;
        while(!(allEqual(userSolution,modelSolution)) && solverAnimator.hasNextOperation()) {
            solverAnimator.redo();
            modelCount--;
        }

        Judgement result = new Judgement(new String[]{"NEVER USED!"},1,modelCount);

        
        
        if(!(allEqual(userSolution,modelSolution)) || modelCount == 0) {
            Note.warning(null, "No state matching the starting state of the user animator found.");
            result.setPoints(0, 0);                
        }
                        
        for (int i=0; i<modelCount; i++) {
            solverAnimator.redo();
            //Note.out(null, "Comparing states "+solverAnimator);
            oldUserState = userSolutionState;
                
            while(!(allEqual(userSolution,modelSolution)) && userAnimator.hasNextOperation() ) {
                userSolutionState++;
                userAnimator.redo();
            }
            if (allEqual(userSolution,modelSolution)){
                //Note.out(null,"state "+i+" correct");
                // Added by Harri 11-08-2003
                // all states must in correct order
//              commenting this line made it skip : //if (i==correctCount)
                    correctCount++;
            }
            else {
                //Note.out(null,"state " + i + " wrong");
                if (!result.errorStateIndexSet()) {
                    result.setErrorStateIndex(i);
                }
                    
                for(; userSolutionState > oldUserState; userSolutionState--)
                    userAnimator.undo();
            }
                
        }
        while(userAnimator.hasNextOperation()) {
            userAnimator.redo();
        }
            
        result.setPoints(correctCount, 0);        
            
        if(isFinished(userAnimator, solverAnimator, userSolution, modelSolution)) {
            result.setComplete();
            //Note.out(null, "exercise is complete");
        }
            
        userAnimator.end();
            
        return result;
    }           

    /**
     * This method creates the judgement for a simulation exercise when we want
     * to grade the different user structures separately of each other.
     * 
     * IMPORTANT NOTE! : This method uses the skipping strategy
     * 
     * @param user          The structures used by the user
     * @param userAnimator  The Animator used by the user
     * @param exercise      The exercise being solved
     * @return              The result of the judging
     */
  
    public Judgement getJudgementSeparateFDTs(FDT[] user, Animator a, SimulationExerciseModel exercise) {

        Animator userAnimator = a; // userAnimator is a class variable
        Animator solverAnimator = new Animator();

        int modelCount   = 0; //number of correct states is modelSolution
        int userCount    = 0; //number of states in user's sulution
        int correctCount = 0; // number of correct states in user solution
        int userSolutionState;
        
        Animator.setActiveAnimator(solverAnimator);
        
        FDT[] modelSolution= exercise.solve();
        FDT[] userSolution = user;
        
        if(modelSolution.length != userSolution.length) {
            Note.warning(null, "The user solution has a different number of structures with "+
                               "modelSolution, stopping judgement.");
            return new Judgement(modelSolution.length);
        }
        //Note.out(null, "rewinding model solution");
        modelCount = solverAnimator.rewind();
        Judgement result = new Judgement(exercise.getStructureNames(), modelSolution.length, modelCount); 

        //Note.out(null, "rewinding user solution");
        // GUI handles placing the beginning state of the user animator to the right place 
        
        // the states of the user animator are evaluated only until the current state.
        // If there are any more states in the user animator, they are removed
            
        // Panu commented this out 2003-01-21
        //Stack redoStack = userAnimator.disposeRedo(); 
            
        userCount = userAnimator.rewind();
        //Note.out(null, "starting judgement");
        int oldUserState = 0; //state of user solution in last iteration
            
        // For-loop executed for each FDT in modelSolution
        for(int j = 0; j < modelSolution.length; j++) {
            solverAnimator.rewind();
            userAnimator.rewind();
            //userState = -userCount;
            userSolutionState = 0;
            correctCount = 0;
            //Note.out(null, "COMPARE " + userSolution[j] + " with " + modelSolution[j]);
            boolean cmpResult = userSolution[j].equals(modelSolution[j]);
            while((cmpResult == false) && solverAnimator.hasNextOperation()) {
                solverAnimator.redo();
                cmpResult = userSolution[j].equals(modelSolution[j]);
            }
            if (cmpResult == false) {
                //Note.out(null, "No state matching the starting state of the user animator found");
                result.setPoints(0, j);
                continue;
            }
            for (int i=0; i<modelCount; i++) {
                solverAnimator.redo();
                //Note.out(null, "Comparing states");
                oldUserState = userSolutionState;
                                
                /** 
                 * Search model solution's current state from user's solution. Search starts
                 * from position defined by Animator
                 */
                while(!(userSolution[j].equals(modelSolution[j])) && userAnimator.hasNextOperation() ) {
                    userSolutionState++;
                    userAnimator.redo();
                }
                
                if (userSolution[j].equals(modelSolution[j])) {
                    //Note.out(null,"state "+i+" correct");
                    // Added by Harri 11-08-2003
                    // all states must in correct order
 //commenting this line made it skip : if (i==correctCount)
                        correctCount++;
                }
                else {
                    //Note.out(null,"state " + i + " wrong");
                    if (!result.errorStateIndexSet()) {
                        result.setErrorStateIndex(i);
                    }
                    
                    for(; userSolutionState > oldUserState; userSolutionState--)
                        userAnimator.undo();
                }
                
            }
            while(userAnimator.hasNextOperation()) {
                userAnimator.redo();
            }
            result.setPoints(correctCount, j);
            
        }
        if(isFinished(userAnimator, solverAnimator, userSolution, modelSolution)) {
            result.setComplete();
            //Note.out(null, "exercise is complete");
        }
        //Note.out(null, "judge end");
        userAnimator.end();
        
        return result;
    }
    

}