//Matrix - algorithm simulation system
//Copyright (C) 2000- 2007 Matrix group
//(See the AUTHORS file and the author tag below for more information
//on the authors.)

//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.

//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package content.exercises;

import javax.swing.JButton;
import java.awt.Component;
import java.awt.GridBagConstraints;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Random;

import content.ExerciseProperties;
import content.interfaces.AWTComponentUtilizer;
import content.interfaces.ConfigureVisualType;
import content.interfaces.GradeDialogLayout;
import content.interfaces.JudgeBlocks;
import content.interfaces.ModelAnswerNames;
import content.interfaces.ComparableExercise;
import content.interfaces.StyledExercise;
import content.exercises.structures.ExerLabeledGraph_DH;
import content.exercises.structures.ExerQueue_DH;
import content.exercises.structures.ExerLabeledHeap_DH;
import content.exercises.structures.PriorityExerVertex_DH;

import matrix.animation.Animator;
import matrix.simulation.VisualTypeConf;
import matrix.structures.FDT.*;
import matrix.structures.FDT.probe.Key;
import matrix.structures.FDT.substructures.Vertex;
import matrix.util.Application;
import matrix.visual.VisualKey;
import matrix.visual.VisualType;

/**
 * This is an exercise where the user executes Dijkstra's algorithm for the given graph.
 * This exercise contains some additional structures that are used when solving the exercise.
 * Structures that the user has to use in this exercise are the graph, the heap and the 
 * traversing order list.
 * 
 * @author kstalnac
 */
public class Dijkstra_Heap implements ComparableExercise, ModelAnswerNames, StyledExercise , 
							ConfigureVisualType, JudgeBlocks, /*ButtonExercise,*/ AWTComponentUtilizer,
							GradeDialogLayout {
    private int randomWanted = 6;
    private int randomTries = 100;
    /** Vertices used when creating the model answer. */
    protected PriorityExerVertex_DH modelVertices[];
    /** Graph that the user manipulates */
    protected ExerLabeledGraph_DH usergraph;
    /** Graph used in the model answer. */
    protected ExerLabeledGraph_DH modelgraph;
    /** Graph in adjacency list format. */
    protected String[] NodeList;
    /** Keys used in the graph. */
    protected final String keys = "ABCDEFGHJ";
    /** Heap that the user manipulates. */
    protected ExerLabeledHeap_DH userHeap;
    /** Queue which the user manipulates. */
    protected ExerQueue_DH userQueue;
    /** Heap used in the model answer. */
    protected ExerLabeledHeap_DH modelHeap;
    /** Queue used in the model answer. */
    protected ExerQueue_DH modelQueue;


    public static final boolean DEBUG = false;


    /**
     * Constructor
     * All the initializations are made in the init method of exercise. 
     */
    public Dijkstra_Heap() {}


    // Methods for styled exercise interface.

    public String[] getStructureVisualisations() {
        String[] s = { "Kamada-Kawai point graph", "layered tree", "list" };
        return s;
    }


    public String[] getModelAnswerVisualisations() {
        String[] s = { "Kamada-Kawai point graph", "layered tree", "list" };
        return s;
    }

    // end of styled exercise interface methods

    long seed = 1;

    /**
     * Get the value of seed.
     * @return value of seed.
     */
    public long getSeed() {
        return seed;
    }

    /**
     * Set the value of seed.
     * @param v  Value to assign to seed.
     */
    public void setSeed(long  v) {
        this.seed = v;
    }


    /**
     * Creates symmetric edge between the given vertices.
     * @param u one end of the symmetric edge
     * @param v other end of the symmetric edge
     * @param weight weigth of the symmetric edge
     */
    private void createSymmetricEdge(PriorityExerVertex_DH u, PriorityExerVertex_DH v, int weight) {
        //Note.out(this, "create edge " + u.toString() + "-" + v.toString());
        u.insertSuccessor(v, weight);
        v.insertSuccessor(u, weight);
    }

    /**
     * Creates the adjacency list representation from the graph vertices.
     * @param list list in which the adjacency list representations for each vertex are created
     * @param t array of graph vertices.
     */
    private void createAdjacencyList(String[] list, PriorityExerVertex_DH[] t) {
        for (int i=0; i<t.length; i++)
            NodeList[i] = "" + t[i].getElement() + ": ";
        for (int i=0; i<t.length; i++) {
            PriorityExerVertex_DH u = t[i];
            Vertex v[] = u.getSuccessors();
            for (int j=0; j<v.length; j++) {
                list[i] += v[j].getElement() + "" + t[i].getWeight((PriorityExerVertex_DH)v[j]) + " ";
            }
        }
    }

    /**
     * Init method of the exercise
     * @return array of the initialized structures used in the exercise.
     */
    public FDT[] init() {
        int l = keys.length(); // keys is the string of alphabet nodes in the graph
        Random R = new Random(seed);
        int i, n;

        // V and modelVertices are arrays of nodes that forms the graph for user and for the model answer, respectively
        PriorityExerVertex_DH V[] = new PriorityExerVertex_DH[l]; 
        modelVertices = new PriorityExerVertex_DH[l];
        NodeList = new String[l]; // adjacency list representation of the graph
        usergraph = new ExerLabeledGraph_DH(); //  G=(V,E), i.e., usergraph=(V,E);

        // Create vertices V
        for(i=0; i<l; i++) {
            V[i]=new PriorityExerVertex_DH();
            V[i].setElement(new Key(""+keys.charAt(i)));
            modelVertices[i]=new PriorityExerVertex_DH();
            modelVertices[i].setElement(new Key(""+keys.charAt(i)));
        }

        //First make sure that graph contains edges which make the relaxation guaranteed

        //First we select the first vertex and two random vertex (all are different from each other)
        int first = 0;
        int random1 = 0;
        int random2 = 0;
        while (random1 == first) {
            random1 = Math.abs(R.nextInt()%l);
        }
        while (random2 == first || random2 == random1) {
            random2 = Math.abs(R.nextInt()%l);
        }

        /* Next we create edge weights so that relaxation occurs with these three 
         * nodes (In that case, when the algorithm starts from the first node.)
         * 
         */
        int firstToRandom2Weight = 1;
        int firstToRandom1Weight = 1;
        int random1ToRandom2Weight = 1;
        firstToRandom2Weight = R.nextInt(5) + 11; //possible values are 11-15
        firstToRandom1Weight = R.nextInt(5) + 1; //possible values are 1-5
        random1ToRandom2Weight = R.nextInt(5) + 1; //possible values are 1-5

        //finally we add the edges with selected weights to graph and modelgraph
        createSymmetricEdge(V[first], V[random2], firstToRandom2Weight);
        createSymmetricEdge(V[first], V[random1], firstToRandom1Weight);
        createSymmetricEdge(V[random1], V[random2], random1ToRandom2Weight);
        createSymmetricEdge(modelVertices[first], modelVertices[random2], firstToRandom2Weight);
        createSymmetricEdge(modelVertices[first], modelVertices[random1], firstToRandom1Weight);
        createSymmetricEdge(modelVertices[random1], modelVertices[random2], random1ToRandom2Weight);

        /* Then we have to make sure that the resulting graph is unified, that is, from 
         * every node, there is a path to all the other nodes.
         * We do this by adding each node to the already existing graph.
         */

        /* Nodes that are already added to a graph are saved in the temp array.
         * In fact, the index of those nodes in the arrays V and modelVertices is saved
         * in the tempArray.
         */
        int[] tempArray = new int[l];
        tempArray[0] = first;
        tempArray[1] = random1;
        tempArray[2] = random2;

        int nodesInGraph = 3;
        int nextNode; //next node to be added to the graph
        /* Node already in the graph to which the next node is connected.
         * This number points to tempArray and from the tempArray we get the index of the node
         * in the arrays V and modelVertices.*/
        int toNode;  
        int weight;

        for (nextNode = 0; nextNode < l; nextNode++) {
            if (V[nextNode].getSuccessors().length < 1) {
                toNode = Math.abs(R.nextInt())%nodesInGraph;
                while (V[tempArray[toNode]].getSuccessors().length > 2) {
                    toNode = Math.abs(R.nextInt())%nodesInGraph;
                }
                weight = Math.abs(R.nextInt())%15 + 1;
                //now we have a node to which we can connect the next node
                createSymmetricEdge(V[nextNode], V[tempArray[toNode]], weight);
                createSymmetricEdge(modelVertices[nextNode], modelVertices[tempArray[toNode]], weight);
                tempArray[nodesInGraph] = nextNode;
                nodesInGraph++;
            }
        }


        /* Then we create rest of the edges randomly
         * Number of the edges is restricted so that every node has at most four successors
         * (three successors is the most usual number)
         */
        int loopCounter = 0;
        weight = 1;
        for (int k = 0; k < 3; k++) {
            for (int j = 0; j < l; j++) {
                if (V[j].getSuccessors().length < 3) {
                    n = j;
                    loopCounter = 0;
                    while (n == j || V[j].hasSuccessor(V[n]) || V[n].getSuccessors().length >= 3) {
                        loopCounter++;
                        if (loopCounter > randomTries && n != j) {
                            break;
                        }
                        n = Math.abs(R.nextInt()%l);                        
                    }
                    weight = Math.abs(R.nextInt()%15) + 1; //weights can be 1-15
                    createSymmetricEdge(V[j], V[n], weight);
                    createSymmetricEdge(modelVertices[j], modelVertices[n], weight);
                }
            }
        }


        createAdjacencyList(NodeList, V);
        usergraph.setVertices(V);

        //switch the weights visible for the user graph
        usergraph.setReferenceLabelEnabled(true);

        //pass the references between graph, heap and queue
        userHeap = new ExerLabeledHeap_DH(usergraph);
        userQueue = new ExerQueue_DH(usergraph, userHeap);
        usergraph.setHeap(userHeap);

        FDT[]  s = { usergraph, userHeap, userQueue };
        return s;
    }

    public FDT[] makeModelAnswer() {
        return solve();
    }


    public FDT[] init(Object input) {
        return init();
    }


    public String[] getStructureNames() {
        ExerciseProperties lang = ExerciseProperties.getInstance();
        
        String[] s = { lang.get("DIJKSTRA_HEAP_GRAPH_TITLE"),
                "   "+lang.get("DIJKSTRA_HEAP_HEAP_TITLE")+"   ",
                lang.get("DIJKSTRA_HEAP_VISITING_TITLE") };

        return s;
    }


    public String getDescription() {
        /*String S="Show how the Dijkstra's algorithm finds the shortest path spanning tree for the weighted graph below "
            + "starting from node " + keys.charAt(0) +".\n\n";*/
        String S="";
        for (int i = 0; i < keys.length(); i++) {
            S += NodeList[i];
            if (i < keys.length() - 1)
                S += "\n";
        }
        return S;
    }

    public FDT[] solve() {

        //create the model graph, pass the vertices to it, and switch the weight visible for it
        modelgraph = new ExerLabeledGraph_DH();
        modelgraph.setVertices(modelVertices);
        modelgraph.setReferenceLabelEnabled(true); // show weights
        //create the heap and the queue and pass the references between the graph, the heap, and the queue
        modelHeap = new ExerLabeledHeap_DH(modelgraph);
        modelQueue = new ExerQueue_DH(modelgraph, modelHeap);
        modelgraph.setHeap(modelHeap);

        //create instance of the algorithm implementation and pass it all the needed objects
        SSSP_Dijkstra_Heap dijkstra = new SSSP_Dijkstra_Heap(modelgraph, modelHeap, modelQueue, modelVertices[0], modelVertices);

        //run the algorithm
        dijkstra.run();

        return new FDT[] { modelgraph, modelHeap, modelQueue };
    }

    public String[] getModelAnswerNames() {
        ExerciseProperties lang = ExerciseProperties.getInstance();
        
        String[] s = { lang.get("DIJKSTRA_HEAP_GRAPH_MODEL_TITLE"),
                "   "+lang.get("DIJKSTRA_HEAP_HEAP_MODEL_TITLE")+"   ",
                lang.get("DIJKSTRA_HEAP_VISITING_MODEL_TITLE") };
        return s;
    }

    public FDT[] getAnswer() {
        FDT[] s = { usergraph, userHeap, userQueue };
        return s;
    }

    public FDT[] getInitialStructures() {
        throw new RuntimeException("not implemented");
    }

    public VisualTypeConf[] conf() {
        VisualTypeConf graph_vtc = new VisualTypeConf();
        graph_vtc.enable("matrix.visual.VisualKey", VisualTypeConf.HIGHLIGHT_OPERATION);
        graph_vtc.enable("matrix.visual.VisualKey", VisualTypeConf.DRAG_OPERATION);
        VisualTypeConf heap_vtc = new VisualTypeConf();
        heap_vtc.enable("matrix.visual.VisualLayeredTree", VisualTypeConf.HIGHLIGHT_OPERATION);
        heap_vtc.enable("matrix.visual.VisualLayeredTree", VisualTypeConf.DROP_OPERATION);
        heap_vtc.enable("matrix.visual.VisualKey", VisualTypeConf.HIGHLIGHT_OPERATION);
        heap_vtc.enable("matrix.visual.VisualKey", VisualTypeConf.DRAG_OPERATION);
        VisualTypeConf queue_vtc = new VisualTypeConf();
        queue_vtc.enable("matrix.visual.VisualList", VisualTypeConf.HIGHLIGHT_OPERATION);
        queue_vtc.enable("matrix.visual.VisualList", VisualTypeConf.DROP_OPERATION);
        return new VisualTypeConf[] { graph_vtc, heap_vtc, queue_vtc };
    }

    /**
     * This method updates the priority of the key in the heap.
     * The key which priority is to be updated is selected by clicking it
     * in the graph or in the heap.
     *
     */
    public void relax() {
        if (VisualType.read instanceof VisualKey) {
            Key tempKey = (Key)VisualType.read.getStructure();
            int distance = usergraph.getDistance(tempKey);
            // -1 means that key was not successor of lastChosen,
            // 0 means that key is the first candidate for lastChosen (so lastChosen has not been set earlier)
            if (distance == -1 || distance == 0) { 
                return;
            }
            Animator.getActiveAnimator().startOperation();
            userHeap.updatePriorityOfKey(tempKey, distance);
            Animator.getActiveAnimator().endOperation();

            application.validate();
            application.validateAnimator();
        } else {
            //do nothing
        }
    }

    public Component[] getAWTComponents() {
        ExerciseProperties lang = ExerciseProperties.getInstance();

        JButton update = new JButton(lang.get("DIJKSTRA_HEAP_UPDATE_BUTTON_TITLE") );
        update.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent ae) {
                relax();
            }
        });
        return new Component[]{update};
    }

    public GridBagConstraints[] getAWTComponentConstraints() {
        return null;
    }

    private Application application = null;
    public void setApplication(Application a) {
        application = a;
    }
    
    /* ComparableExercise interface */

    public int[] getCompareIndices() {
        return new int[] {0, 1, 2};
    }

    public int[] getGradeIndices() {
        return new int[] {1};
    }

    public boolean canRecover() {
        return true;
    }
    
	public String[] getCompareStructureVisualisations(){
		return null; // use default
	}
	
	public String[] getCompareStructureNames(){
		return null; // use default
	}    
	
	public VisualTypeConf[] getCompareVisualTypeConf(){
		return null; // use default
	}

	/* GradeDialog interface */
	
	public boolean showModelAnswerInGradeDialog(){
		return false;
	}
	
	public boolean showComparisonInGradeDialog(){
		return true;
	}

    private static final long serialVersionUID = 8411516459954712923L;
}
