//Matrix - algorithm simulation system
//Copyright (C) 2000- 2007 Matrix group
//(See the AUTHORS file and the author tag below for more information
//on the authors.)

//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.

//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


package content.exercises;

import java.util.ArrayList;
import java.util.Random;

import content.ExerciseProperties;
import content.exercises.Hash_Operations.HashTask;
import content.exercises.Hash_Operations.SelectableTable;
import content.interfaces.ConfigureVisualType;
import content.interfaces.JudgeBlocks;
import content.interfaces.ModelAnswerNames;
import content.interfaces.ComparableExercise;

import matrix.animation.Animator;
import matrix.simulation.VisualTypeConf;
import matrix.structures.FDT.FDT;
import matrix.structures.FDT.LinkedList;
import matrix.structures.FDT.Struct;
import matrix.structures.FDT.probe.LinkedListImpl;
import matrix.structures.FDT.probe.Table;
import matrix.structures.memory.VirtualBoolean;
import matrix.structures.memory.VirtualInteger;
import matrix.structures.memory.VirtualObject;
import matrix.structures.simulationextensions.Selectable;

/**
 * A hashing excercise where probing method is separate chaining.
 * Exercise contains operations add, search and remove.
 * The amount of add, search and remove operations can be set via exercise properties. 
 * 
 * @author lhakulin
 *
 */
public class Hash_Operations_Chaining implements ComparableExercise, ModelAnswerNames, ConfigureVisualType, JudgeBlocks {

    private static final String EMPTY = "";
	
    private Table inputDataTable;
    private Table currentInput;
    private Table userSolutionTable;
    
    private static final int size = 11;
    private static final int duplicates = 4;
    private static final int moreDuplicates = 2;
    private static final int forcedBeginAdditions = 4; // minimum amount of additions in the beginning
    
    // default values of parameters
	private static final int defaultAddDataSize = 15;
	private static final int defaultSearchDataSize = 4;
	private static final int defaultRemoveDataSize = 2;
	private static final int defaultFailedSearchDataSize = 1;
	private static final int defaultInitialDataSize = 0;

    private String taskAddText; 
    private String taskRemoveText; 
    private String taskSearchText;     
    private int initialDataSize;
    
    private static final int min = 100;
    private static final int max = 999;

    public static final boolean DEBUG = false;

    private long seed = 1;

    public Hash_Operations_Chaining() {
    }

    /**
     * Get the value of seed.
     * @return value of seed.
     */
    public long getSeed() {
        return seed;
    }

    /**
     * Set the value of seed.
     * @param v  Value to assign to seed.
     */
    public void setSeed(long  v) {
        this.seed = v;
    }


    /**
     * Gets the texts used in exercise from exercise properties
     *
     */
    private void initTexts(){
    	ExerciseProperties lang = ExerciseProperties.getInstance();
        taskAddText = lang.get("HASH_OPERATIONS_CHAINING_TASK_ADD_TEXT");
        taskSearchText = lang.get("HASH_OPERATIONS_CHAINING_TASK_SEARCH_TEXT");
        taskRemoveText = lang.get("HASH_OPERATIONS_CHAINING_TASK_REMOVE_TEXT");
    }
    
    /**
     * Gets the amount of add operations from exercise properties.
     * Uses default value if the value from exercise properties is not valid
     * @return amount of add operations
     */
    private int getAddDataSize(){
    	int addDataSize;
        ExerciseProperties lang = ExerciseProperties.getInstance();
        try {
            int value = Integer.parseInt(lang.get("HASH_OPERATIONS_CHAINING_OPERATION_ADD_COUNT").trim());
            addDataSize = Math.max(value, duplicates+1);
            addDataSize = Math.min(addDataSize, size);
		} catch (NumberFormatException e) {
			addDataSize = defaultAddDataSize;
		}
		return addDataSize;
    }

    /**
     * Gets the amount of search operations from exercise properties.
     * Uses default value if the value from exercise properties is not valid
     * @return amount of search operations
     */
    private int getSearchDataSize(int addDataSize){
    	int searchDataSize;
        ExerciseProperties lang = ExerciseProperties.getInstance();
        try {
            int value = Integer.parseInt(lang.get("HASH_OPERATIONS_CHAINING_OPERATION_SEARCH_COUNT").trim());			
            searchDataSize = Math.min(value, addDataSize);
		} catch (NumberFormatException e) {
			searchDataSize = defaultSearchDataSize;
		}
		return searchDataSize;
    }
    
    /**
     * Gets the amount of failed search operations from exercise properties.
     * Uses default value if the value from exercise properties is not valid
     * @return amount of failed search operations
     */
    private int getFailedSearchDataSize(int searchDataSize){
    	int failedSearchDataSize;
        ExerciseProperties lang = ExerciseProperties.getInstance();
        try {
            int value = Integer.parseInt(lang.get("HASH_OPERATIONS_OPERATION_FAILED_SEARCH_COUNT").trim());			
            failedSearchDataSize = Math.min(value, searchDataSize);
		} catch (NumberFormatException e) {
			failedSearchDataSize = defaultFailedSearchDataSize;
		}
		return failedSearchDataSize;
    }
    
    /**
     * Gets the amount of remove operations from exercise properties.
     * Uses default value if the value from exercise properties is not valid
     * @return amount of remove operations
     */
    private int getRemoveDataSize(int addDataSize){
    	int removeDataSize;
        ExerciseProperties lang = ExerciseProperties.getInstance();
        try {
            int value = Integer.parseInt(lang.get("HASH_OPERATIONS_CHAINING_OPERATION_REMOVE_COUNT").trim());
            removeDataSize = Math.min(value, addDataSize);
		} catch (NumberFormatException e) {
			removeDataSize = defaultRemoveDataSize;
		}
    	return removeDataSize;
    }

    /**
     * Gets the amount keys initially in the table from exercise properties.
     * Uses default value if the value from exercise properties is not valid
     * @return amount of initial keys in the table
     */
    private int getInitialDataSize(int addDataSize){
    	int initialDataSize;
        ExerciseProperties lang = ExerciseProperties.getInstance();
        try {
            initialDataSize = Integer.parseInt(lang.get("HASH_OPERATIONS_CHAINING_INITIAL_COUNT").trim());
		} catch (NumberFormatException e) {
			initialDataSize = defaultInitialDataSize;
		}
    	return initialDataSize;
    }

    public FDT[] init() {
    	
    	initTexts();
    	    	
    	int addDataSize = getAddDataSize();
        int searchDataSize = getSearchDataSize(addDataSize);
        int removeDataSize = getRemoveDataSize(addDataSize);
        int failedSearchDataSize = getFailedSearchDataSize(searchDataSize);
        initialDataSize = getInitialDataSize(addDataSize);
            	
        // new random
        Random R = new Random(seed);
        int inputData[] = new int[addDataSize];
        // create random data

        for(int i = 0; i < addDataSize-duplicates; i++) {
            inputData[i] = Math.abs(R.nextInt()%(max-min))+min;
        }

        // Create duplicates
        for(int i = addDataSize-duplicates; i < addDataSize; i++) {

            do {
                inputData[i] = inputData[(0+i+duplicates-addDataSize)%(duplicates-moreDuplicates)];
                inputData[i] -= (R.nextInt(addDataSize))*size;
            } while(inputData[i]<min || inputData[i] > max);
        }
                
        inputDataTable = new Table(initialDataSize + inputData.length + searchDataSize + removeDataSize);
        
        // the initial data 
        for (int i=0; i < initialDataSize; i++){
        	int value = Math.abs(R.nextInt()%(max-min))+min;
    		inputDataTable.setObject(new HashTask(value, taskAddText), i);
        }
        
        int inputInd = 0;
        int searchCount = 0;
        int removeCount = 0;
        int failedSearchCount = 0;
        ArrayList hashTableContent = new ArrayList();
        	        
        for (int i=initialDataSize; i < inputDataTable.size(); i++){
        	int f = R.nextInt(inputDataTable.size());
        	
        	// add
        	if (i < forcedBeginAdditions+initialDataSize ||
        			(searchCount >= searchDataSize && removeCount >= removeDataSize) ||
        			(hashTableContent.size() == 0) || (f < addDataSize && inputInd < inputData.length)){
        			        		
            	hashTableContent.add(new Integer(inputData[inputInd]));
        		inputDataTable.setObject(new HashTask(inputData[inputInd++], taskAddText), i);
        	}
        	// search
        	else if ((removeCount >= removeDataSize) ||
        				(f < addDataSize+searchDataSize && searchCount < searchDataSize) ||
        				(searchCount < searchDataSize && hashTableContent.size() <= 1)){
        		
        		int tmp = R.nextInt(searchDataSize+failedSearchDataSize);
        		// failed search
        		if ((tmp < failedSearchDataSize || searchCount+(failedSearchDataSize-failedSearchCount) >= searchDataSize) &&
        				failedSearchCount < failedSearchDataSize){
        			
        			int value;
        			do {
        				value = Math.abs(R.nextInt()%(max-min))+min;
        			} while(hashTableContent.contains(new Integer(value)));
        			
	        		inputDataTable.setObject(new HashTask(value, taskSearchText), i);
        			failedSearchCount++;
        		}
        		// successfull search
        		else {
	        		int ind = R.nextInt(hashTableContent.size());
	        		inputDataTable.setObject(new HashTask(((Integer)hashTableContent.get(ind)).intValue(), taskSearchText), i);
        		}
        		searchCount++;
        	}
        	// remove
        	else {
        		int ind = R.nextInt(hashTableContent.size());
        		inputDataTable.setObject(new HashTask(((Integer)hashTableContent.remove(ind)).intValue(), taskRemoveText), i);
        		removeCount++;
        	}
        }
        currentInput = new Table(1);
        currentInput.setObject(inputDataTable.getObject(0), 0);
        
        userSolutionTable = new SelectableTable(size, inputDataTable, currentInput);
        
        // add the inital keys
        for (int i=0; i < initialDataSize; i++){
        	HashTask task = ((HashTask)inputDataTable.getObject(i));
        	int hashValue = (((Integer)task.getField(HashTask.VALUE)).intValue()) % size;
            LinkedList currentList = (LinkedListImpl)userSolutionTable.getObject(hashValue);
            while (!((SelectableTable)userSolutionTable).selected((HashCell)currentList.getElement())){
            	currentList = currentList.getNext();
            }
        }
        
        return new FDT[] { currentInput, userSolutionTable };
    }


    public FDT[] init(Object o) {
        return init();
    }

    public String[] getStructureNames() {
        ExerciseProperties lang = ExerciseProperties.getInstance();
        return new String[] { 
                lang.get( "HASH_OPERATIONS_CHAINING_INPUT_TITLE" ),
                lang.get( "HASH_OPERATIONS_CHAINING_HASHTABLE_TITLE" )
        };
    }

    public String getDescription() {
        ExerciseProperties lang = ExerciseProperties.getInstance();
        String prompt1 =  lang.get( "HASH_OPERATIONS_CHAINING_PROMPT_PART_1" ); // "The hash function is h(k) = k "
        String prompt2 =  lang.get( "HASH_OPERATIONS_CHAINING_PROMPT_PART_2" ); // " mod "

        String s = prompt1 + " " + prompt2 + " " + size + "\n";
                
        s += lang.get( "HASH_OPERATIONS_CHAINING_PROBING_TYPE_TEXT" ); // "Probing type: separate chaining";
        return s;
    }
    

    public FDT[] solve() {
        FDT[] initialStructures = getInitialStructures();
        Table input =  (Table)initialStructures[0];
        Table modelSolution = (Table)initialStructures[1];

        for(int i = initialDataSize; i < inputDataTable.size(); i++) {
        	HashTask task = ((HashTask)inputDataTable.getObject(i));
        	int hashValue = (((Integer)task.getField(HashTask.VALUE)).intValue()) % size;
            ((SelectableTable)modelSolution).clearSelections();
            LinkedList currentList = (LinkedListImpl)modelSolution.getObject(hashValue);
            while (!((SelectableTable)modelSolution).selected((HashCell)currentList.getElement())){
            	currentList = currentList.getNext();
            }
        }
        return new FDT[] { input, modelSolution };
    }

    public FDT[] makeModelAnswer() {
        return solve();
    }


    public FDT[] getInitialStructures() {
        Table current = new Table();
        current.setObject(inputDataTable.getObject(0), 0);
        Table resultTable = new SelectableTable(size, inputDataTable, current);    
        
        // add the inital keys
        for (int i=0; i < initialDataSize; i++){
        	HashTask task = ((HashTask)inputDataTable.getObject(i));
        	int hashValue = (((Integer)task.getField(HashTask.VALUE)).intValue()) % size;
            LinkedList currentList = (LinkedListImpl)resultTable.getObject(hashValue);
            while (!((SelectableTable)resultTable).selected((HashCell)currentList.getElement())){
            	currentList = currentList.getNext();
            }
        }
        return new FDT[] { current, resultTable };
    }

    public FDT[] getSimulatedStructures() {
        return new FDT[] { inputDataTable, userSolutionTable };
    }
    
    public String[] getModelAnswerNames() {
        ExerciseProperties lang = ExerciseProperties.getInstance();
        return new String[] { lang.get("HASH_OPERATIONS_CHAINING_INPUT_MODEL_TITLE"), lang.get( "HASH_OPERATIONS_CHAINING_HASHTABLE_MODEL_TITLE" ) };
    }

    public FDT[] getAnswer() {
        return new FDT[] { currentInput, userSolutionTable };
    }

    public ComparableExercise getModelAnswer() {
        return this;
    }

    public VisualTypeConf[] conf() {
        VisualTypeConf tabl = new VisualTypeConf();
        tabl.setVisualConfValue("matrix.visual.VisualArray",VisualTypeConf.INDEXED, "false");
        VisualTypeConf hashtabl = new VisualTypeConf();
        hashtabl.enable("matrix.visual.VisualStruct",VisualTypeConf.HIGHLIGHT_OPERATION);
		hashtabl.setVisualConfValue("matrix.visual.VisualList", VisualTypeConf.ROTATED, "true");
		hashtabl.setVisualConfValue("matrix.visual.VisualList", VisualTypeConf.FLIPPEDY, "true");

        return new VisualTypeConf[] { tabl, hashtabl };
    }

    public int[] getCompareIndices() {
        return new int[] {0, 1};
    }

	public int[] getGradeIndices(){
        return null; // Used only in recovery mode
	}

	public boolean canRecover(){
		return false;
	}
	
	public String[] getCompareStructureVisualisations(){
		return null; // use default
	}
	
	public String[] getCompareStructureNames(){
		return null; // use default
	}    
	
	public VisualTypeConf[] getCompareVisualTypeConf(){
		return null; // use default
	}

    static final long serialVersionUID = -37657283333350211L;

    
    /**
     * The hash table of the exercise.
     * @author lhakulin
     *
     */
    private static final class SelectableTable extends Table {
    	    	
    	private Table input; // all the input data of the exercise
    	private Table currentInput; // current input task
    	private VirtualInteger current;
    	
        private String taskAddText; 
        private String taskSearchText; 
        private String taskRemoveText; 
        private String taskFinishedText; 
    	
    	public SelectableTable(int size, Table inputData, Table ci){
    		super();
    		input = inputData;
    		currentInput = ci;
    		current = new VirtualInteger(0, this, "current index of input table");
    		for (int i=0; i < size; i++){
    			LinkedListImpl list = new LinkedListImpl();
    			HashCell cell = new HashCell(EMPTY, this, list);
    			list.setElement(cell);
    			this.setObject(list, i);
    		}
    		setTaskTexts();
    	}
    	
        private void setTaskTexts(){
            ExerciseProperties lang = ExerciseProperties.getInstance();
            taskAddText = lang.get("HASH_OPERATIONS_CHAINING_TASK_ADD_TEXT");
            taskSearchText = lang.get("HASH_OPERATIONS_CHAINING_TASK_SEARCH_TEXT");
            taskRemoveText = lang.get("HASH_OPERATIONS_CHAINING_TASK_REMOVE_TEXT");
            taskFinishedText = lang.get("HASH_OPERATIONS_CHAINING_TASK_FINISHED_TEXT");
        }
    	
    	/**
    	 * Called when one of the cells is selected
    	 * Performs an operation (add, search or remove) according to the currentInput
    	 * @param cell - the cell which was selected
    	 * @return true if currentInput is updated
    	 */
    	public boolean selected(HashCell cell){
    		
    		if (current.eval() > input.getLast()){
    			return false;
    		}
    		
    		boolean result = false;
    		Animator.getActiveAnimator().startOperation();
    		cell.assignSelect(!cell.selected.eval());
    		HashTask task = (HashTask)input.getObject(current.eval());
    		
    		if (task.getField(HashTask.OPERATION).equals(taskAddText)){
    			if (cell.getField(HashCell.VALUE).toString().equals(EMPTY)){    
    				
    				cell.setField(task.getField(HashTask.VALUE), HashCell.VALUE);
    	    		cell.getList().setElement(cell);// needed to get the latest action highlighted when animating
    	    		
    				LinkedListImpl emptyList  = new LinkedListImpl();
    				HashCell hc = new HashCell(EMPTY, this, emptyList);
    				emptyList.setElement(hc);
    				LinkedList list = cell.getList();
    				while(list.getNext() != null){
    					list = list.getNext();
    				}
    				list.setNext(emptyList);
    			}
    			else {
    				LinkedListImpl newList  = new LinkedListImpl();
    				HashCell hc = new HashCell(task.getField(HashTask.VALUE) + "", this, newList);
    				newList.setElement(hc);
    				LinkedList list = cell.getList();
    				while(!((HashCell)(list.getNext().getElement())).getField(HashCell.VALUE).toString().equals(EMPTY)){
    					list = list.getNext();
    				}
    				newList.setNext(list.getNext());
    				list.setNext(newList);
    			}
	    		clearSelections();
	    		current.inc();
	    		currentInput.setObject(input.getObject(current.eval()), 0);
	    		result = true;
    		}
    		else if (task.getField(HashTask.OPERATION).equals(taskRemoveText)){
    			if (cell.getField(HashCell.VALUE).equals(task.getField(HashTask.VALUE).toString()) ||
    					cell.getField(HashCell.VALUE).equals(EMPTY)){
    				if (!cell.getField(HashCell.VALUE).equals(EMPTY)){
    					removeFromList(cell);
    				}
    	    		clearSelections();
    	    		current.inc();
    	    		currentInput.setObject(input.getObject(current.eval()), 0);
    	    		result = true;
    			}
    		}
    		else if (task.getField(HashTask.OPERATION).equals(taskSearchText)){
    			// not found
    			if (cell.getField(HashCell.VALUE).equals(EMPTY)){
    				clearSelections();
    				current.inc();
    	    		currentInput.setObject(input.getObject(current.eval()), 0);    	    		
    	    		result = true;
    			}
    			// found
    			else if (cell.getField(HashCell.VALUE).toString().equals(task.getField(HashTask.VALUE).toString())){

    				// move the selected cell to the first of the list
    				int index = getTableIndex(cell.getList());
    				if (getListLength(index) > 1){
	    				LinkedList first = cell.getList();
	    				LinkedList prev = findPrevious(cell);
	    				if (prev != null){
		    				while (prev != null){
		    					first = prev;
		    					prev = findPrevious((HashCell)prev.getElement());
		    				}
		
		    				removeFromList(cell);
		    				cell.getList().setNext(first);
		    				this.setObject(cell.getList(), index);
	    				}
    				}    				
    	    		clearSelections();
    	    		current.inc();
    	    		currentInput.setObject(input.getObject(current.eval()), 0);
    	    		
    	    		result = true;
    			}
    		}
    		if (currentInput.getObject(0) == null){
    			currentInput.setObject(taskFinishedText, 0);
    		}
    		Animator.getActiveAnimator().endOperation();
    		return result;
    	}
    	
    	/**
    	 * Removes the cell from the linked list
    	 * @param cell - the cell to be removed
    	 */
    	private void removeFromList(HashCell cell){
			LinkedList previous = findPrevious(cell);
			
			if (previous == null && ((HashCell)cell.getList().getNext().getElement()).getField(HashCell.VALUE).equals(EMPTY)){
				// if cell is the only one in the list
				// set it empty
				cell.setField(EMPTY, HashCell.VALUE);
				cell.getList().setNext(null);
			}
			else if (previous == null){
				moveToFirst(cell.getList().getNext());
			}
			else {
				previous.setNext(cell.getList().getNext());
			}
    	}
    	
    	/**
    	 * Finds the previous node in the linked list
    	 * @param cell - the cell whose previous node is to be found
    	 * @return the previous node
    	 */
    	private LinkedList findPrevious(HashCell cell){
    		for (int i=0; i < this.size(); i++){
    			LinkedList node = (LinkedList)getObject(i);
    			while (node != null && node.getNext() != null){
    				if (cell.equals(node.getNext().getElement())){
    					return node;
    				}
    				node = node.getNext();
    			}    			
    		}
    		return null;
    	}
    	
    	/**
    	 * Moves a node to the beginning of the list
    	 * @param list - node to be moved
    	 */
    	private void moveToFirst(LinkedList list){
    		for (int i=0; i < this.size(); i++){
    			LinkedList node = (LinkedList)getObject(i);
    			while (node != null){
    				if (list.equals(node)){
    					this.setObject(list, i);
    					return;
    				}
    				node = node.getNext();
    			}    			
    		}
    	}
    	
    	/**
    	 * Returns the index of the table the list exists
    	 * @param list
    	 * @return the index or -1 if list is not found
    	 */
    	private int getTableIndex(LinkedList list){
    		for (int i=0; i < this.size(); i++){
    			LinkedList node = (LinkedList)getObject(i);
    			while (node != null){
    				if (list.equals(node)){
    					return i;
    				}
    				node = node.getNext();
    			}    			
    		}    		
    		return -1;
    	}
    	
    	/**
    	 * Get the lenght of a linked list in certain index of the table
    	 * @param index - index of the table
    	 * @return the length of the linked list
    	 */
    	private int getListLength(int index){
			int count = 0;
    		LinkedList node = (LinkedList)getObject(index);
			while (node != null){
				node = node.getNext();
				count++;
			}    			
    		return count;
    	}
    	
        /**
         * Clear cell selections in the table
         *
         */
    	private void clearSelections(){
    		for (int i=0; i < this.size(); i++){
    			LinkedList node = (LinkedList)getObject(i);
    			while (node != null){
    				((HashCell)node.getElement()).assignSelect(false);
    				node = node.getNext();
    			}
    		}
    	}

        static final long serialVersionUID = 8668291687493297645L;   
    }
    
    
    /**
     * A hash task contains the value of the key and operation
     * @author lhakulin
     *
     */
    private static class HashTask implements Struct {

		private static final long serialVersionUID = 6332395742556549987L;

		public static final int VALUE = 1; 
        public static final int OPERATION = 2; 

		private VirtualBoolean selected;
        private VirtualInteger value;
        private VirtualObject operation;
        private String[] taskNames;
        private String modString;
        
        public HashTask(int val, String fun){
            value = new VirtualInteger(val, this, "value field");
            operation = new VirtualObject(fun, this, "operation field");
            selected = new VirtualBoolean(false, this, "selected field");
            setTaskNames();
        }
        
        private void setTaskNames(){
            ExerciseProperties lang = ExerciseProperties.getInstance();
            taskNames = new String[4];
            taskNames[0] = "";
            taskNames[1] = lang.get("HASH_OPERATIONS_CHAINING_TASK_KEY_TEXT");
            taskNames[2] = lang.get("HASH_OPERATIONS_CHAINING_TASK_OPERATION_TEXT");
            taskNames[3] = lang.get("HASH_OPERATIONS_CHAINING_TASK_CALCULATION_TEXT");
            
            modString = lang.get( "HASH_OPERATIONS_CHAINING_PROMPT_PART_2" ).trim();
        }
        
        public Object getField(int location) {
            if (location == 0){
                return "";
            }
            else if (location == VALUE){
                return new Integer(value.eval());
            }
            else if (location == OPERATION){
                return operation.toString();
            }
            else {
                int hashValue = (value.eval()) % size;
            	return value.eval() + " " + modString + " " + size + " = " + hashValue;
            }
        }

        public String getFieldName(int location) {
            return taskNames[location];
        }

        public String[] getFieldNames() {
            return taskNames;
        }

        public Object[] getFields() {
            return new Object[]{"", new Integer(value.eval()), operation.toString()};
        }

        public void setField(Object val, int location) {
            if (location == VALUE){
                value.assign(((Integer)val).intValue());
            }
            else if (location == OPERATION){
                operation.setObject(val);
            }
        }

        public Object getElement() {
            return null;
        }

        public void setElement(Object o) {
        }

        public boolean equals(Object o){
            
            if (o instanceof HashTask){
                HashTask r = (HashTask)o;
                return (value.equals(r.value) &&
                        operation.toString().equals(r.operation.toString()) &&
                        selected.equals(r.selected)); 
            }
            return false;
        }

        
    }

    /**
     * A HashCell contains the value of the hashed key.
     * Calls the parent table when selected
     * 
     * @author lhakulin
     *
     */
    private static class HashCell implements Struct, Selectable {

		private static final long serialVersionUID = 6332890425565493645L;

		private VirtualBoolean selected;
		private SelectableTable parent;
	    private LinkedListImpl list;
        private final String[] taskNames = {"",""};
        private VirtualObject value;
        
        public static final int VALUE = 1; 
        
        public HashCell(Object val, SelectableTable p, LinkedListImpl l){
            value = new VirtualObject(val, this, "value field");
            selected = new VirtualBoolean(false, this, "selected field");
            parent = p;
            list = l;
        }
        
        public Object getField(int location) {
            if (location == 0){
                return "";
            }
            else {
                return value.toString();
            }
        }

        public String getFieldName(int location) {
            return taskNames[location];
        }

        public String[] getFieldNames() {
            return taskNames;
        }

        public Object[] getFields() {
            return new Object[]{"", value};
        }

        public void setField(Object val, int location) {
            if (location == VALUE){
                value.setObject(val);
            }
        }

        public Object getElement() {
            return null;
        }

        public void setElement(Object o) {
        }

        public boolean equals(Object o){
            
            if (o instanceof HashCell){
            	HashCell r = (HashCell)o;
                return (value.toString().equals(r.value.toString()) &&
                        selected.equals(r.selected)); 
            }
            return false;
        }
        public long getSelectionTime() {            
            return selected.eval()?1:0;
        }

        public void setSelect(long globalClockTimestamp) {
        	Animator.getActiveAnimator().startOperation();
       		parent.selected(this);            
       		Animator.getActiveAnimator().endOperation();
        }
        
        public void assignSelect(boolean v){
        	selected.assign(v);
        }
        
        public LinkedList getList(){
        	return list;
        }
        
    }

}
