//Matrix - algorithm simulation system
//Copyright (C) 2000- 2007 Matrix group
//(See the AUTHORS file and the author tag below for more information
//on the authors.)

//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.

//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


package content.exercises;

import java.util.ArrayList;
import java.util.Random;

import content.ExerciseProperties;
import content.exercises.Hash_Operations.HashCell;
import content.exercises.Hash_Operations.SelectableTable;
import content.exercises.Hash_Operations.HashTask;
import content.interfaces.ConfigureVisualType;
import content.interfaces.JudgeBlocks;
import content.interfaces.ModelAnswerNames;
import content.interfaces.ComparableExercise;

import matrix.animation.Animator;
import matrix.simulation.VisualTypeConf;
import matrix.structures.FDT.FDT;
import matrix.structures.FDT.probe.Key;
import matrix.structures.FDT.probe.Table;

/**
 * A hashing excercise where the content of the first hash table has
 * to be rehased to a bigger table before completing input operations.
 * Probing method can be linear, quadratic or double hashing.
 * Probing method and the amount of add, search and remove operations can be set
 * via exercise properties. 
 * 
 * @author lhakulin
 *
 */

public class Hash_Operations_Rehash implements ComparableExercise, ModelAnswerNames, ConfigureVisualType, JudgeBlocks {

    private static final int PROBE_TYPE_LINEAR = 0;
    private static final int PROBE_TYPE_QUADRATIC = 1;
    private static final int PROBE_TYPE_DOUBLEHASH = 2;
    private static final String EMPTY = "";
    
    private Table inputDataTable;
    private Table currentInput;
    private Table userSolutionTable;
    private Table userSolutionTable2;
    private String[] firstHashTableContent;

    private static final int size = 11;  // size of the first hashtable
    private static final int secondHashParameter = 7;
    private static final int duplicates = 2;
    private static final int moreDuplicates = 1;
    
    // default values of parameters
	private static final int defaultAddDataSize = 6;
	private static final int defaultSearchDataSize = 2;
	private static final int defaultRemoveDataSize = 2;
    
    static final int min = 100;
    static final int max = 999;

    private String taskAddText; 
    private String taskRemoveText; 
    private String taskSearchText;     
    private String taskRehashText;     
    private String notSelectedText;     
    
    private int probeType = PROBE_TYPE_QUADRATIC; // default is quadratic

    public static final boolean DEBUG = false;

    private long seed = 1;

    public Hash_Operations_Rehash() {
    }

    /**
     * Get the value of seed.
     * @return value of seed.
     */
    public long getSeed() {
        return seed;
    }

    /**
     * Set the value of seed.
     * @param v  Value to assign to seed.
     */
    public void setSeed(long  v) {
        this.seed = v;
    }
  
    /**
     * Gets the texts used in exercise from exercise properties
     *
     */
    private void initTexts(){
    	ExerciseProperties lang = ExerciseProperties.getInstance();
        taskAddText = lang.get("HASH_OPERATIONS_REHASH_TASK_ADD_TEXT");
        taskSearchText = lang.get("HASH_OPERATIONS_REHASH_TASK_SEARCH_TEXT");
        taskRemoveText = lang.get("HASH_OPERATIONS_REHASH_TASK_REMOVE_TEXT");
        taskRehashText = lang.get("HASH_OPERATIONS_REHASH_TASK_REHASH_TEXT");
        notSelectedText = lang.get("HASH_OPERATIONS_REHASH_NOT_SELECTED_TEXT");
    }
    
    /**
     * Gets the amount of add operations from exercise properties.
     * Uses default value if the value from exercise properties is not valid
     * @return amount of add operations
     */
    private int getAddDataSize(){
    	int addDataSize;
        ExerciseProperties lang = ExerciseProperties.getInstance();
        try {
            addDataSize = Integer.parseInt(lang.get("HASH_OPERATIONS_REHASH_OPERATION_ADD_COUNT").trim());
		} catch (NumberFormatException e) {
			addDataSize = defaultAddDataSize;
		}
		return addDataSize;
    }

    /**
     * Gets the amount of search operations from exercise properties.
     * Uses default value if the value from exercise properties is not valid
     * @return amount of search operations
     */
    private int getSearchDataSize(int addDataSize){
    	int searchDataSize;
        ExerciseProperties lang = ExerciseProperties.getInstance();
        try {
            int value = Integer.parseInt(lang.get("HASH_OPERATIONS_REHASH_OPERATION_SEARCH_COUNT").trim());			
            searchDataSize = Math.min(value, addDataSize+(size/2));
		} catch (NumberFormatException e) {
			searchDataSize = defaultSearchDataSize;
		}
		return searchDataSize;
    }
    
    /**
     * Gets the amount of remove operations from exercise properties.
     * Uses default value if the value from exercise properties is not valid
     * @return amount of remove operations
     */
    private int getRemoveDataSize(int addDataSize){
    	int removeDataSize;
        ExerciseProperties lang = ExerciseProperties.getInstance();
        try {
            int value = Integer.parseInt(lang.get("HASH_OPERATIONS_REHASH_OPERATION_REMOVE_COUNT").trim());
            removeDataSize = Math.min(value, addDataSize+(size/2));
		} catch (NumberFormatException e) {
			removeDataSize = defaultRemoveDataSize;
		}
    	return removeDataSize;
    }
    
    /**
     * Sets the probing method got from exercise properties.
     * Uses default value if the value from exercise properties is not valid
     */
    private void setProbeType(){
        ExerciseProperties lang = ExerciseProperties.getInstance();
        try {
            int value = Integer.parseInt(lang.get("HASH_OPERATIONS_REHASH_PROBETYPE").trim());
            if (value == PROBE_TYPE_LINEAR || value == PROBE_TYPE_QUADRATIC || value == PROBE_TYPE_DOUBLEHASH){
            	probeType = value;
            }
		} catch (NumberFormatException e) {
			// using default probeType
		}

    }

    public FDT[] init() {
    	
    	initTexts();
        setProbeType();
    	    	
    	int addDataSize = getAddDataSize();
        int searchDataSize = getSearchDataSize(addDataSize);
        int removeDataSize = getRemoveDataSize(addDataSize);
        int rehashCount = size/2;
    	
        Random R = new Random(seed);
       
        // create rehash data
        int rehashInput[] = new int[rehashCount];

        for(int i = 0; i < rehashCount-duplicates; i++) {
            rehashInput[i] = Math.abs(R.nextInt()%(max-min))+min;
        }

        // Create duplicates for rehash data
        for(int i = rehashCount-duplicates; i < rehashCount; i++) {
            do {
            	rehashInput[i] = rehashInput[(0+i+duplicates-rehashCount)%(duplicates-moreDuplicates)];
                rehashInput[i] -= (R.nextInt(rehashCount))*2*size;
            } while(rehashInput[i]<min || rehashInput[i] > max);
        }
        
        Table rehashData = new Table(rehashInput.length);
        int inputInd = 0;
        for (int i=0; i < rehashData.size(); i++){
        	rehashData.setObject(new HashTask(rehashInput[i], taskAddText, probeType, 2*size), i);
        }
        
        currentInput = new Table(1);

        // first hashing
        userSolutionTable = new SelectableTableRehash(size, rehashData, currentInput);
        firstHashTableContent = new String[size];
        ((SelectableTableRehash)userSolutionTable).setInputSelection(false);
        ((SelectableTableRehash)userSolutionTable).setHashTask(new HashTask(-1, taskRehashText, probeType, 2*size));

        for(int i = 0; i < rehashData.size(); i++) {
        	HashTask task = ((HashTask)rehashData.getObject(i));
        	int hashValue = (((Integer)task.getField(HashTask.VALUE)).intValue()) % size;
            int doubleHashValue = secondHashParameter - (((Integer)task.getField(HashTask.VALUE)).intValue() % secondHashParameter);
            ((SelectableTable)userSolutionTable).clearSelections();          
            probe((SelectableTable)userSolutionTable, hashValue, doubleHashValue, false);
        }
        
        for (int i=0; i < size; i++){
        	String valueStr = ((HashCell)userSolutionTable.getObject(i)).getField(HashCell.VALUE).toString();
        	if (valueStr.equals(EMPTY)){
        		firstHashTableContent[i] = "";
        	}
        	else {
        		firstHashTableContent[i] = valueStr;
        	}
        }
        
        // create the actual input data
        inputDataTable = new Table(addDataSize + searchDataSize + removeDataSize);
        int inputIndex = 0;
        ((SelectableTableRehash)userSolutionTable).setInputSelection(true);
        int inputData[] = new int[addDataSize];
        
        for(int i = 0; i < addDataSize; i++) {
            inputData[i] = Math.abs(R.nextInt()%(max-min))+min;
        }
        
        inputInd = 0;
        int searchCount = 0;
        int removeCount = 0;
        ArrayList hashTableContent = new ArrayList();
        for (int i=0; i < rehashInput.length; i++){
        	hashTableContent.add(new Integer(rehashInput[i]));
        }
        	        
        for (int i=inputIndex; i < inputDataTable.size(); i++){
        	int f = R.nextInt(inputDataTable.size());
        	
        	// add
        	if ((searchCount >= searchDataSize && removeCount >= removeDataSize) ||
        			(hashTableContent.size() == 0) || (f < addDataSize && inputInd < inputData.length)){
        			        		
            	hashTableContent.add(new Integer(inputData[inputInd]));
            	inputDataTable.setObject(new HashTask(inputData[inputInd++], taskAddText, probeType, 2*size), i);
        	}
        	// search
        	else if ((removeCount >= removeDataSize) ||
        				(f < addDataSize+searchDataSize && searchCount < searchDataSize) ||
        				(searchCount < searchDataSize && hashTableContent.size() <= 1)){
        		
        		int ind = R.nextInt(hashTableContent.size());
        		inputDataTable.setObject(new HashTask(((Integer)hashTableContent.get(ind)).intValue(), taskSearchText, probeType, 2*size), i);
        		searchCount++;
        	}
        	// remove
        	else {
        		int ind = R.nextInt(hashTableContent.size());
        		inputDataTable.setObject(new HashTask(((Integer)hashTableContent.remove(ind)).intValue(), taskRemoveText, probeType, 2*size), i);
        		removeCount++;
        	}
        }

        currentInput.setObject(new Key(notSelectedText), 0);
        userSolutionTable2 = new SelectableTableRehash(2*size, inputDataTable, currentInput, userSolutionTable);    
        
        return new FDT[] { currentInput, userSolutionTable, userSolutionTable2 };
    }


    public FDT[] init(Object o) {
        return init();
    }

    public String[] getStructureNames() {
        ExerciseProperties lang = ExerciseProperties.getInstance();
        return new String[] { 
                lang.get( "HASH_OPERATIONS_REHASH_INPUT_TITLE" ),
                lang.get( "HASH_OPERATIONS_REHASH_HASHTABLE_TITLE" ),
                lang.get( "HASH_OPERATIONS_REHASH_HASHTABLE_TITLE" )
        };
    }
    
    public String getDescription() {
        ExerciseProperties lang = ExerciseProperties.getInstance();
        String prompt1 =  lang.get( "HASH_OPERATIONS_REHASH_PROMPT_PART_1" ); // "The hash function is h(k) = k "
        String prompt2 =  lang.get( "HASH_OPERATIONS_REHASH_PROMPT_PART_2" ); // " mod "

        String s = prompt1 + " " + prompt2 + " " + 2*size + "\n";
        
        if (probeType == PROBE_TYPE_DOUBLEHASH){
            String prompt3 =  lang.get( "HASH_OPERATIONS_REHASH_PROMPT_PART_3" ); // "The second hash function is h(k) = k "
            String prompt4 =  lang.get( "HASH_DOUBLEHASH_PROMPT_PART_4" ); // " - (k mod "
            s += prompt3 + " " + prompt4 + " " + secondHashParameter + " )\n";
        }

        s += lang.get( "HASH_OPERATIONS_REHASH_PROBING_TYPE_TEXT" ); // "probing type:"
        
        String typeString = "";
        if (probeType == PROBE_TYPE_LINEAR){
        	typeString = lang.get( "HASH_OPERATIONS_REHASH_LINEAR_PROBING_TEXT" );
        }
        else if (probeType == PROBE_TYPE_QUADRATIC){
        	typeString = lang.get( "HASH_OPERATIONS_REHASH_QUADRATIC_PROBING_TEXT" );
        }
        else if (probeType == PROBE_TYPE_DOUBLEHASH){
        	typeString = lang.get( "HASH_OPERATIONS_REHASH_DOUBLE_PROBING_TEXT" );
        }
        s += " " + typeString;
        
        return s;
    }

    public FDT[] solve() {
        FDT[] initialStructures = getInitialStructures();
        Table input =  (Table)initialStructures[0];
        SelectableTableRehash modelSolution = (SelectableTableRehash)initialStructures[1];
        SelectableTableRehash modelSolution2 = (SelectableTableRehash) initialStructures[2];
        
        for (int i = 0; i < size; i++){
        	HashCell cell = (HashCell)modelSolution.getObject(i);
        	if (!cell.getField(HashCell.VALUE).equals(EMPTY)){
        		modelSolution.setInputSelection(true);
        		modelSolution.selected(cell, true);
        		modelSolution.setInputSelection(false);
        		int keyValue = Integer.parseInt(cell.getField(HashCell.VALUE).toString());
            	int hashValue = keyValue % (2*size);
                int doubleHashValue = secondHashParameter - (keyValue % secondHashParameter);
                modelSolution2.clearSelections();
        		probe(modelSolution2, hashValue, doubleHashValue, true);
        	}
        }
                
        for(int i = 0; i < inputDataTable.size(); i++) {
        	HashTask task = ((HashTask)inputDataTable.getObject(i));
        	int hashValue = (((Integer)task.getField(HashTask.VALUE)).intValue()) % (2*size);
            int doubleHashValue = secondHashParameter - (((Integer)task.getField(HashTask.VALUE)).intValue() % secondHashParameter);

            modelSolution2.clearSelections();
          	probe(modelSolution2, hashValue, doubleHashValue, true);
        }
        modelSolution.undisableAll();
		return new FDT[] { input, modelSolution, modelSolution2 };
    }
    

    /**
     * Does the probing to a table
     * @param modelSolution - the table which is probed
     * @param hashValue
     * @param doubleHashValue
     * @param animatorOn - true if steps are saved in animator
     */
    private void probe(SelectableTable modelSolution, int hashValue, int doubleHashValue, boolean animatorOn){
    	for (int i=0; i < modelSolution.size(); i++){
            if (probeType == PROBE_TYPE_LINEAR && 
            		modelSolution.selected((HashCell)modelSolution.getObject((hashValue+i)%modelSolution.size()), animatorOn)){
            	break;
            }   		
            else if (probeType == PROBE_TYPE_QUADRATIC && 
            		modelSolution.selected((HashCell)modelSolution.getObject((hashValue+i*i)%modelSolution.size()), animatorOn)){
            	break;
            }  		
            else if (probeType == PROBE_TYPE_DOUBLEHASH && 
            		modelSolution.selected((HashCell)modelSolution.getObject((hashValue+i*doubleHashValue)%modelSolution.size()), animatorOn)){
            	break;
            }   		
    	}    	
    }
    
    public FDT[] makeModelAnswer() {
        return solve();
    }

    public FDT[] getInitialStructures() {
    	Table current = new Table();
    	current.setObject(new Key(notSelectedText), 0);
        SelectableTableRehash resultTable = new SelectableTableRehash(size, new Table(), current);
        for (int i=0; i < firstHashTableContent.length; i++){
        	resultTable.setObject(new HashCell(firstHashTableContent[i], resultTable, i), i);
        }
        resultTable.setInputSelection(false);
        resultTable.setHashTask(new HashTask(-1, taskRehashText, probeType, 2*size));
    	SelectableTableRehash resultTable2 = new SelectableTableRehash(2*size, inputDataTable, current, resultTable);
    	return new FDT[] { current, resultTable, resultTable2 };
    }

    public FDT[] getSimulatedStructures() {
        return new FDT[] { currentInput, userSolutionTable, userSolutionTable2 };
    }
    
    public String[] getModelAnswerNames() {
        ExerciseProperties lang = ExerciseProperties.getInstance();
        return new String[] { lang.get( "HASH_OPERATIONS_REHASH_INPUT_MODEL_TITLE"),
        					lang.get( "HASH_OPERATIONS_REHASH_HASHTABLE_MODEL_TITLE" ), 
        					lang.get( "HASH_OPERATIONS_REHASH_HASHTABLE_MODEL_TITLE" )};
    }

    public FDT[] getAnswer() {
        return new FDT[] { currentInput, userSolutionTable, userSolutionTable2 };
    }

    public ComparableExercise getModelAnswer() {
        return this;
    }

    public VisualTypeConf[] conf() {
        VisualTypeConf tabl = new VisualTypeConf();
        tabl.setVisualConfValue("matrix.visual.VisualArray",VisualTypeConf.INDEXED, "false");
        VisualTypeConf hashtabl1 = new VisualTypeConf();
        hashtabl1.enable("matrix.visual.VisualStruct",VisualTypeConf.HIGHLIGHT_OPERATION);
        VisualTypeConf hashtabl2 = new VisualTypeConf();
        hashtabl2.enable("matrix.visual.VisualStruct",VisualTypeConf.HIGHLIGHT_OPERATION);
        return new VisualTypeConf[] { tabl, hashtabl1 , hashtabl2 };
    }

    public int[] getCompareIndices() {
        return new int[] {0, 1, 2};
    }

	public int[] getGradeIndices(){
        return null; // Used only in recovery mode
	}

	public boolean canRecover(){
		return false;
	}
	
	public String[] getCompareStructureVisualisations(){
		return null; // use default
	}
	
	public String[] getCompareStructureNames(){
		return null; // use default
	}     
	
	public VisualTypeConf[] getCompareVisualTypeConf(){
		return null; // use default
	}

    static final long serialVersionUID = -37657283333350211L;
    
    
    /**
     * The hash table of the exercise which handles also rehash operations.
     * @author lhakulin
     *
     */
    private static class SelectableTableRehash extends SelectableTable {
    	       	
        private String taskAddText; 
        private String taskSearchText; 
        private String taskRemoveText; 
        private String taskRehashText;
        private String taskFinishedText; 
        private String graveStoneText; 
        private String notSelectedText;
        private boolean inputSelection;
        private HashTask task=null;
        private SelectableTableRehash initialTable = null;
        private HashCell currentCell;
    	
    	public SelectableTableRehash(int size, Table inputData, Table ci){
    		super(size, inputData, ci);
    		setTaskTexts();
    		inputSelection = false;
    	}
    	
    	public SelectableTableRehash(int size, Table inputData, Table ci, Table initTable){
    		this(size, inputData, ci);
    		setInitialTable(initTable);
    		current.dec();
    	}
    	
    	public void undisableAll(){
    		for (int i=getFirst(); i <= getLast(); i++){
    			((HashCell)getObject(i)).setDisabled(false);
    		}
    	}

    	
        private void setTaskTexts(){
            ExerciseProperties lang = ExerciseProperties.getInstance();
            taskAddText = lang.get("HASH_OPERATIONS_REHASH_TASK_ADD_TEXT");
            taskSearchText = lang.get("HASH_OPERATIONS_REHASH_TASK_SEARCH_TEXT");
            taskRemoveText = lang.get("HASH_OPERATIONS_REHASH_TASK_REMOVE_TEXT");
            taskRehashText = lang.get("HASH_OPERATIONS_REHASH_TASK_REHASH_TEXT");
            taskFinishedText = lang.get("HASH_OPERATIONS_REHASH_TASK_FINISHED_TEXT");
            graveStoneText = lang.get("HASH_OPERATIONS_REHASH_GRAVESTONE_TEXT");
            notSelectedText = lang.get("HASH_OPERATIONS_REHASH_NOT_SELECTED_TEXT");
            
            super.setTaskAddText(taskAddText);
            super.setTaskSearchText(taskSearchText);
            super.setTaskRemoveText(taskRemoveText);
            super.setTaskFinishedText(taskFinishedText);
            super.setGraveStoneText(graveStoneText);
        }
    	
        /**
    	 * Called when one of the cells is selected
    	 * Performs an operation (add, search or remove) according to the currentInput
    	 * @param cell - the cell which was selected
    	 * @return true if currentInput is updated
    	 */
    	public boolean selected(HashCell cell, boolean animatorOn){
    		
    		boolean result = false;

    		if (animatorOn){
				Animator.getActiveAnimator().startOperation();
			}
    		
    		currentCell = cell;
    		
    		if (!inputSelection){
    			if (currentInput.getObject(0).toString().equals(notSelectedText)){
    				result = false;
    			}
    			else if (super.selected(cell, animatorOn) == false){
	        		HashTask task; 
	        		if (currentInput.getObject(0) instanceof HashTask){
	        			task = (HashTask)currentInput.getObject(0);
	        		}
	        		else {
	        			task = (HashTask)input.getObject(current.eval());
	        		}
	        		if (super.currentFinished() == false &&
	        				task.getField(HashTask.OPERATION).equals(taskRehashText) && 
	        				(cell.getField(HashCell.VALUE).toString().equals(EMPTY) ||
	        					cell.getField(HashCell.VALUE).toString().equals(graveStoneText))){
	        			
	    	    		cell.setField(task.getField(HashTask.VALUE), HashCell.VALUE);
	    	    		this.setObject(cell, cell.getIndex()); // needed to get the latest action highlighted when animating
	    	    		clearSelections();
	    	    		if (initialTable != null){
	    	    			initialTable.disableCurrentCell();
	    	    		}
	    	    		if (initialTable != null && initialTable.allDisabledOrEmpty()){
		    	    		current.inc();
		    	    		currentInput.setObject(input.getObject(current.eval()), 0);
	    	    		}
	    	    		result = true;        			
	        		}
	    		}
				else {
					result = true;
				}
    		}
    		else if (!cell.getField(HashCell.VALUE).equals(EMPTY) && !cell.isDisabled()){
    			task.setField(new Integer(Integer.parseInt(cell.getField(HashCell.VALUE).toString())), HashTask.VALUE);
    			currentInput.setObject(task, 0);
    			result = true;
    		}
			if (animatorOn){
				Animator.getActiveAnimator().endOperation();
			}		

			return result;
    	}

    	public void setInputSelection(boolean is){
    		inputSelection = is;
    	}
    	
    	public void setHashTask(HashTask t){
    		task = t;
    	}
    	
    	private void setInitialTable(Table t){
    		if (t instanceof SelectableTableRehash){
    			initialTable = (SelectableTableRehash)t;
    		}
    	}
    	
    	public void disableCurrentCell(){
    		currentCell.setDisabled(true);
    	}

        static final long serialVersionUID = 8668291347493297645L;   
    }
    

}
