// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
package content.exercises.sda;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.GridBagConstraints;
import java.awt.Insets;
import java.awt.Rectangle;
import java.awt.geom.Point2D;
import java.util.Collection;
import java.util.Iterator;
import java.util.Random;
import java.util.TreeSet;

import matrix.animation.Animator;
import matrix.simulation.VisualTypeConf;
import matrix.structures.FDT.FDT;
import matrix.structures.spatial.PaintingStyleDecorator;
import matrix.structures.spatial.FDT.probe.Point;
import matrix.util.Note;
import content.ExerciseProperties;
import content.interfaces.*;
import content.exercises.sda.structures.ExerciseQuadTree;
import content.exercises.sda.structures.PointGenerator;
import content.exercises.sda.structures.SimpleStack;

/**
 * Point-Region Quadtree exercise. In this PRQuadtree implementation, the order
 * of the child nodes is: south-east, north-east, south-west, north-west, which
 * is compatible with the subarea traversal done in morton coding.
 * 
 * @author jtn, Juha Helminen
 */
public class PRQuadTreeInsert implements SimulationExerciseModel, ConfigureVisualType, 
                                         ModelAnswerNames, LayoutExercise, StyledExercise,
                                         SwapBehaviour, ComparableExercise {

    // For debugging
    private static final boolean STATIC_SEED = false;
    private static final long STATIC_SEED_VALUE = 1168851896904L;
    //
    private long seed = System.currentTimeMillis();

    // ** Exercise configuration **
    private static final int TOOLTIP_WAIT_TIME = 1500;
    // Input data coordinate ranges 0-x, 0-y
    private static final int X = 250;
    private static final int Y = X;
    // Minimum distance (along the x and y axis) between the points and 
    // the border of the area visualization.
    private static final int MIN_AREA_BORDER_DISTANCE = 13;
    // Minimum distance (along the x and y axis) between the points and 
    // the split borders.
    private static final int MIN_SPLIT_BORDER_DISTANCE = 3;
    // The minimum distance between the points.
    private static final int MIN_DISTANCE = 25;
    // The number of points to insert.
    private static final int INPUT_SIZE = 8;

    // ** Exercise color and style scheme **
    private static final Color POINT_COLOR = Color.BLUE;
    private static final Color NEXT_POINT_COLOR = Color.MAGENTA;
    private static final Color BORDER_COLOR = Color.DARK_GRAY;
    private static final PaintingStyleDecorator pointDecorator;
    private static final PaintingStyleDecorator nextPointDecorator;
    private static final PaintingStyleDecorator areaDecorator; 
    static {
        pointDecorator = new PaintingStyleDecorator();
        pointDecorator.setLineColor(POINT_COLOR);

        nextPointDecorator = new PaintingStyleDecorator();
        nextPointDecorator.setLineColor(NEXT_POINT_COLOR);

        float[] areaDashPattern = { 4, 4 };
        areaDecorator = new PaintingStyleDecorator();
        areaDecorator.setLineColor(BORDER_COLOR);
        areaDecorator.setFillColor(BORDER_COLOR);
        areaDecorator.setStroke(1, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL, 4, areaDashPattern, 0);
        areaDecorator.setTransparency(0.07);
    }

    // The generated input, the points to be inserted.
    private java.awt.Point[] input;
    // Structures for the student.
    private ExerciseQuadTree studentTree;
    private SimpleStack studentInputStream;

    // Solves the exercise.
    private FDT[] solveHelp() {
        ExerciseQuadTree modelTree = new ExerciseQuadTree(null, 0, X, 0, Y, false);
        modelTree.setPaintingStyleDecorator(areaDecorator);
        SimpleStack modelInputStream = (SimpleStack)this.getInitialStructures()[0];
        Animator a = Animator.getActiveAnimator();
        while (!modelInputStream.ADTIsEmpty()) {
            a.startOperation();
            modelTree.insert(modelInputStream.ADTPop());
            a.endOperation();
        }
        return new FDT[] { modelInputStream, modelTree };
    }

    public FDT[] solve() {
        return new FDT[] { solveHelp()[1] };
    }

    public FDT[] makeModelAnswer() {
        FDT[] s = solveHelp();
        return new FDT[] { s[0], s[0], s[1], s[1] };
    }

    public FDT[] init() {
        // For debugging
        Note.out(this, "The random number generator seed for this exercise is: " + this.getSeed());
        Random r = new Random(this.getSeed());

        TreeSet lastBorders = new TreeSet();
        TreeSet borders = new TreeSet();
        borders.add(new Integer(0));
        int size = X / 2; 
        Iterator iterator;
        Integer border;
        while (size > MIN_DISTANCE) {
            lastBorders = borders;
            borders = new TreeSet();
            iterator = lastBorders.iterator();
            while (iterator.hasNext()) {
                border = (Integer)iterator.next();
                borders.add(border);
                borders.add(new Integer(border.intValue() + size));
            }
            size /= 2;
        }
        borders.add(new Integer(X));
        
        this.input = new java.awt.Point[INPUT_SIZE];
        for (int i = 0; i < INPUT_SIZE; i++) { 
            final int MAX_TRIES = 80;
            int tryCounter = 0;
            while (true) {
                // Generate the input.
                try {
                    this.input[i] = PointGenerator.generatePoint(MIN_AREA_BORDER_DISTANCE, 
                                                                 MIN_AREA_BORDER_DISTANCE,
                                                                 X - MIN_AREA_BORDER_DISTANCE,
                                                                 Y - MIN_AREA_BORDER_DISTANCE,
                                                                 r);
                    checkDistance(this.input[i], this.input, i, MIN_DISTANCE);
                    checkXDistance(this.input[i], borders, MIN_SPLIT_BORDER_DISTANCE);
                    checkYDistance(this.input[i], borders, MIN_SPLIT_BORDER_DISTANCE);
                    
                    // success
                    break;
                } catch (Exception e) {
                    if (tryCounter++ >= MAX_TRIES) {
                        Note.err(this, e.getMessage());
                        Note.show(this, "Could not generate input for the exercise. Please restart the exercise.");
                        return null;
                    }
                }
            }
        }

        this.studentTree = new ExerciseQuadTree(null, 0, X, 0, Y, true);
        this.studentTree.setPaintingStyleDecorator(areaDecorator);
        this.studentInputStream = (SimpleStack)this.getInitialStructures()[0]; 

        return new FDT[] { this.studentInputStream, this.studentInputStream,
                           this.studentTree, this.studentTree };
    }

    private static void checkDistance(Point2D point, Point2D[] points, int size, int distance)
        throws Exception {
        for (int i = 0; i < size; i++) {
            if (point.distance(points[i]) < distance) {
                throw new Exception();
            }
        }
    }
    
    private static void checkXDistance(Point2D point, Collection borders, int distance) throws Exception {
        Iterator iterator = borders.iterator();
        Integer border;
        while (iterator.hasNext()) {
            border = (Integer)iterator.next();
            if (point.distance(border.intValue(), point.getY()) < distance) {
                throw new Exception("Border check failed.");
            }
        }
    }
    
    private static void checkYDistance(Point2D point, Collection borders, int distance) throws Exception {
        Iterator iterator = borders.iterator();
        Integer border;
        while (iterator.hasNext()) {
            border = (Integer)iterator.next();
            if (point.distance(point.getX(), border.intValue()) < distance) {
                throw new Exception("Border check failed.");
            }
        }
    }

    public long getSeed() {
        // Seed might be set to a constant value for debugging purposes.
        if (STATIC_SEED) {
            return PRQuadTreeInsert.STATIC_SEED_VALUE;
        } else {
            return this.seed;
        }
    }

    public void setSeed(long seed) {
        this.seed = seed;
    }

    public String[] getStructureNames() {
    	ExerciseProperties lang = ExerciseProperties.getInstance();
    	String[] s = { lang.get("PRQUADTREE_INPUT_STREAM_TITLE"), lang.get("PRQUADTREE_INPUT_STREAM_AREA_TITLE"), 
    			lang.get("PRQUADTREE_AREA_TITLE"), lang.get("PRQUADTREE_TREE_TITLE")
    	};
    	return s;
    	//return new String[] { "Input stream", "Input stream points",
        //                      "PR-quadtree", "PR-quadtree" };
    }

    public FDT[] getAnswer() {
        return new FDT[] { this.studentTree };
    }

    public FDT[] getInitialStructures() {
        SimpleStack inputStream = new SimpleStack(new Rectangle(X, Y, 0, 0), 
                                                  nextPointDecorator);
        Point p;
        for (int i = 0; i < input.length; i++) {
            // Start indexing from 1.
            p = new PointWithInfo(input[i].x, 
                                  input[i].y, 
                                  String.valueOf(input.length - i));
            p.setPaintingStyleDecorator(pointDecorator);
            inputStream.ADTPush(p);
        }
        return new FDT[] { inputStream };
    }

    public String getDescription() {
        return "Insert the points from the input stream into the PR-quadtree.";
    }

    public VisualTypeConf[] conf() {
        VisualTypeConf stream = new VisualTypeConf();
        VisualTypeConf locations = new VisualTypeConf();
        VisualTypeConf area = new VisualTypeConf();
        VisualTypeConf tree = new VisualTypeConf();
        // Input stream
        stream.setVisualConfValue("matrix.visual.VisualArray", VisualTypeConf.INDEXED, "false");
        stream.enable("matrix.visual.VisualArrayComponent", VisualTypeConf.DRAG_OPERATION);
        stream.enable("matrix.visual.VisualArrayComponent", VisualTypeConf.HIGHLIGHT_OPERATION);
        stream.enable("matrix.visual.VisualStruct", VisualTypeConf.DRAG_OPERATION);
        stream.enable("matrix.visual.VisualStruct", VisualTypeConf.HIGHLIGHT_OPERATION);
        // Quadtree tree visualization
        tree.enable("matrix.visual.VisualLayeredTreeComponent", VisualTypeConf.DROP_OPERATION);
        tree.enable("matrix.visual.VisualLayeredTreeComponent", VisualTypeConf.HIGHLIGHT_OPERATION);
        tree.enable("matrix.visual.VisualStruct", VisualTypeConf.DRAG_OPERATION);
        tree.enable("matrix.visual.VisualStruct", VisualTypeConf.HIGHLIGHT_OPERATION);
        tree.enable("matrix.visual.VisualStruct-matrix.structures.spatial.FDT.probe.Point",
                    VisualTypeConf.MINIMIZED);
        return new VisualTypeConf[] { stream, locations, area, tree };
    }

    public String[] getModelAnswerNames() {
        return getStructureNames();
    }

    public GridBagConstraints[] getConstraints() {
        GridBagConstraints input = new GridBagConstraints();
        input.insets = new Insets(0, 2, 2, 4);
        input.gridy = 1;
        input.gridx = 0;
        GridBagConstraints inputVis = new GridBagConstraints();
        inputVis.insets = new Insets(2, 2, 4, 4);
        inputVis.gridy = 0;
        inputVis.gridx = 0;
        inputVis.gridwidth = 4;
        GridBagConstraints treeVis = new GridBagConstraints();
        treeVis.insets = new Insets(2, 0, 4, 2);
        treeVis.gridy = 0;
        treeVis.gridx = 5;
        GridBagConstraints tree = new GridBagConstraints();
        tree.insets = new Insets(0, 0, 2, 2);
        tree.gridy = 1;
        tree.gridx = 1;
        tree.gridwidth = 6;
        return new GridBagConstraints[] { input, inputVis, treeVis, tree };
    }

    public GridBagConstraints[] getModelAnswerConstraints() {
        return getConstraints();
    }
    
	public GridBagConstraints[] getComparisonConstraints() {
		return null; // use default
	}

    public String[] getStructureVisualisations() {
        return new String[] { "array", "2D", "2D", "layered tree" };
    }

    public String[] getModelAnswerVisualisations() {
        return getStructureVisualisations();
    }

    public boolean getSwapBehaviour() {
        return true;
    }
    
    /* ComparableExercise interface */
    
	public int[] getCompareIndices(){
        return new int[] {0};	
	}

	public int[] getGradeIndices(){
        return null; // Used only in recovery mode
	}

	public boolean canRecover(){
		return false;
	}
	
	public String[] getCompareStructureVisualisations(){
        return new String[] { "layered tree" };
	}
	
	public String[] getCompareStructureNames(){
    	ExerciseProperties lang = ExerciseProperties.getInstance();
    	String[] s = { lang.get("PRQUADTREE_TREE_TITLE")	};
    	return s;
	}
	
	public VisualTypeConf[] getCompareVisualTypeConf(){
		VisualTypeConf[] modelConf = conf();
		return new VisualTypeConf[] { modelConf[3] };
	}

    private static class PointWithInfo extends Point /*implements InformationDecorator*/ {

        private int waitTime;
        
        public PointWithInfo(double x, double y, String key) {
            super(x, y, key);
            this.waitTime = PRQuadTreeInsert.TOOLTIP_WAIT_TIME; 
        }

        public String getInfo() {
            String[] fieldNames = this.getFieldNames();
            Object[] fields = this.getFields();
            StringBuffer str = new StringBuffer();
            str.append("<HTML>");
            // Ignore the id field.
            for (int i = 1; i < fieldNames.length; i++) {
                str.append(fieldNames[i]).append(" ").append(fields[i]).append("<br>");
            }
            str.append("</HTML>");
            return str.toString();
        }

        public int getWaitTime() {
            return this.waitTime;
        }

        public boolean isInfoBoxEnabled() {
            return false;
        }

        public boolean isToolTipInfoEnabled() {
            return true;
        }

        public void setInfo(String s) {
            // disabled
        }

        public void setWaitTime(int t) {
            this.waitTime = t;
        }

        private static final long serialVersionUID = -6535916305400055736L;
    }

    private static final long serialVersionUID = 6483493189023543865L;
}
