// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
package content.exercises.sda.structures;

import java.io.Serializable;
import java.util.HashMap;

import matrix.structures.FDT.advanced.GeometricGraph;
import matrix.structures.FDT.substructures.Vertex;
import matrix.structures.memory.VirtualArray;
import matrix.util.Note;

/**
 * An implementation of the GeometricGraph interface for use in Exercises. This is the generic
 * ExerciseGeometricGraph, if a specific exercise requires some additional or modified functionality,
 * it should extend this class. This class is designed to work together with the DoublyConnectedEdgeList
 * class in order to represent polygon networks. if the two classes are used together, ExerciseGeometricGraph 
 * should be used for visualization, and DoublyConnectedEdgeList for operations on polygon networks.
 * 
 * The vertices of an ExerciseGeometricGraph should be instances of ExerciseGeometricGraphVertex. 
 * 
 * @author jtn
 *
 */
public class ExerciseGeometricGraph implements GeometricGraph, Serializable {

	/**
	 * 
	 */
	private static final long serialVersionUID = 3795220154865241782L;
	private VirtualArray vertices;
	int nro;
	private HashMap vert;
	
	public ExerciseGeometricGraph() {
		vertices = new VirtualArray(this, "Vertices of the graph");
		vert = new HashMap();
		nro = 0;
	}
	
	/**
	 * Returns the X coordinate of the parameter vertex, or -1 if the vertex
	 * has no coordinates set.
	 */
	public int getX(Vertex vertex) {
		if(vertex instanceof ExerciseGeometricGraphVertex) {
			return ((ExerciseGeometricGraphVertex)vertex).getX();
		}
		return -1;
	}

	/**
	 * Returns the Y coordinate of the parameter vertex, or -1 if the vertex
	 * has no coordinates set.
	 */
	public int getY(Vertex vertex) {
		if(vertex instanceof ExerciseGeometricGraphVertex) {
			return ((ExerciseGeometricGraphVertex)vertex).getY();
		}
		return -1;
	}

	/**
	 * Sets the coordinates of the parameter vertex.
	 */
	public void setCoordinates(Vertex vertex, int x, int y) {
		if(vertex instanceof ExerciseGeometricGraphVertex) {
			((ExerciseGeometricGraphVertex)vertex).setCoordinates(x,y);
		}
	}

	/**
	 * Returns the vertices of the graph represented by this object.
	 */
	public Vertex[] getVertices() {
		Vertex[] v = new Vertex[vertices.size()];
		for(int i = 0; i < vertices.size();i++) {
			v[i] = (Vertex)vertices.getObject(i);
		}
		return v;
	}

	/**
	 * Method required by Graph interface. Disabled.
	 */
	public void setVertices(Vertex[] v) {		
		Note.out(this, "setvertices not implemented");
	}

	/**
	 * Method required by Graph interface. Disabled.
	 */
	public void addVertex(Object o) {
		Note.out(this, "addVertex not implemented");
	}

	/**
	 * Adds a new vertex with the given key object to this graph.
	 * @param x
	 * @param y
	 * @param key
	 */
	public Vertex addVertex(int x, int y, Object key) {
		Vertex v = getNewVertex(x,y,key);
		vertices.setObject(v, nro++);
		vert.put(makeVertexKey(x,y), v);
		return v;
	}

	private String makeVertexKey(int i, int j) {
		return "(" + i + "," + j + ")";
	}

	
	/**
	 * Method required by Graph interface. Disabled.
	 */
	public void removeVertex(Vertex v) {
		Note.out(this, "removeVertex not implemented");		
	}

	/**
	 * Method required by Graph interface. Returns dummy value.
	 */
	public Object getElement() {
		return "";
	}

	/**
	 * Method required by FDT interface. Disabled.
	 */
	public void setElement(Object o) {
		Note.out(this, "setElement not implemented");
	}
	
	/**
	 *  Visits a edge defined by the coordinates (x,y) of the origin and goal vertices.
	 *  The vertices are defined by the coordinates in order to make it possible to 
	 *  find them even if the vertex list of the graph is not known to the caller.
	 *  
	 *  Called, for example, from DoublyConnectedEdgeList.
	 *  
	 * @param fromx X-coordinate of the origin
	 * @param fromy Y-coordinate of the origin
	 * @param tox X-coordinate of the goal
	 * @param toy Y-coordinate of the goal
	 */
	public void visit(int fromx, int fromy, int tox, int toy) {
		ExerciseGeometricGraphVertex from = (ExerciseGeometricGraphVertex)vert.get(makeVertexKey(fromx, fromy));
		ExerciseGeometricGraphVertex to =(ExerciseGeometricGraphVertex)vert.get(makeVertexKey(tox, toy));
		if(from == null || to == null) {
			throw new RuntimeException ("Vertex not found!");
		}
		from.selectSuccessor(to);
	}
	
	/**
	 * Creates and returns a new vertex for this graph. This method does not add the vertex to the graph,
	 * and that must be done separately. If a subclass of ExerciseGeometricGraph uses some other vertices
	 * than ExerciseGeometricGraphVertex, this method must be overridden to return instances of the new
	 * vertex class
	 * @param x X-coordinate of the vertex
	 * @param y Y-coordinate of the vertex
	 * @param key Key of the vertex
	 * @return
	 */
	public Vertex getNewVertex(int x, int y, Object key) {
		return new ExerciseGeometricGraphVertex(x,y,key);
	}
	
	public boolean equals(Object o) {
		if(o instanceof ExerciseGeometricGraph) {
			ExerciseGeometricGraph e = (ExerciseGeometricGraph)o;
			Vertex[] vo = getVertices();
			Vertex[] ve = e.getVertices();
			for(int i = 0; i < vo.length; i++) {
				Vertex voo = vo[i];
				Vertex veo = ve[i];
				if(voo == veo) {
					continue;
				}
				if(voo == null && veo != null) {
					return false;
				}
				if(!voo.equals(veo)) {
					return false;
				}
			}
			return true;
		}
		return false;
	}
}
