// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
package content.exercises.sda.structures;

import matrix.structures.FDT.Tree;
import matrix.structures.spatial.Area;
import matrix.structures.spatial.CDT.probe.PRQuadTree;
import matrix.structures.spatial.CDT.probe.PRQuadTreeNode;
import matrix.structures.spatial.FDT.probe.Point;
import matrix.structures.spatial.FDT.probe.SpatialElement;
import matrix.util.Note;

/**
 * PR-Quadtree implementation with extra support for simulation.
 *
 * @author Juha Helminen
 */
public class ExerciseQuadTree extends PRQuadTreeNode {
    
    private boolean showEmptyNodes;
    private int level;

    private ExerciseQuadTree(Point key, boolean showEmptyNodes) {
        super(key);
        this.showEmptyNodes = showEmptyNodes;
        this.level = 0;
    }

    public ExerciseQuadTree(Point key, double minX, double maxX, double minY, double maxY, boolean showEmptyNodes) {
        this(key, minX, maxX, minY, maxY, showEmptyNodes, 0);
	}
	
    private ExerciseQuadTree(Point key, double minX, double maxX, double minY, double maxY, boolean showEmptyNodes, int level) {
        super(key, minX, maxX, minY, maxY);
        this.showEmptyNodes = showEmptyNodes;
        this.level = level;
    }
    
	public void setSubTree(Tree t, int child) {
		switch (child) {
			case 0 : 
				this.getChild(QUADRANT_SW).setElement(t.getElement());
				break;
			case 1 :
				this.getChild(QUADRANT_NW).setElement(t.getElement());
				break;
			case 2 : 
				this.getChild(QUADRANT_SE).setElement(t.getElement());
				break;
			case 3 : 
				this.getChild(QUADRANT_NE).setElement(t.getElement());
				break;
		}
	}
		
    public Tree getNewNode(Object key) {
        if (key instanceof Point) {
            ExerciseQuadTree node = new ExerciseQuadTree((Point)key, this.showEmptyNodes);
            // Inherit the decorator.
            if (this.hasPaintingStyleDecorator()) {
                node.setPaintingStyleDecorator(this.getPaintingStyleDecorator());
            }
            return node;
        } else {
            Note.err(this, "Cannot create a new ExerciseQuadTree with key of type " + key.getClass() + ".");
            return null;
        }
    }

    public PRQuadTreeNode getNewNode(Point key, double minX, double maxX, double minY, double maxY) {
        ExerciseQuadTree node = new ExerciseQuadTree(key, 
                                                     minX, 
                                                     maxX, 
                                                     minY, 
                                                     maxY, 
                                                     this.showEmptyNodes, 
                                                     this.level + 1);
        // Inherit the decorator.
        if (this.hasPaintingStyleDecorator()) {
            node.setPaintingStyleDecorator(this.getPaintingStyleDecorator());
        }
        return node;
    }

    private void split() {
        PRQuadTreeNode node;
        node = this.getNewNode(null,
                               (this.minX + ((this.maxX - this.minX) / 2)),
                               this.maxX,
                               (this.minY + ((this.maxY - this.minY) / 2)),
                               this.maxY);
        this.setChild(node, QUADRANT_NE);
        node = this.getNewNode(null,
                               (this.minX + ((this.maxX - this.minX) / 2)),
                               this.maxX,
                               this.minY,
                               (this.minY + ((this.maxY - this.minY) / 2)));
        this.setChild(node, QUADRANT_SE);
        node = this.getNewNode(null,
                               this.minX,
                               (this.minX + ((this.maxX - this.minX) / 2)),
                               this.minY,
                               (this.minY + ((this.maxY - this.minY) / 2)));
        this.setChild(node, QUADRANT_SW);
        node = this.getNewNode(null,
                               this.minX,
                               (this.minX + ((this.maxX - this.minX) / 2)),
                               (this.minY + ((this.maxY - this.minY) / 2)),
                               this.maxY);
        this.setChild(node, QUADRANT_NW);
    }

	public void setElement(Object o) {
		// If inserting into an empty leaf node
		// split the node to allow more insertions
		if (this.isEmpty() && super.isLeaf()) {
			this.split();
		}
		super.setElement(o);
	}

	public boolean isEmpty() {
		return (this.getKey() == null);
	}

	public int getSubTreeCount() {
        if (this.showEmptyNodes) {
    		// All nodes besides empty leaves are shown to have four children.
    		if (this.isLeaf() && this.isEmpty()) {
    			return 0;
    		} else {
    			return 4;
    		}
        } else {
            return super.getSubTreeCount();
        }
	}

	public SpatialElement[] getSpatialElements() {
		if (this.isEmpty()) {
			return new SpatialElement[0];
		} else {
            SpatialElement[] key = new SpatialElement[1];
			key[0] = (SpatialElement)this.getKey();
			return key;
		}
	}

	public Area[] getSubAreas() {
		if (this.isLeaf()) {
			return new Area[0];
		} else {
			return super.getSubAreas();
		}
	}

	public boolean isLeaf() {
		// Recursively determine whether this node is a leaf, i.e. has no children 
		// (or they and their subtrees are empty)
		return (((this.getNE() == null) || (((ExerciseQuadTree)this.getNE()).isEmpty()) && ((ExerciseQuadTree)this.getNE()).isLeaf()) &&
				((this.getSE() == null) || (((ExerciseQuadTree)this.getSE()).isEmpty()) && ((ExerciseQuadTree)this.getSE()).isLeaf()) &&
				((this.getSW() == null) || (((ExerciseQuadTree)this.getSW()).isEmpty()) && ((ExerciseQuadTree)this.getSW()).isLeaf()) &&
				((this.getNW() == null) || (((ExerciseQuadTree)this.getNW()).isEmpty()) && ((ExerciseQuadTree)this.getNW()).isLeaf()));
	}
    
    public boolean isLabelEnabled() {
        if (this.level == 1) {
            return true;
        } else {
            return false;
        }
    }

    public boolean equals(Object obj) {
		if (obj == null) {
			// obj is an empty tree
			return (this.isEmpty() && this.isLeaf());
		} else if (obj instanceof PRQuadTree) {
			return this.equals(((PRQuadTree)obj).getElement());
		} else {
			if (obj instanceof PRQuadTreeNode) {
				// Compare the reference
				if (this == obj) {
					return true;
				} else {
					PRQuadTreeNode node = (PRQuadTreeNode)obj;
					// Compare the keys
					if (((this.getKey() == null) && (node.getKey() == null)) ||
					    ((this.getKey() != null) && (this.getKey().equals(node.getKey())))) {
						// Compare children/subtrees
						boolean subtreesEqual = false;
						// NE
						if (this.getNE() == null) {
							subtreesEqual = (node.getNE() == null); 
						} else {
							subtreesEqual = ((ExerciseQuadTree)this.getNE()).equals(node.getNE()); 
						}
						if (!subtreesEqual) {
							return false;
						}
						// NW
						if (this.getNW() == null) {
							subtreesEqual = (node.getNW() == null);
						} else {
							subtreesEqual = ((ExerciseQuadTree)this.getNW()).equals(node.getNW()); 
						}
						if (!subtreesEqual) {
							return false;
						}
						// SE
						if (this.getSE() == null) {
							subtreesEqual = (node.getSE() == null); 
						} else {
							subtreesEqual = ((ExerciseQuadTree)this.getSE()).equals(node.getSE()); 
						}
						if (!subtreesEqual) {
							return false;
						}
						// SW
						if (this.getSW() == null) {
							subtreesEqual = (node.getSW() == null); 
						} else {
							subtreesEqual = ((ExerciseQuadTree)this.getSW()).equals(node.getSW()); 
						}
						if (!subtreesEqual) {
							return false;
						} else {
							return true;
						}
					} else {
						return false;
					}
				}
			} else {
				return false;
			}
		}
	}
    
    public String toString() {
        return "";
    }

	private static final long serialVersionUID = 8131091685875118482L;
}
