// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
package content.exercises.sda.structures;

import java.io.Serializable;
import java.util.Iterator;
import java.util.Vector;

import matrix.structures.memory.VirtualArray;
import matrix.structures.memory.VirtualBoolean;
import matrix.structures.memory.VirtualInteger;

/**
 * <p>
 * An implementation of a basic hash table that 
 * uses Matrix memory structures (VirtualArray, VirtualInteger)
 * to implement its functionality, i.e. the state of the table 
 * is recorded in the Matrix animator.
 * </p><p>
 * The Object.hashCode() method is used for computing the hash 
 * value. Keys are compared by executing their equals() method.
 * Therefore keys that are equal in this sense must also have
 * equal hash values as is specified in the Java API. By default 
 * an insert fails if the table contains a value with an equal 
 * key as the one being inserted. An alternative is to turn 
 * on replacing values to have these values replaced with 
 * the new value.
 * </p><p>
 * The size of the table id doubled each time it reaches half
 * full. It uses linear probing and lazy deletion.
 * </p>
 * 
 * @author Juha Helminen
 *
 */
public class HashTable /* TODO implements Map, Set */ implements Serializable {
	protected static final int INITIAL_SIZE = 10;
	protected static final boolean REPLACE_VALUES_DEFAULT = false;
	
	protected final boolean replaceValues;
	protected final VirtualInteger count;
	protected final VirtualInteger size;
	protected final VirtualArray contents;

    public HashTable() {
        this(INITIAL_SIZE, REPLACE_VALUES_DEFAULT);
    }
    
    public HashTable(boolean replaceValues) {
    	this(INITIAL_SIZE, replaceValues);
    }

    public HashTable(int initialSize) {
    	this(initialSize, REPLACE_VALUES_DEFAULT);
    }

    public HashTable(int initialSize, boolean replaceValues) {
    	this.replaceValues = replaceValues;
        this.size = new VirtualInteger(initialSize, null, "size");
        this.contents = new VirtualArray(null, "contents");
        this.contents.setFirst(0);
        this.contents.setLast(this.size.eval() - 1);
        this.count = new VirtualInteger(0, null, "count");
    }

    public boolean put(Object key, Object value) {
        if (this.count.eval() >= (this.size.eval() - this.count.eval())) {
            rehash();
        }
        return this.put(key.hashCode(), key, value);
    }

    private void rehash() {
        Vector vec = new Vector();
        TableEntry entry;
        for (int i = this.contents.getFirst(); i <= this.contents.getLast(); i++) {
            entry = (TableEntry)this.contents.getObject(i);
            if ((entry != null) && entry.isLive()) {
                vec.add(entry);
                this.contents.setObject(null, i);
            }
        }
        this.size.assign(this.size.eval() * 2);
        this.contents.setLast(this.size.eval() - 1);
        this.count.assign(0);
        while (!vec.isEmpty()) {
            entry = (TableEntry)vec.remove(0);
            this.put(entry.getKey().hashCode(), entry.getKey(), entry.getValue());
        }
    }

    private boolean put(int hashCode, Object key, Object value) {
        int index = hashCode % this.size.eval();
        // Linear probing, find an empty or a non-live slot (lazy deletion) in the table,
        // into which to insert the item.
        while ((this.contents.getObject(index) != null) && ((TableEntry)this.contents.getObject(index)).isLive()) {
            if (((TableEntry)this.contents.getObject(index)).getKey() == key) {
            	if (this.replaceValues) {
        			// There exists an entry with the same key and the key must be unique
            		// -> replace the existing value
                    this.contents.setObject(new TableEntry(key, value), index);
                    return true;
            	} else {
                    // An item with the given key already exists.
                    // The key must be unique -> operation failed.
                    return false;
            	}
            }
            index = (index + 1) % this.size.eval();
        }
        this.contents.setObject(new TableEntry(key, value), index);
        this.count.inc();
        return true;
    }

    private TableEntry getEntry(Object key) {
        int hashCode = key.hashCode();
        int index = hashCode % this.size.eval();
        int counter = 0;
        while ((this.contents.getObject(index) != null) && (counter++ < this.size.eval())) {
            if ((this.contents.getObject(index).equals(key)) &&
                ((TableEntry)this.contents.getObject(index)).isLive()) {
                // Found a live entry with a matching key.
                return (TableEntry)this.contents.getObject(index);
            }
            index = (index + 1) % this.size.eval();
        }
        // Entry not found.
        return null;
    }

    public Object[] getKeys() {
        Vector vec = new Vector();
        TableEntry entry;
        for (int i = this.contents.getFirst(); i <= this.contents.getLast(); i++) {
            entry = (TableEntry)this.contents.getObject(i);
            if ((entry != null) && entry.isLive()) {
                vec.add(entry.getKey());
            }
        }
        Object[] keys = new Object[vec.size()];
        vec.toArray(keys);
        return keys;
    }
    
    public Iterator iterator() {
		return new Iterator() {
			private int entriesLeft = HashTable.this.count.eval();
			private int currentEntry = HashTable.this.contents.getFirst() - 1;
			
			public boolean hasNext() {
				return (this.entriesLeft > 0);
			}

			public Object next() {
				TableEntry entry;
		        for (int i = this.currentEntry + 1; i <= HashTable.this.contents.getLast(); i++) {
		            entry = (TableEntry)HashTable.this.contents.getObject(i);
		            if ((entry != null) && entry.isLive()) {
		                this.entriesLeft--;
		                this.currentEntry = i;
		                return new Entry(entry.getKey(), entry.getValue());
		            }
		        }
		        // error
		        return null;
			}
			
			public void remove() {
				TableEntry entry = (TableEntry)HashTable.this.contents.getObject(this.currentEntry);
				HashTable.this.remove(entry.getKey());
			}
		};
    }

    public Iterator iteratorValues() {
        return new Iterator() {
            private int entriesLeft = HashTable.this.count.eval();
            private int currentEntry = HashTable.this.contents.getFirst() - 1;
            
            public boolean hasNext() {
                return (this.entriesLeft > 0);
            }

            public Object next() {
                TableEntry entry;
                for (int i = this.currentEntry + 1; i <= HashTable.this.contents.getLast(); i++) {
                    entry = (TableEntry)HashTable.this.contents.getObject(i);
                    if ((entry != null) && entry.isLive()) {
                        this.entriesLeft--;
                        this.currentEntry = i;
                        return entry.getValue();
                    }
                }
                // error
                return null;
            }
            
            public void remove() { /* not implemented */ }
        };
    }

    public Iterator iteratorKeys() {
        return new Iterator() {
            private int entriesLeft = HashTable.this.count.eval();
            private int currentEntry = HashTable.this.contents.getFirst() - 1;
            
            public boolean hasNext() {
                return (this.entriesLeft > 0);
            }

            public Object next() {
                TableEntry entry;
                for (int i = this.currentEntry + 1; i <= HashTable.this.contents.getLast(); i++) {
                    entry = (TableEntry)HashTable.this.contents.getObject(i);
                    if ((entry != null) && entry.isLive()) {
                        this.entriesLeft--;
                        this.currentEntry = i;
                        return entry.getKey();
                    }
                }
                // error
                return null;
            }
            
            public void remove() { /* not implemented */ }
        };
    }

    public Object get(Object key) {
        TableEntry entry = this.getEntry(key);
        if (entry == null) {
            return null;
        } else {
            return entry.getValue();
        }
    }

    public Object remove(Object key) {
        TableEntry entry = this.getEntry(key);
        if (entry != null) {
            entry.setLive(false);
            this.count.dec();
            return entry.getValue();
        } else {
            return null;
        }
    }
    
    public void clear() {
        for (int i = this.contents.getFirst(); i <= this.contents.getLast(); i++) {
            this.contents.setObject(null, i);
        }
        this.count.assign(0);
    }

    public boolean containsKey(Object key) {
        return (this.getEntry(key) != null);
    }
    
    public int size() {
    	return this.count.eval();
    }

    
    public class Entry {
        private Object key;
        private Object value;
        
        public Entry(Object key, Object value) {
            this.key = key;
            this.value = value;
        }
        
        public Object getKey() {
            return this.key;
        }
        
        public Object getValue() {
            return this.value;
        }
    }
    
    /**
     * A single entry in the hash table.
     */
    private class TableEntry implements Serializable {
		// The unique identifier for this entry.
        private Object key;
        // The value of this entry.
        private Object value;
        // If the entry is not live, it is no longer valid,
        // i.e. it has been deleted. This is used to implement
        // lazy deletion.
        private VirtualBoolean isLive;

        public TableEntry(Object key, Object value) {
            this.key = key;
            this.value = value;
            this.isLive = new VirtualBoolean(true, null, "isLive");
        }

        public Object getKey() {
            return this.key;
        }

        public Object getValue() {
            return this.value;
        }

        public void setLive(boolean value) {
            this.isLive.assign(value);
        }

        public boolean isLive() {
            return this.isLive.eval();
        }

        public boolean equals(Object obj) {
            return (this.getKey().equals(obj));
        }

        public String toString() {
            return "(" + this.getKey() + "," + this.getValue() + ")";

        }
	
        private static final long serialVersionUID = -892058082290706989L;
    }

	private static final long serialVersionUID = -5820760785979613938L;
}
