// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
package content.exercises.sda.structures;

import java.awt.Point;
import java.awt.geom.Rectangle2D;
import java.util.Random;

/**
 * A class for generating random sets of points.
 * 
 * @author Juha Helminen
 */
public class PointGenerator {
    
    /**
     * Value representing an empty box.
     */
    static private final int EMPTY = -1;
    /**
     * Value representing a full box.
     * A full box might not contain a point: the neighbouring 
     * boxes impose constraints on the box which might make it
     * virtually full.
     */
    static private final int FULL = -2;
    /**
     * The number of times a point is tried to reassign into a box 
     * (with different coordinates) until it is deemed to be virtually full.
     */
    static private final int MAX_NUMBER_OF_TRIES = 20;
    /**
     * The length of the sides of the boxes must be <b>at least</b>
     * <code>BOX_DISTANCE_SCALE_FACTOR</code> multiplied with the 
     * minimum allowed distance between two points.  
     */
    static private final double BOX_DISTANCE_SCALE_FACTOR = 1.2;
    
    /**
     * Generates a set of points that satisfies the given set of 
     * constraints.<br><br>
     * 
     * The space assigned for the set of points is divided into
     * a matrix of equally-sized boxes, which each can contain 
     * 0 or 1 point, which results in a somewhat uniform 
     * distribution of the points. Then an attempt is made 
     * to randomly generate the points in the boxes while trying 
     * to satisfy the given constraints.
     * 
     * The area limits (minX, minY, maxX, MaxY) are inclusive.
     * 
     * @param minX              minimum allowed x-coordinate of a point
     * @param minY              minimum allowed y-coordinate of a point
     * @param maxX              maximum allowed x-coordinate of a point
     * @param maxY              maximum allowed y-coordinate of a point
     * @param numberOfPoints    the number of points to generate
     * @param minDistance       the minimum allowed distance between two points
     * @param r                 the random number generator to be used in
     *                          all operations requiring random variables 
     * @return                  a set of points that contains <code>numberOfPoints</code>
     *                          points that satisfy the given constraints
     * @throws IllegalConstraintsException If the constraints are practically unsatisfiable.
     */
    static public Point[] generatePoints(int minX, int minY, int maxX, int maxY, 
                                         int numberOfPoints, 
                                         double minDistance, 
                                         Random r) throws IllegalConstraintsException {
        // TODO Check parameter legality.

        // Divide the given area into equally-sized boxes, 
        // the length of whose sides is relative to the 
        // minimum allowed distance between two points.
        double boxSize = BOX_DISTANCE_SCALE_FACTOR * minDistance;
        int boxRows = (int) ((maxY - minY) / boxSize);
        int boxColumns = (int) ((maxX - minX) / boxSize);
        double boxWidth = boxSize + (((maxX - minX) - (boxColumns * boxSize)) / boxColumns); 
        double boxHeight = boxSize + (((maxY - minY) - (boxRows * boxSize)) / boxRows);
        // Contains references (indices to the points array) to the points
        // that are stored in the boxes or EMPTY/FULL.
        int[] boxes = new int[boxRows * boxColumns];
        for (int i = 0; i < boxes.length; i++) {
            boxes[i] = EMPTY;
        }
        // The points generated.
        Point[] points = new Point[numberOfPoints];

        int fullBoxes = 0;
        for (int i = 0; i < points.length; i++) {
            // Pick a single box at random and
            // if it's empty, assign a new randomly generated random point in it
            // else choose another box.
            int nextBox = r.nextInt(boxes.length);
            boolean constraintsSatisfied = false;
            while (!constraintsSatisfied) {
                while (boxes[nextBox] != EMPTY) {
                    nextBox = chooseNextBox(boxes.length, fullBoxes, r);
                    if (nextBox == -1) {
                        // Failed
                        throw new IllegalConstraintsException("Unable to find a set of points" +
                                                              " that satisfies the constraints" + 
                                                              " with reasonable effort.");
                    }
                }
                fullBoxes++;
                
                int row, column;
                row = nextBox / boxColumns;
                column = nextBox % boxColumns;
                double lowerCornerX = minX + column * boxWidth;
                double lowerCornerY = minY + row * boxHeight;
                points[i] = generatePoint(lowerCornerX, 
                                          lowerCornerY,
                                          lowerCornerX + boxWidth,
                                          lowerCornerY + boxHeight,
                                          r);
                boxes[nextBox] = i;
                constraintsSatisfied = 
                    satisfyConstraints(points, nextBox, boxes, boxRows, boxColumns, minDistance, minX, minY, boxWidth, boxHeight, r);
                if (!constraintsSatisfied) {
                    // Mark this box as (virtually) full, because it was found 
                    // hard to assign a point into it due to the strict constraints.
                    boxes[nextBox] = FULL;
                }
            }
        }
        
        return points;
    }
    
    /**
     * @param limits    The area limits for the points to be generated.
     *                  The limits are inclusive.
     * @see   <code>generatePoints(double, double, double, double, in, double, Random)</code>
     */
    static public Point[] generatePoints(Rectangle2D limits,
                                         int numberOfPoints, 
                                         double minDistance, 
                                         Random r) throws IllegalConstraintsException {
        return generatePoints(limits.getMinX(), limits.getMinY(), limits.getMaxX(), limits.getMaxY(),
                              numberOfPoints, minDistance, r);
    }
    
    
    /**
     * @see   <code>generatePoints(int, int, int, int, int, double, Random)</code>
     */
    static public Point[] generatePoints(double minXdouble, 
                                         double minYdouble, 
                                         double maxXdouble, 
                                         double maxYdouble, 
                                         int numberOfPoints, 
                                         double minDistance, 
                                         Random r) throws IllegalConstraintsException {
        int minX = (int)Math.ceil(minXdouble);
        int minY = (int)Math.ceil(minYdouble);
        int maxX = (int)Math.floor(maxXdouble);
        int maxY = (int)Math.floor(maxYdouble);
        return generatePoints(minX, minY, maxX, maxY, numberOfPoints, minDistance, r);
    }

    static private int chooseNextBox(int numberOfBoxes, int fullBoxes, Random r) {
        if (fullBoxes >= numberOfBoxes) {
            // Failed
            return -1;
        } else {
            // XXX Maybe just choose the next empty box. 
            // That, however, results in clustering.
            return r.nextInt(numberOfBoxes);
        }
    }
    
    static public final Point generatePoint(double minXdouble, 
                                             double minYdouble, 
                                             double maxXdouble, 
                                             double maxYdouble,
                                             Random r) {
        int minX = (int)Math.ceil(minXdouble);
        int minY = (int)Math.ceil(minYdouble);
        int maxX = (int)Math.floor(maxXdouble);
        int maxY = (int)Math.floor(maxYdouble);
        return new Point(minX + r.nextInt(maxX - minX + 1),
                         minY + r.nextInt(maxY - minY + 1));    
    }
    
    static private boolean satisfyConstraints(Point[] points, 
                                              int boxIndex, 
                                              int[] boxes,
                                              int rows,
                                              int columns,
                                              double minDistance,
                                              double minX,
                                              double minY,
                                              double boxWidth,
                                              double boxHeight,
                                              Random r) {
        // Check that the newly generated point satisfies
        // the constraints comparing it to each of its
        // neighbouring non-empty boxes.
        // If ok -> return true
        // else -> apply some technique to satisfy the constraints,
        // if that fails -> return false.
        int tryCounter = 1;
        Point neighbour;
        checkConstraints:
            while (tryCounter < MAX_NUMBER_OF_TRIES) {
                for (int i = -1; i < 2; i++) {
                    for (int j = -1; j < 2; j++) {
                        if (i == j) {
                            continue;
                        } else {
                            int neigbourBoxIndex = boxIndex + i * columns + j;
                            if ((neigbourBoxIndex < 0) || (neigbourBoxIndex >= boxes.length)) {
                                continue;
                            }
                            int neigbourPointIndex = boxes[neigbourBoxIndex];
                            if ((neigbourPointIndex == EMPTY) || (neigbourPointIndex == FULL)) {
                                continue;
                            }
                            neighbour = points[neigbourPointIndex];
                            if (points[boxes[boxIndex]].distance(neighbour) <= minDistance) {
                                // The distance is too small -> 
                                // try to generate a new point in the box.
                                int row = boxIndex / columns;
                                int column = boxIndex % columns;
                                double lowerCornerX = minX + column * boxWidth;
                                double lowerCornerY = minY + row * boxHeight;
                                // TODO Come up with a more sophisticated method instead of 
                                // simply generating a new point randomly without any heuristics.
                                points[boxes[boxIndex]] = generatePoint(lowerCornerX, 
                                                                        lowerCornerY,
                                                                        lowerCornerX + boxWidth,
                                                                        lowerCornerY + boxHeight,
                                                                        r);
                                tryCounter++;
                                continue checkConstraints;
                            }
                        }
                    }
                }
                // The point satisfies all the constraints.
                return true;
            }
        // Unable to generate a point that satisfies the constraints.
        return false;
    }
    
    public static final class IllegalConstraintsException extends Exception {
        IllegalConstraintsException(String message) {
            super(message);
        }
        private static final long serialVersionUID = 6143595021634811579L;
    }

}
