// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
package content.exercises.sda.structures;

import java.awt.Point;
import java.awt.geom.Line2D;
import java.util.Arrays;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.Random;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.Vector;

import matrix.structures.FDT.advanced.GeometricGraph;
import matrix.structures.FDT.substructures.Vertex;
import matrix.util.Note;

/**
 * A class for generating polygons.
 * 
 * @author Juha Helminen
 */
public class PolygonGenerator {

    /**
     * An implementation of the 2 peasants algorithm to generate a simple polygon
     * from a set of points. The given points are reordered to follow the contour 
     * of a simple polygon.
     * 
     * @param points A set of points that will be the vertices of the polygon.
     */
    public static void generatePolygon2Peasants(Point[] points) {
        // Sort the points in ascending order according to the x-coordinate.
        Arrays.sort(points, new Comparator() {
            public int compare(Object arg0, Object arg1) {
                Point p0 = (Point)arg0;
                Point p1 = (Point)arg1;
                return (new Integer(p0.x)).compareTo(new Integer(p1.x));
            }
        });

        // Divide the points to those above and those below 
        // a line from the point with the smallest x-coordinate 
        // to the point with the largest x.
        Line2D l = new Line2D.Double(points[0], points[points.length - 1]);
        LinkedList above = new LinkedList();
        TreeSet below = new TreeSet(new Comparator() {
            // Reverse the order for the points below,
            // that is, sort them in descending order 
            // according to the x-coordinate.
            public int compare(Object arg0, Object arg1) {
                Point p0 = (Point)arg0;
                Point p1 = (Point)arg1;
                int comparison = (new Integer(p0.x)).compareTo(new Integer(p1.x));
                if (comparison == 0) {
                    // Allow duplicates.
                    return -1;
                } else {
                    return -comparison;
                }
            }
        });
        for (int i = 0; i < points.length; i++) {
            if (l.relativeCCW(points[i]) < 0) {
                above.add(new java.awt.Point(points[i].x, points[i].y));
            } else {
                below.add(new java.awt.Point(points[i].x, points[i].y));
            }
        }
        // Construct the complete contour.
        above.addAll(below);
        above.toArray(points);
    }
    
    /**
     * Probability that there is a point between two rectangle points being connected;
     */
    private static final double BETWEEN_PROB = 0.5;
    
    public static GeometricGraph generatePolygonMapAsGraph(int maxx, int maxy, int rectanglex, int rectangley, int mindist) {
    	if(rectanglex % mindist != 0 || rectangley % mindist != 0) {
    		Note.out(null, " rectanglex or rectangley not multiples of mindist, adjusting");
    		rectanglex = mindist*rectanglex / mindist;
    		rectangley = mindist*rectangley / mindist;
    	}
    	/* objects in rectanglePoints must be of type Vector. 
    	 * Keys must be of type "x<coordinate>" or "y<coordinate>" 
    	 */
    	TreeMap rectanglePoints = new TreeMap();
    	ExerciseGeometricGraph g = new ExerciseGeometricGraph();
    	Random rnd = new Random();
    	int currx = 0;
    	int curry = 0;
    	// for loop to go through all boxes in the area according to x axis
    	for(currx = rectanglex; currx <= maxx; currx += rectanglex) {
    		// for loop to go through all boxes in a given x coordinates according to y axis
    		for(curry = rectangley; curry <= maxy; curry += rectangley) {
    			// fetch base points (one point on each side of the rectangle
    			ExerciseGeometricGraphVertex[] bases = getBasePoints(currx,curry,rectanglex, rectangley, mindist, rectanglePoints,rnd,g);
    			Vector polygon = new Vector();
    			for(int i = 0; i < bases.length; i++) {
    				// go through points, connect them if necessary and include points in between if generated
    				ExerciseGeometricGraphVertex curr = bases[i%4];
    				ExerciseGeometricGraphVertex next = bases[(i+1)%4];
    				polygon.add(curr);
    				ExerciseGeometricGraphVertex between = null;
    				if(rnd.nextDouble() < BETWEEN_PROB) {
    					between = getVertexBetween(curr, next, mindist,polygon,rnd,g);    					
    				}
    				if(between != null) {
    					curr.addSuccessor(between);
    					between.addSuccessor(curr);
    					between.addSuccessor(next);
    					next.addSuccessor(between);
    					polygon.add(between);
    				} else {
    					if(!curr.hasSuccessor(next)) {
    						curr.addSuccessor(next);
    					}
    					if(!next.hasSuccessor(curr)) {
    						next.addSuccessor(curr);
    					}
    				}
    			}
    		}
    	}
    	// connect the border points at the edge of the area to one other in order to complete vertices
    	Vector[] vecs = new Vector[4];
    	vecs[0] = (Vector)rectanglePoints.get("x0");
    	vecs[1] = (Vector)rectanglePoints.get("y0");
    	vecs[2] = (Vector)rectanglePoints.get("x" + (currx-rectanglex));
    	vecs[3] = (Vector)rectanglePoints.get("y" + (curry-rectangley));
    	for(int i = 0; i < vecs.length; i++) {
    		Vector v = (Vector)vecs[i];
    		if(v.size() > 1) {
    			ExerciseGeometricGraphVertex prev = (ExerciseGeometricGraphVertex)v.get(0);
    			for(int j = 1; j < v.size(); j++) {
    				ExerciseGeometricGraphVertex curr = (ExerciseGeometricGraphVertex)v.get(j);
    				if(!prev.hasSuccessor(curr)) {
    					prev.addSuccessor(curr);
    				}
    				if(!curr.hasSuccessor(prev)) {
    					curr.addSuccessor(prev);
    				}
    				prev = curr;
     			}
    		}
    	}
    	return g;
    }

    /**
     * Fetch the base points for a given rectangle which is bounded by (currx-x, curry-y), (currx,curry)
     * @param currx
     * @param curry
     * @param x
     * @param y
     * @param storage
     * @param rnd
     * @param g
     * @return
     */
    private static ExerciseGeometricGraphVertex[] getBasePoints(int currx, int curry, 
    		int x, int y, int mindist, TreeMap storage, Random rnd, ExerciseGeometricGraph g) {
    	ExerciseGeometricGraphVertex[] vertices = new ExerciseGeometricGraphVertex[4];
		vertices[0] = getVertex(storage,"x" + (currx-x),curry-y,curry,mindist,rnd,g);
		vertices[1] = getVertex(storage,"y"+ (curry-y),currx-x,currx,mindist,rnd,g);
		vertices[2] = getVertex(storage,"x"+currx,curry-y,curry,mindist,rnd,g);
		vertices[3] = getVertex(storage,"y"+curry,currx-x,currx,mindist,rnd,g);
    	return vertices;
    }

    
    /**
     * Fetch a base point with given x or y coordinate (indicated in the key string) and between the
     * min and max coordinates. If no point exists, new is created
     * @param storage
     * @param key
     * @param min
     * @param max
     * @param rnd
     * @param g
     * @return
     */
    private static ExerciseGeometricGraphVertex getVertex(TreeMap storage, 
    		String key, int min, int max, int mindist,Random rnd,
    		ExerciseGeometricGraph g) {
    	Vector v = (Vector)storage.get(key);
    	if(v == null) {
    		v = new Vector();
    		storage.put(key, v);
    	}
    	ExerciseGeometricGraphVertex vert;
    	for(int i = v.size()-1; i > -1; i--) {
    		vert = (ExerciseGeometricGraphVertex)v.get(i);
    		if(key.charAt(0) == 'x') {
    			if(min < vert.getY() && vert.getY() < max) {
    				return vert;
    			}
    		} else {
    			if(min < vert.getX() && vert.getX() < max) {
    				return vert;
    			}
    		}    		
    	}
    	// First we shift the area for the rnd towards the origo, and then back to the correct coordinates
    	int tmp = rnd.nextInt((max-min)/mindist);
    	int coord = min + tmp*mindist;
    	if(coord == min) {
    		coord += mindist;
    	} else if(coord == max) {
    		coord -= mindist;
    	}
    	if(key.charAt(0) == 'x') {
    		int x = Integer.parseInt(key.substring(1));
    		vert = (ExerciseGeometricGraphVertex)g.addVertex(x,coord,"");
    	} else {
    		int y = Integer.parseInt(key.substring(1));
    		vert = (ExerciseGeometricGraphVertex)g.addVertex(coord,y,"");
    	}
    	v.add(vert);
    	return vert;
    }
    
    
    /**
     * Create an in-between point between two base points.
     * @param curr
     * @param next
     * @return New in-between point or null if none could be generated
     */
    private static ExerciseGeometricGraphVertex getVertexBetween(ExerciseGeometricGraphVertex curr, 
    			ExerciseGeometricGraphVertex next, int mindist,
    			Vector polygon, Random rnd, ExerciseGeometricGraph g) {
    	int minx, miny, maxx,maxy;
    	if(curr.getX() < next.getX()) {
    		minx = curr.getX();
    		maxx = next.getX();
    	} else {
    		minx = next.getX();
    		maxx = curr.getX();
    	}
    	if(curr.getY() < next.getY()) {
    		miny = curr.getY();
    		maxy = next.getY();
    	} else {
    		miny = next.getY();
    		maxy = curr.getY();
    	}
    	if(maxx-minx <= mindist || maxy-miny <= mindist) {
    		return null;
    	}
    	int x, y, count;
    	ExerciseGeometricGraphVertex vert = null;
    	count = 0;
    	mainloop:
    	while(vert == null && count < MAX_COUNT) {
    		int tmp = rnd.nextInt((maxx-minx)/mindist);
    		x = minx + tmp*mindist;
    		tmp = rnd.nextInt((maxy-miny)/mindist);
    		y = miny + tmp*mindist;
    		if(x == maxx) {
    			x -= mindist;
    		} else if(x == minx) {
    			x += mindist;
    		}
    		if(y == maxy) {
    			y -= mindist;
    		} else if(y == miny) {
    			y += mindist;
    		}
    		count++;
    		ExerciseGeometricGraphVertex prev = (ExerciseGeometricGraphVertex)polygon.get(0); 
    		for(int i = 1; i < polygon.size(); i++) {
    			ExerciseGeometricGraphVertex current = (ExerciseGeometricGraphVertex)polygon.get(i);
    			if(intersects(x,curr.getX(),y,curr.getY(),prev.getX(),current.getX(),prev.getY(),current.getY())) {
    				continue mainloop;
    			}
    		}
    		vert = (ExerciseGeometricGraphVertex)g.addVertex(x,y,"");
    	}
    	return vert;
    }
    
    private static final int MAX_COUNT=10;

    /**
 	 * Calculates the intersection point of the line segments represented by the parameter points.
	 * The calculation is done as per instructions at http://mathworld.wolfram.com/Line-LineIntersection.html
	 * It is assumed that the parameters represent separate line segments that intersect in one single point.
	 * If the lines intersect in an area between the parameter points, true is returned.
	 * Points (x1,y1),(x2,y2) define the first line and points (x3,y3),(x4,y4) the second line.
	 *
     * @param x1
     * @param x2
     * @param y1
     * @param y2
     * @param x3
     * @param x4
     * @param y3
     * @param y4
     * @return True if line segments intersect, false otherwise
     */
    public static boolean intersects(int x1, int x2, int y1, int y2, int x3, int x4, int y3, int y4) {
		double x = determinant(determinant(x1,y1,x2,y2), x1-x2, determinant(x3,y3,x4,y4), x3-x4);
		double tmp = determinant(x1-x2, y1-y2, x3-x4, y3-y4);
		// case of parallel lines
		if(tmp == 0) {
			return false;
		}
		x = x / tmp;
		double y = determinant(determinant(x1,y1,x2,y2), y1-y2, determinant(x3,y3,x4,y4), y3-y4);
		tmp = determinant(x1-x2, y1-y2, x3-x4, y3-y4);
		/* Unneeded, the already done with x coordinate
		if(tmp == 0) {
			return false;
		}
		*/
		y = y / tmp;
		// check if intersection inside line segments
		if(((x > x1 && x < x2) || (x < x1 && x > x2)) 
				&& ((y > y1 && y < y2) || (y < y1 && y > y2))
				&& ((x > x3 && x < x4) || (x < x3 && x > x4)) 
				&& ((y > y3 && y < y4) || (y < y3 && y > y4))) {
			return true;			
		}
		return false;
    }
    
	public static double determinant(double a, double b, double c, double d) {
		return a*d - b*c;
	}
    
    
    /**
     * <p>
     * Finds the individual polygons in a polygon map represented by a fully connected geometric 
     * graph.
     * </p><p>
     * The polygons are found by traversing each of the (directed) edges in the graph once in 
     * both clockwise and counterclockwise direction, i.e. the next edge in a polygon is 
     * always the edge that has the smallest angle to the current edge in either cw or ccw 
     * direction.
     * </p>
     * @param   ggraph  A fully connected directed geometric graph representing a polygon map.
     * @return          The set of polygons the graph represents. The polygon that is formed
     *                  from "border" edges is ignored. The polygons are an ordered sequence 
     *                  connected of points, where the last point in the sequence is connected 
     *                  to the first.
     */
    public static java.awt.Point[][] findPolygons(GeometricGraph ggraph) {
        // XXX There seems to be a bug in the generatePolygonMapAsGraph method, which
        // XXX causes the vertices to have as their first successor a vertex whose 
        // XXX coordinates are undefined, i.e. (-1, -1) is returned.
        // XXX As a quick temporary work-around, successors with negative coordinates
        // XXX are completely ignored.
        
        // The polygons are modeled as an ordered sequence connected of points 
        // (the last point in the sequence is connnected to the first).
        java.awt.Point[][] polygons;
        Vertex[] vertices = ggraph.getVertices(), successors;
        Vertex startVertex, lastVertex, currentVertex, nextVertex, successor;
        Vector polys = new Vector(), poly;
        // cw/ccw contains the directed edges that have not been traversed in the clockwise/ccw direction.
        LinkedList cw = new LinkedList(), ccw = new LinkedList();
        Edge edge;
        // True if the current direction is cw.
        boolean isCw;
        
        // Create and store all the edges in cw & ccw.
        for (int i = 0; i < vertices.length; i++) {
            currentVertex = vertices[i];
            successors = currentVertex.getSuccessors();
            for (int j = 0; j < successors.length; j++) {
                successor = successors[j];
                // XXX Ignore successors with negative coordinates.
                if ((ggraph.getX(successor) < 0) || (ggraph.getY(successor) < 0)) {
                    continue;
                }
                if (!cw.contains(new Edge(currentVertex, successor, ggraph))) {
                    edge = new Edge(currentVertex, successor, ggraph);
                    cw.add(edge);
                    ccw.add(edge);
                }
            }
        }
        
        // Find the polygons by traversing all the edges once in both directions (clockwise & ccw). 
        double angle, minAngle;
        boolean edgeVisited = false;
        while (!(cw.isEmpty() && ccw.isEmpty())) {
            // Pick an edge that has not been processed.
            if (!cw.isEmpty()) {
                isCw = true; 
                edge = (Edge)cw.removeFirst();
            } else {
                isCw = false;
                edge = (Edge)ccw.removeFirst();
            }
            // Create a polygon by traversing the selected edge to the selected direction.
            poly = new Vector();
            startVertex = edge.getV0();;
            poly.add(startVertex);
            currentVertex = startVertex;
            nextVertex = edge.getV1();
            do {
                poly.add(nextVertex);
                lastVertex = currentVertex;
                currentVertex = nextVertex;
                successors = currentVertex.getSuccessors();
                // Find the next edge for the polygon by choosing the edge that has the smallest
                // angle to the current traversing direction, cw/ccw.
                minAngle = 2 * Math.PI;
                for (int j = 1; j < successors.length; j++) {
                    successor = successors[j];
                    // XXX Ignore successors with negative coordinates.
                    if ((ggraph.getX(successor) < 0) || (ggraph.getY(successor) < 0)) {
                        continue;
                    }
                    if (isCw) {
                        angle = GeometryFunctions.clockWiseAngle(new Point(ggraph.getX(lastVertex), ggraph.getY(lastVertex)), 
                                                                 new Point(ggraph.getX(currentVertex), ggraph.getY(currentVertex)), 
                                                                 new Point(ggraph.getX(successor), ggraph.getY(successor)));
                    } else {
                        angle = GeometryFunctions.counterClockWiseAngle(new Point(ggraph.getX(lastVertex), ggraph.getY(lastVertex)), 
                                                                 new Point(ggraph.getX(currentVertex), ggraph.getY(currentVertex)), 
                                                                 new Point(ggraph.getX(successor), ggraph.getY(successor)));
                    }
                    if ((angle <= minAngle) && (!lastVertex.equals(successor)) && (!currentVertex.equals(successor))) {
                        minAngle = angle;
                        nextVertex = successor;
                    }
                }
                // Mark the edge visited by removing it from the corresponding lists.
                if (isCw) {
                    edgeVisited = !cw.remove(new Edge(currentVertex, nextVertex, ggraph));
                    ccw.remove(new Edge(nextVertex, currentVertex, ggraph));
                } else {
                    edgeVisited = !ccw.remove(new Edge(currentVertex, nextVertex, ggraph));
                    cw.remove(new Edge(nextVertex, currentVertex, ggraph));
                }
                // If the edge has already been traversed to the given direction -> abort.
                if (edgeVisited) {
                    break;
                }
            } while (!nextVertex.equals(startVertex));
            if (!edgeVisited) {
                // A new polygon was found.
                polys.add(poly);
            }
        }
        
        // Find the polygon with the largest number of vertices,
        // that is the polygon surrounding the polygon map.
        int largestPolygonSize = 0;
        if (polys.size() > 1) {
            for (int i = 0; i < polys.size(); i++) {
                poly = (Vector)polys.get(i);
                if (poly.size() > largestPolygonSize) {
                    largestPolygonSize = poly.size();
                }
            }
        }
         
        // Create the final polygons and return them.
        // polys.size() - 1, because the surrounding polygon is removed
        polygons = new Point[Math.max(1, polys.size() - 1)][];
        for (int i = 0; i < polys.size(); i++) {
            poly = (Vector)polys.get(i);
            if (poly.size() != largestPolygonSize) {
                polygons[i] = new Point[poly.size()];
                for (int j = 0; j < poly.size(); j++) {
                    polygons[i][j] = new Point(ggraph.getX((Vertex)poly.get(j)), ggraph.getY((Vertex)poly.get(j)));
                }
            } else {
                // Skip the polygon with the largest number of vertices.
                // that is the polygon surrounding the polygon map.
                polys.remove(i);
                i--;
            }
        }
        return polygons;
    }
    
    /**
     *  A directed edge in a geometric graph. 
     */
    static private class Edge {
        private Vertex v0;
        private Vertex v1;
        private GeometricGraph ggraph;
        
        public Edge(Vertex v0, Vertex v1, GeometricGraph ggraph) {
            this.v0 = v0;
            this.v1 = v1;
            this.ggraph = ggraph;
        }
        
        public Vertex getV0() {
            return this.v0;
        }
        
        public Vertex getV1() {
            return this.v1;
        }
        
        public GeometricGraph getGraph() {
            return this.ggraph;
        }

        public boolean equals(Object obj) {
            if ((obj instanceof Edge) && (this.getGraph() == ((Edge)obj).getGraph())) {
                Edge edge = (Edge)obj;
                Vertex va0 = this.getV0();
                Vertex va1 = this.getV1();
                Vertex vb0 = edge.getV0();
                Vertex vb1 = edge.getV1();
                if ((this.getGraph().getX(va0) == this.getGraph().getX(vb0)) &&
                    (this.getGraph().getY(va0) == this.getGraph().getY(vb0))) {
                    return ((this.getGraph().getX(va1) == this.getGraph().getX(vb1)) &&
                            (this.getGraph().getY(va1) == this.getGraph().getY(vb1)));
                } else {
                    return false;
                }
            } else {
                return false;
            }
        }
    }
}
