// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
package content.exercises.sda.structures;

import java.awt.Rectangle;
import java.awt.geom.Ellipse2D;
import java.awt.geom.GeneralPath;
import java.awt.geom.Line2D;
import java.awt.geom.PathIterator;
import java.awt.geom.Point2D;
import java.io.Serializable;
import java.util.Iterator;
import java.util.Vector;

import matrix.structures.spatial.Area;
import matrix.structures.spatial.PaintingStyleDecorator;
import matrix.structures.spatial.SpatialComparable;
import matrix.structures.spatial.FDT.probe.LineSegment;
import matrix.structures.spatial.FDT.probe.Point;
import matrix.structures.spatial.FDT.probe.Polygon;
import matrix.structures.spatial.FDT.probe.SpatialElement;
import matrix.util.Note;

public class Triangulation implements Area {

    public static final int VORONOI_NONE = 0;
    public static final int VORONOI_REGIONS = 1;
    public static final int VORONOI_EDGES = 2;
    
    private int drawVoronoi;
    private boolean drawVoronoiVertices;
    private boolean drawDelaunay;
    private boolean drawOnlyCrust;
    private boolean drawOnlyTrianglesVoronoi;
    private boolean drawTriangles;
    private boolean drawVertices;
    
    private final VisualElements visualElements;
    private final int width;
    private final int height;
    
    private HashTable verticesVis;
    private HashTable edgesVis;
    private HashTable voronoiVerticesVis;
    private HashTable trianglesVis;
    private HashTable voronoiEdgesVis;

    private HashTable vertexEdges;
    private HashTable edgeTriangles;

    public Triangulation() {
        this(new VisualElements(), 400, 400);
    }

    public Triangulation(int width, int height) {
        this(new VisualElements(), width, height);
    }

    public Triangulation(VisualElements visualElements, int width, int height) {
        this.drawVoronoi = VORONOI_NONE;
        this.drawVoronoiVertices = true;
        this.drawOnlyTrianglesVoronoi = true;
        this.drawOnlyCrust = false;
        this.drawDelaunay = true;
        this.drawTriangles = true;
        this.drawVertices = true;
        
        this.edgeTriangles = new HashTable(20);
        this.vertexEdges = new HashTable(10);
        
        this.verticesVis = new HashTable();
        this.edgesVis = new HashTable();
        this.voronoiVerticesVis = new HashTable();
        this.trianglesVis = new HashTable();
        this.voronoiEdgesVis = new HashTable();
        
        this.visualElements = visualElements;
        this.width = width;
        this.height = height;
    }

    public void clearEdges() {
        this.vertexEdges.clear();
        this.edgeTriangles.clear();
        this.edgesVis.clear();
        this.voronoiVerticesVis.clear();
        this.trianglesVis.clear();
        this.voronoiEdgesVis.clear();
    }

    public void clear() {
        this.clearEdges();
        this.verticesVis.clear();
    }

    
    
    // *** Visualization settings ***
    
    public void setDrawVoronoi(int drawVoronoi) {
        if (drawVoronoi == VORONOI_NONE) {
            this.visualElements.hideVoronoi();
        }
        this.drawVoronoi = drawVoronoi;
    }

    public void setDrawVertices(boolean drawVertices) {
        this.drawVertices = drawVertices;
    }

    public void setDrawVoronoiVertices(boolean drawVoronoiVertices) {
        this.drawVoronoiVertices = drawVoronoiVertices;
    }

    public void setDrawTriangles(boolean drawTriangles) {
        this.drawTriangles = drawTriangles;
    }

    public boolean isTrianglesDrawn() {
        return this.drawTriangles;
    }

    public boolean isVoronoiDrawn() {
        return (this.drawVoronoi != VORONOI_NONE);
    }

    public boolean isVoronoiVerticesDrawn() {
        return this.drawVoronoiVertices;
    }
    
    public void setDrawTrianglesVoronoi(boolean drawVoronoiTriangles) {
        this.drawOnlyTrianglesVoronoi = drawVoronoiTriangles;
    }

    public boolean isTrianglesVoronoiDrawn() {
        return this.drawOnlyTrianglesVoronoi;
    }

    public void setDrawOnlyCrust(boolean drawOnlyCrust) {
        this.drawOnlyCrust = drawOnlyCrust;
    }

    public boolean isOnlyCrustDrawn() {
        return this.drawOnlyCrust;
    }

    public void setDrawDelaunay(boolean drawDelaunay) {
        if (!drawDelaunay) {
            this.visualElements.hideDelaunay();
        } else {
            this.visualElements.showDelaunay();
        }
        this.drawDelaunay = drawDelaunay;
    }

    public boolean isDelaunayDrawn() {
        return this.drawDelaunay;
    }
    
    
    // *** Triangulation construction ***

    
    public boolean addVertex(java.awt.Point vertex) {
        return this.addVertex(vertex, null);
    }
    

    public boolean addVertex(java.awt.Point vertex, PaintingStyleDecorator psd) {
        if (this.hasVertex(vertex)) {
            return false;
        } else {
            Point point = this.visualElements.newVertex(vertex);
            if (psd != null) {
                point.setPaintingStyleDecorator(psd);
            }
            this.verticesVis.put(vertex, point);
            return true;
        }
    }
    
    public boolean hasVertex(int x, int y) {
        return this.hasVertex(new java.awt.Point(x, y));
    }
    
    public boolean hasVertex(java.awt.Point vertex) {
        return this.verticesVis.containsKey(vertex);
    }

    public void setVertexPaintingStyledecorator(java.awt.Point vertex, PaintingStyleDecorator psd) {
        ((Point)this.verticesVis.get(vertex)).setPaintingStyleDecorator(psd);
    }

    public boolean removeVertex(int x, int y) {
        return this.removeVertex(new java.awt.Point(x, y));
    }
    
    public boolean removeVertex(java.awt.Point vertex) {
        if (this.hasVertex(vertex)) {
            this.deleteVertex(vertex);
            return true;
        } else {
            return false;
        }
    }

    // return the vertices' visual elements
    public Iterator vertices() {
        return this.verticesVis.iteratorValues();
    }
    
    public Iterator voronoiVertices() {
        return this.voronoiVerticesVis.iteratorValues();
    }

    public boolean addEdge(java.awt.Point orig, java.awt.Point dest) {
        return this.addEdge(orig, dest, null);
    }
    
    // duplicate insertion has no effect
    public boolean addEdge(java.awt.Point orig, java.awt.Point dest, PaintingStyleDecorator psd) {
        if (this.hasEdge(orig, dest) || 
            !this.hasVertex(orig) || 
            !this.hasVertex(dest)) {
            return false;
        } else {
            LineSegment edge = this.visualElements.newDelaunayEdge(orig, dest);
            if (psd != null) {
                edge.setPaintingStyleDecorator(psd);
            }
            this.insertEdge(edge);
            return true;
        }
    }

    public boolean hasEdge(java.awt.Point orig, java.awt.Point dest) {
        return this.hasEdge(new Edge(orig, dest));
    }
    
    public LineSegment getEdge(java.awt.Point orig, java.awt.Point dest) {
        return this.getEdge(new Edge(orig, dest));
    }

    private LineSegment getEdge(Edge edge) {
        return (LineSegment)this.edgesVis.get(edge);
    }

    private boolean hasEdge(Edge edge) {
        return this.edgesVis.containsKey(edge);
    }
    
    public LineSegment getVoronoiEdge(java.awt.Point delaunayOrig, java.awt.Point delaunayDest) {
        return (LineSegment)this.voronoiEdgesVis.get(new Edge(delaunayOrig, delaunayDest));
    }
    
    public LineSegment getDelaunayEdge(java.awt.Point voronoiOrig, java.awt.Point voronoiDest) {
        Iterator iterator = this.voronoiEdgesVis.iterator();
        HashTable.Entry entry;
        LineSegment voronoiEdge;
        while (iterator.hasNext()) {
            entry = (HashTable.Entry)iterator.next();
            voronoiEdge = (LineSegment)entry.getValue();
            if ((voronoiEdge.getP1().equals(voronoiOrig) && 
                 voronoiEdge.getP2().equals(voronoiDest)) ||
                (voronoiEdge.getP2().equals(voronoiOrig) && 
                 voronoiEdge.getP1().equals(voronoiDest))) {
                return this.getEdge((Edge)entry.getKey());
            }
        }
        return null;
    }

    public boolean removeEdge(java.awt.Point orig, java.awt.Point dest) {
        return this.removeEdge(new Edge(orig, dest));
    }

    private boolean removeEdge(Edge edge) {
        return this.deleteEdge(edge);
    }

    public Iterator vertexEdges(int x, int y) {
        return this.vertexEdges(new java.awt.Point(x, y));
    }
    
    public Iterator vertexEdges(java.awt.Point vertex) {
        SinglyLinkedList edgeList = (SinglyLinkedList)this.vertexEdges.get(vertex);
        if (edgeList != null) {
            return edgeList.iterator();
        } else {
            return new Iterator(){
                public boolean hasNext() { return false; }
                public Object next() { return null; }
                public void remove() { /* not implemented */ }
            };
        }
    }
    
    // return edges' visual elements
    public Iterator edges() {
        return this.edgesVis.iteratorValues();
    }

    
    public LineSegment[] crust() {
        Vector crust = new Vector();

        // The crust
        Iterator iterator = this.edgesVis.iterator();
        Edge edge;
        HashTable.Entry entry;
        SinglyLinkedList triangleList;
        Triangle triangle1, triangle2;
        Iterator triangles;
        double testValue;
        while (iterator.hasNext()) {
            entry = (HashTable.Entry)iterator.next();
            edge = (Edge)entry.getKey();
            triangleList = (SinglyLinkedList)this.edgeTriangles.get(edge);
            if (triangleList != null) {
                triangles = triangleList.iterator();
                if (triangleList.size() == 2) {
                    triangle1 = (Triangle)triangles.next();
                    triangle2 = (Triangle)triangles.next();
                    testValue = GeometryFunctions.inCircleTest(triangle1.getCenter(), 
                                                               edge.getP1(), 
                                                               edge.getP2(), 
                                                               triangle2.getCenter());
                    if (testValue > 0) {
                        // the point is inside -> anti-crust
                    } else {
                        // the point is outside -> crust
                        crust.add(entry.getValue());
                    }
                } else if (triangleList.size() == 1) {
                    triangle1 = (Triangle)triangles.next();
                    if (edge.getLine2D().relativeCCW(triangle1.getCenter()) < 0) {
                        // the voronoi edge intersects the corresponding delaunay edge
                        // -> crust
                        crust.add(entry.getValue());
                    }
                }
            }
        }

        LineSegment[] crustArray = new LineSegment[crust.size()];
        crust.toArray(crustArray);
        return crustArray;
    }

    public LineSegment[] skeleton() {
        Vector skeleton = new Vector();

        // The skeleton
        Iterator iterator = this.edgesVis.iterator();
        Edge edge;
        HashTable.Entry entry;
        SinglyLinkedList triangleList;
        Triangle triangle1, triangle2;
        Iterator triangles;
        double testValue;
        while (iterator.hasNext()) {
            entry = (HashTable.Entry)iterator.next();
            edge = (Edge)entry.getKey();
            triangleList = (SinglyLinkedList)this.edgeTriangles.get(edge);
            if (triangleList != null) {
                triangles = triangleList.iterator();
                if (triangleList.size() == 2) {
                    triangle1 = (Triangle)triangles.next();
                    triangle2 = (Triangle)triangles.next();

                    testValue = GeometryFunctions.inCircleTest(triangle1.getCenter(), 
                                                               edge.getP1(), 
                                                               edge.getP2(), 
                                                               triangle2.getCenter());

                    if (testValue > 0) {
                        // the point is inside -> anti-crust
                        skeleton.add(this.voronoiEdgesVis.get(edge));
                    } else {
                        // the point is outside -> crust
                    }
                } else if (triangleList.size() == 1) {
                    triangle1 = (Triangle)triangles.next();
                    if (edge.getLine2D().relativeCCW(triangle1.getCenter()) < 0) {
                        // the voronoi edge intersects the corresponding delaunay edge
                        // -> crust
                    } else {
                        skeleton.add(this.voronoiEdgesVis.get(edge));
                    }
                }
            }
        }

        LineSegment[] skeletonArray = new LineSegment[skeleton.size()];
        skeleton.toArray(skeletonArray);
        return skeletonArray;
    }

    
    private void addTriangle(Triangle triangle) {
        if (!this.trianglesVis.containsKey(triangle)) {
            
            this.trianglesVis.put(triangle, this.visualElements.newTriangle(triangle.getPoints()));
            
            // Create the voronoi vertex
            java.awt.Point center = triangle.getCenter();
            this.voronoiVerticesVis.put(center, this.visualElements.newVoronoiVertex(center));

            if (this.drawVoronoi == VORONOI_EDGES) {
                SinglyLinkedList triangleList;
                loop:
                    for (int i = 0; i < 3; i++) {
                        triangleList = (SinglyLinkedList)this.edgeTriangles.get(triangle.getEdge(i));
                        if (triangleList == null) {
                            triangleList = new SinglyLinkedList(triangle);
                            this.edgeTriangles.put(triangle.getEdge(i), triangleList);
                            
                            LineSegment voronoiEdge = 
                                (LineSegment)this.voronoiEdgesVis.remove(triangle.getEdge(i));

                            if (triangle.getEdge(i).getLine2D().relativeCCW(voronoiEdge.getP1()) > 0) {
                                this.voronoiEdgesVis.put(triangle.getEdge(i), 
                                                         this.visualElements.newVoronoiEdge(triangle.getCenter(), 
                                                                                            voronoiEdge.getP1()));
                            } else {
                                this.voronoiEdgesVis.put(triangle.getEdge(i), 
                                                         this.visualElements.newVoronoiEdge(triangle.getCenter(), 
                                                                                            voronoiEdge.getP2()));
                            }
                        } else {
                            triangleList.insert(triangle);
                            
                            if (triangleList.size() == 2) {
                                // Create a voronoi edge if the quadrilateral is correctly triangulated
                                Iterator triangles = triangleList.iterator();
                                Triangle triangle1 = (Triangle)triangles.next();
                                Triangle triangle2 = (Triangle)triangles.next();
        
                                this.voronoiEdgesVis.remove(triangle.getEdge(i));
        
                                for (int j = 0; j < 3; j++) {
                                    if ((!triangle1.getPoint(j).equals(triangle2.getPoint(0))) &&
                                        (!triangle1.getPoint(j).equals(triangle2.getPoint(1))) &&
                                        (!triangle1.getPoint(j).equals(triangle2.getPoint(2))) &&
                                        (GeometryFunctions.inCircleTest(triangle1.getPoint(j), triangle2.getPoint(0), triangle2.getPoint(1), triangle2.getPoint(2)) > 0)) {
                                        // The quadrilateral is incorrect, the vertex lies inside the circumcircle
                                        continue loop;
                                    }
                                }
                                
                                this.voronoiEdgesVis.put(triangle.getEdge(i), 
                                                         this.visualElements.newVoronoiEdge(triangle1.getCenter(), 
                                                                                            triangle2.getCenter()));
                            }
                        }
                    }
            } else {
                SinglyLinkedList triangleList;
                for (int i = 0; i < 3; i++) {
                    triangleList = (SinglyLinkedList)this.edgeTriangles.get(triangle.getEdge(i));
                    if (triangleList == null) {
                        triangleList = new SinglyLinkedList(triangle);
                        this.edgeTriangles.put(triangle.getEdge(i), triangleList);
                    } else {
                        triangleList.insert(triangle);
                    }
                }
            }
        }
    }
    
    
    private void removeTriangle(Triangle triangle) {
        
        this.voronoiVerticesVis.remove(triangle.getCenter());
        
        SinglyLinkedList ntriangles;
        this.trianglesVis.remove(triangle);
        for (int i = 0; i < 3; i++) {
            ntriangles = (SinglyLinkedList)this.edgeTriangles.get(triangle.getEdge(i));
            if (ntriangles != null) {
                ntriangles.delete(triangle);
                if (ntriangles.isEmpty()) {
                    this.edgeTriangles.remove(triangle.getEdge(i));
                }
            }
        }
    }
    
    
    private void deleteVertex(java.awt.Point vertex) {
        this.verticesVis.remove(vertex);
        Iterator edges = this.vertexEdges(vertex);
        Edge edge;
        while (edges.hasNext()) {
            edge = (Edge)edges.next();
            this.removeEdge(edge);
        }
    }
    
    public void flipEdge(java.awt.Point orig, java.awt.Point dest) {
        Edge edge = new Edge(orig, dest);
        if (this.hasEdge(edge)) {
            // Remove the edge and add a flipped edge if possible.
            SinglyLinkedList triangleList = (SinglyLinkedList)this.edgeTriangles.get(edge);
            if ((triangleList == null) || (triangleList.size() != 2)) {
                this.removeEdge(edge);
            } else {
                Iterator triangles = triangleList.iterator();
                java.awt.Point[] newPoints = new java.awt.Point[2];
                Triangle triangle = null;
                for (int i = 0; i < 2; i++) {
                    triangle = (Triangle)triangles.next();
                    for (int j = 0; j < 3; j++) {
                        if (triangle.getEdge(j).equals(edge)) {
                            newPoints[i] = triangle.getPoint((j + 2) % 3);
                        }
                    }
                }
                this.removeEdge(edge);
                if (triangle.intersects(new Edge(newPoints[0], newPoints[1]))) {
                    this.addEdge(newPoints[0], newPoints[1]);
                }
            }
        }
    }

    private void insertEdge(LineSegment edgeVis) {
        java.awt.Point orig = edgeVis.getP1();
        java.awt.Point dest = edgeVis.getP2();

        Edge edge = new Edge(orig, dest);
        this.edgesVis.put(edge, edgeVis);


        SinglyLinkedList edges = (SinglyLinkedList)this.vertexEdges.get(orig);
        if (edges == null) {
            edges = new SinglyLinkedList(edge);
            this.vertexEdges.put(orig, edges);
        } else {
            if (!edges.contains(edge)) {
                edges.insert(edge);
            }
        }
        
        edges = (SinglyLinkedList)this.vertexEdges.get(dest);
        if (edges == null) {
            edges = new SinglyLinkedList(edge);
            this.vertexEdges.put(dest, edges);
        } else {
            if (!edges.contains(edge)) {
                edges.insert(edge);
            }
        }

        createVoronoiEdge(orig, dest);
        
        // Destroy triangles that are broken by this edge and add
        // new triangles that this edge completes
        checkTriangles(edge);
    }

    private void removeTriangles(Edge edge) {
        SinglyLinkedList triangleList = (SinglyLinkedList)this.edgeTriangles.remove(edge);
        if (triangleList != null) {
            Iterator triangles = triangleList.iterator();
            Triangle triangle;
            while (triangles.hasNext()) {
                triangle = (Triangle)triangles.next();
                this.removeTriangle(triangle);
            }
        }
    }
    
    private void createVoronoiEdge(java.awt.Point orig, java.awt.Point dest) {
        Edge edge = new Edge(orig, dest);
        
        double[] midpoint = new double[2];
        midpoint[0] = orig.x + ((dest.x - orig.x) / 2);
        midpoint[1] = orig.y + ((dest.y - orig.y) / 2);
        double[] a = new double[2], b = new double[2];
        double dx = dest.x - orig.x;
        double dy = dest.y - orig.y;
        double length = Math.sqrt(dx * dx + dy * dy);
        // normalize
        int factor = (int)Math.ceil(Math.sqrt(this.width * this.width + this.height * this.height)) + 4;
        dx /= length; dy /= length;

        a[0] = midpoint[0] + factor * dy;
        a[1] = midpoint[1] - factor * dx;
        
        b[0] = midpoint[0] - factor * dy;
        b[1] = midpoint[1] + factor * dx;

        java.awt.Point vorOrig, vorDest;
        if (edge.getLine2D().relativeCCW(new java.awt.Point((int)a[0], (int)a[1])) > 0) {
            vorOrig = new java.awt.Point((int)b[0], (int)b[1]);
            vorDest = new java.awt.Point((int)a[0], (int)a[1]);
        } else {
            vorOrig = new java.awt.Point((int)a[0], (int)a[1]);
            vorDest = new java.awt.Point((int)b[0], (int)b[1]);
        }
        this.voronoiEdgesVis.put(edge, 
                                 this.visualElements.newVoronoiEdge(vorOrig, vorDest));
        
    }

    private void checkTriangles(Edge newEdge) {
        
        Iterator triangles = this.trianglesVis.iteratorKeys();
        Triangle triangle;
        while (triangles.hasNext()) {
            triangle = (Triangle)triangles.next();
            if (triangle.intersects(newEdge)) {
                this.removeTriangle(triangle);
            }
        }
        
        this.checkTriangle(newEdge);
    }

    
    private boolean deleteEdge(Edge edge) {
        if (this.edgesVis.remove(edge) == null) {
            return false;
        } else {
            this.voronoiEdgesVis.remove(edge);
            
            SinglyLinkedList edges = (SinglyLinkedList)this.vertexEdges.get(edge.getP1());
            edges.delete(edge);
            if (edges.isEmpty()) {
                this.vertexEdges.remove(edge.getP1());
            }
            
            edges = (SinglyLinkedList)this.vertexEdges.get(edge.getP2());
            edges.delete(edge);
            if (edges.isEmpty()) {
                this.vertexEdges.remove(edge.getP2());
            }
    
            this.removeTriangles(edge);
            
            Iterator allEdges = this.edgesVis.iteratorKeys();
            Edge e;
            while (allEdges.hasNext()) {
                e = (Edge)allEdges.next();
                this.checkTriangle(e);
            }
        
            return true;
        }
    }

    
    private void checkTriangle(Edge edge) {
        // Check if a new triangle was completed.
        SinglyLinkedList origEdgeList = (SinglyLinkedList)this.vertexEdges.get(edge.getP1());
        SinglyLinkedList destEdgeList;
        Iterator origEdges = origEdgeList.iterator();
        Iterator destEdges;
        Edge e1, e2;
        Triangle triangle;
        while (origEdges.hasNext()) {
            e1 = (Edge)origEdges.next();
            destEdgeList = (SinglyLinkedList)this.vertexEdges.get(edge.getP2());
            destEdges = destEdgeList.iterator();
            while (destEdges.hasNext()) {
                e2 = (Edge)destEdges.next();
                if (!e1.equals(edge) && !e2.equals(edge)) {
                    // Test if this pair of edges completes a triangle.
                    triangle = null;
                    if (e1.getP1().equals(e2.getP1())) {
                        triangle = new Triangle(e1.getP2(),
                                                e2.getP2(),
                                                e1.getP1(),
                                                true);
                    } else if (e1.getP2().equals(e2.getP1())) {
                        triangle = new Triangle(e1.getP1(),
                                                e2.getP2(),
                                                e1.getP2(),
                                                true);
                    } else if (e1.getP2().equals(e2.getP2())) {
                        triangle = new Triangle(e1.getP1(),
                                                e2.getP1(),
                                                e1.getP2(),
                                                true);
                    } else if (e1.getP1().equals(e2.getP2())) {
                        triangle = new Triangle(e1.getP2(),
                                                e2.getP1(),
                                                e1.getP1(),
                                                true);
                    }
                    if (triangle != null) {
                        // Check validity
                        Iterator edges;
                        Edge e;
                        boolean isValid = true;
                        edges = this.edgesVis.iteratorKeys();
                        while (edges.hasNext()) {
                            e = (Edge)edges.next();
                            if (!e.equals(triangle.getE0()) && 
                                !e.equals(triangle.getE1()) &&
                                !e.equals(triangle.getE2()) &&
                                triangle.intersects(e)) {
                                isValid = false;
                                break;
                            }
                        }
                        if (isValid) {
                            // Add the triangle for all the three edges that make the triangle.
                            addTriangle(triangle);
                        }
                    }

                }
            }
        }

    }
    
    
    //  ** Implementation of interface Area **

    public Rectangle getBoundingPolygon() {
        return new Rectangle(this.width, this.height, 0, 0);
    }


    public SpatialElement[] getSpatialElements() {
        Vector vector = new Vector();
        
        if (this.drawVoronoi == VORONOI_REGIONS) {
            // Draw voronoi regions
            Iterator vertices = this.vertexEdges.iterator();
            SinglyLinkedList edgeList;
            java.awt.geom.Area voronoi;
            LineSegment bs;
            Iterator edges;
            Edge edge;
            HashTable.Entry next;
            while (vertices.hasNext()) {
                // initialize the area with the complete area
                voronoi = new java.awt.geom.Area(new Rectangle(-10, -10, this.width + 20, this.height + 20));
                next =  (HashTable.Entry)vertices.next();
                edgeList = (SinglyLinkedList)next.getValue();
                edges = edgeList.iterator();
                while (edges.hasNext()) {
                    edge = (Edge)edges.next();
                    if ((!this.isTrianglesVoronoiDrawn()) || (this.edgeTriangles.get(edge) != null)) {
                        //bs = (PolyLine)this.bisectors.get(edge);
                        bs = (LineSegment)this.voronoiEdgesVis.get(edge);
                        if (bs != null) {
                            // divide the complete area with the bisector
                            // choose the area that has vertices[i] inside it
                            // instersect that area with the voronoi area cosntructed so far
                            Polygon[] polys = split(new Rectangle(-10, -10, this.width + 20, this.height + 20), 
                                                    bs.getLine2D());
                            for (int j = 0; j < polys.length; j++) {
                                if ((polys[j] != null) && polys[j].getShape().contains((java.awt.Point)next.getKey())) {
                                    voronoi.intersect(new java.awt.geom.Area(polys[j].getShape()));
                                }
                            }
                        }
                    }
                }
                PathIterator pi = voronoi.getPathIterator(null);
                double[] coords = new double[6];
                Vector c = new Vector();
                int type;
                double lastx=Double.MIN_VALUE, lasty=Double.MIN_VALUE;
                while (!pi.isDone()) {
                    type = pi.currentSegment(coords);
                    if ((type != PathIterator.SEG_CLOSE) &&
                        (new Point2D.Double(coords[0], coords[1]).distance(lastx, lasty) < 1)) {
                        pi.next();
                        continue;
                    }
                    if ((type == PathIterator.SEG_CLOSE) && (c.size() > 2)) {
                        java.awt.Point[] polypoints = new java.awt.Point[c.size()];
                        c.toArray(polypoints);
                        Polygon voronoiPolygon = this.visualElements.newVoronoiRegion(polypoints);
                        if (voronoiPolygon.contains(new Point((java.awt.Point)next.getKey()))) {
                            vector.add(voronoiPolygon);
                        }
                        c.clear();
                    }
                    c.add(new java.awt.Point((int)coords[0], (int)coords[1]));
                    lastx = coords[0]; lasty = coords[1];
                    pi.next();
                }
            }
        }

        // Vertices
        if (this.drawVertices) {
            Iterator vertices = this.vertices();
            while (vertices.hasNext()) {
                vector.add(vertices.next());
            }
        }
        
        if (this.drawDelaunay) {
            if (this.isOnlyCrustDrawn()) {
                // The crust
                Iterator iterator = this.edgesVis.iterator();
                Edge edge;
                HashTable.Entry entry;
                SinglyLinkedList triangleList;
                Triangle triangle1, triangle2;
                Iterator triangles;
                double testValue;
                while (iterator.hasNext()) {
                    entry = (HashTable.Entry)iterator.next();
                    edge = (Edge)entry.getKey();
                    triangleList = (SinglyLinkedList)this.edgeTriangles.get(edge);
                    if (triangleList != null) {
                        triangles = triangleList.iterator();
                        if (triangleList.size() == 2) {
                            triangle1 = (Triangle)triangles.next();
                            triangle2 = (Triangle)triangles.next();
                            testValue = GeometryFunctions.inCircleTest(triangle1.getCenter(), 
                                                                       edge.getP1(), 
                                                                       edge.getP2(), 
                                                                       triangle2.getCenter());
                            if (testValue > 0) {
                                // the point is inside -> anti-crust
                                if (this.drawVoronoi == VORONOI_EDGES) {
                                    vector.add(this.voronoiEdgesVis.get(edge));
                                }
                            } else {
                                // the point is outside -> crust
                                vector.add(entry.getValue());
                            }
                        } else if (triangleList.size() == 1) {
                            triangle1 = (Triangle)triangles.next();
                            if (edge.getLine2D().relativeCCW(triangle1.getCenter()) < 0) {
                                // the voronoi edge intersects the corresponding delaunay edge
                                // -> crust
                                vector.add(entry.getValue());
                            } else if (this.drawVoronoi == VORONOI_EDGES) {
                                // anti-crust
                                vector.add(this.voronoiEdgesVis.get(edge));
                            }
                        }
                    }
                }
            } else {
                Iterator iterator = this.edgesVis.iteratorValues();
                while (iterator.hasNext()) {
                    vector.add(iterator.next());
                }   
                if (this.drawVoronoi == VORONOI_EDGES) {
                    iterator = this.voronoiEdgesVis.iteratorValues();
                    while (iterator.hasNext()) {
                        vector.add(iterator.next());
                    }
                }
            }

            if (this.isVoronoiVerticesDrawn()) {
                Iterator iterator = this.voronoiVerticesVis.iteratorValues();
                if (this.isOnlyCrustDrawn()) {
                    while (iterator.hasNext()) {
                        vector.add(iterator.next());
                    }
                } else {
                    while (iterator.hasNext()) {
                        vector.add(iterator.next());
                    }
                }
            }

            if (this.isTrianglesDrawn()) {
                Iterator iterator = this.trianglesVis.iteratorValues();
                while (iterator.hasNext()) {
                    vector.add(iterator.next());
                }
            }
            
        }        

        SpatialElement[] spatialElements = new SpatialElement[vector.size()];
        vector.toArray(spatialElements);
        return spatialElements;
    }
    
    private Polygon[] split(Rectangle r, Line2D l) {
        Line2D top = new Line2D.Double(r.getMinX(), r.getMaxY(), r.getMaxX(), r.getMaxY());
        Line2D bottom = new Line2D.Double(r.getMinX(), r.getMinY(), r.getMaxX(), r.getMinY());
        Line2D left = new Line2D.Double(r.getMinX(), r.getMaxY(), r.getMinX(), r.getMinY());
        Line2D right = new Line2D.Double(r.getMaxX(), r.getMaxY(), r.getMaxX(), r.getMinY());
        java.awt.Point is, is2;
        java.awt.Point tl = new java.awt.Point((int)r.getMinX(), (int)r.getMaxY());
        java.awt.Point tr = new java.awt.Point((int)r.getMaxX(), (int)r.getMaxY());
        java.awt.Point br = new java.awt.Point((int)r.getMaxX(), (int)r.getMinY());
        java.awt.Point bl = new java.awt.Point((int)r.getMinX(), (int)r.getMinY());
        Polygon[] poly = new Polygon[2];
        if (top.intersectsLine(l)) {
            is = GeometryFunctions.lineIntersection(top, l);
            if (bottom.intersectsLine(l)) {
                is2 = GeometryFunctions.lineIntersection(bottom, l);
                poly[0] = new Polygon(new java.awt.Point[]{tl, is, is2, bl});
                poly[1] = new Polygon(new java.awt.Point[]{tr, is, is2, br});
            } else if (left.intersectsLine(l)) {
                is2 = GeometryFunctions.lineIntersection(left, l);
                poly[0] = new Polygon(new java.awt.Point[]{tl, is, is2});
                poly[1] = new Polygon(new java.awt.Point[]{tr, is, is2, bl, br});
            } else /*if (right.intersectsLine(l))*/ {
                is2 = GeometryFunctions.lineIntersection(right, l);
                poly[0] = new Polygon(new java.awt.Point[]{tr, is, is2});
                poly[1] = new Polygon(new java.awt.Point[]{tl, is, is2, br, bl});
            }
        } else if (bottom.intersectsLine(l)) {
            is = GeometryFunctions.lineIntersection(bottom, l);
            if (left.intersectsLine(l)) {
                is2 = GeometryFunctions.lineIntersection(left, l);
                poly[0] = new Polygon(new java.awt.Point[]{bl, is, is2});
                poly[1] = new Polygon(new java.awt.Point[]{br, is, is2, tl, tr});
            } else /*if (right.intersectsLine(l))*/ {
                is2 = GeometryFunctions.lineIntersection(right, l);
                poly[0] = new Polygon(new java.awt.Point[]{br, is, is2});
                poly[1] = new Polygon(new java.awt.Point[]{bl, is, is2, tr, tl});
            }
        } else /*if (left.intersectsLine(l))*/ {
            is = GeometryFunctions.lineIntersection(left, l);
            is2 = GeometryFunctions.lineIntersection(right, l);
            poly[0] = new Polygon(new java.awt.Point[]{is, is2, tr, tl});
            poly[1] = new Polygon(new java.awt.Point[]{is, is2, br, bl});
        } /*else if (right.intersectsLine(l)) {
            is[3] = getIntersection(right, l);
        }*/
        return poly;
    }
    
    public boolean hasPaintingStyleDecorator() {
        return false;
    }

    public PaintingStyleDecorator getPaintingStyleDecorator() {
        Note.warning(this, "Operation not implemented.");
        return null;
    }

    public void insert(SpatialComparable e) {
        Note.warning(this, "Operation not implemented.");
    }

    public Area[] getSubAreas() {
        return new Area[]{};
    }

    public void removeSubArea(Area a) {
        Note.warning(this, "Operation not implemented.");
    }

    public void setPaintingStyleDecorator(PaintingStyleDecorator psd) {
        Note.warning(this, "Operation not implemented.");
    }

    public Object getElement() {
        return this;
    }

    public void setElement(Object o) {
        Note.warning(this, "Operation not implemented.");
    }

    public void addSubArea(Area a) {
        Note.warning(this, "Operation not implemented.");
    }

    public void delete(SpatialComparable e) {
        Note.warning(this, "Operation not implemented.");
    }

    
    
    static public class VisualElements implements Serializable {
        
        public VisualElements() {}
        
        protected Point newVertex(java.awt.Point vertex) {
            return new Point(vertex, new PaintingStyleDecorator());
        }
        
        protected LineSegment newDelaunayEdge(java.awt.Point orig, java.awt.Point dest) {
            return new LineSegment(orig, dest, new PaintingStyleDecorator());
        }
        
        protected matrix.structures.spatial.FDT.probe.Triangle newTriangle(java.awt.Point[] vertices) {
            return new matrix.structures.spatial.FDT.probe.Triangle(vertices, new PaintingStyleDecorator());
        }
        
        protected Point newVoronoiVertex(java.awt.Point vertex) {
            return new Point(vertex, new PaintingStyleDecorator());
        }
        
        protected LineSegment newVoronoiEdge(java.awt.Point orig, java.awt.Point dest) {
            return new LineSegment(orig, dest, new PaintingStyleDecorator());
        }
        
        protected Polygon newVoronoiRegion(java.awt.Point[] vertices) {
            return new Polygon(vertices, new PaintingStyleDecorator());
        }

        protected void hideDelaunay() {}
        
        protected void showDelaunay() {}

        protected void hideVoronoi() {}
        
        private static final long serialVersionUID = 6662775633832999400L;
    }        
    
    private class Edge implements Serializable {
        
        java.awt.Point orig;
        java.awt.Point dest;
        
        public Edge(java.awt.Point orig, java.awt.Point dest) {
            this.orig = orig;
            this.dest = dest;
        }
        
        public Line2D getLine2D() {
            // Line2D.Double is not serializable and therefore it's created on the fly,
            // otherwise this class would extend or store a Line2D.Double object.
            return new Line2D.Double(this.orig.x, this.orig.y, this.dest.x, this.dest.y);
        }
        
        public java.awt.Point getP1() {
            return this.orig;
        }

        public java.awt.Point getP2() {
            return this.dest;
        }

        public boolean equals(Object obj) {
            if (obj instanceof Edge) {
                if (this == obj) {
                    return true;
                } else {
                    Edge edge = (Edge)obj;
                    return ((this.getP1().equals(edge.getP1()) &&
                             this.getP2().equals(edge.getP2())) ||
                            (this.getP1().equals(edge.getP2()) &&
                             this.getP2().equals(edge.getP1())));
                }
            } else {
                return false;
            }
        }
        
        public int hashCode() {
            java.awt.Point midp = 
                new java.awt.Point((int)(this.getP1().x + (this.getP2().getX() - this.getP1().x) / 2),
                                   (int)(this.getP1().y + (this.getP2().getY() - this.getP1().y) / 2));
            return new StringBuffer().append(midp.x).append('-').append(midp.y).toString().hashCode();
        }
        
        public String toString() {
            return "(" + this.orig.getX() + "," + this.orig.getY() + ")->(" + this.dest.getX() + "," + this.dest.getY() + ")";
        }
        
        private static final long serialVersionUID = 4641442618711605928L;
    }
    
    private class Triangle implements Serializable {

        private java.awt.Point circumcircleCenter;
        private java.awt.Point v0, v1, v2;
        private Edge e0, e1, e2;

        // ** Constructors **

        // ccw-ordered vertices
        public Triangle(java.awt.Point v1, java.awt.Point v2, java.awt.Point v3) {
            this(v1, v2, v3, false);
        }

        public Triangle(java.awt.Point v0, java.awt.Point v1, java.awt.Point v2, boolean order) {
            java.awt.Point[] vertices = new java.awt.Point[]{v0, v1, v2};
            if (order) {
                GeometryFunctions.ccwOrder(vertices);
            }
            this.e0 = new Edge(vertices[0], vertices[1]);
            this.e1 = new Edge(vertices[1], vertices[2]);
            this.e2 = new Edge(vertices[2], vertices[0]);
            Ellipse2D circumcircle = GeometryFunctions.circumcircle(vertices);
            this.circumcircleCenter = new java.awt.Point((int)circumcircle.getCenterX(), 
                                                         (int)circumcircle.getCenterY());
            this.v0 = v0;
            this.v1 = v1;
            this.v2 = v2;
        }

        // ** Getters **
        
        public java.awt.Point getCenter() {
            return this.circumcircleCenter;
        }

        public Edge getE0() {
            return this.e0;
        }

        public Edge getE1() {
            return this.e1;
        }

        public Edge getE2() {
            return this.e2;
        }
        
        public java.awt.Point getPoint(int pointNumber) {
            switch (pointNumber) {
                case 0 :
                    return this.getE0().getP1();
                case 1 : 
                    return this.getE1().getP1();
                case 2 : 
                    return this.getE2().getP1();
                default : 
                    return null;
            }
        }
        
        public java.awt.Point[] getPoints() {
            return new java.awt.Point[]{ this.getPoint(0), this.getPoint(1), this.getPoint(2) };
        }
        
        public Edge getEdge(int edgeNumber) {
            switch (edgeNumber) {
                case 0 :
                    return this.getE0();
                case 1 : 
                    return this.getE1();
                case 2 : 
                    return this.getE2();
                default : 
                    return null;
            }
        }
        
        public boolean intersects(Edge edge) {
            GeneralPath triangle = new GeneralPath();
            triangle.moveTo(this.v0.x, this.v0.y);
            triangle.lineTo(this.v1.x, this.v1.y);
            triangle.lineTo(this.v2.x, this.v2.y);
            triangle.closePath();
            for (int i = 0; i < 3; i++) {
                if (this.getPoint(i).equals(edge.getP1())) {
                    return (triangle.contains(edge.getP2()) ||
                            this.getEdge((i + 1) % 3).getLine2D().intersectsLine(edge.getLine2D()));
                } else if (this.getPoint(i).equals(edge.getP2())) {
                    return (triangle.contains(edge.getP1()) ||
                            this.getEdge((i + 1) % 3).getLine2D().intersectsLine(edge.getLine2D()));
                }
            }
            return (this.getE0().getLine2D().intersectsLine(edge.getLine2D()) ||
                    this.getE1().getLine2D().intersectsLine(edge.getLine2D()) ||
                    this.getE2().getLine2D().intersectsLine(edge.getLine2D()) ||
                    (triangle.contains(edge.getP2()) &&
                     triangle.contains(edge.getP1())));
        }
        

        public String toString() {
            return "{ " + this.getE0() + " -> " + this.getE1() + " -> " + this.getE2() + " }";
        }

        public boolean equals(Object obj) {
            if (obj instanceof Triangle) {
               if (this == obj)  {
                   return true;
               } else {
                   Triangle triangle = (Triangle)obj;
                   if (((this.getE0().equals(triangle.getE0())) &&
                        (this.getE1().equals(triangle.getE1())) &&
                        (this.getE2().equals(triangle.getE2()))) ||
                       ((this.getE0().equals(triangle.getE1())) &&
                        (this.getE1().equals(triangle.getE2())) &&
                        (this.getE2().equals(triangle.getE0()))) ||
                       ((this.getE0().equals(triangle.getE2())) &&
                        (this.getE1().equals(triangle.getE0())) &&
                        (this.getE2().equals(triangle.getE1())))) {
                       return true;
                   } else {
                       return false;
                   }
               }
            } else {
                return false;
            }
        }
        
        public int hashCode() {
            return this.getE0().hashCode() + this.getE1().hashCode() + this.getE2().hashCode();
        }
        
        private static final long serialVersionUID = 4386273839625701398L;
        
    }
    
    private static final long serialVersionUID = 2561138936837404202L;
    
}
