// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/* $Id: ExerHeap.java,v 1.14 2007/09/10 08:09:05 vkaravir Exp $ */

package content.exercises.structures;

import matrix.structures.FDT.BinaryTree;
import matrix.structures.FDT.probe.*;
import matrix.structures.memory.*;
import matrix.structures.util.MatrixComparable;
import matrix.util.Note;

/**
 * This is an implementation of a binary tree. A binary tree is a
 * tree in which each node has two children (either of which can be null).
 */
public class ExerHeap extends BinTrei {
    
    public static final boolean DEBUG = false;
    public static final String EMPTY = " ";
    protected VirtualInteger keys = new VirtualInteger(0, this, "Key count");
    
    // these variables are for checking the validity of exercise input
    private boolean collision = false; // this one-way-flag is set true if duplicates collide in insertions
    private int recursions = 0; // #recursions during all deletions
    private int max_recurions_during_single_call = 0;
    private int singleCounter = 0; // help variable for the previous one
    private boolean interrupted = false; // this one-way-flag is set true if minHeapify do not reach a leaf
    private boolean equalChildren = false; // this one-way-flag is set true if equal children are encountered during minHeapify

    public ExerHeap(int n) {
        super(n);
    }
    
    
    public ExerHeap(Object[] objs) {
        super(objs);
    }
    
    
    // Private constructor for compatibility with BinaryTree
    private ExerHeap(VirtualArray bt, int curr) {
        this.bt = bt;
        this.curr = curr;
    }
    
    /**
     * This method is used to check validity of an exercise.
     * @return true if collision occurs during insertions (two equal keys are compared with each other)
     */
    public boolean hasCollision() {
        return collision; 
    }
    
    /**
     * This method is used to check validity of an exercise.
     * @return the number of recursive calls for minHeapify during deletions
     */
    public int getNumberOfRecursions() {
        return recursions; 
    }
    
    /**
     * This method is used to check validity of an exercise.
     * @return the number of recursive calls for minHeapify during a single deletion
     */
    public int getMaxNumberOfRecursionsDuringSingleCall() {
        return max_recurions_during_single_call; 
    }
    
    /**
     * This method is used to check validity of an exercise.
     * @return true if collision occurs during insertions (two equal keys are compared with each other)
     */
    public boolean hasInterrupted() {
        return interrupted; 
    }
    
    /**
     * This method is used to check validity of an exercise.
     * @return true if equal children are encountered during minHeapify
     */
    public boolean hasEqualChildren() {
        return equalChildren; 
    }
    
    /**
     * Constructs a heap in a bottom-up
     * manner to convert an array to heap
     */
    /*
     public void BuildHeap() {
     VirtulArray bt = (VirtualArray)this.getArray();
     for (int i =bt.getLast()/2; i >=0; i--)
     bubbleDown(i);
     }
     */
    
    public int leftChild(int n) {
        if(n == 0) return 1;
        return 2 * n + 1;
    }
    
    public int rightChild(int n) {
        if(n == 0) return 2;
        return 2 * n + 2;
    }
    
    public int parent(int n) {
        if (n == 0) return 0;
        return (n-1)/2;
    }
    
    /**
     * This method returns true if the whole heap fills the Min-Heap property
     * 
     * @return true if the structure fills the Min-Heap property, false otherwise.
     */
    public boolean isMinHeap(){
        for (int i=1; i<getSize(); i++){
            MatrixComparable p      = (MatrixComparable)getObject(i);
            MatrixComparable father = (MatrixComparable)getObject(parent(i));
            
            if (p.lt(father))
                return false;
        }
        return true;
    }
    
    public void minHeapify(int i) {
        recursions++;
        singleCounter++;
        if (DEBUG){
            Note.out(this, "Heapify: " + i);
            Note.out(this, "lc: " + leftChild(i) + " rc: " + rightChild(i));
        }
        // if(i >= keys)
        if(i >= getSize())
            return;
        MatrixComparable p = (MatrixComparable)getObject(i);
        MatrixComparable lc = (MatrixComparable)getObject(leftChild(i));
        /*
         if(leftChild(i) >= keys) {
         lc = new Key(EMPTY);
         }
         */
        MatrixComparable rc = (MatrixComparable)getObject(rightChild(i));
        /*
         if(rightChild(i) >= keys) {
         rc = new Key(EMPTY);
         }
         */
        // This code handles situations where the father node is null or empty
        // such a situation never happens in normal cases, but intentionally broken
        // algorithms might have such states.
        
        if (p==null || p.equals(EMPTY)){ 
            
            if ((lc == null || lc.equals(EMPTY)) &&
                    (rc == null || rc.equals(EMPTY)))
                return;
            if (lc == null || lc.equals(EMPTY)){
                swap(i, rightChild(i));
                minHeapify(rightChild(i));
                return;
            }
            
            if (rc == null || rc.equals(EMPTY)){
                swap(i, leftChild(i));
                minHeapify(leftChild(i));
                return;
            }
            
            if(lc.gt(rc)) {                
                swap(i, rightChild(i));
                minHeapify(rightChild(i));
                return;
            }
            else{                
                swap(i, leftChild(i));
                minHeapify(leftChild(i));
                return;
            }
        } // null-father code ends
        
        if(lc == null || lc.equals(EMPTY))
            return;
        if(rc == null || rc.equals(EMPTY)) { 
            if(p.gt(lc) && (lc!=null && !lc.equals(EMPTY))) {
                swap(i, leftChild(i));
                minHeapify(leftChild(i));
            } else interrupted = true;
            return;
        }
        if(lc.gt(rc)) {
            if(p.gt(rc) && (rc!=null && !rc.equals(EMPTY))) {
                swap(i, rightChild(i));
                minHeapify(rightChild(i));
            } else interrupted = true;
        } else {
            if (lc.eq(rc))
                equalChildren = true;
            if(p.gt(lc) && (lc!=null && !lc.equals(EMPTY))) {
                swap(i, leftChild(i));
                minHeapify(leftChild(i));
            } else interrupted = true;
        }
    }
    
    /*
     * Insert a new object into this heap.
     * @return the number of swaps performed 
     */
    public int insert(Object o) {
        setObject(o, keys.eval());
        keys.inc();
        return bubble(keys.eval()-1);
        
    }
    
    public void deleteMin() {
        MatrixComparable tmp = (MatrixComparable)getObject(keys.eval() - 1);
        setObject(tmp, 0);
        setObject(new Key(EMPTY), keys.eval() - 1);
        keys.dec();
        singleCounter = 0;
        minHeapify(0);
        if (singleCounter > max_recurions_during_single_call)
            max_recurions_during_single_call = singleCounter;
    }
    
    /**
     * Empties the root node.
     * Does not "bubble" or swap.
     */
    public void deleteRoot() {
        //Note.out(this, "deleteRoot "+bt);
        setElement(new Key(EMPTY));
        //Note.out(this, "deleted "+bt);
        /*	Key k = new Key(" ");
         setObject(k, 0);*/
    }
    
    /**
     * Deletes the last element from this heap and decreases the heapsize.
     * Does not "bubble" or swap.
     */
    public void decrementHeapsize() {
        //Note.out(this, "decrease Heapsize "+bt);
        if (keys.eval() > 0) {
            setObject(new Key(EMPTY), keys.eval() - 1);
            keys.dec();
        } else
            Note.warning(this, "Cannot decrement anymore (heapsize == 0)");
        /*  Key k = new Key(" ");
         setObject(k, 0);*/
    }
 
    /*
     * Called by inner class DeletableKey in order to delete key.
     * Search and delete the key that invoked this method.
     */
    public void deleteRoot(Key k) {
        if (k == getElement())
            deleteRoot();
        else {
            synchronized ( getClass() ) { // the following temporarily change the inner state
                int tmp = curr; // backup inner state
                for (curr=0; curr<size(); curr++)
                    if (k == getElement()) {
                        deleteRoot();
                        curr=tmp; // restore inner state
                        return;
                    }
                }
            matrix.util.Note.warning(this, "Not found "+k);
        }
    }
    
    BinaryTree left = null;
    public BinaryTree getLeft() {
        BinaryTree nbt = (left == null) ? left = new ExerHeap(bt, leftChild(curr)) : left;
        return (nbt.getElement() == null) ? null : nbt;
    }
    
    BinaryTree right = null;
    public BinaryTree getRight() {
        BinaryTree nbt = (right == null) ? right = new ExerHeap(bt, rightChild(curr)) : right;
        return (nbt.getElement() == null) ? null : nbt;
    }
    
    // return the number of swaps done
    protected int bubble(int n) {
        MatrixComparable p = (MatrixComparable)getObject(parent(n));
        MatrixComparable t = (MatrixComparable)getObject(n);
        
        if ((n != parent(n)) & (p!=null) && (p.eq(t)))
            collision = true;
        if (p == null || p.gt(t)) {
            swap(n,parent(n));
            return bubble(parent(n)) + 1;
        } else return 0;
    }
    
    public class DeletableKey extends Key {
        /**
         * Keys can be wrapped into this in order to be deleted through reflection
         */
        private static final long serialVersionUID = 1L;

        public DeletableKey(Object o) {
            super(o);
        }
        
        // called by reflection in order to delete this key from the data structure
        public void deleteRoot() {
            //matrix.util.Note.out(null, "deleteRoot for key "+this);
            ExerHeap.this.deleteRoot(this);
        }
        
        // called by reflection in order to delete this key from the data structure
        public void decrementHeapsize() {
            //matrix.util.Note.out(null, "decrement heapsize invoked for key "+this);
            ExerHeap.this.decrementHeapsize();
        }
    }
    
    static final long serialVersionUID = 141322807213670974L;
}
