// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
package content.exercises.structures;

import java.awt.Color;

import matrix.decoration.StyleSheet;
import matrix.decoration.StyleSheetAdapter;
import matrix.structures.FDT.FDT;
import matrix.structures.FDT.probe.Key;
import matrix.structures.FDT.substructures.Vertex;
import matrix.structures.memory.VirtualInteger;
import matrix.structures.memory.VirtualObject;

/**
 * This is a priority vertex class that is used in the Dijkstra_Heap exercise.
 * 
 * @author kstalnac
 *
 */
public class PriorityExerVertex_DH extends PriorityExerVertex {

    /** Distance of the this node from the start node of the Dijkstra's algorithm. */
    private VirtualInteger distanceOfNode = new VirtualInteger(Integer.MAX_VALUE, this, "Distance from the start node");

    /**
     * Returns the label for this node.
     * @return '-' if the distance of the node is Integer.MAX_VALUE, else it returns 
     * the actual distance of the node 
     */
    public String getLabel() { 
        if (distanceOfNode.eval() == Integer.MAX_VALUE) {
            return "-";
        } else {
            return distanceOfNode.eval() + "";
        }
    }

    /** Is reference labels visible or not. */
    private boolean isReferenceLabelEnabled = false;

    /**
     * Returns whether the reference labels are visible or not. 
     */
    public boolean isReferenceLabelEnabled() {  
        return isReferenceLabelEnabled;
    }
    
    /**
     * Sets the visibility of the reference labels.
     */
    public void setReferenceLabelEnabled(boolean b) {
        isReferenceLabelEnabled = b;
    }

    /**
     * Returns the reference label of the given edge.
     */
    public String getReferenceLabel(int i) {
        if (i<children.eval())
            return ""+getWeight((PriorityExerVertex)spanningtree.getObject(i));
        else if (children.eval() == 0) {
            if (succ == null)
                succ = getAdjacentVertices(); //get adjacent vertices in the succ variable
            return ""+getWeight((PriorityExerVertex)succ[i]);
        }
        return "?";
    }

    /**
     * Set the distance of the node from the start node.
     * @param distance distance of this node from the start node
     */
    public void setDistanceOfNode(int distance) {
        distanceOfNode.assign(distance);
    }
    
    /**
     * Returns the distance of this node from the start node.
     * @return the distance of this node from the start node, 
     * returns -1 if distance hasn't yet been decided.
     */
    public int getDistanceOfNode() {
        if (distanceOfNode.eval() == Integer.MAX_VALUE) {
            return -1;
        } else {
            return distanceOfNode.eval();
        }
    }

    /**
     * Returns the successor node which key corresponds to the given key.
     * @param key key which node is wanted
     * @return node which key corresponds to the given key
     */
    public PriorityExerVertex_DH getSuccessor(Key key) {
        if (succ == null) {
            succ = getAdjacentVertices();
        }
        for (int i = 0; i < succ.length; i++) {
            if(((Key)succ[i].getElement()).equals(key))
                return (PriorityExerVertex_DH)succ[i];
        }
        return null;
    }
    
    /**
     * Set color of the certain reference..
     * @param i number of the reference
     * @param color new color for the reference
     */
    public void setReferenceColor(int i, Color color) {
        edgeColor.setObject(color, i);
    }
    
    /**
     * This method finds out whether the reference has already been visited.
     * The reference is colored with the different color according to information
     * of this method.
     */
    public boolean isVisitedReferenceTo(FDT to) {
        Vertex[] v = super.getSuccessors();
        for (int i=0; i<v.length; i++)
            if (v[i] == to) {
                Object c = edgeColor.getObject(i);
                return c != null ? (c.equals(Color.black) || c.equals(Color.gray)) : false;
            }
        return false;
    }
    
    
    /** 
     * This object saves the information about the father node from which the spanning 
     * tree extends or is going to extend to this node.  */
    private VirtualObject father = new VirtualObject(this, "Father of this node in the spanning tree");

    /**
     * Set father of this node in the spanning tree.
     * @param f father of this node in the spanning tree
     */
    public void setFrom(PriorityExerVertex_DH f) {
        father.setObject(f);
    }
    
    /**
     * Get father of this node in the spanning tree.
     * @return father of this node in the spanning tree
     */
    public PriorityExerVertex_DH getFrom() {
        return (PriorityExerVertex_DH)father.getObject();
    }

    /** StyleSheet for this node */
    private StyleSheet ss = null;
    
    public StyleSheet getStyleSheet() {
        if(ss == null)
            ss = new StyleSheetAdapter() {
            
            /**
             * Returns the color of the selected reference.
             */
            public Color getModifiedReferenceColor(FDT f) {
                Vertex[] vertices = getAdjacentVertices();
                for (int i = 0; i < vertices.length; i++) {
                    if (vertices[i] == f) {
                        Object c = edgeColor.getObject(i);
                        if (c.equals(Color.black)) {
                            return Color.black;
                        } else {
                            return Color.gray;
                        }
                    }
                }
                return Color.blue; //default value returned if the given fdt is not adjacent to this node
            }
            
            private static final long serialVersionUID = -7416534679517446047L;
        };
     
        ss.setDefaultBackgroundColor(getColor());
        return ss;

    }
    private static final long serialVersionUID = -1494636103871789615L;
}
