// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/* $Id: AnimatorStepMapper.java,v 1.3 2007/09/07 13:32:45 vkaravir Exp $ */

package matrix.animation;

/**
 * The {@link Animator} class has "folding" abilities, and each marker operation
 * conforms an state by itself. So generally, there are always more states than
 * steps (e.g. an operation performed by the user on the visual structures).
 * This class maps associates the two concepts: it provides a way to find the
 * step to which a state belongs and vice versa.
 * 
 * @author Isaac Jurado
 */
public class AnimatorStepMapper {
    
    private Animator animator;
    private int[]    stepFromState;
    private int[]    stateFromStep;
    private int      stepCount;
    private int      stateCount;
    
    /**
     * Creates a new mapper from a given animator.
     * 
     * @param a
     *            The {@link Animator} from which states and steps have to be
     *            retrieved and associated.
     */
    public AnimatorStepMapper(Animator a) {
        a.rewind();
        stateCount = a.getMaximum() + 1;
        stepCount  = 1;
        while (a.hasNextOperation()) {
            a.redo();
            stepCount++;
        }
        
        stepFromState = new int[stateCount];
        stateFromStep = new int[stepCount];
        
        int step = 0, state;
        a.rewind();
        // It's a lame loop construct, but it avoids code duplication
        while (true) {
            state = a.getState();
            stepFromState[state] = step;
            stateFromStep[step]  = state;
            if (!a.hasNextOperation())
                break;
            a.redo();
            step++;
        }
        
        // As there are more states than steps (at least, three times more),
        // the gaps in the stepFromState array have to be filled.
        // The strategy is to propagate the steps backwards
        for (int i = stateCount - 1; i > 0; i--)
            if (stepFromState[i] > 0)
                step = stepFromState[i];
            else
                stepFromState[i] = step;
        
        animator = a;
    }
    
    /**
     * Get the total number of states of the animator.
     * 
     * @return the number of states.
     */
    public int getStateCount() {
        return stateCount;
    }
    
    /**
     * Get the total number of steps of the animator.
     * 
     * @return the number of steps.
     */
    public int getStepCount() {
        return stepCount;
    }
    

    /**
     * Get the last step of the animator.
     * 
     * @return the step number.
     */
    public int getLastStep() {
        return stepCount - 1;
    }
    
    /**
     * Get the last state of the animator.
     * 
     * @return the state number
     */
    public int getLastState() {
        return stateCount - 1;
    }

    /**
     * Get the current state of the animator. This method is only provided for
     * convenience
     * 
     * @return the current state.
     */
    public int getCurrentState() {
        return animator.getState();
    }

    /**
     * Get the current step of the animator. Convenience method to directly
     * retrieve the current step without having to explicitly convert.
     * 
     * @return the current step
     */
    public int getCurrentStep() {
        return stepFromState[animator.getState()];
    }

    /**
     * Convert the given state to its corresponding step.
     * 
     * @param state
     *            the sate to be converted.
     * @return the corresponding step.
     */
    public int getStepFromState(int state) {
        return stepFromState[state];
    }

    /**
     * Convert the given step to its corresponding first state. It wouldn't be
     * safe to return other state because the operation folding could be screwed
     * up.
     * 
     * @param step
     *            the step to ve converted.
     * @return the fist state of the step.
     */
    public int getStateFromStep(int step) {
        return stateFromStep[step];
    }

}
