// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package matrix.simulation;

import java.util.Hashtable;
import java.util.Properties;

/**
   A class for configuring visual types.
   @author Panu Silvasti
 */
public class VisualTypeConf {

    /* User interface operations that can be configured for VisualTypes */
    public static final int DROP_OPERATION = 1;
    public static final int DRAG_OPERATION = 2;
    public static final int HIGHLIGHT_OPERATION = 4;
    public static final int POP_UP_MENU_OPERATION = 8;
    public static final int MINIMIZED = 16;

    /* This Hashtable contains configuration information of user interface 
     * operations for VisualTypes. */
    private Hashtable opHash = new Hashtable();

    
    /* Other properties that can be configured through VisualTypeConf */

    /* VisualArray properties */
    public static final String INDEXED = "indexed";
    public static final String RANGE_SELECTION = "range";
    
    /* VisualContainer properties */
    
    public static final String FLIPPEDY = "flippedy";
    public static final String FLIPPEDX = "flippedx";
    public static final String ROTATED = "rotated";

    /* VisualArea properties */
    public static final String AXIS_VISIBLE = "axis_visible";
    public static final String X_AXIS_VISIBLE = "x_axis_visible";
    public static final String Y_AXIS_VISIBLE = "y_axis_visible";

    
    /* This Hashtable contains all the other properties that can be configured
     * for VisualTypes. The properties and their values for certain VisualType
     * are saved inside a java.util.Properties object which is then saved to
     * this Hashtable by using the name of the VisualType class as a key.*/
    private Hashtable visualConfs = new Hashtable();

    /**
       Enables user interface operations for certain visual types.
     */
    public void enable(String visualTypeClass, int operation) {
        // XXX Should check that visualTypeClass exists!
        Object op = opHash.get(visualTypeClass);
        if (op == null) {
            opHash.put(visualTypeClass, new Integer(operation));
        } else if ((((Integer)op).intValue() & operation)==0) {
            opHash.put(visualTypeClass, new Integer(((Integer)op).intValue()+operation));
        }
    }
    
    
    /** 
	Is given operation enabled for given visual type class.
    */
    public boolean isEnabled(String visualTypeClass, int operation) {
        Object op = opHash.get(visualTypeClass);
        if (op == null) {
            return false;
        } else {
            int opInt = ((Integer)op).intValue();
            // bitwise and operation
            return (opInt & operation) != 0;
        }
    }
    
    /**
     * Sets the value of the property for given VisualType class.
     *  
     * @param visualTypeClass name of the VisualType class
     * @param property property which value is being set
     * @param value value of the property
     */
    public void setVisualConfValue(String visualTypeClass, String property, String value) {
        Properties props = (Properties)visualConfs.get(visualTypeClass);
        if (props == null) {
            props = new Properties();
            props.setProperty(property, value);
            visualConfs.put(visualTypeClass, props);
        } else {
            props.setProperty(property, value);
        }
    }

    /**
     * Returns the value of the property of given VisualType class.
     * 
     * @param visualTypeClass name of the VisualType class
     * @param property property which value is wanted
     * @return Value of the property of given VisualType class. Returns null, if
     * there are no properties found at all for given VisualType class or if the 
     * given property is not set for VisualType class.
     */
    public String getVisualConfValue(String visualTypeClass, String property) {
        Properties props = (Properties)visualConfs.get(visualTypeClass);
        if (props == null) {
            //No Properties object found for VisualType class. Value is not fixed.
            return null;
        } else {
            return props.getProperty(property);
        }
    }
    
}
