// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package matrix.structures.CDT.probe;

import matrix.structures.CDT.CDT;
import matrix.structures.FDT.*;
import matrix.structures.FDT.probe.Key;
import matrix.structures.util.MatrixComparable;

/**
 * <p>Implementation of the RedBlackTree. This is implemented by encapsulating
 * <code>RedBlackNodes</code> inside this <code>RBTree</code>.
 * RebBlackTree is balanced binary search tree with following properties:
 * <ul>
 * <li> Root is black.</li>
 * <li> Every route from root to leaf contains same number of black nodes. </li>
 * <li> Red node can't have red childrens.</li>
 * </ul></p>
 *
 * <p>Every node has parent information so when using
 * methods inherited from <code>BinSearchTree</code> parent information may also need
 * some updating. <code>Rotations</code> class can be used with this class if parents
 * fixing method is called before rotation. These methods are:
 * <ul>
 * <li><code>newParentsZigRight()</code></li>
 * <li><code>newParentsZigLeft()</code></li>
 * <li><code>newParentsZigZagRight()</code></li>
 * <li><code> newParentsZigZagLeft()</code></li>
 * </ul></p>
 *
 * @see matrix.decoration.Styled
 * @see matrix.structures.CDT.probe.BinSearchTree
 * @see matrix.structures.other.Rotations
 * @see RedBlackNode
 *
 * @author Petri Tenhunen
 */
public class RBTree extends BinSearchTree {

  /** Constructs a new RBTree */
  public RBTree() {
    this(null);
  }

  public CDT getNewInstance(){
    return new RBTree();
}

  
  /**
   * Constructs a new RBTree
   * @param o the key to the root node
   */
  public RBTree(Object o) {
    super(o);
  }

  public FDT getNewStructure() {
    return new RedBlackNode() ;
  }

  public Tree getNewNode(Object key) {
    return new RedBlackNode((MatrixComparable)key);
  }

  // Methods for interface CDT

  public CDT insert(MatrixComparable o) {
      setElement(((RedBlackNode)getElement()).insert(o));
      return this;
  }

  public CDT insert(Object o) {
      if (o instanceof MatrixComparable)
          return insert((MatrixComparable)o);
      setElement(((RedBlackNode)getElement()).insert(new Key(o)));
      return this;
  }

  public CDT delete(Object o) {
	  if (o instanceof MatrixComparable) {
		  setElement(((RedBlackNode)getElement()).delete(o));
	  } else {
		  // Wrap the object o inside a Key object
		  setElement(((RedBlackNode)getElement()).delete(new Key(o)));
	  }
	  return this;
  }

  public boolean equals(Object o) {
      if (o instanceof RBTree)
          return getElement().equals(((RBTree)o).getElement());
      else return super.equals(o);
  }



  static final long serialVersionUID = 2860991368734163098L;
}
