// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/* $Id: RadixSearchTrie.java,v 1.9 2007/09/07 13:37:05 vkaravir Exp $ */

package matrix.structures.CDT.probe;

import java.util.BitSet;

import matrix.structures.CDT.CDT;
import matrix.structures.FDT.*;
import matrix.structures.FDT.probe.*;
import matrix.structures.memory.VirtualObject;
import matrix.util.*;
//import matrix.visual.LabelDecorator;
//import matrix.visual.stylesheet.*;

/**
 * A radix search trie. A radix search trie is a dictionary where keys are
 * stored in a binary tree according to their bit values. All keys are
 * stored in leaf nodes and internal nodes are empty. This class uses
 * the matrix.util.KeyBuilder -class to create the bit representations
 * of the keys.
 *
 * @see matrix.util.KeyBuilder
 * @author Jussi Nikander
 */
public class RadixSearchTrie extends BinSearchTree implements CDT /*, Styled */ {


    //    private StyleSheet styleSheet = null;

    private VirtualObject root;

    /** label of this tree
    private String label = "";
    */
    /**
     * Constructs a new <code>RadixSearchTrie</code>.
     */
    public RadixSearchTrie() {
        this(null);
    }

    public CDT getNewInstance(){
        return new RadixSearchTrie();
    }

    
    /**
     * Constructs a new <code>RadixSearchTrie</code>.
     * @param o The contents of the new root of the new tree.
     */
    public RadixSearchTrie(Object o) {
        root = new VirtualObject(this, "root of the trie");
        setElement(getNewNode(o));
    }

    /**
     * tests for the equality of this object and the parameter object.
     * @param o the parameter object
     * @return true if this and o are equal
     */
    public boolean equals(Object o) {
        if(o instanceof BinaryTree) {
            /*
            Note.out(this, "This tree");
            Traverse.printTree((BinaryTree)getElement());
            Note.out(this, "Comparison tree");
            Traverse.printTree((BinaryTree)o);
            return Comparer.compareIdentical((BinaryTree)getElement(), (BinaryTree)o);
            */
        }
        return false;
    }

    // Interface FDT

    public void setElement(Object o) {
        root.setObject(o);
    }

    public Object getElement() {
        return root.getObject();

    }

    /**
     * Creates a new node in the RadixSearchTrie
     * @param key the key of the new node
     * @return a new <code>Tree</code> node
     */
  public Tree getNewNode(Object key) {
    return new LabeledBinTree(key);
  }


    /**
     * Tests if <code>this</code> tree is empty.
     * @return True if tree is empty, false otherwise
     */
    public boolean empty() {
        BinaryTree tree = (BinaryTree) getElement();
        return tree.getElement() == null && tree.getRight() == null && tree.getLeft() == null;
    }

    /**
     * Constructs a new node for a <code>RadixSearchTrie</code> and returns it.
     * @param key the contents of the new node
     * @return the new <code>LabeledBinTree</code>
     */
    public LabeledBinTree newTrie(Object key) {
        return (LabeledBinTree)getNewNode(key);
    }

    /**
     * Tests if a node is not a leaf node.
     * @param t the <code>BinaryTree</code> to be tested
     * @return true if the node is not a leaf
     */
    public boolean internalNode(BinaryTree t) {
        if(t == null)
            return false;
        else if(t.getLeft() == null && t.getRight() == null)
            return false;
        else
            return true;
    }

    /**
     * Inserts a new object into the tree.
     * @param o the object to be inserted
     * @return the root of the tree
     */
    public CDT insert(Object o) {
        if(o == null)
            return this;
        if(o instanceof Table) {
            Table t = (Table)o;
            for(int i = 0; i < t.size(); i++) {
                insert(t.getObject(i));
            }
            return this;
        }
        if(empty()) {
            setElement(getNewNode(o));
            return this;
        } else {
            BitSet tmp = getDigitalKey(o.toString());
            setElement(inserthelp((LabeledBinTree)getElement(), o, tmp, 0));
        }
        return this;
    }


    /**
     * Deletes the node containing the param object from the tree.
     * @param o the object to be deleted
     * @return root of the tree
     */
    public CDT delete(Object o) {
        if(((LabeledBinTree)getElement()).isLeaf()) {
            setElement(getNewNode(""));
            return this;
        }
        BinaryTree tmp = (BinaryTree)getElement();
        BitSet key = getDigitalKey(o.toString());
        if(key.get(0)) {
            if(tmp.getRight().isLeaf()) {
                if(tmp.getLeft().isLeaf()) {
                    tmp.setElement(tmp.getLeft().getElement());
                    tmp.setLeft(null);
                    tmp.setRight(null);
                } else
                    tmp.setRight(null);
            } else {
                BinaryTree tmp2 = deletehelp((LabeledBinTree)tmp.getRight(), o, key, 1);
                if(tmp2 != null)
                    tmp.setRight(tmp2);
            }
        } else {
            if(tmp.getLeft().isLeaf()) {
                if(tmp.getRight().isLeaf()) {
                    tmp.setElement(tmp.getRight().getElement());
                    tmp.setLeft(null);
                    tmp.setRight(null);
                } else
                    tmp.setLeft(null);
            } else {
                BinaryTree tmp2 = deletehelp((LabeledBinTree)tmp.getLeft(), o, key, 1);
                if(tmp2 != null)
                    tmp.setLeft(tmp2);
            }
        }
        return this;
    }

    public Object search() {
        return null;
    }

    /**
     * Searches the tree for node containing the parameter object.
     * @param o the object to be searched
     * @return the corresponding object in the tree.
     */
    public Object search(Object o) {
        BitSet key = getDigitalKey(o.toString());
        return searchhelp((BinaryTree)getElement(), key, 0);
    }

    /**
     * Counts the number of subtrees this node has.
     * @return 0 if node is a leaf, 2 otherwise
    public int getSubTreeCount(LabeledBinTree t) {
        if(t.isLeaf())
            return 0;
        return 2;
    }
     */

    /**
     * Gets the bit representation of a string key.
     */
    private BitSet getDigitalKey(String key) {
        return KeyBuilder.getDigitalKey(key);
    }

    /** Gets the lenght of the bit representation of a digital key */
    private int getDigitalKeyLength(String key) {
        return KeyBuilder.getDigitalKeyLength(key);
    }
    /**
     * Helper function used by delete. It handles the deletion of an object, unless the
     * height of the tree is one or two (if root or its immediate children are leaves).
     * @param trie the node currently being examined
     * @param o the key of the node to be deleted
     * @param key the bit representation of o
     * @param bitNro the index of the bit currently being examined
     */
    private BinaryTree deletehelp(LabeledBinTree trie, Object o, BitSet key, int bitNro) {
        BinaryTree tmp = null;
        if(internalNode(trie)) {
            if(key.get(bitNro++)) {
                if(trie.getRight().isLeaf()) {
                    trie.setRight(null);
                    if(trie.getLeft().isLeaf()) {
                        tmp = trie.getLeft();
                        trie.setLeft(null);
                        return tmp;
                    }
                    return null;
                } else {
                    tmp = deletehelp((LabeledBinTree)trie.getRight(), o, key, bitNro);
                    if(tmp != null) {
                        if(trie.getLeft() != null)
                            trie.setRight(tmp);
                        else {
                            trie.setRight(null);
                            trie.setLeft(null);
                            return tmp;
                        }
                    }
                    return null;
                }
            } else {
                if(trie.getLeft().isLeaf()) {
                    trie.setLeft(null);
                    if(trie.getRight().isLeaf()) {
                        tmp = trie.getRight();
                        trie.setRight(null);
                        return tmp;
                    }
                    return null;
                } else {
                    tmp = deletehelp((LabeledBinTree)trie.getLeft(), o, key, bitNro);
                    if(tmp != null) {
                        if(trie.getRight() != null)
                            trie.setLeft(tmp);
                        else {
                            trie.setLeft(null);
                            trie.setLeft(null);
                            return tmp;
                        }
                    }
                    return null;
                }
            }
        } else {
            Note.err(this, "delete recursed into leaf");
            return null;
        }
    }



    /**
     * Help function used by insert. Handles the insert operation
     * until a leaf node is found. If the leaf is empty, it inserts
     * the param object in it. If the node is not empty, it calls comparingInsert.
     * @param trie the node currently being examined
     * @param o the object to be inserted
     * @param newKey the bit representation of the key
     * @param bitNro the index of the bit currently examined
     * @return the current node being examined
     */
    private BinaryTree inserthelp(BinaryTree trie, Object o, BitSet newKey, int bitNro) {

        // empty leaf found
        if(trie == null)
            return newTrie(o);

        if(internalNode(trie)) {
            if(newKey.get(bitNro++))
                trie.setRight(inserthelp(trie.getRight(), o, newKey, bitNro));
            else
                trie.setLeft(inserthelp(trie.getLeft(), o, newKey, bitNro));
            return trie;
        } else {
            Object old = trie.getElement();
            // equality of objects is tested only here.
            if(old.equals(o)) {
                Note.out(this, o + "key already in tree. Cannot insert.");
                return trie;
            }
            BitSet oldKey = getDigitalKey(old.toString());
            trie.setElement("");
            return comparingInsert(trie, old, oldKey, o, newKey, bitNro);

        }
    }

    /**
     * Compares two objects, one of which was already in the tree, and other which
     * is being inserted in the tree, and finds them new places in the tree.
     * Called by inserthelp. It compares the bit representations of the two
     * parameter objects until they differ. When it found the bit that's different
     * in the objects, it places the objects in the tree in the appropriate place.
     * @param trie the node currently being examined
     * @param old the old object that was already in tree
     * @param oldkey the bit representation of the old object's key
     * @param o the new object
     * @param newKey the bit key of the new object
     * @param bitNro the bit currently being examined
     * @return the node currently being examined
     */
    private BinaryTree comparingInsert(BinaryTree trie, Object old, BitSet oldKey, Object o, BitSet newKey, int bitNro){
        if(newKey.get(bitNro) ^ oldKey.get(bitNro)) {
            if(newKey.get(bitNro)) {
                trie.setRight(newTrie(o));
                trie.setLeft(newTrie(old));
            } else {
                trie.setRight(newTrie(old));
                trie.setLeft(newTrie(o));
            }
        } else if(newKey.get(bitNro++)) {
            trie.setRight(newTrie(""));
            trie.setRight(comparingInsert(trie.getRight(), old, oldKey, o, newKey, bitNro));
        } else {
            trie.setLeft(newTrie(""));
            trie.setLeft(comparingInsert(trie.getLeft(), old, oldKey, o, newKey, bitNro));
        }
        return trie;
    }


    /**
     * Helper fuction that searches the tree for a given bitkey.
     */
    private Object searchhelp(BinaryTree tree, BitSet key, int bitNro) {
        if(tree == null)
            return null;
        if(internalNode(tree)) {
            if(key.get(bitNro++))
                return searchhelp(tree.getRight(), key, bitNro);
            else
                return searchhelp(tree.getLeft(), key, bitNro);
        }
        return tree.getElement();
    }

    // methods for interface LabelDecorator

    /*
     * Tells if label is enabled for this node or not.
     * @return true if label is enabled

    public boolean isLabelEnabled() {
        if(getElement() != null) {
            String key = getElement().toString();
            if(key != "" && key != null)
                return true;
        }
        return false;
    }
     */
    /*
     * Sets the label of this node.
     * @param s the new label

    public void setLabel(String s) {
        label = s;
        return;
    }
     */
    /*
     * Gets the label of this node.
     * @return this node's label

    public String getLabel() {
        return label;
    }
     */
    /*
     * Tells if referencelabels are enabled.
     * @return true if reference labels are enabled

    public boolean isReferenceLabelEnabled() {
        return false;
    }
     */
    /*
     * Sets the i'th reference label of this node.
     * @param i the number of the reference
     * @param s the new label

    public void setReferenceLabel(int i, String s) {
        return;
    }
     */
    /*
     * Returns the i'th reference label of this node.
     * @param i the number of the reference
     * @return the label of the i'th reference

    public String getReferenceLabel(int i) {
        return "";
    }
     */
    // Method for creating a a bit representation of the key

    /**
     * Creates a new label for this node.

    private void makeLabel() {
        Object key = getElement();
        label = "";
        if(key != null) {
            BitSet tmp = getDigitalKey(key.toString());
            int length = getDigitalKeyLength(key.toString());
            for(int i = 0; i < length; i++) {
                if(tmp.get(i))
                    label += "1";
                else
                    label += "0";
            }
        }
    }
    */

    static final long serialVersionUID = 6922386879514894746L;
}
