// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/* $Id: SimpleRedBlackNode.java,v 1.2 2007/09/07 13:37:04 vkaravir Exp $ */

package matrix.structures.CDT.probe;

import java.awt.Color;

import matrix.decoration.StyleSheet;
import matrix.decoration.StyleSheetAdapter;
import matrix.decoration.Styled;
import matrix.decoration.Visitable;
import matrix.structures.FDT.probe.BinTree;
import matrix.structures.FDT.probe.Key;
import matrix.structures.memory.VirtualBoolean;
import matrix.structures.util.MatrixComparable;

/**
 * This class is an extension to the binary tree node class
 * to support coloring needs of a red-black tree implementation. 
 * 
 * The class does not hold any red-black tree algorithms.
 * These are defined in the CDT classes such as TopDownRBTree.
 * 
 * @author Otto Seppl
 * @see matrix.structures.CDT.probe.TopDownRBTree
 *
 */

public class SimpleRedBlackNode extends BinTree implements Styled, Visitable{
    private static final long serialVersionUID = -2095988549773362515L;
    
    /**
     * Color constants used with the method setColor and the constructor
     * 
     * @see SimpleRedBlackNode(MatrixComparable,boolean)
     * @see setColor(boolean)
     */
    public static boolean BLACK = true;
    public static boolean RED   = false;

    private VirtualBoolean blackness;      // Color of the node
    private VirtualBoolean focus;          // Highlighting for showing traversal
    protected StyleSheet   nodeColouring;  // Stylesheet to convey the coloring 
                                            // information to the visualization
    
    /**
     * Creates a new node holding the given key.
     * The node will be colored red.
     *   
     * @param e     the value for the new node
     */

    public SimpleRedBlackNode(MatrixComparable e){
        this(e, false);
    }

    /**
     * Creates a new node holding the given key, with the given initial color.
     * Use this constructor instead of calling the more simple constructor and 
     * then setting the color separately as that will add a step in the animator.
     *   
     * @param e     the value for the new node
     * @param color the color of the new node
     */
    
    public SimpleRedBlackNode(MatrixComparable e, boolean color){
        super(e);
        blackness = new VirtualBoolean(color, this, "color (true=>BLACK)");
        focus     = new VirtualBoolean(false, this, "is the node focused");
    }

    /**
     * Sets the color of the node. The two colors used are defined as 
     * constants RED and BLACK.
     * @param color the color to be set
     */
    public void setColor(boolean color){
        blackness.assign(color);
    }
    
    /**
     * Tells if the node is black
     * @return is the node black
     */
    
    public boolean isBlack(){
        return blackness.eval();
    }

    /**
     * Tells if the node is red
     * @return is the node red
     */

    public boolean isRed(){
        return !blackness.eval();
    }
        

    /**
     * Returns the currents stylesheet. The colors depend on the red-black property of each node.
     * Additionally the visited-color is used to highlight the border to show how the tree is 
     * traversed in the algorithms.
     * 
     * @see matrix.decoration.Styled;
     */
    
    public StyleSheet getStyleSheet() {
        if (nodeColouring != null)
            return nodeColouring;
        else {
            nodeColouring = new StyleSheetAdapter() {
                private static final long serialVersionUID = -7324514050080341486L;

                public Color getDefaultBorderColor() {
                    return (empty() || !isRed()) ? Color.black: Color.red;
                }
                public Color getDefaultBackgroundColor() {
                    return (isRed()) ? Color.red : Color.black;
                }
                public Color getFocusedBackgroundColor() {
                    return (empty() || !isRed()) ? Color.black: Color.red;
                }
                public Color getFocusedBorderColor() {
                    return (empty() || !isRed()) ? Color.red : Color.black;
                }

                public Color getVisitedFocusedBackgroundColor() {
                    return (empty() || !isRed()) ? Color.black: Color.red;
                }
                public Color getVisitedBackgroundColor() {
                    return (empty() || !isRed()) ? Color.black: Color.red;
                }

                public Color getVisitedFocusedBorderColor() {
                    return Color.yellow;
                }
                public Color getVisitedBorderColor() {
                    return Color.yellow;
                }

            };
            return nodeColouring;
        }
    }

    
    /*
     * Visited bit is used by the red-black tree algorithms to show how
     * the tree is traversed in the algorithms.
     */
    
    /**
     * Returns whether the visited-bit is currently on
     * 
     * matrix.decoration.Visitable;
     */
    public boolean isVisited() {
        return focus.eval();
    }

    /**
     * Sets the visited-bit
     * 
     * matrix.decoration.Visitable;
     */
    public void setVisited(boolean val) {
        focus.assign(val);
    }
    
    /**
     * Clears the visited-bit
     * 
     * matrix.decoration.Visitable;
     */
    public void clearVisited(){
        focus.assign(false);
    }

    /*
     * Inner nodes always have data
     */
    public boolean empty(){
        return false;
    }
    
    public boolean equals(Object o){
               
        if (o != null && o instanceof SimpleRedBlackNode){
            SimpleRedBlackNode s = (SimpleRedBlackNode)o;
            
            // Color has to match
            if (s.isRed() != isRed())
                return false;
            
            // children must both exist
            if (s.getLeft()==null || s.getRight()==null)
                return false;
            
            // elements must both be null or non-null
            if (getElement() == null ^ s.getElement() == null)
                return false;
            
            // if elements exist they have to match
            if (getElement() != null && !getElement().equals(s.getElement()))
                return false;            
            
            // both children must match
            if (getLeft().equals(s.getLeft()) && getRight().equals(s.getRight()))
                return true;
        }
        
        return false;
    }
}

class ZNode extends SimpleRedBlackNode{

    private static final long serialVersionUID = -8236620415603269978L;

    public ZNode() {
        super(Key.Empty, true); // Empty black node
    }   

    public boolean empty(){
        return true;
    }
    
    public boolean isLeaf(){
        return true;
    }
    
    public int getSubTreeCount(){return 0;}
    
    public boolean equals(Object o){
        
        //Any z-node will suffice
        if (o != null && o instanceof ZNode){
            return true;
        }
        return false;
    }
}