// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/* $Id: TopDownRBTree.java,v 1.3 2007/09/07 13:37:04 vkaravir Exp $ */

package matrix.structures.CDT.probe;


import matrix.animation.Animator;
import matrix.structures.CDT.CDT;
import matrix.structures.FDT.*;
import matrix.structures.FDT.probe.Key;
import matrix.structures.FDT.probe.LexInf;
import matrix.structures.util.MatrixComparable;

/**
 * A top-down implementation of the red-black tree.
 * 
 * A red-black tree is a balanced binary search tree with the 
 * following properties:
 * 
 * <ul>
 * <li> Every node is either red or black</li>
 * <li> The root node is black.</li>
 * <li> Both children of a red node are black.</li>
 * <li> All leaf nodes are black.<li>
 * <li> Every path from root to a leaf contains the same number of black nodes. </li>
 * </ul>
 *
 * @author Otto Seppl
 */



public class TopDownRBTree extends BinSearchTree {

    private static final long serialVersionUID = 1290415807751403939L;

    // Two sentinels are used to simplify insert and delete
    protected SimpleRedBlackNode lowerSentinel;
    protected SimpleRedBlackNode upperSentinel;

    
    /** Constructs a new RBTree */
    public TopDownRBTree() {
        
        lowerSentinel = new SimpleRedBlackNode(LexInf.getInstance(), SimpleRedBlackNode.BLACK);
        lowerSentinel.setLeft(createZNode());

        upperSentinel = new SimpleRedBlackNode(LexInf.getInstance(), SimpleRedBlackNode.BLACK);
        upperSentinel.setLeft(lowerSentinel);
  //      upperSentinel.setRight(lowerSentinel);
    }

    /**
     * Constructs a new RBTree
     * @param o the key to the root node
     */
    public TopDownRBTree(Object o) {
                
        lowerSentinel = new SimpleRedBlackNode(LexInf.getInstance(), SimpleRedBlackNode.BLACK);

        upperSentinel = new SimpleRedBlackNode(LexInf.getInstance(), SimpleRedBlackNode.BLACK);
        upperSentinel.setLeft(lowerSentinel);
//        upperSentinel.setRight(lowerSentinel);
        
        SimpleRedBlackNode actualRoot = new SimpleRedBlackNode((MatrixComparable)o, SimpleRedBlackNode.BLACK);

        actualRoot.setLeft( createZNode());
        actualRoot.setRight(createZNode());

        lowerSentinel.setLeft(actualRoot);
        lowerSentinel.setRight(createZNode());
    }

    // The root is always the left child of the sentinel
    public Object getElement(){
        return lowerSentinel.getLeft();
    }
    
    public void setElement(Object o){
        // TODO probably should be left empty
    }
    
    public CDT getNewInstance(){
        return new TopDownRBTree();
    }

    public FDT getNewStructure() {
        return new SimpleRedBlackNode(new Key(null)) ;
    }

    public Tree getNewNode(Object key) {
        return new SimpleRedBlackNode((MatrixComparable)key);
    }

    // Methods for interface CDT

    public CDT insert(MatrixComparable o) {
                        
        // cursor traverses the tree. Originally this is the left child of the sentinel which
        // is the actual root of the red-black tree
        
        SimpleRedBlackNode cursor = (SimpleRedBlackNode)lowerSentinel.getLeft();

        // Left and right are the left and right children of cursor
        
        SimpleRedBlackNode left;
        SimpleRedBlackNode right;
        
        // F2 is the grandfather and F1 the father of cursor.
        
        // The two sentinels simplify the insertion algorithm by eliminating all root-related special cases
        // References to the sentinels are placed in the beginning in the father and grandfather variables
        
        SimpleRedBlackNode f2     =upperSentinel; // grandfather of cursor in the algorithm
        SimpleRedBlackNode f1     =lowerSentinel; // father of cursor in the algorithm
        
        // The algorithm needs to keep track of the branches taken from the grandfather and father
        // to choose the appropriate rotation to be applied
        
        boolean f1left      = true; // must have been the left branch as f2 is lexical infinity
        boolean currentLeft = true; // must have been the left branch as f1 is lexical infinity

        Animator.getActiveAnimator().startOperation();
        cursor.setVisited(true);
        Animator.getActiveAnimator().endOperation();

        
        while (!(cursor instanceof ZNode)){
                        
            left   = (SimpleRedBlackNode)cursor.getLeft();
            right  = (SimpleRedBlackNode)cursor.getRight();
            
            // Fix potential color conflicts on the way down 

            if ( cursor.isBlack() && left.isRed() && right.isRed() ){
                Animator.getActiveAnimator().startOperation("Red-black tree color change");

                cursor.setColor(SimpleRedBlackNode.RED);
                left.setColor(SimpleRedBlackNode.BLACK);
                right.setColor(SimpleRedBlackNode.BLACK);

                Animator.getActiveAnimator().endOperation();

                if (f1 == lowerSentinel){ //Is the node root?
                    Animator.getActiveAnimator().startOperation("Red-black tree color change");
                    cursor.setColor(SimpleRedBlackNode.BLACK);
                    Animator.getActiveAnimator().endOperation();                    
                }
                    
                
                if ( f1.isRed()){
                    // rotations needed

                    Animator.getActiveAnimator().startOperation();
                    cursor.clearVisited();
                    Animator.getActiveAnimator().endOperation();
                    
                    if (f1left){
                        if (currentLeft){
                            Animator.getActiveAnimator().startOperation("Red-black single rotation right");

                            //Single rotation right

                            Object dataTemp = f2.getElement();
                            f2.setElement(f1.getElement());
                            f1.setElement(dataTemp);

                            f1.setLeft(f1.getRight());
                            f1.setRight(f2.getRight());
                            f2.setRight(f1);
                            f2.setLeft(cursor);

                            // move cursor down

                            // F2 same
                            f1 = cursor;
                            f1left = true;

                            // Move cursor to the appropriate branch
                            if (o.geq((MatrixComparable)cursor.getElement())){
                                cursor = right;
                                currentLeft = false;
                            }else{
                                cursor = left;
                                currentLeft = true;
                            }

                            Animator.getActiveAnimator().endOperation();

                        }else{
                            Animator.getActiveAnimator().startOperation("Red-black double rotation right");

                            //Double rotation right

                            Object dataTemp = f2.getElement();
                            f2.setElement(cursor.getElement());
                            cursor.setElement(dataTemp);

                            f1.setRight(cursor.getLeft());
                            cursor.setLeft(cursor.getRight());
                            cursor.setRight(f2.getRight());
                            f2.setRight(cursor);

                            // move cursor down

                            // F2 same

                            // Move cursor to the appropriate branch
                            if (o.geq((MatrixComparable)f2.getElement())){ //f2 has the old cursor data
                                f1left      = false;
                                currentLeft = true;

                                f1 = cursor;
                                cursor = right;
                            }else{
                                f1left      = true;
                                currentLeft = false;

                                cursor = left;
                            }

                            Animator.getActiveAnimator().endOperation();

                        }
                    }else{
                        if (currentLeft){

                            Animator.getActiveAnimator().startOperation("Red-black double rotation left");

                            //Double rotation left

                            Object dataTemp = f2.getElement();
                            f2.setElement(cursor.getElement());
                            cursor.setElement(dataTemp);

                            f1.setLeft(cursor.getRight());
                            cursor.setRight(cursor.getLeft());
                            cursor.setLeft(f2.getLeft());
                            f2.setLeft(cursor);

                            // move cursor down

                            // F2 same
                            f1left = false;

                            // Move cursor to the appropriate branch
                            if (o.geq((MatrixComparable)f2.getElement())){ //f2 has the old cursor data
                                cursor = right;
                                currentLeft = false;
                            }else{
                                f1 = cursor;
                                cursor = left;
                                currentLeft = true;
                            }
                            
                            Animator.getActiveAnimator().endOperation();

                        }else{

                            Animator.getActiveAnimator().startOperation("Red-black single rotation left");

                            //Single rotation left

                            Object dataTemp = f2.getElement();
                            f2.setElement(f1.getElement());
                            f1.setElement(dataTemp);

                            f1.setRight(f1.getLeft());
                            f1.setLeft(f2.getLeft());
                            f2.setLeft(f1);
                            f2.setRight(cursor);

                            // move cursor down

                            // F2 same
                            f1 = cursor;
                            f1left      = false;

                            // Move cursor to the appropriate branch
                            if (o.geq((MatrixComparable)cursor.getElement())){
                                cursor = right;
                                currentLeft = false;

                            }else{
                                cursor = left;
                                currentLeft = true;
                            }

                            Animator.getActiveAnimator().endOperation();

                        }
                    }

                    // highlight the new cursor
                    Animator.getActiveAnimator().startOperation();
                    cursor.setVisited(true);
                    Animator.getActiveAnimator().endOperation();

                }

            }
            
            // Move cursor down
            
            f2 = f1;
            f1 = cursor;

            f1left = currentLeft; // change branch information as we proceed downwards

            if (o.geq((MatrixComparable)cursor.getElement())){
                cursor = (SimpleRedBlackNode)cursor.getRight();
                currentLeft = false;
            }else{
                cursor = (SimpleRedBlackNode)cursor.getLeft();
                currentLeft = true;
            }

            // move coloring down in visualization
            Animator.getActiveAnimator().startOperation();
            f1.clearVisited();
            cursor.setVisited(true);
            Animator.getActiveAnimator().endOperation();

        }

        // Clear the cursor highlight on exit from the loop
        Animator.getActiveAnimator().startOperation();
        cursor.clearVisited();
        Animator.getActiveAnimator().endOperation();

        
        //Perform the actual insertion
        
        SimpleRedBlackNode newNode = new SimpleRedBlackNode(o, SimpleRedBlackNode.RED) ;
        
        if (currentLeft){
            Animator.getActiveAnimator().startOperation("Adding the key");
            newNode.setLeft(cursor);
            newNode.setRight(createZNode());
            f1.setLeft(newNode);
            Animator.getActiveAnimator().endOperation();
            
            cursor = newNode;
        }
        else{
            Animator.getActiveAnimator().startOperation("Adding the key");
            newNode.setLeft(cursor);
            newNode.setRight(createZNode());
            f1.setRight(newNode);
            Animator.getActiveAnimator().endOperation();

            cursor = newNode;
        }

        if (f1 == lowerSentinel){ //Is the node root - if yes, force black
            Animator.getActiveAnimator().startOperation("Red-black tree color change");
            ((SimpleRedBlackNode)lowerSentinel.getLeft()).setColor(SimpleRedBlackNode.BLACK);
            Animator.getActiveAnimator().endOperation();                    
        }
        
        // Fix color conflicts if f1 was red.
        
        if (f1.isRed()){
            
            Animator.getActiveAnimator().startOperation("final rotation");
            if (f1left){
                if (currentLeft){                    
                    //Single rotation right

                    //repeated code... refactor?
                    Object dataTemp = f2.getElement();
                    f2.setElement(f1.getElement());
                    f1.setElement(dataTemp);

                    f1.setLeft(f1.getRight());
                    f1.setRight(f2.getRight());
                    f2.setRight(f1);
                    f2.setLeft(cursor);

                }else{
                    //Double rotation right
                    Object dataTemp = f2.getElement();
                    f2.setElement(cursor.getElement());
                    cursor.setElement(dataTemp);

                    f1.setRight(cursor.getLeft());
                    cursor.setLeft(cursor.getRight());
                    cursor.setRight(f2.getRight());
                    f2.setRight(cursor);
                }
                
            }else{
                if (currentLeft){
                    //Double rotation left
                    Object dataTemp = f2.getElement();
                    f2.setElement(cursor.getElement());
                    cursor.setElement(dataTemp);

                    f1.setLeft(cursor.getRight());
                    cursor.setRight(cursor.getLeft());
                    cursor.setLeft(f2.getLeft());
                    f2.setLeft(cursor);

                }else{
                    //Single rotation left
                    Object dataTemp = f2.getElement();
                    f2.setElement(f1.getElement());
                    f1.setElement(dataTemp);

                    f1.setRight(f1.getLeft());
                    f1.setLeft(f2.getLeft());
                    f2.setLeft(f1);
                    f2.setRight(cursor);
                }
            }
            Animator.getActiveAnimator().endOperation();
        }        
        
        return this; // Root never changes
    }

    public CDT insert(Object o) {
        if (o instanceof MatrixComparable){
            insert((MatrixComparable)o);
            return this;
        }

        // Wrap object inside a key to make it MatrixComparable
        insert(new Key(o));
        return this;
    }

    public CDT delete(MatrixComparable o){
        
        // ADD CODE HERE
        
        return this;
    }
    
    public CDT delete(Object o) {
        if (o instanceof MatrixComparable) {
            delete((MatrixComparable)o);
            return this;
        }
        // Wrap object inside a key to make it MatrixComparable
        delete(new Key(o));
        return this;
    }

    //TODO FIXME This just does not work yet...

    public boolean equals(Object o) {
        if (o instanceof TopDownRBTree)
            return getElement().equals(((TopDownRBTree)o).getElement());
        else return super.equals(o);
    }

    public ZNode createZNode(){
        return new ZNode();
    }

}
