// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/* $Id: StaticLinkedListImpl.java,v 1.11 2007/09/10 08:13:16 vkaravir Exp $ */

package matrix.structures.FDT.probe;

// Source code example for "A Practical Introduction
// to Data Structures and Algorithm Analysis"
// by Clifford A. Shaffer, Prentice Hall, 1998.
// Copyright 1998 by Clifford A. Shaffer
// Modified for Matrix by Ari Korhonen, 1999

import matrix.structures.FDT.*;
import matrix.structures.FDT.indexed.StaticLinkedList;
import matrix.structures.memory.*;
import matrix.structures.other.Comparer;
import matrix.util.*;

public class StaticLinkedListImpl implements StaticLinkedList {// Array-based list implementation

    private VirtualInteger numInList = new VirtualInteger(this, "number of nodes in list");   // Actual number of Objects in list
    private VirtualInteger curr = new VirtualInteger(this, "position of current node");        // Position of current Object
    private VirtualArray listArray;        // Array holding list Objects

    public StaticLinkedListImpl() { // Constructor: use default size
        setup();
    }

    /** Constructor for turning a VirtualArray into a list. */
    public StaticLinkedListImpl(Array array) {
        setup((VirtualArray)array);
    }

    public boolean equals(StaticLinkedList l) {
        return Comparer.compareList(this, l);
    }

    // FDT interface Methods
    public FDT getNewStructure() { return new StaticLinkedListImpl(); }

    public void setElement(Object o) {
        Note.err(this,
                 "Element: Cannot perform setElement("+o+") for LinkedList");
    }

    public Object getElement() {
        return this;
    }

    // CDT

    private void setup(VirtualArray array) { // Do actual initialization work
        numInList.assign(array.getLast()-array.getFirst()+1);
        curr.assign(0);
        listArray = array;
    }

    private void setup() {    // Do actual initialization work
        setup(new VirtualArray(this, "nodes of the list"));   // Create listArray
    }

    public void clear() {             // Remove all Objects from list
        numInList.assign(curr.assign(0));  // Simply reinitialize values
    }

    public void insert(Object it) { // Insert Object at current position
        Assert.notFalse((curr.eval() >= 0) &&
                        (curr.eval() <= numInList.eval()),
                        "Bad value for curr");
        // Shift Objects to make room
        for (int i=numInList.eval(); i>curr.eval(); i--)
            listArray.setObject(listArray.getObject(i-1), i);
        listArray.setObject(it, curr);
        numInList.inc();            // Increment list size
    }

    public void append(Object it) { // Insert Object at tail of list
        listArray.setObject(it, numInList); // Increment list size
        numInList.inc();
    }

    public Object remove() {        // Remove and return current Object
        Assert.notFalse(!isEmpty(), "Can't delete from empty list");
        Assert.notFalse(isInList(), "No current element");
        Object it = listArray.getObject(curr);  // Hold removed Object
        // Shift elements down
        for(int i=curr.eval(); i<numInList.eval()-1; i++)
            listArray.setObject(listArray.getObject(i+1), i);
        numInList.dec();                  // Decrement list size
        return it;
    }

    public void setFirst() { curr.assign(0); } // Set curr to first position
    public void prev() { curr.dec(); }  // Move curr to previous position

    public void next() { curr.inc(); }  // Move curr to next position

    public int length()                // Return length of list
    { return numInList.eval(); }

    public void setPos(int pos)     // Set curr to specified position
    { curr.assign(pos); }

    public void setValue(Object it) { // Set current Object's value
        Assert.notFalse(isInList(), "No current element");
        listArray.setObject(it, curr);
    }

    public Object currValue() {     // Return current Object's value
        Assert.notFalse(isInList(), "No current element " + curr);
        return listArray.getObject(curr);
    }

    public boolean isEmpty()        // Return true if list is empty
    { return numInList.eval() == 0; }

    public boolean isInList()       // True if curr is within list
    { return (curr.eval() >= 0) && (curr.eval() < numInList.eval()); }

    public void print_() {           // Print all list's Objects
        if (isEmpty()) System.out.println("()");
        else {
            System.out.print("( ");
            for (setFirst(); isInList(); next())
                System.out.print(currValue() + " ");
            System.out.println(")");
        }
    }

    public void print() {
        Note.out(this, "print");
        //Arrayesenter.Represent(this);
    }

    // interface Array

    public Object getObject(int i) { return listArray.getObject(i); }
    public void setObject(Object o, int i) { setPos(i); setValue(o); }
    public int getFirst() { return 0; }
    public int getLast() { return numInList.eval() - 1; }
    public void setFirst(int i) { listArray.setFirst(i); }
    public void setLast(int i) { listArray.setLast(i); }
    /*
    public FDT getKey() {
        Note.out(this, "Error: cannot get from LinkedList"); return null; }
    public void setKey(FDT s) {
        Note.out(this, "Error: cannot set to LinkedList"); }
    */

    static final long serialVersionUID = -5723735629198457873L;
}
