// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package matrix.structures.algorithms;

//import java.util.Stack;
import java.awt.Color;
import java.util.Hashtable;

import matrix.animation.Animator;
import matrix.decoration.*;
import matrix.structures.FDT.*;
import matrix.structures.FDT.substructures.Vertex;
import matrix.util.Note;

/**
 * Graph_Traverse is the abstract superclass of all graph traversing 
 * algorithms. 
 * It contains all functionality needed to store different parts of the 
 * algorithm to the animator. While it is not currently forced in the code,
 * all graph traversing algorithms are supposed to start at the node 
 * returned by the getElement() -method of the Graph traversed.
 * The source Vertex can be set in the GUI by using the setSource -command
 * from the popupmenu of a graph's component.
 *
 * @author Jussi Nikander
 */

public abstract class Graph_Traverse {

    /** Default color for vertices that have not been visited by the algorithm */
    public static final Color UNVISITED_COLOR = Color.white;

    /** Default color for vertices that have been visited by the algorithm */
    public static final Color VISITED_COLOR = Color.gray;

    /** Default color for vertices that are no longer used in the algoritm */
    public static final Color FINISHED_COLOR = Color.black;
  
    /** Default color for vertices that are in the border */
    public static final Color BORDER_COLOR = Color.gray;
  
  /** 
   * The visited table contains nodes visited by the
   * algorithm, the border table contains nodes in
   * the border and finished contains nodes
   * that have been finished.
   * 
   * A call to init() will destroy the contents of
   * these tables. 
   *
   * Used by methods visit(), finish(), etc.
   */
    protected Hashtable visited, border, finished;

    /** Animator used to store the operations. */
    protected Animator a;

    /** graph traversed */
    protected Graph dg;

    public Graph_Traverse() {
    }

    /**
     * Initialises the Dynamic graph by coloring all its vertices white
     * and creating the required hash tables.
     */
    private void init() {
	//a = Animator.getActiveAnimator();
	Vertex[] vertices = dg.getVertices();
	visited = new Hashtable();
	for(int i = 0; i < vertices.length; i++) {
	    if(vertices[i] == null)
		continue;
	    if(visited.contains(vertices[i]))
	       continue;
	    visited.put(vertices[i].getElement(), vertices[i]);
	    recurse_init(vertices[i]);
	}
	visited = new Hashtable();
	border = new Hashtable();
	finished = new Hashtable();
    }

    /** Recursively initialises the whole graph */
    private void recurse_init(Vertex v) {
	if(v == null)
	    return;
	initColor(v);
	Vertex[] succ = v.getSuccessors();
	for(int i = 0; i < succ.length; i++) {
	    if(visited.contains(succ[i]))
		continue;
	    visited.put(succ[i].getElement(), succ[i]);
	    recurse_init(succ[i]);
	}
    }

    /**
     * Initialises the color of vertices. Differs from changeColor in the
     * fact that it doesn't store color changes separately in the animator.
     */
    private void initColor(Vertex v) {
      changeColor(v, UNVISITED_COLOR);
    }

    /**
     * Changes the DefaultBackgroundColor of the param vertex to the the param Color.
     * @param v the vertex to be modified
     * @param c the new Background color of the param vertex
     */
    public void changeColor(Vertex v, Color c) {
	if(v instanceof Styled) {
	    a.startOperation();
	    ((Styled)v).getStyleSheet().setDefaultBackgroundColor(c);
	    a.endOperation();
	    return;
	}
	if(v instanceof ColorDecorator) {
	    a.startOperation();
	    ((ColorDecorator)v).setDefaultBackgroundColor(c);
	    a.endOperation();
	    return;
	}
	Note.err(v, "neither Styled nor Colordecorator. Cannot set color.");
    }

    /**
     * Runs a particular graph traversing
     * algorithm. Called by the traverse -method and should never be called
     * by user code.
     */
    public abstract void run();


    /**
     * Method called by the visualisation when 
     * an traversing algorithm is used.
     */
    public final void traverse(Graph dg) {
	this.dg = dg;
	a = Animator.getActiveAnimator();
	init();
	a.startOperation();
	run();
	a.endOperation();
	a.startOperation();
	init();
	a.endOperation();
    }

    public final void traverse(FDT fdt) {
	if(fdt instanceof Graph) 
	    traverse((Graph)fdt);
	else
	    Note.err(fdt, "not an instance of Graph, cannot traverse");
    }

    /** 
     * Returns the Graph used by this algorithm.
     * @return the graph
     */    
    public Graph getGraph() {
	return dg;
    }

  /**
   * Visits a vertex and changes its color
   * to VISITED_COLOR.
   *
   * @param v the vertex to be visited
   */
  public void visit(Vertex v) {
    if(visited != null) {
      visited.put(v, v);
      changeColor(v, VISITED_COLOR);
    } else
      Note.err(this, "visited -table not initialized, cannot visit");
  }

  /**
   * Tests if the parameter vertex has been visited.
   * 
   * @param v the vertex to be tested
   * @return true if the vertex has been visited
   */
  public boolean isVisited(Vertex v) {
    if(v == null)
      return true;
    return visited.containsKey(v);
  }

  /**
   * Adds the parameter vertex to border
   * and changes its color to BORDER_COLOR.
   *
   * @param v the vertex to be added
   */
  public void addToBorder(Vertex v) {
    if(visited != null) {
      border.put(v,v);
      changeColor(v, BORDER_COLOR);
    } else
      Note.err(this, "border not initialized, cannot add to border");
  }

  /**
   * Removes the parameter vertex from border
   * 
   * @param v the vertex to be removed
   */
  public void removeFromBorder(Vertex v) {
    border.remove(v);
  }

  /**
   * Tests if the parameter vertex is in border
   * 
   * @param v the vertex to be tested
   * @return true if the vertex is in border
   */
  public boolean isInBorder(Vertex v) {
    if(v == null)
      return true;
    return border.containsKey(v);
  }

  /**
   * Adds the vertex to the list of finished 
   * vertices. 
   * 
   * @param v the vertex to finish
   */
  public void finish(Vertex v) {
    finished.put(v,v);
    changeColor(v, FINISHED_COLOR);
  }

  /**
   * Tells if the vertex is in the list of finished vertices
   * 
   * @param v the vertex to be tested
   */
  public boolean isFinished(Vertex v) {
    if(v == null)
      return true;
    return finished.containsKey(v);
  }

    /**
     * Tells if a node has been visited or not.
     * @return true if node has NOT been visited
     */
    public boolean notVisited(Vertex v) {
	if(v == null)
	    return false;
	/*
	if(v instanceof Styled) 
	    return ((Styled)v).getStyleSheet().getDefaultBackgroundColor().equals(UNVISITED_COLOR);
	if(v instanceof ColorDecorator) 
	    return ((ColorDecorator)v).getDefaultBackgroundColor().equals(UNVISITED_COLOR);
	Note.err(this, v + " neither Styled nor ColorDecorator.");
	return false;
      */
      return !(visited.containsKey(v));
    }
}

