// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
package matrix.structures.spatial;

import java.awt.Rectangle;
import matrix.structures.FDT.FDT;
import matrix.structures.spatial.FDT.probe.SpatialElement;

public interface Area extends FDT {

	/**
	 * Fetches the subareas of this Area. The subareas may overlap, but
	 * they must all be contained inside this Area.
	 * 
	 * @return array containing the subareas
	 */
	public Area[] getSubAreas();
	/**
	 * Fetches the spatial data elements contained in this Area. The spatial
	 * elements contained in this Area do not include the elements contained
	 * in subareas. The elements contained in this area may, however, overlap
	 * subareas or elements contained in them.
	 * 
	 * @return array containing the spatial elements
	 */
	public SpatialElement[] getSpatialElements();

	/**
	 * Inserts a new spatial element to this area. Implementation details of 
	 * this method are dependent of the implementation. The actual area where 
	 * the element is inserted to may be this Area or any of its subareas.
	 * 
	 * This is basically an ADT insert method, similar to one defined for
	 * all dictionaries.
	 * 
	 * @param e the element to be added
	 */	
	public void insert(SpatialComparable e);
	
	/**
	 * Removes one of the spatial elements contained in this Area.
	 * The element may be included in this instance of Area or any
	 * of its subareas, depending on the implementation. 
	 *
	 * This is basically an ADT delete method, similar to the
	 * one defined for all dictionaries.
	 * 
	 * @param e the element to be deleted
	 */
	public void delete(SpatialComparable e);

	/**
	 * Adds a new subarea to this Area. The subarea must be completely
	 * contained in this area, but can overlap existing subareas.
	 * The subarea is added to this instance of Area, and not to any
	 * existing subarea.
	 * 
	 * @param a the subarea to be added
	 */
	public void addSubArea(Area a);

	/**
	 * Removes one of the subareas of this instance of Area.
	 * 
	 * @param a the subarea to be removed.
	 */
	public void removeSubArea(Area a);

	/**
	 * Gets the bounds of this Area as a polygon.
	 * 
	 * @return a matrix.structues.spatial.FDT.Polygon that contains this area
	 */
	public Rectangle getBoundingPolygon();
    
    /**
     * @return  True if this area has an associated <code>PaintingStyleDecorator</code> 
     *          that can be obtained with getPaintingStyleDecorator(), false otherwise.
     */
    public boolean hasPaintingStyleDecorator();

    /**
     * Sets the <code>PaintingStyleDecorator</code> associated with this area.
     * 
     * @param psd The decorator to be used.
     */
    public void setPaintingStyleDecorator(PaintingStyleDecorator psd);

    /**
     * @return  The <code>PaintingStyleDecorator</code> object associated with this
     *          area.
     */
    public PaintingStyleDecorator getPaintingStyleDecorator();

}
