// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
/**
 * 
 */
package matrix.structures.spatial.CDT.probe;


import matrix.structures.CDT.CDT;
import matrix.structures.FDT.Tree;
import matrix.structures.FDT.probe.Key;
import matrix.structures.memory.VirtualObject;
import matrix.structures.spatial.Area;
import matrix.structures.spatial.FDT.probe.Point;
import matrix.structures.util.Iterator;
import matrix.util.Note;

/**
 * An implementation of the Point-Region-quadtree data structure.
 * The structure is used to store 2-dimensional data identified 
 * by a point, i.e. (x,y)-coordinates.
 * 
 * @see PRQuadTreeNode, QuadTree
 * @author Juha Helminen
 */
public class PRQuadTree extends QuadTree implements Iterator, Area {

	// -- Constructors --
	
	/**
	 * Create an empty tree.
	 */
	public PRQuadTree() {
		super(null);
	}

	/**
	 * Create an empty tree and insert Object o into it.
	 * @param o The data to be inserted in the root.
	 */
	public PRQuadTree(Point o) {
		super(o);
	}

	/**
	 * Create an empty tree corresponding to the rectangular area given as parameter.
	 * @param minX The x-coordinate of the lower left corner of the rectangular area.
	 * @param maxX The x-coordinate of the upper right corner of the rectangular area.
	 * @param minY The y-coordinate of the lower left corner of the rectangular area.
	 * @param maxY The y-coordinate of the upper right corner of the rectangular area.
	 */
	public PRQuadTree(int minX, int maxX, int minY, int maxY) {
        this.root = new VirtualObject(this, "root");
        this.root.setObject(new PRQuadTreeNode(null, minX, maxX, minY, maxY));
	}

	
	// -- Implementations of inherited abstract methods --
	
	public Tree getNewNode(Object key) {
		if (key instanceof Point) {
			return new PRQuadTreeNode((Point)key); 
		} else if (key == null) {
            return new PRQuadTreeNode(); 
		} else if (key instanceof Key) {
			return getNewNode(((Key)key).getObject());
		} else {
			Note.warning(this, "Cannot insert object of type " + key.getClass());
			return null;
		}
	}

    public CDT insert(Object o) {
		((PRQuadTreeNode)getElement()).insert(o);
		return this;
	}
	
	public CDT delete(Object o) {
		((PRQuadTreeNode)getElement()).delete(o);
		return this;
	}

	public Object search(Object o) {
		if (o instanceof Point) {
			return ((PRQuadTreeNode)getElement()).search((Point)o);
		} else {
			return null;
		}
	}

	public CDT getNewInstance() {
		return new PRQuadTree();
	}
	
	public void setElement(Object o) {
        if (o instanceof PRQuadTreeNode)
            root.setObject(o);
        else if (o == null) {
            root.setObject(getNewNode(null));
        } else {
            Note.err(this, "Cannot set non-PRQuadTree as root.");
        }
	}
	
	public boolean equals(Object obj) {
		if (obj instanceof PRQuadTree) {
			return super.equals(obj);
		} else {
			return false;
		}
	}
	
	private static final long serialVersionUID = 7169319677410848058L;
}
