// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
package matrix.structures.spatial.CDT.probe;

import java.awt.Color;
import java.awt.Rectangle;
import java.awt.geom.Point2D;
import java.util.LinkedList;
import java.util.Vector;
import matrix.structures.FDT.FDT;
import matrix.structures.memory.VirtualObject;
import matrix.structures.spatial.Area;
import matrix.structures.spatial.PaintingStyleDecorator;
import matrix.structures.spatial.SpatialComparable;
import matrix.structures.spatial.FDT.probe.Point;
import matrix.structures.spatial.FDT.probe.PolyLine;
import matrix.structures.spatial.FDT.probe.Polygon;
import matrix.structures.spatial.FDT.probe.SpatialElement;

/**
 * @author Juha Helminen
 */
public class QuadEdge implements FDT, Area {

    private VirtualObject paintingStyleDecorator = new VirtualObject(null, "PaintingStyleDecorator");
	// The starting point for queries, that is,
    // the entry point to the quad-edge structure.
    // This stores a branch corresponding to the 
    // starting vertex of an edge in the structure.
	private VirtualObject anchor;
	
    // ** Constructors **

    public QuadEdge() {
        this.anchor = new VirtualObject(this, "anchor");
    }

    public QuadEdge(Point2D[] polygonVertices) {
        this();
        this.setAnchor(makePolygon(polygonVertices));
    }

    public QuadEdge(int[][] adjacencyList, Point2D[] vertices) {
        this();
        this.setAnchor(makeGraph(adjacencyList, vertices));
    }

    // ** Implementation of interface FDT **

    public Object getElement() {
        return this;
    }

    public void setElement(Object o) {
        // Disabled.
    }
    
    
    // ** Implementation of interface Area **
    
    public Rectangle getBoundingPolygon() {
        // TODO
        return new Rectangle(300, 300, 0, 0);
    }

    public SpatialElement[] getSpatialElements() {
        return (SpatialElement[])this.BFS(new GetSpatialElements());
    }

    public Area[] getSubAreas() {
        return new Area[0];
    }

    public void insert(SpatialComparable e) {
        // Disabled.
    }

    public void delete(SpatialComparable e) {
        // Disabled.
    }

    public void addSubArea(Area a) {
        // Disabled.
    }

    public void removeSubArea(Area a) {
        // Disabled.
        
    }

    public boolean hasPaintingStyleDecorator(){
        return (paintingStyleDecorator.getObject() != null);
    }

    public void setPaintingStyleDecorator(PaintingStyleDecorator psd){
        paintingStyleDecorator.setObject(psd);
    }

    public PaintingStyleDecorator getPaintingStyleDecorator(){
        return (PaintingStyleDecorator)paintingStyleDecorator.getObject();
    }
    
    
    private Object BFS(Operation operation) {
        if (this.getAnchor() != null) {
            LinkedList open = new LinkedList();
            LinkedList closed = new LinkedList();
            Branch branch;
            open.add(this.getAnchor());
            do {
                branch = (Branch)open.remove(0);
                if (operation.execute(branch)) {
                    break;
                }
                closed.add(branch);
                closed.add(branch.symmetric());
                if (!closed.contains(branch.next())) {
                    open.add(branch.next());
                }
                if (!closed.contains(branch.symmetric().next())) {
                    open.add(branch.symmetric().next());
                }
            } while (!open.isEmpty());
        }
        return operation.returnValue();
    }
    
    private interface Operation {
        public boolean execute(Branch branch);
        public Object returnValue();
    }
    
//    // XXX testing
//    int counter = 0;
//    //
    
    private class GetSpatialElements implements Operation {
        private Vector spatialElements;
        
        public GetSpatialElements() {
            this.spatialElements = new Vector();
//            // XXX testing
//            counter = 0;
//            //
        }
        
        public boolean execute(Branch branch) {
            Point2D p0 = branch.getVertex();
            Point2D p1 = branch.symmetric().getVertex();
            this.spatialElements.add(new PolyLine(new Point[]{ new Point(p0), new Point(p1)}));
            SpatialComparable[] branchElements = branch.getSpatialElements();
            for (int i = 0; i < branchElements.length; i++) {
                this.spatialElements.add(branchElements[i]);
            } 
            branchElements = branch.symmetric().getSpatialElements();
            for (int i = 0; i < branchElements.length; i++) {
                this.spatialElements.add(branchElements[i]);
            } 
            branchElements = branch.rotate().getSpatialElements();
            for (int i = 0; i < branchElements.length; i++) {
                this.spatialElements.add(branchElements[i]);
            } 
            branchElements = branch.rotate().symmetric().getSpatialElements();
            for (int i = 0; i < branchElements.length; i++) {
                this.spatialElements.add(branchElements[i]);
            } 
            
//            // XXX testing
//            this.spatialElements.add(new Point(branch.symmetric().getVertex(), String.valueOf(counter++)));
//            //
            
            return false;
        }
        
        public Object returnValue() {
            SpatialComparable[] spatialElements = new SpatialComparable[this.spatialElements.size()];
            this.spatialElements.toArray(spatialElements);
            return spatialElements;
        }
    }
    
    private class FindVertexBranch implements Operation {
        private final Point2D vertex;
        private Branch branch;
        
        public FindVertexBranch(Point2D vertex) {
            this.vertex = vertex;
            this.branch = null;
        }
        
        public boolean execute(Branch branch) {
            if (this.vertex.equals(branch.getVertex())) {
                this.branch = branch;
                return true;
            } else if (this.vertex.equals(branch.symmetric().getVertex())) {
                this.branch = branch.symmetric();
                return true;
            } else {
                return false;
            }
        }
        
        public Object returnValue() {
            return this.branch;
        }
    }
    
    private void setAnchor(Branch branch) {
        this.anchor.setObject(branch);
    }
    
    private Branch getAnchor() {
        return (Branch)this.anchor.getObject();
    }
    
    private boolean isEmpty() {
        return (this.anchor.getObject() == null);
	}
	
    // The given graph must be connected, otherwise all edges won't
    // be included in the quad-edge structure.
    // There may be no edges from a node to the same node.
    private Branch makeGraph(int[][] adjacencyList, Point2D[] vertices) {
        for (int i = 0; i < adjacencyList.length; i++) {
            for (int j = 0; j < adjacencyList[i].length; j++) {
                this.addEdge(vertices[i], vertices[adjacencyList[i][j]]);
            }
        }
        return this.getAnchor();
    }
    
    private void addEdge(Point2D p0, Point2D p1) {
        if ((p0 == null) ||
            (p1 == null) ||
            (p0 == p1)) {
            return;
        } else if (this.isEmpty()) {
            this.setAnchor(this.makeEdge(p0, p1));
        } else {
            Branch vertexBranch;
            Branch newEdge = this.makeEdge(p0, p1);
            vertexBranch = (Branch)BFS(new FindVertexBranch(newEdge.getVertex()));
            if (vertexBranch != null) {
                vertexBranch.insert(newEdge);
            }
            vertexBranch = (Branch)BFS(new FindVertexBranch(newEdge.symmetric().getVertex()));
            if (vertexBranch != null) {
                vertexBranch.insert(newEdge.symmetric());
            }
        }
    }

    private Branch makePolygon(Point2D[] polygonVertices) {
        Branch anchor = null, previous = null, next = null;
        if (polygonVertices.length > 1) {
            anchor = makeEdge(polygonVertices[0], polygonVertices[1]);
            previous = anchor;
            for (int i = 1; i < polygonVertices.length - 1; i++) {
                next = makeEdge(polygonVertices[i], polygonVertices[i + 1]);
                previous.symmetric().splice(next);
                previous = next;
            }
            next = makeEdge(polygonVertices[polygonVertices.length - 1], 
                            polygonVertices[0]);
            next.symmetric().splice(anchor);
        }
        return anchor;
    }
    
//    private Branch makePolygon(Point2D[] polygonVertices) {
//        if (polygonVertices.length > 1) {
//            addEdge(polygonVertices[0], polygonVertices[1]);
//            for (int i = 1; i < polygonVertices.length - 1; i++) {
//                addEdge(polygonVertices[i], polygonVertices[i + 1]);
//            }
//            addEdge(polygonVertices[polygonVertices.length - 1], 
//                    polygonVertices[0]);
//        }
//        return this.getAnchor();
//    }

    private Branch makeEdge(Point2D p0, Point2D p1) {
		// Create four branches that linked together 
        // represent a quad-edge object.
		Branch q0 = new Branch(), 
               q1 = new Branch(), 
               q2 = new Branch(), 
               q3 = new Branch();
		// Link the branches together creating a single quad-edge.
		q0.setNextBranch(q1);
		q1.setNextBranch(q2);
		q2.setNextBranch(q3);
		q3.setNextBranch(q0);
        // Create the links representing the topology.
		// Link 0 & 2 to themselves (= around the vertices) and
		// 1 & 3 to each other (= around the polygon)
		q0.setNext(q0);
		q2.setNext(q2);
		q1.setNext(q3);
		q3.setNext(q1);
		// Set the pointers to the vertices.
		q0.setVertex((Point2D)p0.clone());
		q2.setVertex((Point2D)p1.clone());
		// XXX set the pointers to the polygons
        // Create the visualizations for the branches.
        q0.createPolygonVis();
        q1.createPolygonVis();
        q2.createPolygonVis();
        q3.createPolygonVis();
//        q0.updateReferenceVis();
//        q1.updateReferenceVis();
//        q2.updateReferenceVis();
//        q3.updateReferenceVis();
		return q0;
	}
    
	/**
	 * Class representing a branch of a quad-edge. 
	 * Each quad-edge structure consists of four branches 
     * cyclically linked together.
	 */
	private static class Branch {
		// The next quad-edge counterclockwise. 
		// Points to the branch in the next quad-edge counterclockwise 
        // around the data (a vertex or a polygon).
		private VirtualObject nextEdge;
		// The next branch of this quad-edge counterclockwise.
        // These connect the four branches of a quad-edge in 
        // counterclockwise order.
		private VirtualObject nextBranch;
        // Reference to the vertex or polygon this branch 
        // corresponds to.
		private VirtualObject data;
		
        private Polygon polygon;
        private Point2D point;
        private PolyLine reference;
        
		private Branch() {
			this.nextEdge = new VirtualObject(null, "nextEdge");
			this.nextBranch = new VirtualObject(null, "nextBranch");;
			this.data = new VirtualObject(null, "data");
//            this.reference = new VirtualObject(null, "reference");
		}
		
		protected boolean isEmpty() {
			return (this.getVertex() == null);
		}
		
		public void splice(Branch b) {
			Branch alpha, beta, thisNext, bNext, alphaNext, betaNext;
			alpha = this.next().rotate();
			beta = b.next().rotate();
			thisNext = this.next();
			bNext = b.next();
			alphaNext = alpha.next();
			betaNext = beta.next();
			
			this.setNext(bNext);
			b.setNext(thisNext);
			alpha.setNext(betaNext);
			beta.setNext(alphaNext);
            
//            this.updateReferenceVis();
//            b.updateReferenceVis();
//            alpha.updateReferenceVis();
//            beta.updateReferenceVis();
		}
		
		public void delete() {
			// Disconnect the edge
			this.splice(this.previous());
			this.symmetric().splice(this.symmetric().previous());
		}
		
        public void insert(Branch branch) {
            // TODO check for operation legality, vertices must be equal
            Branch spliceBranch = this.chooseBranch(branch);
            if (spliceBranch != null) {
                spliceBranch.splice(branch);
            }
        }
        
        private Branch chooseBranch(Branch branch) {
            if ((this.symmetric().getVertex().equals(branch.symmetric().getVertex())) ||
                (this.next().symmetric().getVertex().equals(branch.symmetric().getVertex()))) {
                // The branch already exists.
                return null;
            } else {
                if (this == this.next()) {
                    // Currently there is only one branch.
                    return this;
                } else {
                    Point2D center = this.getVertex();
                    Point2D thisVertex = this.symmetric().getVertex();
                    Point2D branchVertex = branch.symmetric().getVertex();
                    double centerX = center.getX();
                    double centerY = center.getY();
                    double thisX = thisVertex.getX();
                    double thisY = thisVertex.getY();
                    double branchX = branchVertex.getX();
                    double branchY = branchVertex.getY();
                    double thisAngle = Math.atan2(thisY - centerY, thisX - centerX);
                    double branchAngle = Math.atan2(branchY - centerY, branchX - centerX);
                    if (branchAngle == thisAngle) {
                        return this;
                    } else if (thisAngle < branchAngle) {
                        return this.next().chooseBranch(branch, thisAngle, true, this);
                    } else {
                        return this.previous().chooseBranch(branch, thisAngle, false, this);
                    }
                }
            }
        }
        
        private Branch chooseBranch(Branch branch, double previousAngle, boolean directionNext, Branch first) {
            Point2D center = this.getVertex();
            Point2D thisVertex = this.symmetric().getVertex();
            Point2D branchVertex = branch.symmetric().getVertex();
            double centerX = center.getX();
            double centerY = center.getY();
            double thisX = thisVertex.getX();
            double thisY = thisVertex.getY();
            double branchX = branchVertex.getX();
            double branchY = branchVertex.getY();
            double thisAngle = Math.atan2(thisY - centerY, thisX - centerX);
            double branchAngle = Math.atan2(branchY - centerY, branchX - centerX);
            if (directionNext) {
                if ((thisAngle >= branchAngle) || (thisAngle < previousAngle) || (this == first)) {
                    return this.previous();
                } else {
                    return this.next().chooseBranch(branch, thisAngle, true, first);
                }
            } else {
                if ((thisAngle <= branchAngle) || (thisAngle > previousAngle) || (this == first)) {
                    return this;
                } else {
                    return this.previous().chooseBranch(branch, thisAngle, false, first);
                }
            }
        }

        protected void setNextBranch(Branch nextBranch) {
			this.nextBranch.setObject(nextBranch);
		}
		
		protected void setNext(Branch next) {
			this.nextEdge.setObject(next);
		}
		
		protected void setVertex(Point2D o) {
			this.data.setObject(o);
		}
        
        public void createPolygonVis() {
            // XXX refactor!
            // Create the rectangle visualizing this branch.
            if (this.getVertex() != null) {
                Point2D p0 = this.getVertex();
                Point2D p1 = this.symmetric().getVertex();
                int ab = 4;
                double dx = p1.getX() - p0.getX();
                double dy = p1.getY() - p0.getY();
                Point2D center = new Point2D.Double(p0.getX() + (dx / 2),
                                                    p0.getY() + (dy / 2));
                double length = Math.sqrt(dx * dx + dy * dy);
                Point2D unitVector = new Point2D.Double(dx / length, dy / length);
                Point2D normalVector = new Point2D.Double(-dy / length, dx / length);
                Point[] square = new Point[]{new Point(center.getX() - 3 * ab * unitVector.getX() - ab * normalVector.getX(), 
                                                       center.getY() - 3 * ab * unitVector.getY() - ab * normalVector.getY()),
                                             new Point(center.getX() - 3 * ab * unitVector.getX() + ab * normalVector.getX(), 
                                                       center.getY() - 3 * ab * unitVector.getY() + ab * normalVector.getY()), 
                                             new Point(center.getX() - ab * unitVector.getX() + ab * normalVector.getX(), 
                                                       center.getY() - ab * unitVector.getY() + ab * normalVector.getY()),
                                             new Point(center.getX() - ab * unitVector.getX() - ab * normalVector.getX(), 
                                                       center.getY() - ab * unitVector.getY() - ab * normalVector.getY())};
                this.polygon = new Polygon(square);
                this.point = new Point2D.Double(center.getX() - 2 * ab * unitVector.getX(), 
                                                center.getY() - 2 * ab * unitVector.getY());
            } else {
                // This branch refers to a polygon face.
                Point2D p0 = this.rotate().getVertex();
                Point2D p1 = this.rotate().symmetric().getVertex();
                int ab = 4;
                double dx = p1.getX() - p0.getX();
                double dy = p1.getY() - p0.getY();
                Point2D center = new Point2D.Double(p0.getX() + (dx / 2),
                                                    p0.getY() + (dy / 2));
                double length = Math.sqrt(dx * dx + dy * dy);
                Point2D unitVector = new Point2D.Double(dx / length, dy / length);
                Point2D normalVector = new Point2D.Double(-dy / length, dx / length);
                Point[] square = new Point[]{new Point(center.getX() - ab * unitVector.getX() + ab * normalVector.getX(), 
                                                       center.getY() - ab * unitVector.getY() + ab * normalVector.getY()),
                                             new Point(center.getX() - ab * unitVector.getX() + 3 * ab * normalVector.getX(), 
                                                       center.getY() - ab * unitVector.getY() + 3 * ab * normalVector.getY()), 
                                             new Point(center.getX() + ab * unitVector.getX() + 3 * ab * normalVector.getX(), 
                                                       center.getY() + ab * unitVector.getY() + 3 * ab * normalVector.getY()),
                                             new Point(center.getX() + ab * unitVector.getX() + ab * normalVector.getX(), 
                                                       center.getY() + ab * unitVector.getY() + ab * normalVector.getY())};
                this.polygon = new Polygon(square);
                PaintingStyleDecorator psd = new PaintingStyleDecorator();
                //psd.setStroke(new BasicStroke(1));
                this.polygon.setPaintingStyleDecorator(psd);
                this.point = new Point2D.Double(center.getX() + 2 * ab * normalVector.getX(), 
                                                center.getY() + 2 * ab * normalVector.getY());
            }
            PaintingStyleDecorator psd = new PaintingStyleDecorator();
            psd.setLineColor(Color.WHITE);
            this.polygon.setPaintingStyleDecorator(psd);
        }
		
        public void updateReferenceVis() {
            if (this.getVertex() != null) {
                Point2D center = this.getVertex();
                Point2D p0 = this.point;
                Point2D p1 = this.next().point;

                Point[] points = new Point[]{new Point(p0),
                                             new Point(p1)};
                PolyLine ref = new PolyLine(points);
                PaintingStyleDecorator psd = new PaintingStyleDecorator();
                psd.setLineColor(Color.RED);
                //psd.setStroke(new BasicStroke(1));
                ref.setPaintingStyleDecorator(psd);
//                this.reference.setObject(ref);
                reference = ref;
//                this.reference.setObject(new PolyLine(new Point[0]));
                // XXX
            } else {
                Point2D p0 = this.point;
                Point2D p1 = this.next().point;
                int ab = 4;
                double dx = p1.getX() - p0.getX();
                double dy = p1.getY() - p0.getY();
                double length = Math.sqrt(dx * dx + dy * dy);
                Point2D unitVector = new Point2D.Double(dx / length, dy / length);
                Point2D normalVector = new Point2D.Double(-dy / length, dx / length);
                Point[] points = new Point[]{new Point(p0), 
                                             new Point(p1)};
                PolyLine ref = new PolyLine(points);
                PaintingStyleDecorator psd = new PaintingStyleDecorator();
                psd.setLineColor(Color.BLUE);
                //psd.setStroke(new BasicStroke(1));
                ref.setPaintingStyleDecorator(psd);
//                this.reference.setObject(ref);
                reference = ref;

            }
        }

        public SpatialComparable[] getSpatialElements() {
            this.updateReferenceVis();
            return new SpatialComparable[]{ this.polygon, this.reference };
        }
        
		public Point2D getVertex() {
			return (Point2D)this.data.getObject();
		}
		
		public Branch rotate() {
			return (Branch)this.nextBranch.getObject();
		}
		
		public Branch next() {
			return (Branch)this.nextEdge.getObject();
		}
		
		public Branch previous() {
			return this.rotate().next().rotate();
		}
		
		public Branch symmetric() {
			return this.rotate().rotate();
		}
	} 

}
