// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
package matrix.structures.spatial.CDT.probe;

import java.awt.Rectangle;
import matrix.structures.CDT.CDT;
import matrix.structures.FDT.FDT;
import matrix.structures.FDT.Tree;
import matrix.structures.memory.VirtualObject;
import matrix.structures.spatial.Area;
import matrix.structures.spatial.PaintingStyleDecorator;
import matrix.structures.spatial.SpatialComparable;
import matrix.structures.spatial.FDT.probe.SpatialElement;
import matrix.structures.util.Iterator;

/**
 * An abstract class that defines the basic functionality of a quadtree.
 * Quadtree is a tree data structure, in which each non-leaf, i.e. internal 
 * node has four children.
 * 
 * @see QuadTreeNode
 * @author Juha Helminen
 */
public abstract class QuadTree implements CDT, Tree, FDT, Iterator, Area {

	// The root node of the tree.
	protected VirtualObject root;

	
	// -- Constructors --
	
	/**
	 * Contructs an empty tree.
	 */
	public QuadTree() {
		this(null);
	}
	
	/**
	 * Contructs an empty tree and inserts object o into it.
	 * 
	 * @param o The data to be inserted into the new tree.
	 */
	public QuadTree(Object o) {
        root = new VirtualObject(this, "root");
        root.setObject(getNewNode(o));
	}

	
	// -- Implementation of interface Tree -- 
	
	abstract public Tree getNewNode(Object key);
	
	public int getSubTreeCount() {
        return ((Tree)getElement()).getSubTreeCount();
	}

	public Tree[] getSubTrees() {
        return ((Tree)getElement()).getSubTrees();
	}

	public void setSubTree(Tree t, int child) {
        ((Tree)getElement()).setSubTree(t, child);
	}


	// -- Implementation of interface Area --

	public SpatialElement[] getSpatialElements() {
		return ((QuadTreeNode)this.getElement()).getSpatialElements();
	}

	public Area[] getSubAreas() {
		return ((QuadTreeNode)this.getElement()).getSubAreas();
	}

	public Rectangle getBoundingPolygon() {
		return ((QuadTreeNode)this.getElement()).getBoundingPolygon();
	}

	public void addSubArea(Area a) {
		((QuadTreeNode)this.getElement()).addSubArea(a);
	}

	public void removeSubArea(Area a) {
		((QuadTreeNode)this.getElement()).removeSubArea(a);
	}

	public void insert(SpatialComparable e) {
		this.insert((Object)e);
	}

	public void delete(SpatialComparable e) {
		this.delete((Object)e);
	}

    public boolean hasPaintingStyleDecorator(){
        return ((PRQuadTreeNode)getElement()).hasPaintingStyleDecorator();
    }

    public void setPaintingStyleDecorator(PaintingStyleDecorator psd){
        ((PRQuadTreeNode)getElement()).setPaintingStyleDecorator(psd);
    }

    public PaintingStyleDecorator getPaintingStyleDecorator(){
        return ((PRQuadTreeNode)getElement()).getPaintingStyleDecorator();
    }


	// -- Implementation of interface CDT -- 

	abstract public CDT delete(Object o);

	abstract public CDT insert(Object o);

	abstract public Object search(Object o);

	abstract public CDT getNewInstance();

	
	// -- Implementation of interface FDT -- 
	
	public Object getElement() {
		return root.getObject();
	}

	abstract public void setElement(Object o);

	
	// -- Implementation of interface Iterator --

	public boolean hasNext() {
		return ((QuadTreeNode)this.getElement()).hasNext();
	}

	public Object next() {
		return ((QuadTreeNode)this.getElement()).next();
	}

	public void reset() {
		((QuadTreeNode)this.getElement()).reset();
	}

	
	// -- Overridden methods inherited from class Object --
	
	public boolean equals(Object obj) {
	      if (obj instanceof QuadTree) {
	          return this.getElement().equals(((QuadTree)obj).getElement());
	      } else {
	    	  return false;
	      }
	}

	public String toString() {
		return this.getElement().toString(); 
	}
	
}
