// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
package matrix.structures.spatial.CDT.probe;

import java.awt.Color;
import java.awt.Rectangle;
import java.util.ArrayList;

import matrix.decoration.LabelDecorator;
import matrix.structures.FDT.FDT;
import matrix.structures.FDT.Tree;
import matrix.structures.FDT.probe.Key;
import matrix.structures.FDT.probe.Table;
import matrix.structures.memory.VirtualArray;
import matrix.structures.memory.VirtualInteger;
import matrix.structures.memory.VirtualObject;
import matrix.structures.spatial.Area;
import matrix.structures.spatial.PaintingStyleDecorator;
import matrix.structures.spatial.SpatialComparable;
import matrix.structures.spatial.FDT.probe.SpatialElement;
//import matrix.structures.spatial.FDT.probe.Rectangle;
import matrix.structures.util.Iterator;
import matrix.util.Note;

/**
 * An abstract class that defines the basic functionality and structure 
 * of a node in a quadtree. Each non-leaf, i.e. internal node has four 
 * children. Additionally all nodes contain a data/key field and 
 * coordinates, which define what rectangular area the node corresponds 
 * to in 2-dimensional space.
 * 
 * @see QuadTree
 * @author Juha Helminen
 */
public abstract class QuadTreeNode implements Tree, FDT, Iterator, Area, LabelDecorator {

    private VirtualObject paintingStyleDecorator = new VirtualObject(null, "PaintingStyleDecorator");
	// Literal constants for the node's four children,
	// which all represent a quadrant of the node's area.
	// (Replace with enum in Java 5)
	protected static final int QUADRANT_NE = 0;
	protected static final int QUADRANT_SE = 1;
	protected static final int QUADRANT_SW = 2;
	protected static final int QUADRANT_NW = 3;
	protected static final int OUTSIDE = -1;
	protected static final int INDIVISIBLE = -2;
	// Default values for the 2-dimensional area of a new node.
	protected static final double DEFAULT_MINX = 0;
	protected static final double DEFAULT_MAXX = 128;
	protected static final double DEFAULT_MINY = 0;
	protected static final double DEFAULT_MAXY = 128;
	

	/** 
    * The children, quadrants:
    * children[0] is the 1st child = NorthEast quadrant
    * children[1] is the 2nd child = SE quadrant
    * children[2] is the 3rd child = SW quadrant
    * children[3] is the 4th child = NW quadrant
    * For leaf nodes: children[0] = children[1] = children[2] = children[3] = null.
    * */
	private VirtualArray children;
	/**
	 * Contains the keys/data stored in the node.
	 */
	protected Table keys;
	protected VirtualInteger numberOfKeys;
	// These define the rectangular area, which the node corresponds to.
	protected double minX, maxX, minY, maxY;
	private Rectangle bounds;

	// -- Constructors --

	/**
	 * Constructs an empty node corresponding to the default area.
	 */
	public QuadTreeNode() {
		this(null, DEFAULT_MINX, DEFAULT_MAXX, DEFAULT_MINY, DEFAULT_MAXY);
	}

	/**
	 * Constructs an empty node corresponding to the default area and
	 * insert object key into it.
	 * @param key The key/data to be inserted in the node.
	 */
	public QuadTreeNode(SpatialComparable key) {
		this(key, DEFAULT_MINX, DEFAULT_MAXX, DEFAULT_MINY, DEFAULT_MAXY);
	}

	/**
	 * Constructs an empty node corresponding to the area given 
	 * as parameter and inserts object key in it.
	 * 
	 * @param key The key/data to be inserted in the node.
	 * @param minX The x-coordinate of the lower left corner of the rectangular area.
	 * @param maxX The x-coordinate of the upper right corner of the rectangular area.
	 * @param minY The y-coordinate of the lower left corner of the rectangular area.
	 * @param maxY The y-coordinate of the upper right corner of the rectangular area.
	 */
	public QuadTreeNode(SpatialComparable key, double minX, double maxX, double minY, double maxY) {
		this.children = new VirtualArray(this, "QuadTreeNode.children");
		this.keys = new Table();
		this.numberOfKeys = new VirtualInteger(0, this, "numberOfKeys");
		this.keys.setObject(key, 0);
		if (key != null) {
			this.numberOfKeys.assign(1);
		}
		this.minX = minX;
		this.maxX = maxX;
		this.minY = minY;
		this.maxY = maxY;
		bounds = new Rectangle((int)this.minX, (int)this.minY, (int)(this.maxX - this.minX), (int)(this.maxY - this.minY));
	}

	
	
	// -- Getters = accessors --
	
	/**
	 * @return The (first) key/data element stored in the node.
	 */
	public SpatialComparable getKey() {
		return this.getKey(0);
	}
	
	/**
	 * @return The array of keys.
	 */
	public Table getKeys() {
		return this.keys;
	}

	/**
	 * @param index
	 * @return The key/data element stored in the node at index.
	 */
	public SpatialComparable getKey(int index) {
		return (SpatialComparable)this.keys.getObject(index);
	}

	/**
	 * @return The reference to the NE-child of the node or null if node is a leaf.
	 */
	public QuadTreeNode getNE() {
		return (QuadTreeNode)children.getObject(0);
	}
	
	/**
	 * @return The reference to the SE-child of the node or null if node is a leaf.
	 */
	public QuadTreeNode getSE() {
		return (QuadTreeNode)children.getObject(1);
	}

	/**
	 * @return The reference to the SW-child of the node or null if node is a leaf.
	 */
	public QuadTreeNode getSW() {
		return (QuadTreeNode)children.getObject(2);
	}

	/**
	 * @return The reference to the NW-child of the node or null if node is a leaf.
	 */
	public QuadTreeNode getNW() {
		return (QuadTreeNode)children.getObject(3);
	}

	

	// -- Setters = mutators --
	
	/**
	 * Set the key(s)/data element(s) in the node to object key.
	 * @param key The object to be stored in the node.
	 */
	public void setKey(Object key) {
		this.numberOfKeys.assign(1);
		this.setKey(key, 0);
	}

	/**
	 * Stores key in the node at index.
	 * @param key The object to be stored in the node.
	 * @param index
	 */
	public void setKey(Object key, int index) {
		if ((key == null) || (key instanceof SpatialComparable)) {
			keys.setObject(key, index);
		} else if (key instanceof Key) {
			this.setKey(((Key)key).getObject());
		} else {
			Note.warning(this, "Cannot insert object of class " + key.getClass());
		}
	}

	/**
	 * Set the NE-child of the node to the node given as parameter.
	 * @param qtNode The node to be set as a child.
	 */
	protected void setNE(QuadTreeNode qtNode) {
		children.setObject(qtNode, 0);
		if (qtNode != null) {
			qtNode.setLabel("NE");
		}
	}
	
	/**
	 * Set the SE-child of the node to the node given as parameter.
	 * @param qtNode The node to be set as a child.
	 */
	protected void setSE(QuadTreeNode qtNode) {
		children.setObject(qtNode, 1);
		if (qtNode != null) {
			qtNode.setLabel("SE");
		}
    }

	/**
	 * Set the SW-child of the node to the node given as parameter.
	 * @param qtNode The node to be set as a child.
	 */
	protected void setSW(QuadTreeNode qtNode) {
		children.setObject(qtNode, 2);
		if (qtNode != null) {
			qtNode.setLabel("SW");
		}
    }

	/**
	 * Set the NW-child of the node to the node given as parameter.
	 * @param qtNode The node to be set as a child.
	 */
	protected void setNW(QuadTreeNode qtNode) {
		children.setObject(qtNode, 3);
		if (qtNode != null) {
			qtNode.setLabel("NW");
		}
	}
	

	
	// -- Implementation of interface FDT --
	
	public Object getElement() {
		// If the node contains only one key, return it,
		// otherwise return the array of keys.
		if (this.numberOfKeys.eval() > 1) {
			return this.getKeys();
		} else {
			return this.getKey();
		}
	}

	public void setElement(Object o) {
		if (o == null) {
			this.numberOfKeys.assign(0);
			this.setKey(o); 
		} else if (o instanceof SpatialComparable) {
			this.setKey(o); 
		} else if (o instanceof Key) {
			this.setElement(((Key)o).getObject());
		} else {
			Note.warning(this, "Cannot insert object of class " + o.getClass());
		}
	}


	
	// -- Implementation of interface Tree --
	
	public int getSubTreeCount() {
		if (this.isLeaf()) {
			// Leaves have no children.
			return 0;
		} else {
			// Internal nodes have 4 (though possibly empty) children.
			return 4;
		}
	}

	public Tree[] getSubTrees() {
        Tree[] subTrees = new Tree[4];
        /*
         * The quadrants are visualized in the tree in the
         * following order. 		
         * 	   __o__
         * 	  / / \ \
         *   / |   | \ 
         *  SW NW SE NE
         */
        subTrees[0] = this.getSW();
        subTrees[1] = this.getNW();
        subTrees[2] = this.getSE();
        subTrees[3] = this.getNE();
        return subTrees;
	}

	public void setSubTree(Tree t, int child) {
        if ((t == null) || (t instanceof QuadTreeNode)) {
        	switch (child) {
        		case 0: this.setSW((QuadTreeNode)t); break;
        		case 1: this.setNW((QuadTreeNode)t); break;
        		case 2: this.setSE((QuadTreeNode)t); break;
        		case 3: this.setNE((QuadTreeNode)t); break;
        		default: Note.warning(this, "Index out of bounds.");
        	}
        } else {
        	Note.warning(this, "Parameter " + t + " is not a quadtree node.");
        }
	}

	abstract public Tree getNewNode(Object key);
	
	
	
	// -- Implementation of interface Area --

	public SpatialElement[] getSpatialElements() {
        SpatialElement[] keys = new SpatialElement[this.numberOfKeys.eval()];
		for (int i = 0; i < this.numberOfKeys.eval(); i++) {
			keys[i] = (SpatialElement)this.getKey(i);
		}
		return keys;
	}

	public Area[] getSubAreas() {
		Area[] children = new Area[4];
		for (int i = 0; i < 4; i++) {
			children[i] = (Area)this.getChild(i);
		}
		return children;
	}

	public Rectangle getBoundingPolygon() {
		return bounds;
	}

	public void addSubArea(Area a) {
		Note.warning(this, "Method addSubArea() not implemented.");
	}

	public void removeSubArea(Area a) {
		Note.warning(this, "Method removeSubArea() not implemented.");
	}

	abstract public void insert(SpatialComparable e);

	abstract public void delete(SpatialComparable e);

    public boolean hasPaintingStyleDecorator(){
        return (paintingStyleDecorator.getObject() != null);
    }

    public void setPaintingStyleDecorator(PaintingStyleDecorator psd){
        paintingStyleDecorator.setObject(psd);
    }

    public PaintingStyleDecorator getPaintingStyleDecorator(){
        return (PaintingStyleDecorator)paintingStyleDecorator.getObject();
    }
	

	// -- Implementation of interface Iterator --
	
	private java.util.Iterator keyIterator = null;

	/**
	 * Recursively traverse the tree starting from this node
	 * and insert the found keys/data into the list given as 
	 * parameter.
	 * @param keys The list of keys found so far.
	 */
	private void recursiveTraverse(ArrayList keys) {
		if (!this.isEmpty()) {
			for (int i = 0; i < this.numberOfKeys.eval(); i++) {
				keys.add(this.getKey(i));
			}
		} else {
			if (this.getNE() != null) this.getNE().recursiveTraverse(keys);
			if (this.getSE() != null) this.getSE().recursiveTraverse(keys);
			if (this.getSW() != null) this.getSW().recursiveTraverse(keys);
			if (this.getNW() != null) this.getNW().recursiveTraverse(keys);
		}
	}
	
	public boolean hasNext() {
		if (keyIterator == null) {
			ArrayList keys = new ArrayList();
			// Find the keys stored in the subtree starting from this node
			// and insert them into the list key.
			this.recursiveTraverse(keys);
			// Generate an iterator for the newly created list of keys.
			keyIterator = keys.iterator();
		}
		return keyIterator.hasNext();
	}

	public Object next() {
		return keyIterator.next();
	}

	public void reset() {
		keyIterator = null;
	}

	
	
	// -- Helper methods for implementing quadtree operations --
	
	/**
	 * @param quadrant Specifies the child to get. Literal constants for quadrants 
	 * are defined in the class QuadTreeNode. Legal values are 0-3.
	 * @return The child corresponding to parameter quadrant or null if an illegal parameter.
	 */
	protected QuadTreeNode getChild(int quadrant) {
		switch(quadrant) {
			case QUADRANT_NE: return this.getNE();
			case QUADRANT_SE: return this.getSE();
			case QUADRANT_SW: return this.getSW();
			case QUADRANT_NW: return this.getNW();
			default: return null;
		}
	}

	/**
	 * @param child The node to be set as the new child.
	 * @param quadrant Specifies the child in which to store the node. 
	 * Literal constants for quadrants are defined in the class 
	 * QuadTreeNode. Legal values are 0-3.
	 */
	protected void setChild(QuadTreeNode child, int quadrant) {
		switch(quadrant) {
			case QUADRANT_NE: this.setNE(child); break;
			case QUADRANT_SE: this.setSE(child); break;
			case QUADRANT_SW: this.setSW(child); break;
			case QUADRANT_NW: this.setNW(child); break;
			default: /* do nothing */ break; 
		}
	}

	/**
	 * @return The number of actual children (internal or data containing leaf nodes)
	 * this node has. 0 for leaf nodes and 1-4 for internal nodes.
	 */
	protected int getChildCount() {
		int count = 4;
		if ((this.getNE() == null) || (this.getNE().isLeaf() && this.getNE().isEmpty())) count--;
		if ((this.getSE() == null) || (this.getSE().isLeaf() && this.getSE().isEmpty())) count--;
		if ((this.getSW() == null) || (this.getSW().isLeaf() && this.getSW().isEmpty())) count--;
		if ((this.getNW() == null) || (this.getNW().isLeaf() && this.getNW().isEmpty())) count--;
		return count;
	}
	
	public boolean isEmpty() {
		return (this.numberOfKeys.eval() == 0);
	}
	
	/**
	 * Destroys all references to children.
	 */
	protected void deleteChildren() {
		this.setNE(null);
		this.setSE(null);
		this.setSW(null);
		this.setNW(null);
	}
	
	public boolean isLeaf() {
		return ((this.getNE() == null) &&
				(this.getSE() == null) &&
				(this.getSW() == null) &&
				(this.getNW() == null));
	}


	
	// -- Overridden methods inherited from class Object --

	public String toString() {
		return "\n" + this.toString("\n ");
	}

	protected String toString(String indent) {
		String ne, se, sw, nw;
		if (this.isLeaf()) {
			return ((this.getElement() == null) ? 
		               "-o" : 
		               "-" + this.getElement().toString());
		} else {
			if (this.getSW() == null) {
				sw = "-o";
			} else {
				sw = this.getSW().toString(indent + "|  ");
			}
			if (this.getNW() == null) {
				nw = "-o";
			} else {
				nw = this.getNW().toString(indent + "|  ");
			}
			if (this.getSE() == null) {
				se = "-o";
			} else {
				se = this.getSE().toString(indent + "|  ");
			}
			if (this.getNE() == null) {
				ne = "-o";
			} else {
				ne = this.getNE().toString(indent + "   ");
			}
			return ((this.getElement() == null) ? 
		               "-o" : 
		               "-" + this.getElement().toString()) + indent + "|-" +
				   sw + indent + "|-" +
				   nw + indent + "|-" +
				   se + indent + "\\-" +
				   ne + indent;
		}
	}
	
	public boolean equals(Object obj) {
		if (obj == null) {
			return false;
		} else if (obj instanceof QuadTreeNode) {
			// Compare the reference
			if (this == obj) {
				return true;
			} else {
				// Compare the key
				if ((((this.getElement() == null) &&
					 ((QuadTreeNode)obj).getElement() == null)) || 
				    (this.getElement().equals(((QuadTreeNode)obj).getElement()))) {
					// Recursively compare the subtrees
					return (((this.getNE() == null) ? 
								(((QuadTreeNode)obj).getNE() == null) : 
									this.getNE().equals(((QuadTreeNode)obj).getNE())) &&
							((this.getSE() == null) ? 
									(((QuadTreeNode)obj).getSE() == null) : 
										this.getSE().equals(((QuadTreeNode)obj).getSE())) &&
							((this.getSW() == null) ? 
									(((QuadTreeNode)obj).getSW() == null) : 
										this.getSW().equals(((QuadTreeNode)obj).getSW())) &&
							((this.getNW() == null) ? 
									(((QuadTreeNode)obj).getNW() == null) : 
										this.getNW().equals(((QuadTreeNode)obj).getNW())));
				} else {
					return false;
				}
			}
		} else {
			return false;
		}
	}


	// -- Implementation of interface LabelDecorator --

	private transient boolean labelEnabled = true;
    private String label = "";
    
    public boolean isLabelEnabled() {
        return this.labelEnabled;
    }

    public void setLabelEnabled(boolean b) {
        this.labelEnabled = b;
    }

    public void setLabel(String s) {
        label = s;
    }

    public String getLabel() {
        return label;
    }

    public boolean isReferenceLabelEnabled() {
        return false;
    }

    public void setReferenceLabelEnabled(boolean b) {
    }

    public void setReferenceLabel(int i, String s) {
    }

    public String getReferenceLabel(int i) {
        return null;
    }
    
	public Color getLabelColor() {
    	// Return null if the color of the label is not defined
    	return null;
    }
    
}
