// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
package matrix.structures.spatial.CDT.probe;

import java.awt.Color;
import java.awt.Rectangle;

import matrix.decoration.LabelDecorator;
import matrix.structures.CDT.CDT;
import matrix.structures.FDT.FDT;
import matrix.structures.FDT.Tree;
import matrix.structures.memory.VirtualObject;
import matrix.structures.spatial.Area;
import matrix.structures.spatial.PaintingStyleDecorator;
import matrix.structures.spatial.SpatialComparable;
import matrix.structures.spatial.FDT.probe.SpatialElement;
import matrix.structures.util.Iterator;
import matrix.util.Note;

/**
 * Class defining the basic functionality of an R-tree, which is B-tree-like 
 * data structure used to index spatial data.
 *  
 * @see RTreeNode
 * @author Juha Helminen
 */
public class RTree implements CDT, Tree, FDT, Iterator, Area, LabelDecorator {

	// The root node of the tree.
	protected VirtualObject root;
	private static final Rectangle DUMMY_BOUNDS = new Rectangle(0, 0, 0, 0);
	
	// -- Constructors --
	
	public RTree() {
        root = new VirtualObject(this, "root");
        root.setObject(new RTreeNode());
	}

	/**
	 * @param minEntries The minimum allowed number of entries in a node.
	 * @param maxEntries The maximum allowed number of entries in a node.
	 */
	public RTree(int minEntries, int maxEntries) {
        root = new VirtualObject(this, "root");
        root.setObject(new RTreeNode(minEntries, maxEntries));
	}

	// -- Implementation of interface Tree -- 
	
	public Tree getNewNode(Object key) {
		return ((RTreeNode)this.getElement()).getNewNode(key);
	}
	
	public int getSubTreeCount() {
        return ((Tree)this.getElement()).getSubTreeCount();
	}

	public Tree[] getSubTrees() {
        return ((Tree)this.getElement()).getSubTrees();
	}

	public void setSubTree(Tree t, int child) {
        ((Tree)this.getElement()).setSubTree(t, child);
	}

	
	// -- Implementation of interface Area -- 

	public SpatialElement[] getSpatialElements() {
		return ((RTreeNode)this.getElement()).getSpatialElements();
	}

	public Area[] getSubAreas() {
		return ((RTreeNode)this.getElement()).getSubAreas();
	}
	public Rectangle getBoundingPolygon() {
		if (((RTreeNode)this.getElement()).isEmpty()) {
			// The tree is empty and has no real bounds.
			return RTree.DUMMY_BOUNDS;
		} else {
			return ((RTreeNode)this.getElement()).getBoundingPolygon();
		}
	}

	public void insert(SpatialComparable e) {
		this.insert((Object)e);
	}

	public void delete(SpatialComparable e) {
		this.delete((Object)e);
	}

	public void addSubArea(Area a) {
		((RTreeNode)this.getElement()).addSubArea(a);
	}
	
	public void removeSubArea(Area a) {
		((RTreeNode)this.getElement()).removeSubArea(a);
	}

    public PaintingStyleDecorator getPaintingStyleDecorator() {
        return ((RTreeNode)this.getElement()).getPaintingStyleDecorator();
    }

    public boolean hasPaintingStyleDecorator() {
        return ((RTreeNode)this.getElement()).hasPaintingStyleDecorator();
    }

    public void setPaintingStyleDecorator(PaintingStyleDecorator psd) {
        ((RTreeNode)this.getElement()).setPaintingStyleDecorator(psd);
    }
    

	// -- Implementation of interface CDT -- 

    public CDT insert(Object o) {
		// Set the new root, in case it's changed
		this.setElement(((RTreeNode)this.getElement()).insert(o));
		return this;
	}

	public CDT delete(Object o) {
		this.setElement(((RTreeNode)this.getElement()).delete(o));
		return this;
	}

	/** Obsolete method. */
	public Object search(Object o) {
		return null;
	}

	public CDT getNewInstance() {
		return new RTree();
	}

	
	// -- Implementation of interface FDT -- 
	
	public Object getElement() {
		return root.getObject();
	}

	public void setElement(Object o) {
        if (o instanceof RTreeNode)
            root.setObject(o);
        else if (o == null) {
            root.setObject(getNewNode(null));
        } else {
            Note.err(this, "Cannot set non-RTree as root.");
        }
	}

	
	// -- Implementation of interface iterator --

	public boolean hasNext() {
		return ((RTreeNode)this.getElement()).hasNext();
	}

	public Object next() {
		return ((RTreeNode)this.getElement()).next();
	}

	public void reset() {
		((RTreeNode)this.getElement()).reset();
	}

    
    // -- Implementation of interface LabelDecorator --
	
	public String getLabel() {
        return ((RTreeNode)this.getElement()).getLabel();
    }

    public String getReferenceLabel(int i) {
        return ((RTreeNode)this.getElement()).getReferenceLabel(i);
    }

    public boolean isLabelEnabled() {
        return ((RTreeNode)this.getElement()).isLabelEnabled();
    }

    public boolean isReferenceLabelEnabled() {
        return ((RTreeNode)this.getElement()).isReferenceLabelEnabled();
    }

    public void setLabel(String s) {
        ((RTreeNode)this.getElement()).setLabel(s);
    }

    public void setLabelEnabled(boolean b) {
        ((RTreeNode)this.getElement()).setLabelEnabled(b);
    }

    public void setReferenceLabel(int i, String s) {
        ((RTreeNode)this.getElement()).setReferenceLabel(i, s);
    }

    public void setReferenceLabelEnabled(boolean b) {
        ((RTreeNode)this.getElement()).setReferenceLabelEnabled(b);
    }

	public Color getLabelColor() {
    	// Return null if the color of the label is not defined
    	return null;
    }
    
    // -- Overridden methods inherited from class Object --

    public boolean equals(Object obj) {
	      if (obj instanceof RTree) {
	          return this.getElement().equals(((RTree)obj).getElement());
	      } else {
	    	  return false;
	      }
	}

	public String toString() {
		return this.getElement().toString(); 
	}

	private static final long serialVersionUID = -8495402418477451004L;
}
