// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
package matrix.structures.spatial.FDT.probe;

import matrix.structures.spatial.SpatialComparable;

import java.awt.geom.Rectangle2D;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Point2D;
import java.awt.geom.Area;

import matrix.structures.FDT.Struct;
import matrix.structures.FDT.probe.Key;
import matrix.structures.memory.VirtualObject;
import matrix.structures.memory.VirtualDouble;
import matrix.structures.memory.VirtualInteger;

/**
 * A spatial data element representing an ellipse.  It is important to note that
 * the (x,y) coordinate pair indicates the <strong>center</strong> of the ellipse instead of the 
 * upper left corner, as is the case in some graphical applications (including java class java.awt.geom.Ellipse2D). 
 * The defined coordinates are placed at the center of the ellipse in order to keep the data structure
 * representation closer to mathematical representation of an ellipse. This is not an element of computer graphics, but
 * a data structure. 
 * 
 * This has the additional effect that the (x,y) coordinate pair stored in the Ellipse is not the
 * same as what is in the Ellipse2D instance contained in this object.  
 * 
 * Since this object is a data element in Matrix, it has a string key which is used to identify this object
 * in the framework (outside the spatial context).
 * 
 * @author Jussi Nikander
 *
 */
public class Ellipse extends SpatialElement implements SpatialComparable, Struct {

	//private Ellipse2D ellipse;
	private Point center;
	private VirtualObject height, width;

	private static String[] fieldNames = {"id", "center", "width", "height" };
	
	public Ellipse() {
		this(0,0,0,0,"");
	}
	
	/**
	 * Constructor for an spatial data element representing an ellipse.
	 * @param x The x coordinate of the center of the ellipse
	 * @param y The y coordiante of the center of the ellipse
	 * @param w The width of the ellipse
	 * @param h The height of the ellipse
	 * @param key The key (string identifier) of the ellipse
	 */
	public Ellipse(double x, double y, double w, double h, String key) {
		super(key);
		center = new Point(x,y,"");
		width = new VirtualObject(new Double(w), this, "width");
		height = new VirtualObject(new Double(h), this, "height");
		// The x,y for Ellipse2D refers to the upper left coordinate of the ellipse
		//ellipse = new Ellipse2D.Double(x - (w / 2), y - (h / 2), w, h);
	}

	
	/**
	 * This Ellipse object contains an instance of java.awt.geom.Ellipse2D, which represents
	 * the same ellipse as this object. This method returns that Ellipse2D.
	 * @return an Ellipse2D object representing the same ellipse as this object.
	 */
	public Ellipse2D getEllipse2D() {
		double x, y, w, h;
		x = center.getX();
		y = center.getY();
		w = ((Double)width.getObject()).doubleValue();
		h = ((Double)height.getObject()).doubleValue();
		return new Ellipse2D.Double(x - (w / 2), y - (h / 2), w, h);
	}
	
	/**
	 * Returns the center point of this ellipse.
	 * @return return Point representing the center of this ellipse
	 */
	public Point getCenter() {
		return center;
	}
	
	public double getHeight() {
		return ((Double)height.getObject()).doubleValue();
	}
	
	public double getWidth() {
		return ((Double)width.getObject()).doubleValue();
	}
	public boolean intersects(SpatialComparable sc) {
		if(sc instanceof Point) {
			Point2D p = ((Point)sc).getPoint2D();
			return getEllipse2D().contains(p);
		} else {
			Area a = new Area(getEllipse2D());
			Area ca = new Area(sc.getShape());
			a.intersect(ca);
			return !a.isEmpty();
		} 
	}

	public boolean contains(SpatialComparable sc) {
		if(sc instanceof Point) {
			Point2D p = ((Point)sc).getPoint2D();
			return getEllipse2D().contains(p);
		} else {
			Area a = new Area(getEllipse2D());
			Area ca = new Area(sc.getShape());
			ca.subtract(a);
			return a.isEmpty();
		} 
	}

	public Rectangle2D getBoundingBox() {
		return new Rectangle((int)(this.getCenter().getX() - (getWidth() / 2)), 
				 (int)(this.getCenter().getY() - (getHeight() / 2)), 
				 (int)getWidth(), 
				 (int)getHeight());
	}

	public Shape getShape() {
		return getEllipse2D();
	}

	public String[] getFieldNames() {
		return fieldNames;
	}

	public String getFieldName(int location) {
		if(location < 4 && location > -1) {
			return fieldNames[location];
		}
		return "";
	}

	public Object[] getFields() {
		Object[] o = new Object[4];
		o[0] = getVisualizationString();
		o[1] = center;
		o[2] = width.getObject();
		o[3] = height.getObject();
		return o;
	}

	public Object getField(int location) {
		if(location == 0) {
			return getVisualizationString();
		}else if(location == 1) {
			return center;
		} else if(location == 2) {
			return width.getObject();
		} else if(location == 3) {
			return height.getObject();
		}
		return null;
	}

	public void setField(Object value, int location) {
		if(location == 0) {
			setKey(value);
		} else if(location == 1) {
			if(value instanceof Point) {
				center = (Point)value;
			}
		} else if(location == 2) {
			setField(width, value);
		} else if(location == 3) {
			setField(height, value);
		}
	}
	
	private void setField(VirtualObject field, Object value) {
		if(value instanceof Double) {
			field.setObject(value);
		} else if(value instanceof VirtualDouble) {
			field.setObject(new Double(((VirtualDouble)value).eval()));
		} else if( value instanceof VirtualInteger) {
			field.setObject(new Double(((VirtualInteger)value).eval()));
		} else if(value instanceof Key) {
			try {
				field.setObject(new Double(((Key)value).getVisualizationString()));
			} catch(NumberFormatException e) {
				
			}
		} else {
			try {
				field.setObject(new Double(value.toString()));
			}catch(NumberFormatException e) {
				
			}
		}
		
	}

}
