// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
package matrix.structures.spatial.FDT.probe;

import java.awt.Shape;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.awt.geom.QuadCurve2D;

import content.exercises.sda.structures.GeometryFunctions;

import matrix.structures.spatial.PaintingStyleDecorator;
import matrix.structures.spatial.SpatialComparable;

/**
 * A class for storing and visualizing parabolic curves 
 * in the Area visualization.
 * 
 * @author Juha Helminen
 */
public class Parabola extends SpatialElement {
	
    private final QuadBezier bezier;
	
	public Parabola(Point2D focus, Line2D directrix) {
		this(focus, directrix, directrix.ptLineDist(focus), directrix.ptLineDist(focus), null);
	}

	public Parabola(Point2D focus, Line2D directrix, double w1, double w2, PaintingStyleDecorator psd) {
        // When the parabola arcs have high/low curvature they might be drawn 
        // incorrectly due to lacking precision.
	    if (Math.abs(w1 + w2) < 8) {
	        // An arc of zero width.
	        this.bezier = null;
	        return;
	    }
//	    System.err.println(w1 + " - " + w2 + " = " + (w1 + w2));
		double length = Math.sqrt(sqr(directrix.getX2() - directrix.getX1()) + 
								  sqr(directrix.getY2() - directrix.getY1()));
		double h = directrix.ptLineDist(focus);
		double nx, ny, dx, dy;
		if (directrix.relativeCCW(focus) < 0) {
			nx = (directrix.getY2() - directrix.getY1()) / length;  
			ny = (directrix.getX1() - directrix.getX2()) / length;
			dx = (directrix.getX2() - directrix.getX1()) / length;
			dy = (directrix.getY2() - directrix.getY1()) / length;
		} else {
			nx = -(directrix.getY2() - directrix.getY1()) / length;  
			ny = -(directrix.getX1() - directrix.getX2()) / length;
            dx = (directrix.getX1() - directrix.getX2()) / length;
            dy = (directrix.getY1() - directrix.getY2()) / length;
		}

		// Compute the tangents that define the parabola and create the 
		// equivalent bezier curve for visualization through QuadCurve2D.
        // The parabola is drawn as a quadratic bezier curve with the other
        // end points of the tangents and their intersection point as 
        // control points.
		Line2D tan1, tan2;
		Point2D control;
        tan1 = tangent(focus, dx, dy,
                       -w1, h, nx, ny);
        tan2 = tangent(focus, dx, dy,
                       w2, h, nx, ny);
        control = GeometryFunctions.lineIntersection(tan1, tan2);

        this.bezier = new QuadBezier(tan2.getX1(), tan2.getY1(), 
                                     control.getX(), control.getY(), 
                                     tan1.getX1(), tan1.getY1());
		
		
        if (psd != null) {
            this.setPaintingStyleDecorator(psd);
        }
	}
	
	
	private static Line2D tangent(Point2D focus, double dx, double dy,
	                              double w, double h, double nx, double ny) {
	    final int extend = 200;
	    double vx, vy;
	    if (w == 0) {
	        return new Line2D.Double(focus.getX() + h / 2 * nx, 
	                                 focus.getY() + h / 2 * ny, 
	                                 focus.getX() + h / 2 * nx - extend * dx, 
                                     focus.getY() + h / 2 * ny - extend * dy);
	    } else {
	        vx = w * dx;
	        vy = w * dy;
	    }
        double[] result = new double[2];
        double[] args = 
            new double[]{sqr(h)*sqr(nx) - 2*h*nx*vx + sqr(vx) + sqr(h)*sqr(ny) - 2*h*ny*vy + sqr(vy),
                         -2*h*sqr(nx) + 2*nx*vx - 2*h*sqr(ny) + 2*ny*vy,
                         sqr(nx) + sqr(ny) - 1};
        QuadCurve2D.solveQuadratic(args, result);
        double d = result[0];

        Point2D start = new Point2D.Double(focus.getX() + vx + h*nx - d*nx,
                                           focus.getY() + vy + h*ny - d*ny);
        double r1x = -vx + 2*d*nx - h*nx;
        double r1y = -vy + 2*d*ny - h*ny;
        return new Line2D.Double(start.getX(), start.getY(), 
                                 start.getX() + extend * r1x, start.getY() + extend * r1y);
	}
	
	private static double sqr(double number) {
		return (number * number);
	}

    public Shape getShape() {
        if (this.bezier == null) {
            return new java.awt.Rectangle();
        } else {
            return this.bezier.getShape();
        }
    }

    public boolean contains(SpatialComparable sc) {
        if (this.bezier == null) {
            return false;
        } else {
            return this.bezier.contains(sc);
        }
	}
	
	public boolean intersects(SpatialComparable sc) {
        if (this.bezier == null) {
            return false;
        } else {
            return this.bezier.intersects(sc);
        }
	}

    private static final long serialVersionUID = 2031934277253204313L;
}
