// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
package matrix.structures.spatial;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Stroke;
import java.io.Serializable;

/**
 * This class is used to set the graphical properties of elements using the area 
 * visualization. The properties that can be set are: line color, fill color, 
 * line stroke style and fill transparency. The line and fill colors are set 
 * separately for three different states: normal, in focus and selected. Also,
 * the outlines and/or filling can be enabled/disabled. The fill and 
 * transparency settings do not affect the drawing of elements with no defined
 * fillable interior area, e.g. polylines.  
 * 
 * The PaintingStyleDecorator can be attached to any SpatialElement and 
 * structure implementing the Area interface using their setPaintingStyleDecorator() 
 * method.
 * 
 * One decorator object can (and should) be used for all the objects with the 
 * same graphical style.
 * 
 * The style and color changes applied through a decorator object aren't recorded 
 * in the animator.
 *   
 * @author oseppala, jhhelmi2
 *
 */

public class PaintingStyleDecorator implements Serializable {
    
    // ** The default color and style scheme **
    
    private transient static final Stroke DEFAULT_STROKE = new BasicStroke(3, BasicStroke.CAP_ROUND, 
                                                                           BasicStroke.JOIN_ROUND, 10.0f, 
                                                                           new float[]{4, 4}, 0);
    private static final Color DEFAULT_LINE_COLOR = Color.GREEN;
    private static final Color DEFAULT_FOCUS_LINE_COLOR = Color.CYAN;
    private static final Color DEFAULT_SELECTION_LINE_COLOR = Color.YELLOW;
    private static final Color DEFAULT_SELECTION_FOCUS_LINE_COLOR = mixColors(DEFAULT_SELECTION_LINE_COLOR, 
                                                                              DEFAULT_FOCUS_LINE_COLOR, 
                                                                              .5f);
    private static final Color DEFAULT_FILL_COLOR = Color.GREEN;
    private static final Color DEFAULT_FOCUS_FILL_COLOR = Color.CYAN;
    private static final Color DEFAULT_SELECTION_FILL_COLOR = Color.YELLOW;
    private static final float DEFAULT_TRANSPARENCY = 0.1f;
    
    private transient Stroke stroke = DEFAULT_STROKE;
    private float width = 3f;
    private int cap = BasicStroke.CAP_ROUND;
    private int join = BasicStroke.JOIN_ROUND;
    private float miterlimit = 10.0f;
    private float[] dashPattern = new float[]{4, 4};
    private float dashPhase = 0f;
    private Color lineColor = DEFAULT_LINE_COLOR;
    private Color focusLineColor = DEFAULT_FOCUS_LINE_COLOR;
    private Color selectionLineColor = DEFAULT_SELECTION_LINE_COLOR;
    private Color selectionFocusLineColor = DEFAULT_SELECTION_FOCUS_LINE_COLOR;
    private Color fillColor = DEFAULT_FILL_COLOR;
    private Color focusFillColor = DEFAULT_FOCUS_FILL_COLOR;
    private Color selectionFillColor = DEFAULT_SELECTION_FILL_COLOR;
    private double transparency = DEFAULT_TRANSPARENCY;
    private boolean fill = true;
    private boolean outline = true;
    
    
    //-----LINE
    /**
     * @return True if the outline is to be drawn.
     */
    public boolean hasOutline() {
        return this.outline;
    }
    
    /**
     * Select whether the outline of the element is drawn or not.
     * 
     * @param outline True if the outline should be drawn.
     */
    public void setOutline(boolean outline) {
        this.outline = outline;
    }
    
    /**
     * @return The color for drawing the outlines of an element
     *         which isn't in focus or selected, not null.
     */
    public Color getLineColor() {
        return this.lineColor;
    }
    
    /**
     * @param lineColor The color to be used for drawing the outlines of an element
     *                  which isn't in focus or selected.
     */
    public void setLineColor(Color lineColor) {
        if (lineColor != null) {
            this.lineColor = lineColor;
        }
    }
    
    /**
     * @return The color for drawing the outline of an element in focus, not null. 
     */
    public Color getFocusLineColor() {
        return this.focusLineColor;
    }
    
    /**
     * @param focusLineColor The color to be used for drawing the outlines of 
     *                       an element in focus.
     */
    public void setFocusLineColor(Color focusLineColor) {
        if (focusLineColor != null) {
            this.focusLineColor = focusLineColor;
        }
    }
    
    /**
     * @return The color for drawing the outlines of a selected element, not null.
     */
    public Color getSelectionLineColor() {
        return this.selectionLineColor;
    }
    
    /**
     * @param selectionLineColor The color to be used for drawing the outlines of 
     *                           a selected element.
     */
    public void setSelectionLineColor(Color selectionLineColor) {
        if (selectionLineColor != null) {
            this.selectionLineColor = selectionLineColor;
        }
    }

    /**
     * @return The color for drawing the outlines of a selected focused element, not null.
     */
    public Color getSelectionFocusLineColor() {
        return this.selectionFocusLineColor;
    }
    
    /**
     * @param selectionFocusLineColor The color to be used for drawing the outlines of 
     *                                a selected focused element.
     */
    public void setSelectionFocusLineColor(Color selectionFocusLineColor) {
        if (selectionFocusLineColor != null) {
            this.selectionFocusLineColor = selectionFocusLineColor;
        }
    }

    
    //-----FILL
    /**
     * @return True if the element is to be filled, that is, its interior is to be colored.
     */
    public boolean hasFill() {
        return this.fill;
    }
    
    /**
     * Select whether the element's interior will be filled or not.
     * 
     * @param fill True if the element is to be filled, that is, its interior is to be colored.
     */
    public void setFill(boolean fill) {
        this.fill = fill;
    }
    
    /**
     * @return The color used for filling, when the element is not in focus or selected, 
     *         not null.
     */
    public Color getFillColor() {
        return this.fillColor;
    }
    
    /**
     * @param fillColor Set the color used for filling, when the element is not in focus or
     *                  selected, should not be null.
     */
    public void setFillColor(Color fillColor) {
        if (fillColor != null) {
            this.fillColor = fillColor;
        }
    }

    /**
     * @return The color used for filling when the element is in focus, not null.
     */
    public Color getFocusFillColor() {
        return this.focusFillColor;
    }

    /**
     * @param focusFillColor Set the color used for filling when the element is in focus,
     *                       should not be null.
     */
    public void setFocusFillColor(Color focusFillColor) {
        if (focusFillColor != null) {
            this.focusFillColor = focusFillColor;
        }
    }

    /**
     * @return The color used for filling when the element is selected, not null.
     */
    public Color getSelectionFillColor() {
        return this.selectionFillColor;
    }
    
    /**
     * @param selectionFillColor Set the color used for filling when the element is 
     *                           selected, should not be null.
     */
    public void setSelectionFillColor(Color selectionFillColor) {
        if (selectionFillColor != null) {
            this.selectionFillColor = selectionFillColor;
        }
    }
    
    
    //-----STROKE
    /**
     * @return The stroke for drawing the outline, not null.
     */
    public Stroke getStroke() {
        if (this.stroke == null) {
            // The stroke is null if the object was serialized and the stroke object
            // was lost in the process because it isn't serializable
            // -> recreate the stroke object.
            this.createStroke();
        }
        return this.stroke;
    }
    
    /**
     * Creates the stroke object based on the stored properties.
     */
    private void createStroke() {
        if (this.dashPattern == null) {
            this.stroke = new BasicStroke(this.width, this.cap, this.join, this.miterlimit);
        } else {
            // Create a dashed stroke.
            this.stroke = new BasicStroke(this.width, this.cap, this.join, this.miterlimit, this.dashPattern, this.dashPhase);
        }
    }
    
    /**
     * Creates a stroke object of class BasicStroke to be used when drawing the outline.
     * See class BasicStroke for allowed values.
     * 
     * @param width         The width of the BasicStroke.
     * @param cap           The decoration of the ends of a BasicStroke
     * @param join          The decoration applied where path segments meet.
     */
    public void setStroke(float width, int cap, int join) {
        this.width = width;
        this.cap = cap;
        this.join = join;
        this.miterlimit = 10f;
        this.dashPattern = null;
        this.dashPhase = 0f;
        this.createStroke();
    }
    
    /**
     * Creates a stroke object of class BasicStroke to be used when drawing the outline.
     * See class BasicStroke for allowed values.
     * 
     * @param width         The width of the BasicStroke.
     * @param cap           The decoration of the ends of a BasicStroke
     * @param join          The decoration applied where path segments meet.
     * @param miterlimit    The limit to trim the miter join.
     */
    public void setStroke(float width, int cap, int join, float miterlimit) {
        this.width = width;
        this.cap = cap;
        this.join = join;
        this.miterlimit = miterlimit;
        this.dashPattern = null;
        this.dashPhase = 0f;
        this.createStroke();
    }

    /**
     * Creates a stroke object of class BasicStroke to be used when drawing the outline.
     * See class BasicStroke for allowed values.
     * 
     * @param width         The width of the BasicStroke.
     * @param cap           The decoration of the ends of a BasicStroke
     * @param join          The decoration applied where path segments meet.
     * @param miterlimit    The limit to trim the miter join.
     * @param dashPattern   The array representing the dashing pattern.
     * @param dashPhase     The offset to start the dashing pattern.
     */
    public void setStroke(float width, int cap, int join, float miterlimit, float[] dashPattern, float dashPhase) {
        this.width = width;
        this.cap = cap;
        this.join = join;
        this.miterlimit = miterlimit;
        this.dashPattern = dashPattern;
        this.dashPhase = dashPhase;
        this.createStroke();
    }

    
    //-----TRANSPARENCY
    /**
     * @return The value of the transparency or -1 if the there is none.
     */
    public double getTransparency() {
        return this.transparency;
    }
    
    /**
     * @param transparency The level of transparency to use for filling, 
     *                     should be greater or equal to 0.
     */
    public void setTransparency(double transparency){
        if (transparency >= 0) {
            this.transparency = transparency;
        }
    }
    
    /**
     * @return True if the element's fill color has some level of transparency.
     */
    public boolean hasTransparency() {
        return this.transparency >=0;
    }

    /**
     * Disable transparency, that is, use solid filling.
     */
    public void clearTransparency() {
        this.transparency = -1;
    }

    public static Color mixColors(Color original, Color layer, float alpha) {
        return new Color(
                ((1f - alpha) * original.getRed() + alpha * layer.getRed()) / 255f,
                ((1f - alpha) * original.getGreen() + alpha * layer.getGreen()) / 255f,
                ((1f - alpha) * original.getBlue() + alpha * layer.getBlue()) / 255f);    
    }
    
    private static final long serialVersionUID = -2209570917892367906L;
}
