// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
package matrix.structures.util;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.IOException;

public class EditDistance{
	
    /**
     * This methods return the minimum of its three parameters
     * 
     * @param a first parameter
     * @param b second parameter
     * @param c third parameter
     * @return the value of the smallest parameter
     */
	public final static int min3(int a, int b, int c){
		if (a<b && a<c)
			return a;
		if (b<c)
			return b;
		else
			return c;
	}
	
    /**
     * This method calculates the edit distance between two arrays of Strings.
     * The edit distance is the smallest number of replacements, insertions and deletions
     * required to change the structures identical.
     * 
     * @param a first array
     * @param b second array
     * @return the edit distance
     */
	public final static int distance(
			final String[] a,
			final String[] b){
		
		int[][] distances = new int[a.length][b.length];
		return distanceAux(a,b,0,0,distances);
	}
	
    /**
     * This method doeas nost of the dirty work. The acursor and bcursor parameters hold
     * the cursor positions to the two arrays. The distances array holds all the already known
     * distances between subarrays.
     * Using the array drops the complexity of the algorithm greatly.
     * 
     * @param a the first array
     * @param b the second array
     * @param acursor the position of the cursor in the first array
     * @param bcursor the position of the cursor in the second array
     * @param distances the edit distances already known.
     * @return
     */
    
	private final static int distanceAux(final String[] a,
			final String[] b,
			int acursor,
			int bcursor,
			int[][] distances){
		
		if (a.length == acursor)
			return b.length - bcursor;
		if (b.length == bcursor)
			return a.length - acursor;
		
		if (distances[acursor][bcursor] == 0){
			distances[acursor][bcursor] = min3(
					1 + distanceAux(a, b, acursor+1, bcursor  , distances),
					1 + distanceAux(a, b, acursor,   bcursor+1, distances),
					((int)(
                            ((a[acursor]==null && b[bcursor]==null) || 
                            (a[acursor]!=null && b[bcursor]!= null && a[acursor].equals(b[bcursor])))?0:1)
					)+
					distanceAux(a, b, acursor+1, bcursor+1, distances)
			);
		}
		return distances[acursor][bcursor];
	}
	
    /**
     * This method calculates the editdistances for each pair of 1D-string arrays
     * held in the 2D-arrays given as params.
     * 
     * @param a the first array of arrays
     * @param b the second array of arrays
     * @return distances between the pairs of arrays
     */
    
    public static int[][] fdtDistances(String[][] a, String[][] b){
        int[][] table = new int[b.length][a.length];
        
        for (int i=0; i<a.length; i++)
            for (int j=0; j<b.length; j++)
                table[j][i] = EditDistance.distance(a[i],b[j]);
        
        return table;
    
    }

//**** The following methods are only used for testing distance measures manually ****
    
    // Turns a String into an array of Strings by cutting each letter to its own respective String
	public static String[] cut(String input){
		String[] result=new String[input.length()];
		for (int i=0; i<result.length; i++)
			result[i]=input.substring(i,i+1);
		return result;
	}
	
    //Test method - Reads in lines from the keyboard and prints out distances between them.
	public static void main(String[] args){
		BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
		String rivi1, rivi2;
		
		try{
			rivi1 = br.readLine();
			while (rivi1.equals("lopeta") == false){
				rivi2 = br.readLine();
				
				System.out.println(EditDistance.distance(cut(rivi1),
						cut(rivi2)));
				rivi1 = br.readLine();
			}
		}catch(IOException e){}
		
	}
		
}
