// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
/**
 * SelectionVisitor class for seraching for selected entities
 * Implemented using the visitor pattern.
 * 
 * Typically used by calling the <b>getSelectedInOrder</b> method with
 * the FDT to start the search from.
 * 
 * @author Otto Seppl
 * @version 1.0
 * Created on June 20, 2006
 *
 */
package matrix.structures.util;

import matrix.structures.FDT.Array;
import matrix.structures.FDT.LinkedList;
import matrix.structures.FDT.Graph;
import matrix.structures.FDT.Struct;
import matrix.structures.FDT.Tree;
import matrix.structures.FDT.FDT;
import matrix.structures.FDT.indexed.StaticLinkedList;
import matrix.structures.FDT.substructures.Vertex;

import matrix.structures.simulationextensions.Selectable;
import matrix.structures.util.FDTVisitAssistant;
import matrix.structures.util.FDTVisitor;

import java.util.*;

public class SelectionVisitor implements FDTVisitor{
    
    private HashSet alreadyVisited;
    private ArrayList selectedItems;
    
    /**
     * Use this method to recursively search for (and clear) selections from an FDT 
     * 
     * @param startingPoint The starting point of the search (e.g. root of a tree, node in a graph)
     * @param clearSelections whether the selections should be cleared
     * @return
     */
    
    
    public static ArrayList getSelectedInOrder(FDT startingPoint, boolean clearSelections){
        ArrayList selected = findSelected(startingPoint);
        Collections.sort(selected, new Comparator(){

			public int compare(Object arg0, Object arg1) {				
				return (int)(((Selectable)arg0).getSelectionTime() - ((Selectable)arg1).getSelectionTime());
			}
			}
        );
        
        if (clearSelections){
        	for (int i=0; i<selected.size(); i++){
        		((Selectable)selected.get(i)).setSelect(0);
        	}
        }
        return selected;
    }
    
    public SelectionVisitor(){
        alreadyVisited = new HashSet(); // to avoid endless recursion 
        selectedItems  = new ArrayList();
    } 
    
    private static ArrayList findSelected(FDT fdt){
    	SelectionVisitor visitor = new SelectionVisitor();
        FDTVisitAssistant.accept(fdt, visitor);
        
        return visitor.getSelectedArray();
    }
    
    public ArrayList getSelectedArray(){
    	return selectedItems;
    }
    
    public void inspect(Object o){
    	if (o instanceof Selectable){
    		long selectionTime = ((Selectable)o).getSelectionTime();
    		if (selectionTime != 0)
    			selectedItems.add(o);
    	}    	
    }
    
    public Object visitNonFDT(Object o){
    	if (alreadyVisited.contains(o))
    		return null;
    	alreadyVisited.add(o);
    	
    	inspect(o);    	
        return null;
    }
    
    public Object visit(FDT fdt){    	
        if (alreadyVisited.contains(fdt))
            return null;        
        alreadyVisited.add(fdt);
        
        inspect(fdt);
        return null;
    }
    
    public Object visit(Array a) {
        
        if (alreadyVisited.contains(a))
            return null;        
        alreadyVisited.add(a);
        
        inspect(a);
        
        int aFirst = a.getFirst();
        int aLast  = a.getLast();
        
        Object aObj;
        
        for(int i = aFirst; i <= aLast; i++) {
            aObj = a.getObject(i);
            FDTVisitAssistant.accept(aObj, this);
        }
        
        return null;
    }
    
    public Object visit(Struct s){
        if (alreadyVisited.contains(s))
            return null;        
        alreadyVisited.add(s);

        inspect(s);
        
        Object[] fields = s.getFields();
        for (int i=0; i<fields.length; i++){
        	FDTVisitAssistant.accept(fields[i], this);
        }
        return null;
    }
    
    public Object visit(LinkedList a) {
                
        if (alreadyVisited.contains(a))
            return null;
        alreadyVisited.add(a);
        
        inspect(a);
        
        Object aObj = a.getElement();
        if(aObj == null)
            ;// do nothing
        else
            FDTVisitAssistant.accept(aObj, this);
        
        while(a.hasNext()){
        	a    = a.getNext();
        	aObj = a.getElement();
            
        	if(aObj == null)
        		;// do nothing
        	else
                FDTVisitAssistant.accept(aObj, this);
        }
        
        return null;
    }
    
    
    public Object visit(StaticLinkedList a) {
                
        if (alreadyVisited.contains(a))
            return null;
        alreadyVisited.add(a);

        inspect(a);
        
        a.setFirst();
        Object aObj;
        
        while(a.isInList()) {
            aObj = a.currValue();
            if(aObj == null)
                ;// do nothing
            else
                FDTVisitAssistant.accept(aObj, this);
            a.next();
        }
        return null;
    }
    
    /**
     * Compares two common trees. This is an identical comparison, so the children of
     * the nodes must be in the same order in both trees.
     * @param a the first tree to be compared
     * @param b the second tree to be compared
     * @return true if and only if the contents of the two trees are equal
     */
    public Object visit(Tree a) {
        
        if (alreadyVisited.contains(a))
            return null;
        alreadyVisited.add(a);
        inspect(a);
        
        if (a==null){
            return null;
        }
        
        Object aObj = a.getElement();
        if(aObj == null)
            ; //do nothing
        else
            FDTVisitAssistant.accept(aObj, this);
        
        Tree[] aTree = a.getSubTrees();
        for(int i = 0; i < aTree.length; i++){
            if (aTree[i] == null)
            	continue;
            FDTVisitAssistant.accept(aTree[i], this);
        }
        return null;
    }
    
    public Object visit(Graph a) {
        
        if (alreadyVisited.contains(a))
            return new ArrayList();
        alreadyVisited.add(a);
        
        if(a == null)
            return null;

        inspect(a);
        
        Vertex[] aVertices = a.getVertices();
        Object ao;
        
        for(int i = 0; i < aVertices.length; i++) {
            ao = aVertices[i].getElement();
            if(ao == null)
                ; // do nothing
            else
                FDTVisitAssistant.accept(ao, this);
            
        }
        return null;
    }
    
}
