// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/* $Id: ComparisonPanel.java,v 1.12 2009/06/18 10:12:52 lhakulin Exp $ */

package matrix.uitools;

import java.awt.*;

import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;

//import applications.trakla2.recovery.DistanceGraph;

import java.awt.event.*;

import content.assessment.AssessmentStrategy;
import content.interfaces.*;
import matrix.animation.*;
import matrix.visual.*;
import matrix.structures.FDT.FDT;
import matrix.structures.util.EditDistance;
import matrix.structures.util.Linearizer;
import matrix.simulation.VisualTypeConf;

/**
 * A panel which shows two visual animators and two structure panels along with
 * some more information. There are also two animators: the master and the
 * slave. They can be synchronized in two modes: normal and recovery. The master
 * animator is the one that controls both structure panels when the view is
 * synchronized. In normal mode only equal steps (between animators) can be
 * matched. In recovery mode every master step can be synchronized to the
 * closest matching state frorm the slave.
 * 
 * @author Isaac Jurado
 */
public class ComparisonPanel extends JPanel implements ActionListener, ChangeListener {

    private Animator            masterAnimator, slaveAnimator;
    private AnimatorStepMapper  masterMapper,   slaveMapper;
    private StructurePanel      masterPanel,    slavePanel;
    private VisualAnimator      masterVisualAnimator;
    private SlaveVisualAnimator slaveVisualAnimator;
    
    private int[] slaveClosestStep, slaveClosestStepDistance;
    private FDT[] masterFDTs, slaveFDTs;
    
    private boolean recoveryEnabled;
    private int     masterLastCorrectStep, slaveLastCorrectStep;         
    
    private JLabel    lbl_distance, lbl_masterStepInfo, lbl_slaveStepInfo;
    private JCheckBox chk_in_sync;
    
    private static final Color green  = new Color(  0, 200, 0);
    private static final Color red    = new Color(220,   0, 0);
    private static final Color yellow = new Color(190, 190, 0);

    
    /////////////////////
    //                 //
    //   CONSTRUCTOR   //
    //                 //
    /////////////////////
    
    public ComparisonPanel (Animator userAnim, ComparableExercise ex) {
        super(new GridBagLayout());

        // Create animators and FDT[] arays
        masterAnimator = new Animator();
        slaveAnimator  = userAnim;

        Animator a = Animator.getActiveAnimator();
        Animator.setActiveAnimator(masterAnimator);

        if (ex instanceof MultipleQuestions){
        	int questionCount = ((MultipleQuestions)ex).numberOfQuestions();
    		((MultipleQuestions)ex).changeAnimator(0);
        	int fdtCount = ex.getAnswer().length; // number of fdts in one question
        	masterFDTs = new FDT[questionCount*fdtCount];
        	slaveFDTs = new FDT[questionCount*fdtCount];
        	
        	for (int i=0; i < questionCount; i++){
        		((MultipleQuestions)ex).changeAnimator(i);
                Animator.setActiveAnimator(masterAnimator);
                FDT[] tmpMasterFDTs = ex.solve();
                for (int j=0; j < tmpMasterFDTs.length; j++) {
					masterFDTs[(i*fdtCount)+j] = tmpMasterFDTs[j];
				}
    	        Animator.setActiveAnimator(a);  // Restore previous
    	        FDT[] tmpSlaveFDTs = ex.getAnswer();
                for (int j=0; j < tmpSlaveFDTs.length; j++) {
					slaveFDTs[(i*fdtCount)+j] = tmpSlaveFDTs[j];
				}
        	}
        }
        else {
	        masterFDTs = ex.solve();
	        Animator.setActiveAnimator(a);  // Restore previous
	        slaveFDTs  = ex.getAnswer();
        }

        masterMapper = new AnimatorStepMapper(masterAnimator);
        slaveMapper  = new AnimatorStepMapper(slaveAnimator);

        recoveryEnabled = ex.canRecover();
        if (recoveryEnabled)
            calculateDistances(ex);
        else
            gradeExercise();
        //DistanceGraph.distanceGraphs(slaveFDTs, slaveAnimator, masterFDTs, masterAnimator);

        masterAnimator.rewind();
        slaveAnimator.rewind();
        
        // There are some misterious issues about creating the visual animators
        // before judging, so they have to ve creaated after evaluation
        masterVisualAnimator = new VisualAnimator(masterAnimator);
        slaveVisualAnimator  = new SlaveVisualAnimator(slaveAnimator);

        masterPanel  = newConfiguredStructurePanel(ex, masterAnimator, masterFDTs);
        slavePanel   = newConfiguredStructurePanel(ex, slaveAnimator, slaveFDTs);

        // Link structure panels with visual animators
        masterPanel.setVisualAnimator(masterVisualAnimator);
        slavePanel.setVisualAnimator(slaveVisualAnimator);

        // Disable play, reset and visual structure manipulation through
        // dragging
        masterVisualAnimator.setPlayButtonVisible(false);
        slaveVisualAnimator.setPlayButtonVisible(false);
        masterPanel.disableVisualComponents();
        slavePanel.disableVisualComponents();

        // Put structure panels into scrollable panels
        JScrollPane masterSP = new JScrollPane();
        JScrollPane slaveSP  = new JScrollPane();
        masterSP.setViewportView(masterPanel);
        slaveSP.setViewportView(slavePanel);
        
        //
        // INTERFACE CONSTRUCTION
        //
        JPanel masterBox = new JPanel();
        JPanel slaveBox  = new JPanel();
        masterBox.setLayout(new BoxLayout(masterBox, BoxLayout.PAGE_AXIS));
        slaveBox.setLayout(new BoxLayout(slaveBox, BoxLayout.PAGE_AXIS));

        // Visual animator borders
        Border b = BorderFactory.createEtchedBorder(EtchedBorder.LOWERED);
        TitledBorder tb = BorderFactory.createTitledBorder(b, "Model Answer");
        tb.setTitleJustification(TitledBorder.CENTER);
        masterBox.setBorder(tb);
        if (ex.canRecover())
            tb = BorderFactory.createTitledBorder(b, "Your Answer ( recovery enabled )");
        else
            tb = BorderFactory.createTitledBorder(b, "Your Answer");
        tb.setTitleJustification(TitledBorder.CENTER);
        slaveBox.setBorder(tb);
        
        // Sync checkbox and labels holding useful information
        chk_in_sync        = new JCheckBox("Synchronize views");
        lbl_distance       = new JLabel(" ");
        lbl_masterStepInfo = new JLabel();
        lbl_slaveStepInfo  = new JLabel();
        
        // Lay out widgets
        masterBox.add(masterVisualAnimator);
        masterBox.add(chk_in_sync);
        masterBox.add(lbl_masterStepInfo);
        slaveBox.add(slaveVisualAnimator);
        slaveBox.add(lbl_distance);
        slaveBox.add(lbl_slaveStepInfo);
        
        GridBagConstraints c = new GridBagConstraints();

        c.fill   = GridBagConstraints.NONE;
        c.insets = new Insets(10, 10, 10, 10);
        add(masterBox, c);
        
        c.gridwidth = GridBagConstraints.REMAINDER;
        add(slaveBox, c);
        
        c.gridwidth = 1;
        c.fill      = GridBagConstraints.BOTH;
        c.anchor    = GridBagConstraints.NORTH;
        c.weightx   = 1;
        c.weighty   = 1;
        add(masterSP, c);

        c.gridwidth = GridBagConstraints.REMAINDER;
        add(slaveSP, c);

        // Manage events
        masterPanel.addActionListener(this);
        slavePanel.addActionListener(this);
        chk_in_sync.addChangeListener(this);

        chk_in_sync.setSelected(true);
        slaveVisualAnimator.setEnabled(false);
        syncSlaveWithMaster();
        updateStepInfo();
    }

    
    ////////////////////////
    //                    //
    //   EVENT HANDLERS   //
    //                    //
    ////////////////////////

    public void actionPerformed (ActionEvent evt) {       
        if (chk_in_sync.isSelected())
            if (!recoveryEnabled && masterMapper.getCurrentStep() > masterLastCorrectStep) {
                chk_in_sync.setSelected(false);
                // Only redo slave if possible
                if (slaveMapper.getCurrentStep() < slaveMapper.getLastStep())
                    slaveAnimator.redo();
                slaveVisualAnimator.setEnabled(true);
                slaveVisualAnimator.validate();
            } else {
                syncSlaveWithMaster();
            }      
        updateStepInfo();
    }


    public void stateChanged(ChangeEvent evt) {
        if (recoveryEnabled) {
            /*
             * In recovery mode we can always match a master step with a slave
             * step with approximation
             */
            if (chk_in_sync.isSelected()) {
                // From false to true
                slaveVisualAnimator.setEnabled(false);
                lbl_distance.setEnabled(true);
                syncSlaveWithMaster();
                updateStepInfo();
            } else {
                // From true to false
                slaveVisualAnimator.setEnabled(true);
                slaveVisualAnimator.validate();
                lbl_distance.setEnabled(false);
            }
        } else {
            /*
             * In normal compare mode we can only match correct steps, so when
             * syncronization is requested, we go back to the last matching
             * state if necessary
             */
            if (chk_in_sync.isSelected()) {
                // From false to true
                if (masterMapper.getCurrentStep() > masterLastCorrectStep)
                    masterAnimator.GoTo(masterMapper.getStateFromStep(masterLastCorrectStep), 0);
                syncSlaveWithMaster();
                masterVisualAnimator.validate();
                slaveVisualAnimator.setEnabled(false);
            } else {
                // From true to false
                slaveVisualAnimator.setEnabled(true);
                slaveVisualAnimator.validate();
            }
        }
    }

    
    ///////////////////////////
    //                       //
    //   AUXILIARY METHODS   //
    //                       //
    ///////////////////////////
    
    /*
     * Update the "Step X of Y" labels and the colored label indicating the
     * correctness of the current slaveAnimator step.
     */
    private void updateStepInfo() {
        int mstp, sstp;
        
        mstp = masterMapper.getCurrentStep();
        sstp = slaveMapper.getCurrentStep();

        lbl_masterStepInfo.setText("Step "+mstp+" of "+masterMapper.getLastStep());
        lbl_slaveStepInfo.setText("Step "+sstp+" of "+slaveMapper.getLastStep());
            
        if (recoveryEnabled) {
            // Recovery mode
            if (chk_in_sync.isSelected()) {
            	if (slaveMapper.getCurrentStep() == 0) {
            		lbl_distance.setForeground(Color.BLACK);
            		lbl_distance.setText("InitialStep");
            	} else if (slaveClosestStepDistance[mstp] > 0) {
            		lbl_distance.setForeground(red);
            		lbl_distance.setText("Step incorrect");
            	} else if (mstp > masterLastCorrectStep) {
            		lbl_distance.setForeground(yellow);
            		lbl_distance.setText("Matching step");
            	} else {
            		lbl_distance.setForeground(green);
            		lbl_distance.setText("Step correct");
            	}
            }
        } else {
            // Normal mode
        	if (slaveMapper.getCurrentStep() == 0) {
        		lbl_distance.setForeground(Color.BLACK);        		
        		lbl_distance.setText("Initial step");
        	} else if (slaveMapper.getCurrentStep() > slaveLastCorrectStep) {
                lbl_distance.setForeground(red);
                lbl_distance.setText("Step incorrect");
            } else {
                lbl_distance.setForeground(green);
                lbl_distance.setText("Step correct");
            }
        }

    }
    
    
    /*
     * Find the state of the salveAnimator witch matches the current state of
     * the masterAnimator.
     */
    private void syncSlaveWithMaster() {
        int s = slaveClosestStep[masterMapper.getCurrentStep()];
        slaveAnimator.GoTo(slaveMapper.getStateFromStep(s), 0);
        slaveVisualAnimator.validate();           
    }
    
    
    /*
     * Only invoked in recovery mode. Creates the distance matrix and looks for
     * the minimum distances between each master and slave step.
     */
    private void calculateDistances(ComparableExercise ex) {
    	int fdt_index = ex.getGradeIndices()[0];
        // These do all the dirty work
        String[][] model     = Linearizer.linearize(masterFDTs[fdt_index], masterAnimator);
        String[][] user      = Linearizer.linearize(slaveFDTs[fdt_index], slaveAnimator);
        int[][]    distances = EditDistance.fdtDistances(user, model);
        int        sum       = 0;
        
        slaveClosestStep         = new int[distances.length];
        slaveClosestStepDistance = new int[distances.length];

        // Now filter the necessary information
        for (int mi = 0; mi < distances.length; mi++) {
            int min_distance = Integer.MAX_VALUE;
            int ui = (mi > 0 ? slaveClosestStep[mi - 1] : 0);
 
            while (ui < distances[mi].length - 1 && distances[mi][ui] != 0) {
                if (distances[mi][ui] < min_distance) {
                    slaveClosestStep[mi] = ui;
                    min_distance = distances[mi][ui];
                }
                ui++;
            }
            if (distances[mi][ui] < min_distance) {
                slaveClosestStep[mi] = ui;
                min_distance = distances[mi][ui];
            }
            slaveClosestStepDistance[mi] = min_distance;
            sum += min_distance;
            if (sum == 0)
            	masterLastCorrectStep = mi;
        }
        
    }
    
    
    /*
     * There is a manual grading here because the Jury does not provide the
     * information needed here. Here what is needed is the mapping between
     * correct master steps and student steps. Also the last correct master step
     * and the last correct student step in order to update the lbl_distance
     * accordingly to the current slaveAnimator step.
     */
    private void gradeExercise() {
        slaveClosestStep = new int[masterMapper.getStepCount()];

        masterAnimator.rewind();
        slaveAnimator.rewind();

        int ms = 0, ss = 0;  // Master Step, Slave Step
        slaveLastCorrectStep = 0;
        boolean keep_going = true;

        while (keep_going) {
            ss = slaveLastCorrectStep;
            slaveAnimator.GoTo(slaveMapper.getStateFromStep(ss), 0);
            while (slaveAnimator.hasNextOperation()
                    && !AssessmentStrategy.allEqual(masterFDTs, slaveFDTs)) {
                slaveAnimator.redo();
                ss++;
            }
            
            keep_going = false;
            if (AssessmentStrategy.allEqual(masterFDTs, slaveFDTs)) {
                slaveClosestStep[ms] = ss;
                slaveLastCorrectStep = ss;
                masterLastCorrectStep = ms;
                if (masterAnimator.hasNextOperation()) {
                    masterAnimator.redo();
                    ms++;
                    keep_going = true;
                }
            }
        }
    }
       
    
    /*
     * Create a StructurePanel with all the FDTs added and properly configured.
     */
    private static StructurePanel newConfiguredStructurePanel(ComparableExercise e, Animator a, FDT[] fdts) {
        StructurePanel p = new StructurePanel(a);
        VisualType[] vt = null;
        int[] ind = e.getCompareIndices();

        GridBagConstraints[] gbc = null;
        
        if(e instanceof LayoutExercise) {
        	gbc = ((LayoutExercise)e).getComparisonConstraints();
        	if (gbc == null){ // use default
        		GridBagConstraints[] modelgbc = ((LayoutExercise)e).getModelAnswerConstraints();
        		gbc = new GridBagConstraints[ind.length];
	            for (int i = 0; i < ind.length; i++){
	            	gbc[i] = modelgbc[ind[i]];
	            }
        	}
        }
        if (e instanceof StyledExercise) {
        	String[] visualisations;
        	if (e.getCompareStructureVisualisations() == null){  // use default
	            String[] vis = ((StyledExercise)e).getModelAnswerVisualisations();
	            visualisations = new String[ind.length];
	            for (int i = 0; i < ind.length; i++){
	            	visualisations[i] = vis[ind[i]];
	            }
        	}
        	else {
        		visualisations = e.getCompareStructureVisualisations();
        	}
        	if (gbc == null){
        		vt = p.addStructures(fdts, visualisations);
        	} 
        	else {
        		vt = p.addStructures(fdts, visualisations, gbc);
        	}
        } 
        else {
        	if (gbc == null){
        		vt = p.addStructures(fdts);
        	}
        	else {
        		vt = p.addStructures(fdts, gbc);        		
        	}
        }

    	if (e.getCompareStructureNames() != null){
            String[] names = e.getCompareStructureNames();
            for (int i = 0; i < names.length; i++)
                vt[i].setName(names[i]);
    	}
    	else if (e instanceof ModelAnswerNames) { // use default
            String[] names = ((ModelAnswerNames)e).getModelAnswerNames();
            for (int i = 0; i < vt.length; i++){
                vt[i].setName(names[ind[i]]);
            }
        }
    	
        if (e instanceof HasRotatedContainers) {
            boolean[] rotateInfo = ((HasRotatedContainers)e).getComparisonRotations();
            if (rotateInfo == null){ // use default
            	boolean[] modelRotateInfo = ((HasRotatedContainers)e).getModelAnswerRotations();
            	rotateInfo = new boolean[ind.length];
                for (int i = 0; i < vt.length; i++){
                	rotateInfo[i] = modelRotateInfo[ind[i]];
                }
            }
            for (int i = 0; i < vt.length; i++) {
                if (vt[i] instanceof matrix.visual.VisualContainer)
                    ((matrix.visual.VisualContainer)vt[i]).setRotated(rotateInfo[i]);
            }	  
        }

        for (int i = 0; i < vt.length; i++) {
            if (vt[i] instanceof VisualContainer) {
                ((VisualContainer)vt[i]).setHotSpotEnabled(false);
                ((VisualContainer)vt[i]).setNullHotSpotEnabled(false);
                ((VisualContainer)vt[i]).setSearchHotSpotEnabled(false);
            }
            if (vt[i] instanceof VisualArray)
                ((VisualArray) vt[i]).setResizeHotSpotEnabled(false);
        }

        if (e instanceof ConfigureVisualType) {
            VisualTypeConf[] conf = e.getCompareVisualTypeConf();
            if (conf == null){ // use default
                conf = ((ConfigureVisualType)e).conf();
                for (int i = 0; i < vt.length; i++){
                    vt[i].configure(conf[ind[i]]);
                }
            }
            else {
	            for (int i = 0; i < vt.length; i++){
	                vt[i].configure(conf[i]);
	            }
            }
        }

        return p;
    }

    private static final long serialVersionUID = 1700137183116989271L;
}

