// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package matrix.uitools;

import java.awt.*;
import javax.swing.*;
import java.awt.event.*;
import java.io.*;
import java.util.zip.GZIPOutputStream;

import matrix.animation.*;
import matrix.structures.FDT.FDT;
import matrix.util.*;
import matrix.util.export.*;
import matrix.util.io.*;

/**
 * MenuFrame is a subclass of MatrixFrame that contains all functionality required
 * by the "default" Menus included in the prototype application. All menu-related
 * stuff is here and not in MatrixFrame for two reasons. <p>
 *
 * <ol>
 * <li> If someone wishes to create an application where there are no menus, it
 * is possible to do so (and you do not need to use a class that contains
 * all the menu functionality</li>
 * <li> It is easier to maintain menu functionality when it has been separated
 * from the other functionality.</li>
 * </ol>
 */

public class MenuFrame extends MatrixFrame {

    public static final int DEFAULT_WIDTH = 10000;  
    public static final int DEFAULT_HEIGHT = 10000;

    private JMenu fontMenu;

    // quick-hack version of history frame
    private boolean state = false;

    public MenuFrame() {
	super();
    }

    public MenuFrame(Animator a) {
	super(a);
    }

    public MenuFrame(StructurePanel p) {
	super(p);
    }

    public MenuFrame(Animator a, StructurePanel p) {
	super(a,p);
    }

    public MenuFrame(VisualAnimator va, Animator a) {
	super(va,a);
    }
      
    public MenuFrame(VisualAnimator va, Animator a, StructurePanel p) {
	super(va,a,p);
    }

    public MenuFrame(FDT[] s, Animator a) {
	super(s,a);
    }

    public JFrame getNewFrame(StructurePanel p) {
	return new MenuFrame(p);
    }

    public JFrame getNewFrame() {
	return new MenuFrame();
    }

    public void open() {
        JFileChooser fd = new JFileChooser();
        int returnVal = fd.showOpenDialog(this);
        if (returnVal == JFileChooser.CANCEL_OPTION) {
            return;
        }
        File selectedFile = fd.getSelectedFile();
        String fileName = selectedFile.getAbsolutePath();

        if (!selectedFile.exists()) {
            getApplication().showMessage("File not found", "File " + fileName + " not found.");
            return;
        }
        try {
            structureFromFile(fileName);
        } catch (SecurityException e) {
            Note.show(this, "Applet cannot read local filesystem. " +
                    "See http://www.cs.hut.fi/Research/Matrix/applet/security.html." +
                    "Original exception was:" + e.getMessage());
        }
    }

    protected void structureFromFile(String fileName) {
	int type = FileUtil.sniffType(fileName);
	MatrixFrame mf = null;
	switch(type) {
	case FileUtil.FDT_CLASS:
	    //mf = new MatrixFrame();
	    mf = this;
	    FDT temp = FileUtil.instantiateFDTClass(fileName);
	    if (temp != null) { //we don't want to add null structure to application
		mf.getApplication().addStructure(temp);
		mf.setVisible(true);
	    }
	    break;
	case FileUtil.GRAPH_DESCRIPTION:
	    //mf = new MatrixFrame();
	    mf = this;
	    //mf.getApplication().addStructure(FileUtil.loadGraphFromFile(fileName));
	    FileUtil.loadVisualStructuresFromFile(fileName, mf);
	    mf.setVisible(true);
	    break;
	case FileUtil.SERIALIZED_ANIMATION:
	    MatrixSerialization ms = FileUtil.openAnimation(fileName);
	    //mf = new MatrixFrame(ms.getStructures(), ms.getAnimator());
	    mf = new MenuFrame(ms.getStructures(), ms.getAnimator());
	    mf.getAnimator().rewind();
	    if (ms.getLayout() != null) 
		ms.getLayout().restoreProperties(mf.getStructurePanel());
	    mf.getStructurePanel().validate();
	    mf.validate();
	    mf.setVisible(true);
	    break;
	case FileUtil.TABLE_DESCRIPTION:
	    //mf = new MatrixFrame();
	    mf = this;
	    //mf.getApplication().addStructure(FileUtil.loadTableFromFile(fileName));
	    FileUtil.loadVisualStructuresFromFile(fileName, mf);
	    mf.setVisible(true);
	    break;
	case FileUtil.STRUCTURES_DESCRIPTION:
	    mf = this;
	    //FDT[] temp = FileUtil.loadStructuresFromFile(fileName);
	    //mf.addStructures(temp);
	    FileUtil.loadVisualStructuresFromFile(fileName, mf);
	    mf.setVisible(true);
	    break;
	case -1: //the file is not in supported format,(sniffType() returned -1)
	    Note.show(this, "The file " + fileName + " is not in file format supported by Matrix.");
	}
    }


    public void save() {
	save(false);
    }

    public void saveLayout() {
	save(true);
    }

    /* The parameter chooses whether the layout information is saved along the structures or not. */
    private void save(boolean layout) {
        boolean ok = true;
        File saveFile;
        do {
            JFileChooser fd = new JFileChooser();
            fd.setDialogTitle("Save as serialized...");
            int returnVal = fd.showSaveDialog(this);
            if (returnVal == JFileChooser.CANCEL_OPTION) {
                return;
            }
            saveFile = fd.getSelectedFile();
            String fileName = saveFile.getAbsolutePath();
            
            if (saveFile.exists()) {
                YesNoDialog yesNo = new YesNoDialog(this, "File " + fileName + " exists! Overwrite?", "File exists");
                ok = yesNo.getResult();
            } else {
                ok = true;
            }
        } while(!ok);
        getAnimator().end();
        getAnimator().rewind();
        getStructurePanel().validate();
        validate();
        try {
            MatrixSerialization.save(getAnimator(),
                    getApplication().getStructures(),
                    saveFile,
                    layout,
                    getStructurePanel());
        } catch (SecurityException e) {
            Note.show(this, "Applet cannot write local filesystem. " +
                    "See http://www.cs.hut.fi/Research/Matrix/applet/security.html." +
                    "Original exception was:" + e.getMessage());
            
        }
    }


    public void saveToASCII() {
        boolean ok = true;
        File saveFile;
        do {
            JFileChooser fd = new JFileChooser();
            fd.setDialogTitle("Save as ASCII...");
            int returnVal = fd.showSaveDialog(this);
            if (returnVal == JFileChooser.CANCEL_OPTION) {
                return;
            }
            saveFile = fd.getSelectedFile();
            String fileName = saveFile.getAbsolutePath();

            if (saveFile.exists()) {
                YesNoDialog yesNo = new YesNoDialog(this, "File " + fileName + " exists! Overwrite?", "File exists");
                ok = yesNo.getResult();
            } else {
                ok = true;
            }
        } while(!ok);

        MatrixFrame f = this;
        StructurePanel sp = f.getStructurePanel();
        Component[] cs = sp.getComponents();
        ASCIISaver as = new ASCIISaver(cs, saveFile);
        as.saveStructuresToFile();
    }
    

    public void clear() {
	StructurePanel s = getStructurePanel();
	s.removeComponents();
	getAnimator().end();
	getAnimator().disposeUndo();
	s.validate();
	s.repaint();
	validate();
	repaint();
    }
  
    public void export(String format) {
        JPanel p = getStructurePanel();
        MatrixFrame f = this;
        MatrixGraphics matrix = new MatrixGraphics(new Rectangle(0, 0, 
                DEFAULT_WIDTH, DEFAULT_HEIGHT), p.getGraphics());
        boolean ok = true;
        File saveFile;
        do {
            JFileChooser fd = new JFileChooser();
            fd.setDialogTitle("Export...");
            int returnVal = fd.showSaveDialog(this);
            if (returnVal == JFileChooser.CANCEL_OPTION) {
                return;
            }
            saveFile = fd.getSelectedFile();
            String fileName = saveFile.getAbsolutePath();

            if (saveFile.exists()) {
                YesNoDialog yesNo = new YesNoDialog(this, "File " + fileName + " exists! Overwrite?", "File exists");
                ok = yesNo.getResult();
            } else {
                ok = true;
            }           
        } while(!ok);
        
        RepaintManager.currentManager(getStructurePanel()).setDoubleBufferingEnabled(false);
        try {
            if (format.equals("tex")) {      // latex
                p.paint(matrix);
                LatexWriter tolatex = new LatexWriter(matrix.getContents());
                tolatex.write(saveFile);
                return;
            } else if (format.equals("svg") || format.equals("svgz")) {
                Animator animator = f.getAnimator();
                animator.rewind();
                p.validate();
                f.validate();
                p.paint(matrix);
                boolean svg = format.equals("svg");
                int steps = 0;
                matrix.nextStep();
                while (animator.hasNextOperation()) {
                    animator.redo();
                    p.validate();
                    f.validate();
                    p.paint(matrix);
                    matrix.nextStep();
                    steps++;
                } 
                SVGWriter toSVG = new SVGWriter(matrix.getContents());
                toSVG.setMaxStep(steps);
                toSVG.setBackground(p.getBackground());
                matrix.setScale(1.3);
                if (svg)                        // svg
                    toSVG.write(saveFile);
                else {                          // gzipped svg
                    try {
                        File tempFile = new File(saveFile + ".tmp");
                        toSVG.write(tempFile);
                        FileInputStream fis =
                            new FileInputStream(tempFile);
                        GZIPOutputStream zos = new GZIPOutputStream(
                                new FileOutputStream(saveFile));
                        int n = -1;
                        while( (n=fis.read())!=-1)
                            zos.write(n);
                        zos.flush();
                        zos.close();
                        tempFile.delete();
                    } catch (IOException e) {
                        Note.out(e,"Error: " + e.toString()); 
                        e.printStackTrace();
                    }
                }	    
            }
        } catch (SecurityException e) {
            Note.show(f, "Applet cannot write local filesystem. " +
                    "See http://www.cs.hut.fi/Research/Matrix/applet/security.html." +
                    "Original exception was:" + e.getMessage());
        }
        RepaintManager.currentManager(getStructurePanel()).setDoubleBufferingEnabled(true);

        f.validate();    
        p.validate();
    }

    public void aboutDialog() {
	JDialog d = new AboutDialog(this);
	d.pack();
	d.setVisible(true);
    }


    public JMenu createFontMenu() {
        fontMenu = new JMenu("Font");
        fontMenu.setMnemonic(KeyEvent.VK_F);
        Font defFont = getStructurePanel().getFont();
        String defName; //name of the default font
        ButtonGroup fontGroup = new ButtonGroup(); //button group where radio buttons are added to
        if (defFont != null) {
            defName = defFont.getFamily();
        } else {
            /* If default font is null, set default font to "dialog".
             */
            defName = "dialog";
            getStructurePanel().setFont(new Font(defName, 0, 12));
        }
        String[] fontNames = GraphicsEnvironment.getLocalGraphicsEnvironment().getAvailableFontFamilyNames();
        for (int j = 0; j < fontNames.length; j++) {
            //Default font crashes on Linux
            if (!fontNames[j].equalsIgnoreCase("default")) {
                JRadioButtonMenuItem mi = new JRadioButtonMenuItem(fontNames[j], false);
                if(fontNames[j].equals(defName)) {
                    mi.setSelected(true);
                }
                // XXX Could we manage with just one ActionListener?
                mi.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        String comm = e.getActionCommand();
                        int size = getStructurePanel().getFont().getSize();
                        int style = getStructurePanel().getFont().getStyle();
                        getStructurePanel().setFont(new Font(comm,style,size));
                        invalidateComponents();
                        // are these really required??
                        validateComponents();
                        validate();
                    }
                });
                fontGroup.add(mi); //adding radio button to button group
                fontMenu.add(mi);
            }
        }
        return fontMenu;
    }

    /**
     * Changes the used font size of the StructurePanel and then 
     * validates components to show them on the screen with new font.
     * This is called when selecting new size from the font size menu.
     * 
     * @param size new size of the font
     */
    public void changeFontSize(int size) {
        Font fnt = getStructurePanel().getFont();
        int style = fnt.getStyle();
        getStructurePanel().setFont(new Font(fnt.getName(),style,size));
        invalidateComponents();
        //XXX are these really necessary?
        validateComponents();
        validate();
    }

    /**
     * This method should be called when initializing MatrixFrame, because 
     * otherwise validate and validateComponents -methods are called and they cause 
     * problems when done during MatrixFrame initialization. (Call to this method comes thus
     * from matrixconf.xml from initialization of font menu.)
     * 
     * @param size new size of the font
     */
    public void changeFontSizeDuringInitialization(int size) {
        Font fnt = getStructurePanel().getFont();
        int style = fnt.getStyle();
        getStructurePanel().setFont(new Font(fnt.getName(),style,size));
    }
    
    public void newStructure(FDT s) {
	getApplication().addStructure(s);
	validate();
    }

    /**
     * Adds new structure to the StructurePanel.
     *
     * @param s FDT which is going to be added
     * @param name name of the structure
     */
    public void newStructure(FDT s, String name) {
	((StructurePanel)getApplication()).addStructure(s, name);
	validate();
    }

    public void dump() {
	ByteArrayOutputStream buffer = new ByteArrayOutputStream();
	PrintStream out = new PrintStream(buffer);
	getStructurePanel().dump(out);
	getStructurePanel().showMessage("Dump", buffer.toString());
    }

    public void dumpAnimator() {
	ByteArrayOutputStream buffer = new ByteArrayOutputStream();
	PrintStream out = new PrintStream(buffer);
	getAnimator().dump(out);
	getStructurePanel().showMessage("Dump", buffer.toString());

    }



    public void showHistory() {
	state = !state;
	History.getHistory(getStructurePanel()).setVisible(state);
    }

    public void back() {
  	if (getAnimator().hasPreviousOperation()) {
  	    getAnimator().undo();
  	    getStructurePanel().invalidateComponents();
	    validateComponents();
	    validate();
	}
    }

    public void forward() {
  	if (getAnimator().hasNextOperation()) {
  	    getAnimator().redo();
  	    getStructurePanel().invalidateComponents();
	    validateComponents();
	    validate();
	}
	
    }

    public void begin() {
	getAnimator().rewind();
	getStructurePanel().invalidateComponents();
	validateComponents();
	validate();
    }

    public void end() {
	getAnimator().end();
	getStructurePanel().invalidateComponents();
	validateComponents();
	validate();
    }

    public void play() {
	CommandThread commandThread;
	commandThread = new CommandThread(100, new Command() {
		public void execute() {
		    CommandThread ct;
		    while (getAnimator().hasNextOperation()) {
			ct = new CommandThread(500, new Command() {
				public void execute() {
				    getAnimator().redo(); }
			    });
			ct.start();
			try { ct.join(); }
			catch (InterruptedException ex) {}
			getStructurePanel().invalidateComponents();
			validateComponents();
			validate();
		    }
		}
	    });
	commandThread.start();
    }

    public void setBegin() {
	getAnimator().disposeUndo();
	validateComponents();
	validate();
    }

    public void setEnd() {
	getAnimator().disposeRedo();
	validateComponents();
	validate();
    }


}
