// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package matrix.util;

import java.awt.*;
import javax.swing.*;
import java.awt.event.*;

import matrix.animation.Animator;
import matrix.simulation.Picker;
import matrix.structures.FDT.FDT;
import matrix.visual.*;

/**
 * Application interface for all Applications using matrix framework. There are several
 * tasks invoked by the framework where tasks are delivered to the user interface. For 
 * example framework wants to add some functionality to menubar or to some pop-up menu. 
 * Direct reference from package matrix to the application build on matrix is not a 
 * good idea - so there are some demands for the user interface defined here. </p><p>
 *
 * Application word might be a bit confusing here. Matrix provides tools for creating 
 
 * 
 * @see ApplicationAdapter
 * @author Jussi Nikander
 * @author Panu Silvasti
 * @author Petri Tenhunen
 */
public interface Application{
  
    /**
     * Connect a new information to given VisualType
     */
    public void showInfoMessage(VisualType vt, String info);


    /**
     * Remove additional information of given VisualType
     */ 
    public void removeInfoMessage(VisualType vt);


    /**
     * Shows the argument string to user
     * 
     * @param title the title of the message
     * @param message the body of the message
     */
    public void showMessage(String title, String message);

    
    /**
     * Asks input from the user
     * @param title the question asked
     * @param default the default answer
     */
    public String getInput(String title, String defaultString);
  

    /**
     * This method adds the parameter application to the action
     * listeners of this application and vice versa. <p>
     * 
     * engage should be called when two applications are
     * representing the same structure and modifications of the 
     * structure should be updated to both Applications.
     */
    public void engage(Application app);

  
    /**
     * This method is called when there might be changes in structure
     * and visual view has to be validated. 
     */
    public void fireActionEvent(ActionEvent a);


    /**
     * Creates a new application.
     */
    public Application getNewInstance(FDT[] fdt, Animator a);


    /**
     * Creates a new window that contains the current application.
     */
    public JFrame getNewWindow();


    /**
     * Adds the given structure to be shown with this application.
     */
    public void addStructure(FDT fdt);

  /**
   * Changes the representation of the parameter visual container to
   * the type given in the parameter string
   */
  public void changeRepresentation(VisualContainer oldType, String newType);

    /**
     * Returns the structures in this Application
     */
    public FDT[] getStructures();

    public Picker getPicker();

    /**
     * Removes the given visual type from this application.
     */
    public void delete(VisualType vt);


    /**
     * Validates this application and all its components.
     */
    public void validate();

  /**
   * Adds the parameter ActionListener to the ActionListeners of
   * this Application.
   */
  public void addActionListener(ActionListener a);

  /**
   * Removes the parameter ActionListener from the ActionListeners of
   * this Application.
   */
  public void removeActionListener(ActionListener a);

    /**
     * There can be one visual animator connected to this application.
     * validateAnimator makes the animator validate itself, if there
     * is one. <p>
     *
     * Note that the visual animator should not be validated in the
     * validate method of this application, or there could be continuous
     * loop.
     */
    public void validateAnimator();
    

    public Dimension getSize();


    /**
     * Repants this application and all its subcomponents.
     */
    public void repaint();


    /**
     * Updates the references which have been manipulated in the GUI.
     */
    public void updateReferences();


    public Animator getAnimator();

}

