// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/* $Id: KeyBuilder.java,v 1.12 2007/09/07 13:39:38 vkaravir Exp $ */

/**
 * Creates digital keys from <code>String</code>s. Keys are represented as <code>BitSet</code>s, so it
 * is easy to do bitwise operations. There are different character encodings to choose from.
 *
 * @author Jussi Nikander
 */
package matrix.util;
import java.util.BitSet;

public class KeyBuilder {


    /**
     * creates a digital key form letters a-z, A-Z, 0-9, ignoring 
     * all other characters. The bit keys created by this method are
     * six-bit. 0 has the smallest key value (000001) and < the largest
     * (111110). There are no input characters that have key values of
     * (000000) or (111111). 
     *
     * @param s A <code>String</code> to be converted
     * @return a <code>BitSet</code> containing the digital keys to be used.
     */
    public static BitSet getDigitalKey(String s) {
	
	BitSet tmp = new BitSet((s.length()) * 6);
	int i = 0;
	int pos = 0;
	char a;
	
	for(int n = 0; n < s.length(); n++) {
	    a = s.charAt(n);
	    if(Character.isDigit(a) || Character.isUpperCase(a)) {
		i = Character.getNumericValue(a) + 1;
	    } else if(Character.isLowerCase(a)) {
		i = Character.getNumericValue(a) + 27;
	    } else {
		break;
	    }
	    for(int j = 0; j < 6; j++) {
		if((i & 32) == 32) 
		    tmp.set(pos++);
		else
		    tmp.clear(pos++);
		i <<= 1;
	    }
	}
	tmp.set(pos);
	return tmp;
    }

    /**
     * Creates a five-bit digital key from letters a-z, A-Z, ignoring case.
     * a and A have the smallest key value (00011) and z and Z have the largest
     * (11100).
     *
     * @param s the <code>String</code> to be converted
     * @return a <code>BitSet</code> containing the bit representation
     */
    public static BitSet getFiveBitDigitalKey(String s) {
	BitSet tmp = new BitSet((s.length() * 5) +1);
	int i;
	int pos = 0;
	char a;

	for(int n = 0; n < s.length(); n++) {
	    a = s.charAt(n);
	    if(Character.isLowerCase(a) || Character.isUpperCase(a)) 
		i = Character.getNumericValue(a) - 9;
	    else
		break;
	    for(int j = 0; j < 5; j++) {
		if((i & 16) == 16) 
		    tmp.set(pos++);
		else
		    tmp.clear(pos++);
		i <<= 1;
	    }
	}
	tmp.set(pos);
	return tmp;
    }

    /**
     * Tells the length of a digital key created from the argument
     * <code>String</code>.
     *
     * @param s The <code>String</code>.
     * @return the length of the digital key created from the argument <code>String</code>
     */
    public static int getDigitalKeyLength(String s) {
	int tmp = 0;
	char a;
	for(int n = 0; n < s.length(); n++) {
	    a = s.charAt(n);
	    if(Character.isLetterOrDigit(a))
		tmp += 6;
	}
	return tmp;
    }


}





