// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/* $Id: MG_Oval.java,v 1.16 2007/09/07 13:34:44 vkaravir Exp $ */

package matrix.util.export;

import java.awt.*;
import java.io.*;

/**
 * Represents oval in matrix graphics. 
 */
public class MG_Oval extends MG_Element {

    protected int x, y, width, height;
    protected Color fill;

    public MG_Oval(int x, int y, int width, int height, Color color) {
	this(x, y, width, height, color, null);
    }

    public MG_Oval(int x, int y, int width, int height, Color color,
		   Color fill) {
	super();
	this.x = x;
	this.y = y;
	this.width = width;
	this.height = height;
	this.color = color;
	this.fill = fill;
    }

    public MG_Oval(Rectangle rect, Color color, Color fill) {
	this(rect.x, rect.y, rect.width, rect.height, color, fill);
    }

    public MG_Oval(Rectangle rect, Color color,
		   Color fill, int step) {
	this(rect, color, fill);
	int i = 0;
	setOpacity(0);
	while (i <= step) {
	    addStep(this, i);
	    i++;
	}
    }

    public Rectangle getBounds() { 
	return new Rectangle(x, y, width, height); 
    }
    public int[] getCoord() { return new int[] { x, y }; }
    public int[] getSize() { return new int[] { width, height }; }
    public Color getFillColor() { return fill; }

    public String toString() {
	return "MG_Oval x:" + x + " y:" + y + " w:" + width + " h:" +
	    height + " color:" + color + " fill:" + fill + 
	    " opacity:" + getOpacity();
    }

    public MG_Element getStep(int step) {
	MG_Oval temp = (MG_Oval)steps.get(new Integer(step));
	if (temp == null) return null;
	return new MG_Oval(temp.getBounds(), temp.getColor(), 
			   temp.getFillColor());
    }

    public String toLatex() {
	float gLevel = ColorConverter.toGrayLevel(color);
	float lh = (float)(height/2.0);
	float lw = (float)(width/2.0);
 	float yl = -y - lh;
	float xl = x + lw;
	String tex = "\\move(" + xl + " " + yl + ")\n";
	if (lh == lw) {
	    if (fill != null) {
		float fLevel = ColorConverter.toGrayLevel(color);
		tex += "\\fcir f:" + fLevel + " r:" + lw + "\n";
		return tex;
	    } else {
		tex += "\\lcir r:" + lw + "\n";
		if (gLevel == 1)
		    tex = "\\setgray 1\n" + tex + "\\setgray 0\n";
	    }
	} else {
	    if (fill != null) {
		lh -= 0.5;
		lw -= 0.5;
		float fLevel = ColorConverter.toGrayLevel(color);
		double xp = 0;
		for (double i = 0; i <= lh; i += 0.2) {
		    double fact = (i*i*1.0)/(lh*lh*1.0);
		    xp = lw*Math.sqrt(1 - fact);
		    tex += "\\move (" + (xl - xp) + " " + (yl-i) + 
			")\n" + "\\lvec (" + (xl + xp) + " " + (yl-i) +
			")\n" + "\\move (" + (xl - xp) + " " + 
			(yl + i) + ")\n" + "\\lvec (" + 
			(xl + xp) + " " + (yl + i) + ")\n";
		}
		if (gLevel != 0)
		    tex = "\\setgray " + gLevel + "\n" + tex + "\\setgray 0\n";
	    } else {
		tex += "\\lellip rx:" + lw + " ry:" + lh + "\n";
		if (gLevel == 1)
		    tex = "\\setgray 1\n" + tex + "\\setgray 0\n";
	    }
	}
	return tex;
    }

    public void toSVG(Writer out, SVGUtil util) throws IOException {
	float rx = (float)(width/2.0);
	float ry = (float)(height/2.0);
	float sx = x + rx;
	float sy = y + ry;
	int[] colors = ColorConverter.toRGB(color);
	String tagName;
	out.write("<ellipse cx=\"" + sx + "\" cy=\"" + sy + "\" rx=\"" + 
		  rx + "\" ry=\"" + ry + "\"");
	if (getOpacity() != 1)
	    out.write(" opacity=\"" + getOpacity() + "\"");
	if (fill != null) {
	    int[] fillc = ColorConverter.toRGB(fill);
	    out.write(" fill=\"rgb(" + fillc[0] + "," + fillc[1] + 
		"," + fillc[2] + ")\" stroke=\"none\"");
	} else
	    out.write(" fill=\"none\"" + " stroke=\"rgb(" + colors[0] 
		+ "," + colors[1] + "," + colors[2] + ")\"");
	out.write(">");
	animate(out, util);
	out.write("</ellipse>\n");
    }

    public int getType() {
	return MG_Element.MG_OVAL;
    }

    /**
     * Creates the SVG elements needed to animate this element.
     *
     * @param out Output Writer
     * @param util SVGUtil used exporting this animation
     * @see SVGUtil
     */
    private void animate(Writer out, SVGUtil util) throws IOException {
	int i = 0;
	MG_Oval curr = this;
	while (hasStep(i+1)) {
	    MG_Oval next = (MG_Oval)steps.get(new Integer(i+1));
	    float rx = (float)(curr.width/2.0);
	    float ry = (float)(curr.height/2.0);
	    float sx = curr.x + rx;
	    float sy = curr.y + ry;
	    float rxto = (float)(next.width/2.0);
	    float ryto = (float)(next.height/2.0);
	    float sxto = next.x + rxto;
	    float syto = next.y + ryto;
	    if (sx != sxto) 
		out.write(util.animate("cx", i, "" + sx, "" + sxto));
	    if (sy != syto)
		out.write(util.animate("cy", i, "" + sy, "" + syto));
	    if (rx != rxto)
		out.write(util.animate("rx", i, "" + rx, "" + rxto));
	    if (ry != ryto)
		out.write(util.animate("ry", i, "" + ry, "" + ryto));
	    if (!curr.color.equals(next.color)) {
		int[] c1 = ColorConverter.toRGB(curr.color);
		int[] c2 = ColorConverter.toRGB(next.color);
		String from = "rgb(" + c1[0] + "," + c1[1] + "," + c1[2] + ")";
		String to = "rgb(" + c2[0] + "," + c2[1] + "," + c2[2] + ")";
		out.write(util.animate("animateColor", "stroke", i, 
				      from, to));
	    }
	    if (curr.fill != null && next.fill != null) 
		if (!curr.fill.equals(next.fill)) {
		    int[] c1 = ColorConverter.toRGB(curr.fill);
		    int[] c2 = ColorConverter.toRGB(next.fill);
		    String from = "rgb(" + c1[0] + "," + c1[1] + "," + c1[2] + ")";
		    String to = "rgb(" + c2[0] + "," + c2[1] + "," + c2[2] + ")";
		    out.write(util.animate("animateColor", "fill", i, 
					   from, to));
		}
	    if (curr.getOpacity() != next.getOpacity())         // opacity
		out.write(util.animate("opacity", i, "" + curr.getOpacity(), 
				       "" + next.getOpacity()));
	    i++;
	    curr = next;
	}
	float pause = util.getPause();
	float stepLength = util.getStepLength();
	float endTime = (i)*(pause + stepLength) + pause;
	if (!util.addPanel()) {
	    String opac = "0";
	    if (i == util.getMaxStep())
		opac = "0.5";
	    out.write(SVGUtil.animateAdditive("opacity", endTime, stepLength, 
					      opac, "sum"));
	}
	else if (i < util.getMaxStep() && util.addPanel())
	    out.write(util.animate("opacity", i, ""+curr.getOpacity(), "0"));
    }
}
