// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/* $Id: MG_Text.java,v 1.20 2007/09/07 13:34:44 vkaravir Exp $ */

package matrix.util.export;

import java.awt.*;
import java.io.*;

/**
 * Represents text in matrix graphics. 
 */
public class MG_Text extends MG_Element {

    /** Text contents of this object. */
    private String str;
    /** Coordinates of this object. */
    private int x, y;
    /** Font. */
    private Font font;
    /** Fontmetrics. */
    private FontMetrics fm;
    /** Special characters replaced to export SVG */
    private static final String[] svgSpecial = new String[] {"&", "<"};
    /** Strings used in replacing special characters in SVG */
    private static final String[] svgReplace = new String[] {"&amp;", "&lt;"};
    /** Special characters replaced to export LaTeX */
    private static final String[] latexSpecial = new String[] {"\\", "{", "}", "^", "$", "&", "#", "_", "~", ">"};
    /** Strings used in replacing special characters in LaTeX*/
    private static final String[] latexReplace = new String[] {"$\\backslash$", "\\{", "\\}", "\\^{}", "\\$", "\\&", "\\#", "\\_", "\\~{}", "$>$"};

    /**
     * Constructs a new <code>MG_Text</code> element using the
     * given initial values. 
     *
     * @param str the string of this text element
     * @param x the x coordinate
     * @param y the y coordinate
     * @param color the color of the text
     * @param font the font of the text
     * @param fm the fontmetrics
     */
    public MG_Text(String str, int x, int y, Color c, Font f, 
		   FontMetrics fm) {
	super();
	this.str = str;
	this.x = x;
	this.y = y;
	color = c;
	font = f;
	this.fm = fm;
    }

    /**
     * Constructs a new <code>MG_Text</code> element using the
     * given initial values for all steps before the given step.
     * 
     *
     * @param str the string of this text element
     * @param coord the x (index 0) and y (index 1) coordinates
     * @param color the color of the text
     * @param font the font of the text
     * @param fm the fontmetrics
     * @param step the step that these properties are used in
     */
    public MG_Text(String str, int[] coord, Color c, Font f, 
		   FontMetrics fm, int step) {
	this(str, coord[0], coord[1], c, f, fm);
	int i = 0;
	setOpacity(0);
	while (i <= step) {
	    addStep(this, i);
	    i++;
	}
    }

    public String toString() {
	return "MG_Text str:" + str + " x:" + x + " y:" + y + 
	    " color:" + color + " opacity:" + getOpacity();
    }

    /**
     * Returns the string contents of this element.
     *
     * @return String contents of this element.
     */
    public String getString() {
	return str;
    }

    /**
     * Returns the coordinates of this element.
     *
     * @return array of this elements coordinates, index 0 is x-coordinate
     *         and index 1 is y-coordinate.
     */
    public int[] getCoord() {
	return new int[]{x, y};
    }

    /**
     * Returns the <code>Font</code> of this element.
     *
     * @return the font of this element
     */
    public Font getFont() {
	return font;
    }

    /**
     * Return the <code>FontMetrics</code> of this element.
     *
     * @return the <code>FontMetrics</code> of this element
     */
    public FontMetrics getFontMetrics() {
	return fm;
    }

    /**
     * Generates Texdraw representation of this text element
     * using the given scaling for fonts. 
     *
     * @param scale the scaling factor
     * @return Texdraw representation of this element as 
     * <code>String</code>
     */
    public String toLatex(double scale) {
	int yl = -y + fm.getHeight()/2 - 2; // y in LaTeX coordinates
	if ((str.trim()).equals(""))
	    return "";
	float gLevel = ColorConverter.toGrayLevel(color);
	String tex = "\\htext (" + x + " " + yl + ") {\\matrixfontselect " + 
            ((gLevel == 1.0) ? "\\color{white}{" : "") + // write white text white
	    replaceString(str, latexSpecial, latexReplace) +
            ((gLevel == 1.0) ? "}" : "") +
            "}\n";
	return tex;
    }

    // interface MG_Element

    public MG_Element getStep(int step) {
	MG_Text temp = (MG_Text)steps.get(new Integer(step));
	if (temp == null) return null;
	return new MG_Text(temp.str, temp.x, temp.y, temp.color,
			   temp.font, temp.fm);
    }

    public String toLatex() {
	return toLatex(1);
    }

    public void toSVG(Writer out, SVGUtil util) throws IOException {
	int[] colors = ColorConverter.toRGB(color);
        out.write("<text text-anchor=\"start\"");
	out.write(" x=\"" + x + "\"");
	out.write(" y=\"" + y + "\" stroke=\"none\"");
	out.write(" fill=\"rgb(" + colors[0] + "," + colors[1] + "," + 
	    colors[2] + ")\"");
	if (getOpacity() != 1)
	    out.write(" opacity=\"" + getOpacity() + "\"");
	out.write(" font-size=\"" + font.getSize() // -
		  /*      Math.round(font.getSize()*0.15) - 1)*/ + "\"");
	out.write(" font-family=\"" + font.getFamily() + "\"");
	out.write(">" + replaceString(str, svgSpecial, svgReplace));
	animate(out, util);
	out.write("</text>\n");
    }

    public int getType() {
	return MG_Element.MG_TEXT;
    }

    // interface MG_Element ending

    /**
     * Creates the SVG elements needed to animate this element.
     *
     * @param out Output Writer
     * @param util SVGUtil used exporting this animation
     * @see SVGUtil
     */
    private void animate(Writer out, SVGUtil util) throws IOException {
	int i = 0;
	MG_Text curr = this;
	float stepLength = util.getStepLength();
	float pause = util.getPause();
	while (hasStep(i+1)) {
	    MG_Text next = (MG_Text)steps.get(new Integer(i+1));
	    out.write(animateStep(curr, next, i, util));
	    if (!curr.str.equals(next.str)) {                   // text
		float beginTime = (i)*(pause + stepLength) + pause;
		out.write(util.animate("opacity", i, "" +curr.getOpacity(),
				       "0"));
		out.write("</text>");
		int[] colors = ColorConverter.toRGB(curr.color);
		out.write("<text text-anchor=\"start\"");
		out.write(" x=\"" + curr.x + "\"");
		out.write(" y=\"" + curr.y + "\" stroke=\"none\"");
		out.write(" fill=\"rgb(" + colors[0] + "," + colors[1] + "," + 
		    colors[2] + ")\"");
		out.write(" font-size=\"" + font.getSize() //-
			  // Math.round(font.getSize()*0.15)
			  /* - 1)*/ + "\"");
		out.write(" font-family=\"" + font.getFamily() + "\"");
		out.write(" opacity=\"0\"");
		out.write(">" + replaceString(next.str, svgSpecial, svgReplace));
		out.write(animateStep(curr, next, i, util));
		out.write(util.animate("opacity", i, "0", "1"));
	    }
	    i++;
	    curr = next;
	}
	// hide element at the end
	float endTime = (i)*(pause + stepLength) + pause;
	if (!util.addPanel()) {
	    String opac = "0";
	    if (i == util.getMaxStep())
		opac = "0.5";
	    out.write(SVGUtil.animateAdditive("opacity", endTime, stepLength, 
					      opac, "sum"));
	} else if (i < util.getMaxStep() && util.addPanel())
	    out.write(util.animate("opacity", i, ""+curr.getOpacity(), "0"));
    }

    /**
     * Animate the given elements. The current step and the next 
     * step are given as parameters.
     *
     * @param curr the element of the current step
     * @param next the element of the next step
     * @param stepLength the length of one step of animation
     * @param beginTime the time this step begins
     * @return string representing the needed animations
     */
    private String animateStep(MG_Text curr, MG_Text next, int step,
			       SVGUtil util) {
	String anim = "";
	if (curr.x != next.x)                   // x coordinate
	    anim += util.animate("x", step, "" + curr.x, 
				 "" + next.x);
	if (curr.y != next.y)                   // y coordinate
	    anim += util.animate("y", step, "" + curr.y, 
				 "" + next.y);
	if (!curr.color.equals(next.color)) {          // color
	    int[] c1 = ColorConverter.toRGB(curr.color);
	    int[] c2 = ColorConverter.toRGB(next.color);
	    String from = "rgb(" + c1[0] + "," + c1[1] + "," + c1[2] + ")";
	    String to = "rgb(" + c2[0] + "," + c2[1] + "," + c2[2] + ")";
	    anim += util.animate("animateColor", "fill", step, from, to);
	}
	if (curr.getOpacity() != next.getOpacity())         // opacity
	    anim += util.animate("opacity", step, "" + curr.getOpacity(), 
				 "" + next.getOpacity());
	if (curr.font.getSize() != next.font.getSize())
	    anim += util.animate("font-size", step, "" + (curr.font.getSize() -
				 Math.round(curr.font.getSize()*0.15)), 
				 "" + (next.font.getSize() - 
				 Math.round(next.font.getSize()*0.15)));
	if (!curr.font.getFamily().equals(next.font.getFamily()))
	    anim += util.animate("font-family", step, curr.font.getFamily(),
				 next.font.getFamily());
	return anim;
    }

    /**
     * Utility method to replace strings. Needed when taking care
     * of special characters in SVG and LaTeX. Strings in array oldS
     * at index <code>x</code> is replaced with string in array
     * newS at index <code>x</code> in the original string.
     *
     * @param orig the original string
     * @param oldS an array of the strings to replace
     * @param newS an array of the strings to replace with
     * @return string with the wanted replacements
     */
    private String replaceString(String orig, String[] oldS, String[] newS) {
	String replaced = orig;
	for (int n = 0; n < oldS.length; n++) {
	    int i = replaced.indexOf(oldS[n]);
	    while (i != -1) {
		String temp = replaced;
		replaced = temp.substring(0, i);
		replaced += newS[n];
		replaced += temp.substring(i + oldS[n].length());
		i = replaced.indexOf(oldS[n], i + newS[i].length() );
	    }
	}
	return replaced;
    }
}
