// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/* $Id: SVGWriter.java,v 1.24 2007/09/07 13:34:43 vkaravir Exp $ */

package matrix.util.export;

import java.awt.*;
import java.io.*;
import java.util.Enumeration;

import matrix.util.Note;

/** 
 * Class used to export SVG animations from Matrix.
 */
public class SVGWriter {

    /** Default length of one animation step, 2s. */
    public static final float DEFAULTSTEPLENGTH = 2;
    /** Default length of pause between steps, 1s. */
    public static final float DEFAULTPAUSE = 1;
    /** Default value representing wheteher or not add animator panel, true. */
    public static final boolean DEFAULTANIMATORPANEL = true;
    /** Maximum step used in animation. Initially 0. */
    private static int maxStep = 0;
    /** Contents that are to be exported to SVG. */
    private MatrixGraphicsContents mgc;
    private SVGUtil util;
    /** Background color to be used with the animation. Default value is
     * lightgray (rgb(211, 211, 211)). */
    private Color bgcolor = new Color(211, 211, 211);

    /**
     * Constructs a new SVGWriter for exporting the given 
     * <code>MatrixGraphicsContents</code>.
     */
    public SVGWriter(MatrixGraphicsContents mgc) {
	this.mgc = mgc;
    }

    /** 
     * Writes SVG animation of the contents this object contains
     * to the given file. Default stepLength and pause times are
     * used. Same as write(DEFAULTSTEPLENGTH, DEFAULTPAUSE, f,
     * DEFAULTANIMATORPANEL).
     * 
     * @param f Output file
     */
    public void write(File f) {
	write(DEFAULTSTEPLENGTH, DEFAULTPAUSE, f, DEFAULTANIMATORPANEL);
    }

    /** 
     * Writes SVG animation of the contents this object contains
     * to the given file using the given stepLength and pause
     * times. Same as write(stepLength, pause, f, DEFAULTANIMATORPANEL).
     * 
     * @param stepLength Length of one step of animation (seconds)
     * @param pause Pause between steps (seconds)
     * @param f Output file
     */
    public void write(float stepLength, float pause, File f) {
	write(stepLength, pause, f, DEFAULTANIMATORPANEL);
    }	    

    public void write(File f, boolean panel) {
	write(DEFAULTSTEPLENGTH, DEFAULTPAUSE, f, panel);
    }

    public void write(float stepLength, float pause, File f, boolean panel) {
	try {
	    write(stepLength, pause, new FileWriter(f), panel);
	} catch (IOException e) {
	    Note.out(e,"Error: " + e.toString());
	    e.printStackTrace();
	}
    }

    /** 
     * Writes SVG animation of the contents this object contains
     * to the given <code>OutputStream</code> using the default 
     * stepLength and pause times.
     * 
     * @param stepLength Length of one step of animation (seconds)
     * @param pause Pause between steps (seconds)
     * @param f Output file
     */
    public void write(OutputStream out) {
	write(out, DEFAULTANIMATORPANEL);
    }

    public void write(OutputStream out, boolean panel) {
	write(new OutputStreamWriter(out), panel);
    }

    public void write(float stepLength, float pause, OutputStream out,
		      boolean panel) {
	write(stepLength, pause, new OutputStreamWriter(out), panel);
    }
    
    /** 
     * Writes SVG animation of the contents of this object
     * to the given <code>Writer</code> using the default 
     * stepLength and pause times.
     * 
     * @param stepLength Length of one step of animation (seconds)
     * @param pause Pause between steps (seconds)
     * @param f Output file
     */
    public void write(Writer out) {
	write(out, DEFAULTANIMATORPANEL);
    }
    
    public void write(Writer out, boolean panel) {
	write(DEFAULTSTEPLENGTH, DEFAULTPAUSE, out, panel);
    }

    public void write(float stepLength, float pause, Writer out, boolean panel) {
	util = new SVGUtil(stepLength, pause, maxStep, panel);
	float scale = (float)mgc.getScale();
	Rectangle bounds = mgc.getBounds();
	try {
	    out.write("<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>\n");
	    out.write("<!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 1.0//EN\" ");
	    out.write("\"http://www.w3.org/TR/2001/REC-SVG-20010904/DTD/" +
		      "svg10.dtd\">\n");
	    out.write("<svg width=\"" + 
		      ((panel)?Math.max(bounds.width*scale, 300):bounds.width*scale) +
		      "\" height=\"" +
		      ((panel)?(bounds.height*scale+105):(bounds.height*scale)) + 
		      "\" viewBox=\"0 0 " + 
		      ((panel)?Math.max(bounds.width*scale, 300):bounds.width*scale) +
		      " " + ((panel)?bounds.height*scale+105:bounds.height*scale) + 
		       "\" xmlns=\"http://www.w3.org" +
		      "/2000/svg\" preserveAspectRatio=\"xMidYMid meet\" " +
		      "version=\"1.0\" zoomAndPan=\"magnify\"" +
		      ((panel)?" onload=\"init(evt)\"" : "") +
		      ">\n\n");
	    String translation;
	    if (bounds.x > 0 || bounds.y > 0) {
		if (bounds.width*scale < 300 && panel)
		    translation = "" + ((300 - bounds.width*scale)/2 - bounds.x*scale);
		else
		    translation = "-" + Math.max(bounds.x, 0)*scale;
		translation += " -" + Math.max(bounds.y, 0)*scale;
	    } else
		translation = "1, 1";
	    out.write("<g transform=\"translate(" + translation + ")" +
		      " scale(" + scale + ")\" stroke-width=\"1\">\n");
	    out.write("<desc>Data structures</desc>\n");
	    double rectx = bounds.x;
	    if (panel && bounds.width*scale < 300)
		rectx = (bounds.x*scale - (300 - bounds.width*scale)/2)/scale;
	    out.write("<rect x=\"" + rectx + "\" y=\"" + bounds.y + 
		      "\" width=\"" +
		      ((panel)?Math.max(bounds.width, 300/scale):bounds.width) + 
		      "\" height=\"" + bounds.height + "\" ");
	    //if (bgcolor == null)
	    //out.write("fill=\"lightgray\"/>\n");
	    //	    else {
		int[] colors = ColorConverter.toRGB(bgcolor);
		out.write("fill=\"rgb(" + colors[0] + "," + colors[1] + "," + 
		  colors[2] + ")\"/>\n");
		//}
	    writeGroup(mgc.getRoot(), out);
	    out.write("</g>\n");
	    if (panel && maxStep > 0) {
		addScript(out);
		addControlPanel(out, bounds);
	    }
	    out.write("</svg>\n");
	    out.close();
	} catch (IOException e) {
	    Note.out(e,"Error: " + e.toString()); 
	    e.printStackTrace();
	}
    }

    /**
     * Writes SVG-representation of the given group and all of it's
     * subgroups to the given <code>Writer</code>. This method should
     * only be called to the root of the group hierarchy.
     * 
     * @param grp Group which is to be exported.
     * @param out The <code>Writer</code>.
     */
    private void writeGroup(MG_Group grp, Writer out)
             throws IOException {
	Enumeration subs = grp.getGroups();
	Enumeration contents = grp.getContents();
	out.write("<g>\n");//<desc>" + grp.getName() + "</desc>\n");
	while (contents.hasMoreElements()) {
	    Object obj = contents.nextElement();
	    if (obj instanceof MG_Element)
		((MG_Element)obj).toSVG(out, util);
	}
	while (subs.hasMoreElements())
	    writeGroup((MG_Group)subs.nextElement(), out);
	out.write("</g>\n");
    }

    /** 
     * Sets the maximum step used in the animation this object is used
     * to export. This must be set before writing!
     *
     * @param maxStep the new maximum step
     */
    public void setMaxStep(int maxStep) { 
	SVGWriter.maxStep = maxStep;
    }

    /** Returns the maximum step used in the animation. */
    public int getMaxStep() { return maxStep; }


    /**
     * Writes scripts needed to control the animation with animator panel
     * to the given <code>Writer</code>.
     *
     * @param out Output writer
     */
    private void addScript(Writer out) throws IOException{
	int firstStep = util.getFirstStep() + 1;
	int lastStep = util.getLastStep();
	out.write("\n<!-- Scripts needed to control the animation -->\n\n");
	out.write("<defs>\n<script type=\"text/ecmascript\"" +
		  /* Viewer specific, for Adobe SVG Viewer 3.0+. Sets this
		     file to use ASV scripting engine. */
		  /*
		    " xmlns:a3=\"http://ns.adobe.com/AdobeSVGViewerExtensions/3.0/\"" +
		    " a3:scriptImplementation=\"Adobe\"" + 
		  */
		  "><![CDATA[\n");
	out.write("\tvar svgdoc;\n\tvar playB;\n\tvar forwardB;\n" +
		  "\tvar beginB;\n\tvar backwardB;\n" +
		  "\tvar step = 1;\n" +
		  "\tvar speed = 100;\n\tvar ind;\n" +
		  "\tvar stepLength = " + util.getStepLength() + ";\n" +
		  "\tvar play = false;\n\tvar animPlaying = false;\n" +
		  "\tvar steps = new Array(" + util.getUsedSteps() + ");\n" +
		  "\tvar maxStep = steps.length;\n");
	
	out.write("\n\tfunction init(evt) {\n" +
		  "\t\tsvgdoc = evt.getTarget().getOwnerDocument();\n" +
		  "\t\tplayB = svgdoc.getElementById(\"plButton\");\n" +
		  "\t\tforwardB = svgdoc.getElementById(\"forwButton\");\n" +
		  "\t\tbeginB = svgdoc.getElementById(\"begButton\");\n" +
		  "\t\tbackwardB = svgdoc.getElementById(\"backwButton\");\n" +
		  /* Viewer specific, for Adobe SVG Viewer 3.0+. Disables
		     pause from ASV popupmenu. */
		  /*
		    "\t\tvar mNodes = window.contextMenu.getElementsByTagName" +
		    "(\"item\");\n\t\tvar i = 0;\n\t\tvar ok = false;\n" +
		    "\t\twhile (i < mNodes.length && ok == false) {\n" +
		    "\t\t\tif (mNodes.item(i).getAttribute(\"action\") " +
		    "== \"Pause\") {\n\t\t\t\tmNodes.item(i).setAttribute" +
		    "(\"enabled\", \"no\");\n\t\t\t\tok = true;\n\t\t\t}\n" +
		    "\t\t\ti++;\n\t\t}\n" +
		  */
		  "\t\tind = svgdoc.getElementById(\"ind\");\n" +
		  "\t}\n");
	out.write("\n\tfunction forwardClick() {\n" +
		  "\t\tif (play == false) forward();\n\t}\n");
	out.write("\n\tfunction forward() {\n" +
		  "\t\tif (animPlaying == true) return;\n" +
		  "\t\tif (play == false)" +
		  " backwardB.setAttribute(\"opacity\", 1);\n" +
		  "\t\tif (step >= maxStep) return;\n" +
		  "\t\tind.setAttribute(\"x1\", " +
		  "(step)*(178/(maxStep - 1)) + 6);\n" +
		  "\n\t\tind.setAttribute(\"x2\", " +
		  "(step)*(178/(maxStep - 1)) + 6);\n" +
		  "\t\tanimPlaying = true;\n" +
		  "\t\tsvgdoc.getElementById(\"speed\").setAttribute" +
		  "(\"opacity\", 0.2);\n" +
		  "\t\tsvgdoc.getElementById(\"anim\" + steps[step])" +
		  ".beginElement();\n\t\tstep++;\n\t}\n");
	out.write("\n\tfunction backwardClick() {\n" +
		  "\t\tif (play == false) backward();\n\t}\n");
	out.write("\n\tfunction backward() {\n" +
		  "\t\tif (animPlaying == true) return;\n" +
		  "\t\tforwardB.setAttribute(\"opacity\", 1);\n" +
		  "\t\tplayB.setAttribute(\"opacity\", 1);\n" +
		  "\t\tif (step <= 1) return;\n\t\tstep--;\n" +
		  "\t\tif (step == 1) " +
		  "backwardB.setAttribute(\"opacity\", 0.2);\n" +
		  "\t\tind.setAttribute(\"x1\", (step - 1)*" +
		  "(178/(maxStep)) + 6);\n" +
		  "\t\tind.setAttribute(\"x2\", (step - 1)*" +
		  "(178/(maxStep)) + 6);\n" +
		  "\t\tanimPlaying = true;\n" +
		  "\t\t\tsvgdoc.getElementById(\"speed\").setAttribute" +
		  "(\"opacity\", 0.2);\n" +
		  "\t\tsvgdoc.getElementById(\"animb\" + steps[step])." +
		  "beginElement();\n\t}\n");
	out.write("\n\tfunction playAnimation() {\n" +
		  "\t\tif (step >= maxStep) return;\n" +
		  "\t\tif (play == false) {\n\t\t\tplay = true;\n" +
		  "\t\t\tplayB.firstChild.data = \"Stop\";\n" +
		  "\t\t\tforwardB.setAttribute(\"opacity\", 0.2);\n" +
		  "\t\t\tbackwardB.setAttribute(\"opacity\", 0.2);\n" +
		  "\t\t\tstartNextElement();\n" +
		  "\t\t} else {\n\t\t\tplay = false;\n" +
		  "\t\t\tplayB.firstChild.data = \"Play\";\n" +
		  "\t\t\tif (step < maxStep) forwardB.setAttribute" +
		  "(\"opacity\", 1);\n" +
		  "\t\t\tif (step > 1) backwardB.setAttribute" +
		  "(\"opacity\", 1);\n" +
		  "\t\t\tsvgdoc.getElementById(\"speed\").setAttribute" +
		  "(\"opacity\", 1);\n\t\t}\n\t}\n");
	out.write("\n\tfunction startNextElement() {\n" +
		  "\t\tif (play == true) forward();\n" +
		  "\t\telse svgdoc.getElementById(\"speed\").setAttribute(" +
		  "\"opacity\", 1);\n\t}\n");
	out.write("\n\tfunction animationEnd() {\n" +
		  "\t\tanimPlaying = false;\n" +
		  "\t\tif (step >= maxStep) {\n" +
		  "\t\t\tplayB.setAttribute(\"opacity\", 0.2);\n" +
		  "\t\t\tforwardB.setAttribute(\"opacity\", 0.2);\n" +
		  "\t\t\tbackwardB.setAttribute(\"opacity\", 1);\n" +
		  "\t\t\tsvgdoc.getElementById(\"speed\").setAttribute" +
		  "(\"opacity\", 1);\n\t\t\tif (play == true) {\n" +
		  "\t\t\t\tplay = false;\n" +
		  "\t\t\t\tplayB.firstChild.data = \"Play\";\n\t\t\t}\n" +
		  "\t\t} else startNextElement();\n\t}\n");
	out.write("\n\tfunction changeSpeed(raise) {\n" +
		  "\t\tif (animPlaying == true) return;\n" +
		  "\t\tvar sliderElem = svgdoc.getElementById(\"slider\");\n" +
		  "\t\tvar oldX = sliderElem.getAttribute(\"x1\");\n" +
		  "\t\tif (raise == true) {\n" +
		  "\t\t\tif (speed == 200) return;\n\t\t\tspeed += 20;\n" +
		  "\t\t\tif (speed == 21) speed--;\n" +
		  "\t\t\tsliderElem.setAttribute(\"x1\", oldX - 12);\n" +
		  "\t\t\tsliderElem.setAttribute(\"x2\", oldX - 12);\n" +
		  "\t\t} else {\n\t\t\tif (speed == 1) return;\n" +
		  "\t\t\tspeed -= 20;\n" +
		  "\t\t\tsliderElem.setAttribute(\"x1\", oldX - (-12));\n" +
		  "\t\t\tsliderElem.setAttribute(\"x2\", oldX - (-12));\n" +
		  "\t\t\tif (speed == 0) speed = 1;\n\t\t}\n" +
		  "\t\tvar newSL = (speed/100)*stepLength;\n" +
		  "\t\tvar animNodes = svgdoc.getElementsByTagName(" +
		  "\"animateColor\");\n\t\tvar i = 0;\n" +
		  "\t\twhile (i < animNodes.length)\n" +
		  "\t\t\tanimNodes.item(i++).setAttribute(\"dur\", newSL);\n" +
		  "\t\tanimNodes = svgdoc.getElementsByTagName" +
		  "(\"animate\");\n\t\ti = 0;\n" +
		  "\t\twhile (i < animNodes.length) {\n" +
		  "\t\t\tanimNodes.item(i).setAttribute(\"dur\", newSL);\n" +
		  "\t\t\ti++;\n\t\t}\n\t}\n");
	out.write("]]></script>\n</defs>\n\n");
    }

    /**
     * Writes svg elements for the animator panel to the given 
     * <code>Writer</code>. 
     *
     * @param out Output writer
     * @param bounds Bounds of the graphics used. Can be obtained
     *               from <code>MatrixGraphicsContent</code>.
     */
    private void addControlPanel(Writer out, Rectangle bounds) 
                 throws IOException {
        float scale = (float)mgc.getScale();
	double xTransl = Math.max((bounds.width*scale - 300)/2, 5);
	double yTransl = bounds.height*scale + 10;
	out.write("<g transform=\"translate(" + xTransl + "," + yTransl +
		  ") scale(1.5)\" stroke-width=\"1\">\n");
	out.write("<title>Animator Panel</title>\n");
	out.write("<desc>Buttons needed to control the animation.</desc>\n");
	out.write("<g onclick=\"backwardClick()\">\n" +
		  "<rect x=\"0\" y=\"0\" width=\"40\" height=\"15\" " +
		  "fill=\"lightgrey\" stroke=\"black\"/>\n" +
		  "<text x=\"2\" y=\"10\" fill=\"black\" font-size=\"8\" " +
		  "id=\"backwButton\" opacity=\"0.2\">Backward</text>\n" +
		  "</g>\n");
	out.write("<g onclick=\"forwardClick()\">\n" +
		  "<rect x=\"50\" y=\"0\" width=\"40\" height=\"15\" " +
		  "fill=\"lightgrey\" stroke=\"black\"/>\n" +
		  "<text x=\"55\" y=\"10\" fill=\"black\" font-size=\"8\" " +
		  "id=\"forwButton\">Forward</text>\n</g>\n");
	out.write("<g onclick=\"reload()\">\n" +
		  "<rect x=\"100\" y=\"0\" width=\"40\" height=\"15\" " +
		  "fill=\"lightgrey\" stroke=\"black\"/>" +
		  "<text x=\"110\" y=\"10\" fill=\"black\" font-size=\"8\" " +
		  "id=\"begButton\">Begin</text>\n</g>\n");
	out.write("<g onclick=\"playAnimation()\">\n" +
		  "<rect x=\"150\" y=\"0\" width=\"40\" height=\"15\" " +
		  "fill=\"lightgrey\" stroke=\"black\"/>\n" +
		  "<text x=\"162\" y=\"10\" fill=\"black\" font-size=\"8\" " +
		  "id=\"plButton\">Play</text>\n</g>\n");
	out.write("<g>\n<text x=\"0\" y=\"40\" font-size=\"8\" " +
		  "fill=\"black\">Animation speed:</text>\n" +
		  "<rect x=\"0\" y=\"45\" width=\"30\" height=\"12\" " +
		  "fill=\"lightgrey\" stroke=\"black\" " +
		  "onclick=\"changeSpeed(true)\"/>\n" +
		  "<rect x=\"160\" y=\"45\" width=\"30\" height=\"12\" " +
		  "fill=\"lightgrey\" stroke=\"black\" " +
		  "onclick=\"changeSpeed(false)\"/>\n" +
		  "<g id=\"speed\"\n>" +
		  "<line x1=\"12\" y1=\"51\" x2=\"18\" y2=\"51\" " +
		  "stroke=\"black\" fill=\"none\" " +
		  "onclick=\"changeSpeed(true)\"/>\n" +
		  "<line x1=\"172\" y1=\"51\" x2=\"178\" y2=\"51\" " +
		  "stroke=\"black\" fill=\"none\" " +
		  "onclick=\"changeSpeed(false)\"/>\n" +
		  "<line x1=\"175\" y1=\"48\" x2=\"175\" y2=\"54\" " +
		  "stroke=\"black\" fill=\"none\" " +
		  "onclick=\"changeSpeed(false)\"/>\n</g>\n" +
		  "<line x1=\"35\" y1=\"51\" x2=\"155\" y2=\"51\" " +
		  "stroke=\"black\" fill=\"none\"/>\n " +
		  "<line x1=\"95\" y1=\"48\" x2=\"95\" y2=\"54\" " +
		  "stroke=\"black\" fill=\"none\" id=\"slider\"/>\n" +
		  "</g>\n");
	out.write("<g>\n<polygon fill=\"lightgrey\" stroke=\"black\" " +
		  "points=\"0,25 5,22 5,28\" " +
		  "onclick=\"backwardClick()\"/>\n" +
		  "<polygon fill=\"lightgrey\" stroke=\"black\" " +
		  "points=\"185,22 190,25 185,28\" " +
		  "onclick=\"forwardClick()\"/>\n" +
		  "<line x1=\"5\" y1=\"25\" x2=\"185\" y2=\"25\" " +
		  "stroke=\"black\" fill=\"none\"/>\n" +
		  "<line x1=\"6\" y1=\"22\" x2=\"6\" y2=\"28\" " +
		  "fill=\"none\" stroke=\"black\" id=\"ind\"/>\n" +
		  "</g>\n");
	out.write("</g>\n");
    }


    /**
     * Sets the background color of the created animation. The default color
     * is lightgray (rgb(211, 211, 211). 
     *
     * @param cl The new background color.
     */
    public void setBackground(Color cl) {
	this.bgcolor = cl;
    }
}
	    
