// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package matrix.util.io;

import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.io.*;
import java.net.*;
import java.util.Vector;

import matrix.animation.Animator;
import matrix.structures.FDT.FDT;
import matrix.util.*;

import matrix.uitools.StructurePanel;
import matrix.visual.*;

/**
 * A library class to save and load matrix animations from and to a file. Instance of
 * this class is created internally to store the animator and the FDT Vector. This 
 * instance is then serialized into a file. 
 *
 * @author Petri Tenhunen
 */
public class MatrixSerialization implements Serializable {

    private Animator animator;
    private FDT[] savedFDT;
    private MatrixLayout layout;
    /** 
     * Instance of this class is only created to encapsulate animator and fdt vector
     * together. This capsulation is serialized into a file when save is called.
     *
     * @param a animator to save
     * @param v fdt vector to save
     */
    protected MatrixSerialization(Animator a, Vector v) {
	animator = a;
	this.savedFDT = new FDT[v.size()];
	v.copyInto(savedFDT);
    }

  /** 
     * Instance of this class is only created to encapsulate animator and fdt vector
     * together. This capsulation is serialized into a file when save is called.
     *
     * @param a animator to save
     * @param ss fdt array to save
     */
    protected MatrixSerialization(Animator a, FDT[] ss) {
        animator = a;
        savedFDT = ss;
    }


    public void setLayout(MatrixLayout ml) {
	layout = ml;
    }

    public MatrixLayout getLayout() {
	return layout;
    }

    /** This inner class is for storing the layout while saving serialized Matrix objects.
     * The layout variables (i.e., the name of the representation, title name, orientation, etc.) 
     * are stored into the java.util.Properties as a name/value pairs and restored while the
     * structures is loaded back. 
     */
    public static class MatrixLayout implements Serializable {
	java.util.Properties prop[];
	/* Stored properties */
	static final String NAME = "name"; // title name of the container
	static final String REPRESENTATION = "representation"; // the name of the representation 
	
	static final String MINIMIZED = "minimized"; // boolean flag
	static final String ALIVE = "alive"; // boolean flag
	static final String ENABLED = "enabled"; // boolean flag

	static final String HOTSPOTENABLED = "hotspotenabled"; // boolean flag
	static final String LABELED = "labeled"; // boolean flag
	static final String TITLED = "titled"; // boolean flag
	static final String ROTATED = "rotated"; // boolean flag
	static final String FLIPPEDX = "flippedX"; // boolean flag
	static final String FLIPPEDY = "flippedY"; // boolean flag

	static final String NOTDEFINED = "###notdefined###"; // boolean flag
        
        // Values of the layout settings start here.
        // for the documentation, see jawa.awt.GridBagConstraints
        static final String GRIDX = "gridx"; // int value
        static final String GRIDY = "gridy"; // int value
        static final String GRIDHEIGHT = "gridheight"; // int value
        static final String GRIDWIDTH = "gridwidth"; // int value
        static final String IPADX = "ipadx"; // int value
        static final String IPADY = "ipady"; // int value
        static final String WEIGHTX = "weightx"; // double value
        static final String WEIGHTY = "weighty"; // double value
        static final String BOTTOM = "bottom"; // int value
        static final String TOP = "top"; // int value
        static final String LEFT = "left"; // int value
        static final String RIGHT = "right"; // int value
        static final String ANCHOR = "anchor"; // not supporteed yet
        static final String FILL = "fill"; // not supported yet
        // Values of the layout settings end here.
 
	public MatrixLayout(FDT[] ss, StructurePanel panel) {
	    prop = new java.util.Properties[ss.length];
	    for (int i=0; i<ss.length; i++) {
		prop[i] = new java.util.Properties();
		storeProperties(ss[i], prop[i], panel, i);
	    }
	}    
	
	public void storeProperties(FDT fdt, java.util.Properties prop, StructurePanel p, int i) {
	    try { 
		VisualContainer c = (VisualContainer)p.getComponent(i);
		//Note.out(this, "Saving properties for " + c);
		// java.awt.Container methods
		prop.setProperty(NAME, c.getName() != null ? c.getName() : "no name");
		// matrix.VisualContainer (extends java.awt.Container) methods
		prop.setProperty(HOTSPOTENABLED, ""+c.isHotSpotEnabled());
		prop.setProperty(REPRESENTATION, c.getNameOfRepresentation() != null ? c.getNameOfRepresentation() : NOTDEFINED);
		//Note.out(this, "representationname=<"+c.getNameOfRepresentation()+">");
		//Note.out(this, "l="+c.getNameOfRepresentation().length());
		prop.setProperty(MINIMIZED, ""+c.isMinimized());
		prop.setProperty(ALIVE, ""+c.isAlive());
		prop.setProperty(ENABLED, ""+c.isSubComponentsEnabled());

		//isLabeled returns always true and setLabeled not implemented
		//prop.setProperty(LABELED, ""+c.isLabeled());
		prop.setProperty(TITLED, ""+c.isTitled());
		prop.setProperty(ROTATED, ""+c.isRotated());
		prop.setProperty(FLIPPEDX, ""+c.isFlippedX());
		prop.setProperty(FLIPPEDY, ""+c.isFlippedY());
        
                // store the properties of the GridBagLayout
                GridBagLayout layout = (GridBagLayout)p.getLayout();
                GridBagConstraints gbc = layout.getConstraints(c);
                prop.setProperty(GRIDX, ""+gbc.gridx);
                prop.setProperty(GRIDY, ""+gbc.gridy);
                prop.setProperty(IPADX, ""+gbc.ipadx);
                prop.setProperty(IPADY, ""+gbc.ipady);
                prop.setProperty(GRIDHEIGHT, ""+gbc.gridheight);
                prop.setProperty(GRIDWIDTH, ""+gbc.gridwidth);
                prop.setProperty(WEIGHTX, ""+gbc.weightx);
                prop.setProperty(WEIGHTY, ""+gbc.weighty);
                //prop.setProperty(ANCHOR, ""+gbc.anchor);
                //prop.setProperty(FILL, ""+gbc.fill);
                Insets in = gbc.insets;
                prop.setProperty(TOP, ""+in.top);
                prop.setProperty(BOTTOM, ""+in.bottom);
                prop.setProperty(LEFT, ""+in.left);
                prop.setProperty(RIGHT, ""+in.right);
	    } 
	    catch (ClassCastException e) {  
		Note.err(null, "MatrixSerialization storeProperties fails: "+e.toString()); 
		return; 
	    }
	}

	/** 
	 * Restore representation first and after that other layout variables. 
	 */
	public void restoreProperties(StructurePanel p) {
	    restoreProperties1(p);
	    restoreProperties2(p);
	}

	/** 
	 * Restore representation first and after that other layout variables. 
	 */
	private void restoreProperties1(StructurePanel p) {
	    for (int i=0; i<prop.length; i++) {
		java.awt.Component c = p.getComponent(i);
		java.util.Properties pi = prop[i];
		try { 
		    if (c instanceof VisualContainer) {
			VisualContainer vc = (VisualContainer)c;
			//Note.out(this, "Restoring properties for " + vc);
			// java.awt.Container methods
			if (pi.getProperty(NAME) != null) vc.setName(pi.getProperty(NAME));
			// matrix.VisualContainer (extends java.awt.Container) methods
			if ((pi.getProperty(REPRESENTATION) != null) && (pi.getProperty(REPRESENTATION).length()>0) && (pi.getProperty(REPRESENTATION) != NOTDEFINED)) {
			    //Note.out(this, "representation = <" + pi.getProperty(REPRESENTATION) + ">");
			    vc.changeRepresentation(pi.getProperty(REPRESENTATION));
			}
		    }
		} 
		catch (ClassCastException e) {  
		    Note.err(null, "MatrixSerialization restoreProperties fails: "+e.toString()); 
		    return; 
		}
	    }
	}

	/** 
	 * Restore representation first and after that other layout variables. 
	 */
	private void restoreProperties2(StructurePanel p) {
	    for (int i=0; i<prop.length; i++) {
		java.awt.Component c = p.getComponent(i);
		java.util.Properties pi = prop[i];
		try { 
		    if (c instanceof VisualContainer) {
			VisualContainer vc = (VisualContainer)c;
			//Note.out(this, "Restoring subproperties for " + vc);
			// matrix.VisualType (extends java.awt.Container) methods
			vc.setMinimized(truefalse(pi.getProperty(MINIMIZED)));
			vc.setAlive(truefalse(pi.getProperty(ALIVE)));
			vc.enableSubComponents(truefalse(pi.getProperty(ENABLED)));
			// matrix.VisualContainer (extends VisualType) methods
			vc.setHotSpotEnabled(truefalse(pi.getProperty(HOTSPOTENABLED)));
			//isLabeled returns always true and setLabeled not implemented
			//vc.setLabeled(truefalse(pi.getProperty(LABELED))); 
			vc.setTitled(truefalse(pi.getProperty(TITLED))); 
			vc.setRotated(truefalse(pi.getProperty(ROTATED))); 
			vc.setFlippedX(truefalse(pi.getProperty(FLIPPEDX))); 
			vc.setFlippedY(truefalse(pi.getProperty(FLIPPEDY))); 
            
                        // restore the properties of the GridBagLayout
                        GridBagLayout layout = (GridBagLayout)p.getLayout();
                        GridBagConstraints gbc = layout.getConstraints(c);
                        gbc.gridx = Integer.parseInt(pi.getProperty(GRIDX));
                        gbc.gridy = Integer.parseInt(pi.getProperty(GRIDY));
                        gbc.ipadx = Integer.parseInt(pi.getProperty(IPADX));
                        gbc.ipady = Integer.parseInt(pi.getProperty(IPADY));
                        gbc.gridheight = Integer.parseInt(pi.getProperty(GRIDHEIGHT));
                        gbc.gridwidth = Integer.parseInt(pi.getProperty(GRIDWIDTH));
                        gbc.weightx = Double.parseDouble(pi.getProperty(WEIGHTX));
                        gbc.weighty = Double.parseDouble(pi.getProperty(WEIGHTY));
                        Insets in = new Insets(0, 0, 0, 0);
                        in.top = Integer.parseInt(pi.getProperty(TOP));
                        in.bottom = Integer.parseInt(pi.getProperty(BOTTOM));
                        in.left = Integer.parseInt(pi.getProperty(LEFT));
                        in.right = Integer.parseInt(pi.getProperty(RIGHT));
                        //gbc.anchor = gbc.getClass().getField(pi.getProperty(ANCHOR)).getInt(gbc);
                        //gbc.fill = gbc.getClass().getField(pi.getProperty(FILL)).getInt(gbc);
                        gbc.insets = in;
                        layout.setConstraints(vc, gbc);
		    }
		} 
		catch (ClassCastException e) {  
		    Note.err(null, "MatrixSerialization restoreProperties fails: "+e.toString()); 
		    Note.err(null, "The item was " + c);
		    return; 
                }
	    }
	}

	private boolean truefalse(String tf) {
	    if (tf.equals("true") || tf.equals("TRUE") || tf.equals("True") || tf.equals("1"))
		return true;
	    if (tf.equals("false") || tf.equals("FALSE") || tf.equals("False") || tf.equals("0"))
		return false;
	    Note.warning(null, "MatrixSerialization truefalse fails: " + tf);
	    return false; 
	}

	static final long serialVersionUID =  5228939065982705422L;
    }
    
    /**
     * If there is a saved matrix animation in given file a new Application window 
     * is opened with saved animation.
     *
     * @param f file to load
     * @see matrix.util.Application
     * @see matrix.util.ApplicationAdapter
     */
    /* UNUSED METHOD?
    public static boolean load(File f) {
	MatrixSerialization loaded;

	try { 
	    FileInputStream istream = new FileInputStream(f);
	    ObjectInputStream p = new ObjectInputStream(istream);
      
	    loaded = (MatrixSerialization)p.readObject();
	    istream.close();
	} 
	catch (IOException e) { Note.warning(null,e.toString()); return false; }
	catch (ClassNotFoundException e) {  Note.warning(null,e.toString()); return false; }

	Application app = ApplicationAdapter.getApplication().getNewInstance(loaded.savedFDT, 
							   loaded.animator);
	MatrixLayout layout = loaded.getLayout();
	if (layout != null)
	    layout.unloadProperties(app);
	else
	    Note.out(null, "No properties found");

	if (app instanceof Frame) {
	  ((Frame)app).show();
	} else {
	  app.getNewWindow().show();
	}
	return true;
    }
    */
    
    /**
     * Save animator and fdt vector together into the argument file.
     *
     * @param a animator to save
     * @param v fdt vector to save
     * @param f saving file
     * @param savelayout if true, save also some representation details
     * @deprecated replaced by save(Animator, FDT[], File, boolean)
     */
    public static void save(Animator a, Vector v, File f, boolean savelayout, StructurePanel panel) {
	FDT[] tmp = new FDT[v.size()];
	for(int i = 0; i < tmp.length; i++) {
	    tmp[i] = (FDT) v.elementAt(i);
	}
	save(a, tmp, f, savelayout, panel);
    }
     
    public static void save(Animator a, FDT[] ss, File f, boolean savelayout, StructurePanel panel) {
	//Note.out(null, "save FDT[] = " + ss);
       MatrixSerialization s = new MatrixSerialization(a, ss);
       
       if (savelayout) {
	   s.setLayout(new MatrixLayout(ss, panel));
       }

       try {
	   FileOutputStream ostream = new FileOutputStream(f);
	   ObjectOutputStream p = new ObjectOutputStream(ostream);
	   p.writeObject(s);
	   p.flush();
	   ostream.close();
       } catch (IOException e) { 
	   Note.err(null, "Error: " + e.toString()); 
	   e.printStackTrace();
       }    
   }
    
    
    public static MatrixSerialization loadFromFile(File f) throws IOException {
	MatrixSerialization loaded;

	try { 
	    FileInputStream istream = new FileInputStream(f);
	    ObjectInputStream p = new ObjectInputStream(istream);
	    
	    loaded = (MatrixSerialization)p.readObject();
	    istream.close();
	} 
	/*catch (IOException e) { Note.warning(null,e.toString()); return null; }*/
	catch (ClassNotFoundException e) {  Note.warning(null,e.toString()); return null; }
	
	return loaded;
    }



    public static MatrixSerialization loadFromNet(URL url) {
	MatrixSerialization loaded;
	try {
	    URLConnection c = url.openConnection();
	    InputStream istream = c.getInputStream();
	    ObjectInputStream p = new ObjectInputStream(istream);
	    
	    loaded = (MatrixSerialization)p.readObject();
	    istream.close();
	} 
	catch (IOException e) { 
	    Note.warning(null,e.toString()); 
	    return null; 
	}
	catch (ClassNotFoundException e) {  
	    Note.warning(null,e.toString()); 
	    return null; 
	}
	return loaded;
    }
    



    public Animator getAnimator() {
	return animator;
    }


    public Vector getFDTs() {
      Vector v = new Vector(savedFDT.length);
      for (int i = 0; i < savedFDT.length; i++) {
	v.addElement(savedFDT[i]);
      }
      return v;
    }


    public FDT[] getStructures() {
      return savedFDT;
    }


    public String toString() {
	StringBuffer buf = new StringBuffer();
	buf.append("MatrixSerialization:\n");
	buf.append("Animator=" + animator);
	buf.append("\nFDTs\n");
	java.util.Enumeration e = getFDTs().elements();
	while (e.hasMoreElements()) {
	    FDT fdt = (FDT)e.nextElement();
	    buf.append(fdt.toString());
	    buf.append("\n");
	}
	return buf.toString();
    }
    

   static final long serialVersionUID = 983793253973548179L;
}

