// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/* $Id: TableFromFile.java,v 1.21 2007/09/07 13:34:42 vkaravir Exp $ */

package matrix.util.io;

import java.io.*;
import java.util.*;

import matrix.structures.FDT.FDT;
import matrix.structures.FDT.probe.Key;
import matrix.structures.FDT.probe.Table;

/**
 * Creates a table from data in a file.
 * <p>
 * Input file format: The file contains keys of the table and there must be
 * only one key per line.<br>
 * 
 * The first line of the file must be following:
 * "#matrix table" or "#matrix array"
 * <p>
 * example:<br>
 *
 * #matrix array<br>
 * key1<br>
 * key2<br>
 * key3<br>
 * key4<br>
 * etc.
 *
 * <p>
 * This can also open a file which has been saved by Save To ASCII -method 
 * and which contains a table and any number of inner structures.
 * <p>
 * Moreover, lines can contain additional comments, which are skipped when loading the file.
 * Comments must start with string "//". Space characters are also removed from the end of 
 * each line. Empty characters before the comment must be space characters.
 * <p>
 * example:<br>
 * <br>
 * #matrix array          //this is the header of the file<br>
 * key1                   // first key<br>
 * key2                   //second key<br>
 * ...<br>
 */
public class TableFromFile {

    public static final String TABLE_HEADER = "#matrix table";
    public static final String TABLE_HEADER2 = "#matrix array";

    private String[] array;
    private ArrayList arrayList = new ArrayList();
    boolean isTable = false;
    public Table table;

    // after parsing the file this contains structures in strings
    private HashMap structuresInASCII = new HashMap(); 
    
    /**
     * Constructor
     *
     * @param filename file, which contains keys of the table
     */
    public TableFromFile(String fileName) {
	try {
	    parseFile(fileName);

	    String temp = (String)structuresInASCII.get(fileName);
	    String header = null;
	    StringTokenizer st = new StringTokenizer(temp, "\n", false);
	    if (st.hasMoreTokens())
		header = st.nextToken();

	    if (header.trim().equalsIgnoreCase(TABLE_HEADER) ||
		header.trim().equalsIgnoreCase(TABLE_HEADER2)) {
		isTable = true;
		readAndParseKeysOfTable(st);
	    }
	} catch (Exception e) {
	    e.printStackTrace();
	    isTable = false;
	}
	
	if (isTable)
	    makeTable();

    }

    /**
     * Constructor. This is used when opening the inner structures. The purpose of 
     * this is that the original file need not to be parsed again.  
     *
     * @param sia contains all the structures in strings 
     */
     public TableFromFile(HashMap sia) {
	structuresInASCII = sia;
    }

    StructuresFromFile sff = null;

    /**
     * Constructor. This is used when opening the inner structures. The
     * purpose of this is that the original file need not to be parsed
     * again.
     *
     * @param sia contains all the structures in strings
     * @param sff StructureFromFile which starts the saving process
     */
    public TableFromFile(HashMap sia, StructuresFromFile sff) {
	structuresInASCII = sia;
	this.sff = sff;
    }


    /**
     * Returns true, if the file contains a table.
     *
     * @param fileName file which needs to be checked
     * @return true, if the file contains a table
     */
    public static boolean isTable(String fileName) {
	try {
	    BufferedReader reader = new BufferedReader(new FileReader(fileName));
	    String line = reader.readLine();
	    line = handleLine(line);
	    while(!(line.toLowerCase().startsWith("#matrix"))) {
		line = reader.readLine();
		line = handleLine(line);
	    }
	    //file contains the table if the first header in 
	    //the file is TABLE_HEADER or TABLE_HEADER2
	    if(line.trim().equalsIgnoreCase(TABLE_HEADER) ||
	       line.trim().equalsIgnoreCase(TABLE_HEADER2))
		return true;
	    else
		return false;
	} catch (Exception e) {
	    return false;
	}
    }

	

    /**
     * Parses a file and inserts the string representation of the each structure 
     * into hashmap.
     *
     * @param fileName file, which contains the structures and needs to be parsed
     */
    private void parseFile(String fileName) {
	try {
	    String structureName = fileName; //the name of the top structure is the name of the file
	    String tia = ""; //table in ascii
	    BufferedReader reader = new BufferedReader(new FileReader(fileName));
	    String line = null;
	    while((line = reader.readLine()) != null) {
		line = handleLine(line);
		while (line.equals("")) {
		    line = reader.readLine();
		    line = handleLine(line);
		}
		if(line.equals("#EOS")) { 
		    // the description of last structure has ended and 
		    // the description of next structure is beginning
		    structuresInASCII.put(structureName, tia);
		    line = reader.readLine(); //this contains now name of the new structure
		    if(line != null) {
			line = handleLine(line);
			while (line.equals("")) {
			    line = reader.readLine();
			    line = handleLine(line);
			}
			structureName = line;
			tia = "";
		    }
		    //System.out.println("creating new structureInASCII: " + line);
		}
		else if(line.startsWith("#representation") ||
			line.startsWith("#rotated") ||
			line.startsWith("#minimized")) {
		    //do nothing, skip these lines
		    System.out.println("Skipped: " + line);
		}
		else {
		    if(line.equals(""))
			tia = tia.concat(" \n");
		    else
			tia = tia.concat(line + "\n");
		    //System.out.println("adding line into tia");
		}
	    }
	    structuresInASCII.put(structureName, tia); //adding the last structure
	    
	} catch(Exception e) {
	    System.out.println("Error in parsing the file: " + fileName);
	    e.printStackTrace();
	}
    }


    private static String handleLine(String str) {
	String temp = str;
	temp = removeSpacesFromBeginning(temp);
	temp = removeCommentsFromLine(temp);
	return temp;
    }

    /**
     * Removes space characters from the beginning of the string.
     *
     * @param str string which need to be handled.
     * @return string where the possible space characters are removed
     */
    private static String removeSpacesFromBeginning(String str) {
	String temp = str;
	while(temp.indexOf(" ") == 0) {
	    if (temp.length() == 1)
		return "";
	    temp = temp.substring(1);
	}
	return temp;
    }
    

    /**
     * Removes comments from the string. Comments starts with "//".
     *
     * @param str string which needs to be handled
     * @return string from where possible comments are removed
     */
    private static String removeCommentsFromLine(String str) {
	String temp = str;
	int indexOfComment = temp.indexOf("//");
	if (indexOfComment == -1)
	    ;
	else
	    temp = temp.substring(0, indexOfComment);
	temp = removeSpacesFromEnd(temp);
	return temp;
    }


    /**
     * Removes space characters from the end of a string.
     *
     * @param str string which needs to be handled
     * @return string from where possible space charachters are removed
     */
    private static String removeSpacesFromEnd (String str) {
	String temp = str;
	int indexOfLastSpace = temp.lastIndexOf(" ");
	while (indexOfLastSpace == (temp.length() - 1)) {
	    if (temp.length() == 0)
		return temp;
	    temp = temp.substring(0, indexOfLastSpace);
	    indexOfLastSpace = temp.lastIndexOf(" ");
	}
	if (temp.lastIndexOf("'") == (temp.length() - 1))
	    return temp.concat(" ");
	else
	    return temp;
    }

    
    /**
     * Returns true, if the given file contained the keys of the table.
     *
     * @return true, if the given file contained the keys of the table
     */
    public boolean isTable() {
	return isTable;
    }
    

    /**
     * Reads the keys from the StringTokenizer one token (is same as one key) at a 
     * time and saves them finally to a String array.
     *
     * @param st input 
     */
    private void readAndParseKeysOfTable(StringTokenizer st) {
	String line;
	while(st.hasMoreTokens()) {
	    line = st.nextToken();
	    line = processName(line);
	    arrayList.add(line);
	}
	array = (String[])arrayList.toArray(new String[0]);
    }


    /**
     * Processes the name of the key. This removes all the escape characters "'"
     * from the key and returns back original keys. With help of escape character
     * it is possible that the key can contain space characters and other special
     * characters (eg. "_", "#", "'") which can have a special meaning when 
     * parsing the file.
     *
     * @param nam name of the key in a saved form
     * @return name of the original key where all the markup characters have been removed
     */
    private String processName(String name) {
	String tempName = "";
	String temp = null;
	int index;
	if (name != "") {
	    for(index = 0; index < name.length(); index++) {
		if(index == (name.length() - 1))
		    temp = name.substring(index);
		else
		    temp = name.substring(index, index + 1);
		
		if(temp.equals("'")){
		    if(index == name.length() - 1)
			;
		    else {
			index++;
			tempName = tempName.concat(name.substring(index, index + 1));
		    }
		}
		else if(temp.equals(" "))
		    return tempName;
		else {
		    tempName = tempName.concat(temp);
		}
	    }
	}
	return tempName;		       
    }


    /**
     * Makes the Table from String array.
     *
     */
    private void makeTable() {
	String key = null;
	FDT[] temp = new FDT[array.length];
	for(int i=0; i < array.length; i++) {
	    key = array[i];
	    if (structuresInASCII.get(key) != null) {
		temp[i] = getStructureFromASCII(key);
	    }
	    else {
		temp[i] = new Key(array[i]);
	    }
	}
	table = new Table("");
	for (int i=0; i < temp.length; i++) {
	    table.setObject(temp[i], i);
	}
    }
    
    /**
     * Returns FDT which corresponds the given name of the structure. 
     *
     * @param sname the name of the structure
     * @return returns FDT which corresponds the name of the structure
     */    
    private FDT getStructureFromASCII(String sname) {
	// string representation of the structure is in hashmap
	String temp = (String)structuresInASCII.get(sname);
	StringTokenizer st = new StringTokenizer(temp, "\n", false);
	String visualSettings = ""; //visualization settings
	String line = null;
	line = st.nextToken();
	while(!(line.toLowerCase().startsWith("#matrix"))) {
	    visualSettings = visualSettings.concat(line + "\n");
	    line = st.nextToken();
	}

	//String firstLine = st.nextToken();
	//if (firstLine.equals(TABLE_HEADER)) {
	if(line.trim().equalsIgnoreCase(TABLE_HEADER) ||
	   line.trim().equalsIgnoreCase(TABLE_HEADER2)) {
	    TableFromFile tff = new TableFromFile(structuresInASCII, sff);
	    FDT fdttemp = tff.getTableFromASCII(st);
	    //saves the visualization settings for current fdt
	    sff.setVisualSettingsForFDT(fdttemp, visualSettings);
	    return fdttemp;
	    //Loading of inner structures that are marked with "autopolymorph" tag
	} else if (line.trim().startsWith(StructuresFromFile.AUTO_POLYMORPH_HEADER)) {
	    if (line.trim().length() > StructuresFromFile.AUTO_POLYMORPH_HEADER.length() + 1) {
		line = line.trim().substring(StructuresFromFile.AUTO_POLYMORPH_HEADER.length() + 1);
		FDT fdttemp = sff.getPolyMorphFromASCII(line, st);
		sff.setVisualSettingsForFDT(fdttemp, visualSettings);
		return fdttemp;
	    }
	    return null;
	}
	else {
	    GraphFromFile gff = new GraphFromFile(structuresInASCII, sff);
	    FDT fdttemp = gff.getGraphFromASCII(st);
	    //saves the visualization settings for current fdt
	    sff.setVisualSettingsForFDT(fdttemp, visualSettings);
	    return fdttemp;
	}
    }


    /**
     * Returns the Table.
     *
     * @return Table containing the given keys
     */
    public Table getTable() {
	return table;
    }


    /**
     * Returns the FDT which corresponds the string representation which is 
     * given through the StringTokenizer
     *
     * @param st StringTokenizer which contains the keys of the structure.
     * @return returns the corresponding FDT
     */
    public FDT getTableFromASCII(StringTokenizer st) {
	try {
	    readAndParseKeysOfTable(st);
	    makeTable();
	} catch (Exception e) {
	    e.printStackTrace();
	}
	return getTable();
    }
}
