// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/* $Id: LayeredTreeContours.java,v 1.9 2007/09/10 08:11:13 vkaravir Exp $ */

package matrix.visual;

import java.awt.Dimension;
import java.util.*;

/**
 * Stores left and right contours of subtrees for VisualLayeredTreeComponent.
 *
 * @author Riku Saikkonen
 * @version $Revision: 1.9 $
 * @see VisualLayeredTreeComponent
 */
public class LayeredTreeContours {
    private int cHeight, cLeftExtent, cRightExtent;
    private Vector/*<Dimension>*/ leftCDims;
    private Vector/*<Dimension>*/ rightCDims;

    /** Creates a one-element contour pair
     * @param left left extent (usually < 0)
     * @param right right extent (usually > 0)
     * @param height height of element (> 0)
     */
    public LayeredTreeContours(int left, int right, int height) {
        cHeight = height;

        leftCDims = new Vector(1);
        leftCDims.addElement(new Dimension(-left, height));
        cLeftExtent = left;

        rightCDims = new Vector(1);
        rightCDims.addElement(new Dimension(-right, height));
        cRightExtent = right;
    }

    public int getHeight() {
        //Note.assertCont(cHeight > 0);
        return cHeight;
    }

    public int getLeftExtent() {
        return cLeftExtent;
    }

    public int getRightExtent() {
        return cRightExtent;
    }

    public int getWidth() {
        //Note.assertCont(cRightExtent - cLeftExtent >= 0);
        return cRightExtent - cLeftExtent;
    }

    /**
     * Adds a rectangular object (a new root) on top of this
     * contour pair. O(1).
     *
     * @param left left extent (usually < 0)
     * @param right right extent (usually > 0)
     * @param height height of element (> 0)
     * @param addHeight how much to increase height of the topmost
     *                  elements in the pair
     * @param originTrans translates origin of the contours this much
     *                    to the right
     */
    public void addOnTop(int left, int right, int height,
                         int addHeight, int originTrans) {
        ((Dimension)leftCDims.lastElement()).height += addHeight;
        ((Dimension)leftCDims.lastElement()).width += originTrans + left;
        ((Dimension)rightCDims.lastElement()).height += addHeight;
        ((Dimension)rightCDims.lastElement()).width += originTrans + right;

        leftCDims.addElement(new Dimension(-left, height));
        rightCDims.addElement(new Dimension(-right, height));

        cHeight += height + addHeight;
        cLeftExtent -= originTrans;
        cRightExtent -= originTrans;
        if (left < cLeftExtent)
            cLeftExtent = left;
        if (right > cRightExtent)
            cRightExtent = right;
        //Note.assertCont(cLeftExtent <= 0, "cLeftExtent=" + cLeftExtent);
        //Note.assertCont(cRightExtent >= 0, "cRightExtent=" + cRightExtent);
        //Note.assertCont(cHeight > 0, "cHeight=" + cHeight);
    }

    /**
     * Joins this contour pair with another that lies to the right
     * of this one. This is O(max(dThis,dOther)), where dThis is
     * the depth (number of elements) of this contour pair.
     *
     * @param other the contour to join this with
     * @param hDist horizontal distance that the other contour is
     *              translated by before joining
     */
    public void joinWith(LayeredTreeContours other, int hDist) {
        //Note.out(this,"joinWith: this.w="+getWidth()+", other.w="+other.getWidth()+", hDist="+hDist);
        // Update left contour: add start of other left contour
        // before this one
        if (other.cHeight > cHeight) {
            Vector newLeftC = new Vector();
            int otherLeft = other.cHeight - cHeight;
            Enumeration e = other.leftCDims.elements();

            while (e.hasMoreElements() && otherLeft > 0) {
                Dimension oDim = (Dimension) e.nextElement();
                Dimension dim = new Dimension(oDim);

                otherLeft -= oDim.height;
                if (otherLeft < 0)
                    dim.height += otherLeft;

                newLeftC.addElement(dim);
            }

            Dimension middle = (Dimension) newLeftC.lastElement();
            int thisCDisp = 0, otherCDisp = 0;

            while (e.hasMoreElements())
                otherCDisp += ((Dimension) e.nextElement()).width;

            e = leftCDims.elements();
            while (e.hasMoreElements()) {
                Dimension d = (Dimension) e.nextElement();
                thisCDisp += d.width;
                newLeftC.addElement(d);
            }
                
            // Correct for the difference between cumulative
            // displacements at the join point
            middle.width -= thisCDisp - otherCDisp;
            // and a coordinate transformation
            middle.width -= hDist;
                
            leftCDims = newLeftC;
        }

        // Update right contour: replace the end of this contour
        // with the other right contour
        if (other.cHeight >= cHeight) {
            // Copy other contour directly
            rightCDims = (Vector) other.rightCDims.clone();
        } else {
            Enumeration e = rightCDims.elements();
            int thisLeft = cHeight - other.cHeight;
            int nextIndex = 0;

            while (e.hasMoreElements() && thisLeft > 0) {
                Dimension dim = (Dimension) e.nextElement();

                nextIndex++;
                thisLeft -= dim.height;
                if (thisLeft < 0)
                    dim.height += thisLeft;
            }

            int thisCDisp = 0, otherCDisp = 0;
            while (e.hasMoreElements())
                thisCDisp += ((Dimension) e.nextElement()).width;

            // Discard the rest of rightCDims
            rightCDims.setSize(nextIndex);

            Dimension middle = (Dimension) rightCDims.lastElement();

            // Add other.rightCDims here
            e = other.rightCDims.elements();
            while (e.hasMoreElements()) {
                Dimension d = (Dimension) e.nextElement();
                otherCDisp += d.width;
                rightCDims.addElement(d);
            }

            // Correct for the difference between cumulative
            // displacements at the join point
            middle.width += thisCDisp - otherCDisp;

            // coordinate transformation for topmost original width
            middle.width += hDist;
        }
        // Translate topmost width by -hDist
        ((Dimension)rightCDims.lastElement()).width -= hDist;

        // Update extents
        if (other.cHeight > cHeight)
            cHeight = other.cHeight;
        if (other.cLeftExtent + hDist < cLeftExtent)
            cLeftExtent = other.cLeftExtent + hDist;
        if (other.cRightExtent + hDist > cRightExtent)
            cRightExtent = other.cRightExtent + hDist;
        //Note.out(this, "joinWith done, width = " + getWidth());
        //Note.assertCont(cLeftExtent <= 0, "cLeftExtent=" + cLeftExtent);
        //Note.assertCont(cRightExtent >= 0, "cRightExtent=" + cRightExtent);
        //Note.assertCont(cHeight > 0, "cHeight=" + cHeight);
    }

    /**
     * Calculates the translation between this pair's right
     * contour and the left contour of the pair given as argument.
     * This is O(min(dThis,dOther)), where dThis is the depth
     * (number of elements) of this contour pair.
     *
     * @param other the other contour pair
     * @param wantedDist the minimum horizontal distance wanted
     *                   between these contours
     * @return the calculated horizontal translation for the other
     *         contour
     */
    public int calcTranslation(LayeredTreeContours other,
                               int wantedDist) {
        Vector lc = rightCDims, rc = other.leftCDims;
        int li = lc.size() - 1, ri = rc.size() - 1;
        Dimension lCumD = new Dimension(0, 0);
        Dimension rCumD = new Dimension(0, 0);
        int displacement = wantedDist;

        //Note.out(this, "calcTranslation: " + li + "," + ri);

        while (true) {
            if (li < 0) {
                if (ri < 0 || rCumD.height >= lCumD.height)
                    break;

                Dimension rd = (Dimension) rc.elementAt(ri);
                rCumD.height += rd.height;
                rCumD.width += rd.width;
                ri--;
            } else if (ri < 0) {
                if (lCumD.height >= rCumD.height)
                    break;

                Dimension ld = (Dimension) lc.elementAt(li);
                lCumD.height += ld.height;
                lCumD.width += ld.width;
                li--;
            } else {
                Dimension ld = (Dimension) lc.elementAt(li);
                Dimension rd = (Dimension) rc.elementAt(ri);
                int leftNewHeight = lCumD.height;
                int rightNewHeight = rCumD.height;
                    
                if (leftNewHeight <= rightNewHeight) {
                    lCumD.height += ld.height;
                    lCumD.width += ld.width;
                    li--;
                }
                    
                if (rightNewHeight <= leftNewHeight) {
                    rCumD.height += rd.height;
                    rCumD.width += rd.width;
                    ri--;                    
                }
            }

            // Invariant: displacement >= r.w - l.w + wantedDist
            //Note.out(this, li+","+ri+" = (" + lCumD.width+","+rCumD.width + ")/(" + lCumD.height+","+rCumD.height +"): " + displacement+" ?>=? "+(rCumD.width - lCumD.width + wantedDist));
            if (displacement < rCumD.width - lCumD.width + wantedDist)
                displacement = rCumD.width - lCumD.width + wantedDist;
        }
            
        return displacement;
    }
}
