// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/* $Id: VisualArray.java,v 1.68 2008/01/18 16:07:34 oseppala Exp $ */

package matrix.visual;

import java.awt.*;

import matrix.animation.Animator;
import matrix.simulation.VisualTypeConf;
import matrix.structures.CDT.CDT;
import matrix.structures.FDT.*;
import matrix.structures.FDT.advanced.StyledArray;
import matrix.structures.FDT.probe.Key;
import matrix.structures.FDT.probe.Table;
import matrix.structures.spatial.FDT.probe.SpatialElementTable;
import matrix.util.Note;

public class VisualArray extends VisualContainer {
	//protected Array array; // array which to represent
	protected boolean indexed = true; // component show its index
    protected boolean rangeSelectable = false; // ranges can be selected from the array
    
    private boolean rotateComponents = false; // Rotate VisualArrayComponents
	private boolean resizeHotSpotEnabled = true;
    
	public void setRotateComponents(boolean b) {
		rotateComponents=b;
	}
	
	public boolean getRotateComponents() {
		return rotateComponents;
	}
	
	public VisualArray(Array a) {
		super(a);
		//this.array = a;
		this.indexed = !isNestedContainer(); //indexed; 
	}
	
	public String getRepresentationName() {
		return "Array";
	}
	
	public void addNullHotSpot() { /* do not add! */ }
	
	public void addComponents() {
		super.addComponents();
        
        if (isNestedContainer()) {
                setIndexed(false);
        }
		add(visualResizeHotSpot = new VisualResizeHotSpot(getStructure()));
		visualResizeHotSpot.setVisible(isTitled() && resizeHotSpotEnabled);
	}
    
    /** 
     * Sets the resize hot spot enabled or disabled depending on the
     * given value. If resize hot spot is not added, does nothing.
     */
    public void setResizeHotSpotEnabled(boolean value) {
        resizeHotSpotEnabled = value;
        if (visualResizeHotSpot != null)
            visualResizeHotSpot.setVisible(value);
    }
	
	/**
	 * Renames all the keys in a table.
	 *
	 * Asks user to give new keys and then inserts them into beginning of 
	 * the table and sets the size of the table so that only the given
	 * keys are visible.
	 *
	 */
	public void renameAllKeys() {
		if (getStructure() instanceof Table) {
			String keys = getApplication().getInput("Give new keys: ", "");
			if (keys.length() > 0) {
				int i = -1;

                getAnimator().startOperation();
				while(keys.length() > 0 ){
					int index = keys.indexOf(" ");
					if (index > 0) {
						String key = keys.substring(0, index);
						i++;
						getArray().setObject(new Key(key), i);
						keys = keys.substring(index + 1);
					} 
					else if  (index == -1) {
						i++;
						getArray().setObject(new Key(keys), i);
						keys = "";
					}
					else {
						keys = keys.substring(1);
					}
				}
				getArray().setFirst(0);
				if (i == -1) i = 0;
				getArray().setLast(i);
				getAnimator().endOperation();
                
				setInvalid();
			}
		}
		else {
			Note.show(this, "Not a real array. Cannot rename keys.");
		}
	}
	
	
	
	public void RaiseIndex() {
		Animator.getActiveAnimator().startOperation();
		getArray().setFirst(getArray().getFirst()+1);
		getArray().setLast(getArray().getLast()+1);
		Animator.getActiveAnimator().endOperation();
		setInvalid();
	}
	
	public void LowerIndex() {
		Animator.getActiveAnimator().startOperation();
		getArray().setFirst(getArray().getFirst()-1);
		getArray().setLast(getArray().getLast()-1);
		Animator.getActiveAnimator().endOperation();
		setInvalid();
	}
	
	public Array getArray() {
		if (getStructure() instanceof Array)
			return (Array)getStructure();
		else if (getStructure() != null) {
			if (getStructure().getElement() instanceof Array)
				return (Array)getStructure().getElement();
			else
				Note.err(this, "Cannot determine array for " + getStructure());
		} else Note.err(this, "No element defined: " + getStructure());
		return null;
	}
	
	// operational methods
	public void indexed() {
		indexed = !indexed;
		setInvalid();
	}
	
	public boolean isIndexed() {
		return indexed;
	}
	
	public void setIndexed(boolean val) {
		indexed = val;
	}
	
    public boolean isRangeSelectable() {
        return rangeSelectable;
    }
    
    public void setRangeSelectable(boolean val) {
        rangeSelectable = val;
    }
    
    
	public void increment() {
		Animator.getActiveAnimator().startOperation();
		getArray().setObject(null, getArray().getLast()+1);
		Animator.getActiveAnimator().endOperation();
		setInvalid();
		
		if(getArray() instanceof Table) {
			if(((Table)getArray()).isRandomTable())
				debugRefresh();
		}
	}
	
	public void decrement() {
		Animator.getActiveAnimator().startOperation();
		getArray().setLast(getArray().getLast() - 1);
		Animator.getActiveAnimator().endOperation();
		setInvalid();
	}
	
	// double is a reserved word
	public void Double() {
		int newLast = 2*(getArray().getLast() - getArray().getFirst() + 1);
		Animator.getActiveAnimator().startOperation();
		getArray().setObject(null, getArray().getFirst() + newLast - 1);
		Animator.getActiveAnimator().endOperation();
		setInvalid();
		
		if(getArray() instanceof Table) {
			if(((Table)getArray()).isRandomTable())
				debugRefresh();
		}
		
	}
	
	// hack to make SpatialElementTable work. If possible, should be replaced with a better
	// way to do it. - 16.6.06 jtn
	public void setType(int i) {
		if(getArray() instanceof SpatialElementTable) {
			((SpatialElementTable)getArray()).setType(i);
			setInvalid();
		}
	}
	
	
	public void halve() {
		
		int newLast = (getArray().getLast() - getArray().getFirst() + 1)/2;
		Animator.getActiveAnimator().startOperation();
		getArray().setLast(getArray().getFirst() + newLast - 1);
		Animator.getActiveAnimator().endOperation();
		setInvalid();
	}
	
	public void debugRefresh() {
		if(getArray() instanceof Table) 
			if(((Table) getArray()).isRandomTable()) {
				((Table) getArray()).randomize();
			}
		super.debugRefresh();
	}
	
	public void setTitled(boolean val) {
		super.setTitled(val);
		if (visualResizeHotSpot != null)
			visualResizeHotSpot.setVisible(val);
	}
	
	/** Return the insets for arrays.
	 * We use smaller insets for titled arrays (no extra space around the
	 * components). */
	public Insets getInsets() {
		if (isTitled() && getFont() != null) {
			int h = getMetrics().getHeight();
			int w = 1;
			return new Insets(h, w, h, w);
		} else
			return super.getInsets();
	}
	
    /*
	public void doLayout() {
		super.doLayout();
		visualResizeHotSpot.setBounds(getBounds().width-visualResizeHotSpot.getBounds().width, 
					      1, 
					      visualResizeHotSpot.getBounds().width, 
					      visualResizeHotSpot.getBounds().height);
	}
    */	
	protected LayoutSize doActualLayout() {
		Component[] cmps = getItems();
		LayoutSize sz = new LayoutSize(0, 0);
		sz.width = sz.height = 0;
		
		// compute array width (sz.width) and max height (sz.height)
		for (int i=0; i<cmps.length; i++) {
			VisualComponent vc = (VisualComponent) cmps[i];
			LayoutSize cs = vc.lGetPreferredSize();
			sz.width += cs.width;
			sz.height = Math.max(sz.height, cs.height);
		}
		
		// Set bounds for components. This is done separately from the
		// previous, because the height of all components should be
		// the same.
		int x = 0;
		for (int i=0; i<cmps.length; i++) {
			VisualComponent vc = (VisualComponent) cmps[i];
			LayoutSize cs = vc.lGetPreferredSize();
			vc.lSetBounds(x, 0, cs.width, sz.height);
			if ((i==cmps.length-1)&&(!isRotated())&&(!isFlippedX())) {
				visualResizeHotSpot.setBounds(x, 1, cs.width, getInsets().top-1);
			}
			if ((i==0)&&(!isRotated())&&(isFlippedX())) {
				visualResizeHotSpot.setBounds(x, 1, cs.width, getInsets().top-1);
			}
			x += cs.width;
		}
		if (isRotated()) {
			int vrw=(int)visualResizeHotSpot.getPreferredSize().getWidth();
			visualResizeHotSpot.setBounds(1+x-vrw, 1, vrw, getInsets().top-1);
		}
		return sz;
	}
	
	protected boolean shouldBeRotated() {
		return getParent() instanceof VisualArrayComponent;
	}
	
	VisualArrayComponent empty = null;
	public void validateComponents() {
		super.validateComponents();
		if (isNestedContainer()) {
			if (shouldBeRotated()) {
				VisualArrayComponent vac = (VisualArrayComponent)getParent();
				VisualArray va = vac.getComponentParent();
				setRotated(!va.isRotated());
			} 
		}
		if (isDuplicate()) {
			Component[] cmps; 
			int j = 0;
			for (int i=getArray().getFirst(); i<=getArray().getLast(); i++) {
				cmps = getItems(); 
				
				FDT s = (j < getItemCount()) ? ((VisualType)cmps[j]).getStructure() : null;
				
				Object o = new Key("");
				
				if ((s != null) && (!s.equals(o))) {
					int tmp = 0;
					int k;
					for (k=j+1; k<getItemCount(); k++) {
						if (((VisualType)cmps[k]).getStructure().equals(o)) {
							tmp = k;
							break;
						}
					}
					if (tmp != 0) { 
						Component c = cmps[tmp];
						remove(tmp);
						add(c, j);
					} else {
						add(newComponent(i, o), j);
                        //Removal of the old element to avoid moving of empty keys
                        remove(j+1);
					}
				} else if (s == null) {
					add(newComponent(i, o), j);
				}
				VisualArrayComponent vac = 
					(VisualArrayComponent)getItem(j);
				vac.setIndex(i);    
				j++;
			}
			while (j < getItemCount())
				remove(j);
			return;
		}
		if ((getArray() != null)) {
			Component[] cmps; 
			int j = 0;
			for (int i=getArray().getFirst(); i<=getArray().getLast(); i++) {
				cmps = getItems(); //getComponents();
				Object o = getArray().getObject(i);
				
				FDT s = (j < getItemCount()) ? ((VisualType)cmps[j]).getStructure() : null;
				
				if (o == null) {
					if ((s == null) || (!(s instanceof Key)) ||	(!((Key)s).equals(Key.Empty))){
						add(newComponent(i, o), j);
                    }
				} else if ((s != null) && (!s.equals(o))) {
					int tmp = 0;
					int k;
					for (k=j+1; k<getItemCount(); k++) {
						if (((VisualType)cmps[k]).getStructure().equals(o)) {
							tmp = k;
							break;
						}
					}
					if (tmp != 0) { 
						Component c = cmps[tmp];
						remove(tmp);
						add(c, j);
					} else {
                        add(newComponent(i, o), j);
                        //Removal of the old element to avoid moving of empty keys
                        remove(j+1);
					}
				} else if (s == null) {
					add(newComponent(i, o), j);
				}
				VisualArrayComponent vac = 
					(VisualArrayComponent)getItem(j);
				vac.setIndex(i);    
				//if (getArray() instanceof StyledArray)
				//  vac.setLabeled(true);
				j++;
			}
			while (j < getItemCount()) // remove leftovers
				remove(j);
		}
	}
	
	public String getIndexName(int index) {
		if (getStructure() instanceof StyledArray)
			return ((StyledArray)getStructure()).getIndexName(index);
		else
			return String.valueOf(index);
	}
	
	/** @returns the label that has the maximum width.  A heuristics
	 is used with digits. If the array is a StyledArray, the
	 precise max width is calculated.
	 */
	public String getMaxIndexName() {
		if (getStructure() instanceof StyledArray) { // this is slow process for StyledArrays
			StyledArray sa = (StyledArray)getStructure();
			String label = "";
			int labelWidth = 0;
			FontMetrics fm = getMetrics();
			for (int i=0; i<sa.getLast(); i++) {
				String tmp = sa.getIndexName(i);
				int tmpWidth = fm.stringWidth(tmp);
				if (labelWidth < tmpWidth) {
					label = tmp;
					labelWidth = tmpWidth;
				}
			}
			return label;
		}
		// we just determine how many digits there is in the max index and multiply that by the maxWidth of a digit
		else {
			if(getStructure() instanceof Array) {
				int first = ((Array)getStructure()).getFirst();
				int last = ((Array)getStructure()).getLast();
				int maxIndex = Math.max(Math.abs(first),Math.abs(last));
				if (first < 0) maxIndex *= 10;
				String s = String.valueOf(myPow(10,discreteLog(maxIndex+1))-1); // 0,9,99,999,...
				return s;
			}
			return "";
		}
	}
	
	public static final double log10e = 0.4342944819;
	/* @returns the discrete logarithm for x. */
	private int discreteLog(int x) {
		double d = x;
		return (int)(Math.ceil(log10e * Math.log(d))); 
	}
	
	/* @returns a^b */
	private int myPow(int a, int b) {
		if (b == 0) return 1;
		if (b == 1) return a;
		if ((b % 2) == 1) return a*myPow(a, b-1);
		int tmp = myPow(a, b/2);
		return tmp*tmp;
	}
	
	public void paintComponent(Graphics g) {
		super.paintComponent(g);
		paintFrame(g);
	}
	
	// simulation methods
	public void write(int i) {
		getArray().setObject(read.getStructure(), i);
	}
	
	public boolean canInsert() {
		return getStructure() instanceof CDT;
	}
	
	public void insert(Object structure) {
		if (structure instanceof Key)
			if (((Key)structure).toString() == "")
				return;
		if (getStructure() instanceof CDT) {
			if (!(structure == null))
				((CDT)getStructure()).insert(structure);
		}
	}
	
	public void delete(Object structure) {
		erase(structure);
	}
	
	public void erase(Object component) {
		if(getStructure() instanceof CDT) 
			((CDT)getStructure()).delete(component);
	}
	
	protected void removeKeyHelp(Object o) {
		if(getParent() instanceof VisualType)
			((VisualType)getParent()).removeKeyHelp(o);
		else if(getStructure() instanceof CDT) {
			((CDT)getStructure()).delete(o);
			//this.setStructure(((CDT)getStructure()).delete(o));
		}
		else if(getStructure() instanceof Array) {
			Array a = (Array)getStructure();
			int i = a.getFirst();
			while(i < a.getLast()) {
				if(o.equals(a.getObject(i))) {
					a.setObject(null, i);
					break;
				}
				i++;
			}
		}
		else
			remove();	
	}	    
	
	public VisualArrayComponent newComponent(int place, Object o) {
		return new VisualArrayComponent(place, o, rotateComponents);
	}
    
    /**
     * Configures the properties related to VisualArray and then calls the 
     * configure method of the super class.
     */
    public void configure(VisualTypeConf vtc) {
        if (vtc != null) {
            String value = vtc.getVisualConfValue(this.getClass().getName(), VisualTypeConf.INDEXED);
            //In the following way, no need to check whether value is null or not.
            if ("true".equals(value)) { 
                setIndexed(true);
            } else if ("false".equals(value)) {
                setIndexed(false);
            }

            value = vtc.getVisualConfValue(this.getClass().getName(), VisualTypeConf.RANGE_SELECTION);
            //In the following way, no need to check whether value is null or not.
            if ("true".equals(value)) { 
                setRangeSelectable(true);
            } else if ("false".equals(value)) {
                setRangeSelectable(false);
            }

        }
        super.configure(vtc);
    }
}
