//Matrix - algorithm simulation system
//Copyright (C) 2000- 2007 Matrix group
//(See the AUTHORS file and the author tag below for more information
//on the authors.)

//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.

//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/* $Id $ */

package matrix.visual;

import java.awt.*;
import java.util.*;

import matrix.structures.FDT.*;
import matrix.structures.FDT.substructures.Vertex;
import matrix.util.Note;

/**
 * Circular layout
 * Modified from Ville Karavirta's Kamada-Kawai implementation
 * 
 * @author Ville Karavirta
 */
public class VisualCircleLayoutGraph extends VisualGraph {

    private static final long serialVersionUID = -5908459017930669396L;
    /** Edge-length. */
    protected double len;
    /** Nodes of the component currently being handled. */
    protected VisualGraphComponent[] nodeList;
    /** X-coordinates of the component currently being handled. */
    protected double[] xCoord;
    /** Y-coordinates of the component currently being handled. */
    protected double[] yCoord;
    /** Number of nodes in the component currently being handled. */
    protected int nodes;
    /** Is the layout changed since last calculation. */
    private boolean isChanged = true;
    protected boolean lenAdjust = false;
    
    public static final double EDGE_ADD_LENGTH = 60;

    public VisualCircleLayoutGraph(Graph g) { 
        super(g); 
    }

    public VisualCircleLayoutGraph(Vertex v) { 
        super(v); 
    }


    // These are here just for setting the isChanged value
    protected VisualGraphComponent createNewComponent(Vertex v) {
        isChanged = true;
        return super.createNewComponent(v);
    }

    protected VisualReference createNewReference(VisualComponent v1, 
            VisualComponent v2) {
        isChanged = true;
        return super.createNewReference(v1, v2);
    }

    public void flipX() {
        isChanged = true;
        super.flipX();
    }

    public void flipY() {
        isChanged = true;
        super.flipY();
    }

    public void rotated() {
        isChanged = true;
        super.rotated();
    }

    protected LayoutSize doActualLayout() {
        int wid = 0, hei = 0;
        //Note.out(this, "calculating layout\n");
        Component[] com = getItems();
        VisualGraphComponent[][] coms= getConnectedComponents(com);
        //Note.out(this, "components: " + coms.length);
        for (int i = 0; i < coms.length; i++) {
            nodeList = coms[i];
            nodes = nodeList.length;
            xCoord = new double[nodes];
            yCoord = new double[nodes];
            if (lenAdjust == false) {
                len = calculateEdgeLength();

                // Finding the length of the longest string
                // (weight). Adding this length to edge length to
                // visualize longer strings (weights) so that they
                // don't overlap with the nodes or arrow heads.
                Component[] cmp = getLinks();
                int biggestLength  = 0;
                int tempLength = 0;
                if (isWeighted) {
                    for(int k = 0; k < cmp.length; k++) {
                        tempLength = getMetrics().stringWidth(((VisualReference)cmp[k]).getName()); //name of the reference is the weight
                        if (tempLength > biggestLength)
                            biggestLength = tempLength;
                    }
                }
                len += biggestLength;

            }
            double minimumEdgeLength = calculateMaximumWidth();
            
            circleLayout(minimumEdgeLength + EDGE_ADD_LENGTH);
            LayoutSize lays = assignCoordinates(wid);
            wid += lays.width;
            hei = Math.max(hei, lays.height);
            if (i != coms.length - 1)
                wid += getGapWidth();
        }
        isChanged = false;
        return new LayoutSize(wid, hei);
    }

    protected double calculateEdgeLength() {
        double len;
        double maxR = -1;
        for (int j = 0; j < nodes; j++) {
            LayoutSize nodeSize = nodeList[j].lGetPreferredSize();
            maxR = Math.max(maxR, nodeSize.width/2);
            maxR = Math.max(maxR, nodeSize.height/2);
        }       
        //lucky (or unlucky) guess for edge length
        if (isWeighted) //little bit longer edge length for weighted graphs
            len = maxR*Math.sqrt(nodes)*1.7;//*1.15; // by changing this coefficient it is possible to adjust the length of the edges
        else
            len = maxR*Math.sqrt(nodes)*1.15;
        len = Math.min(len, 200);
        return len;
    }


    /**
     * Lays the nodes on a circle, or if the layout is not changed
     * after previous painting, retrieves the current coordinates.
     * In this method values in xCoord and yCoord are set correctly.
     */
    protected void circleLayout(double edgeLength) {
        
        int w,h;
        w = h = (int)((edgeLength) / Math.tan(Math.PI / nodes));

        for (int i = 0; i < nodes; i++) {
            if (isChanged) {
                // If you want to make this faster uncomment the previous line
                // and comment the next line. Now the layout is always calculated
                // from scratch, while it could be calculated completely again 
                // only if it has changed. Note, however, that making it like 
                // that can cause the layout to look slighty different after
                // each time the layout is calculated.
                //if (true) {
                xCoord[i] = (h/2)*Math.cos(2*Math.PI*((double)i/(double)nodes) 
                        + (w/2));
                yCoord[i] = (h/2)*Math.sin(2*Math.PI*((double)i/(double)nodes)
                        + (h/2));
            } else {
                xCoord[i] = nodeList[i].lGetBounds().x;
                yCoord[i] = nodeList[i].lGetBounds().y;
            }
        }
    }

    /**
     * Assigns coordinates for the nodes. 
     * NOTE: This method should be called only immediately after calling
     * kamadaKawai(..).
     * @param x x-coordinate of the upper-left corner of the area
     * @return <code>LayoutSize</code> of the nodes.
     */    
    protected LayoutSize assignCoordinates(int x) {
        double minX = Double.MAX_VALUE, minY = Double.MAX_VALUE;
        for (int i = 0; i < nodes; i++) {
            minX = Math.min(minX, xCoord[i]);
            minY = Math.min(minY, yCoord[i]);
        }
        int maxW = 0, maxH = 0;
        for (int i = 0; i < nodes; i++) {
            LayoutSize nodeSize = nodeList[i].lGetPreferredSize();
            int w = nodeSize.width;
            maxW = Math.max(maxW, w);
            int h = nodeSize.height;
            maxH = Math.max(maxH, h);
            int xC = (int)(xCoord[i] - minX);
            int yC = (int)(yCoord[i] - minY);
            nodeList[i].lSetBounds(xC + x, yC, w, h);
            maxW = Math.max(maxW, xC + w);
            maxH = Math.max(maxH, yC + h);
        }
        return new LayoutSize(maxW, maxH);
    }


    private Vector checked = null;

    /** 
     * Returns the connected components of this graph. 
     * @param com Components of the graph.
     * @return Connected components in <code>Vector</code> which
     * contains <code>Vectors</code> of the components.
     */
    private VisualGraphComponent[][] getConnectedComponents(Component[] com) {
        Vector ready = new Vector();
        checked = new Vector();
        int nod = com.length;
        VisualGraphComponent wNode;
        for (int i = 0; i < nod; i++) {
            wNode = (VisualGraphComponent)com[i];
            if (!checked.contains(wNode))
                ready.add(findConnectedNodes(wNode, new Vector()));
        }
        VisualGraphComponent[][] coms = new VisualGraphComponent[ready.size()][];
        Iterator it = ready.iterator();
        int i = 0;
        //Note.out(this, "components: " + coms.size());
        while (it.hasNext()) {
            Vector nodesL = (Vector)it.next();
            nodes = nodesL.size();
            coms[i] = new VisualGraphComponent[nodes];
            System.arraycopy(nodesL.toArray(), 0, coms[i], 0, nodes);
            i++;
        }
        checked = null;
        return coms;
    }

    /**
     * Finds the components connected to the given node.
     * @param iNode The node.
     * @param comps Vector containing the other nodes the given node is
     * already known to be connected.
     * @return Vector containing all the nodes connected to the given node.
     */
    private Vector findConnectedNodes(VisualGraphComponent iNode, Vector comps) {
        comps.add(iNode);
        checked.add(iNode);
        Vector presuc = new Vector();
        VisualGraphComponent[] suc = iNode.getSuccessorComponents();
        VisualGraphComponent[] pre = iNode.getPredecessorComponents();
        for (int i = 0; i < suc.length; i++)
            presuc.add(suc[i]);
        for (int i = 0; i < pre.length; i++)
            if (!presuc.contains(pre[i]))
                presuc.add(pre[i]);
        Iterator edgeIter = presuc.iterator();
        VisualGraphComponent nextNode;
        while (edgeIter.hasNext()) {
            nextNode = (VisualGraphComponent)edgeIter.next();
            if (!checked.contains(nextNode))
                comps = findConnectedNodes(nextNode, comps);
        }
        return comps;
    }


    protected double calculateMaximumWidth() {
        double maxFound = 0.0;        
        double sizeA, sizeB;
               
        LayoutSize nodeSize = nodeList[0].lGetPreferredSize();

        sizeA = Math.max(nodeSize.width/2, nodeSize.height/2);

        //first and last
        if (nodes > 1){
            maxFound = sizeA + Math.max(nodeList[nodes-1].lGetPreferredSize().width/2, nodeList[nodes-1].lGetPreferredSize().height/2);
        }
            
        // adjacent
        for (int i = 0; i < nodes - 1; i++) {
            nodeSize = nodeList[i+1].lGetPreferredSize();
            sizeB = Math.max(nodeSize.width/2, nodeSize.height/2);

            maxFound = Math.max(sizeA + sizeB, maxFound);
                
        }
        return maxFound;
    }


}
