// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/* $Id: VisualFRGraph.java,v 1.8 2007/09/10 08:11:06 vkaravir Exp $ */

package matrix.visual;

import java.util.Vector;

import matrix.structures.FDT.*;
import matrix.structures.FDT.substructures.Vertex;

/**
 * A version of the Fruchterman-Reingold graph drawing algorithm from
 * paper "Graph Drawing by Force-directed Placement" in
 * Software - Practice and Experience, Vol. 21(11) (1991) by Thomas 
 * Fruchterman and Edward Reingold. The algorithm has been modified 
 * for non-uniform vertices as described in paper "Drawing Graphs 
 * with Non-Uniform Vertices" in Proceedings of Working Conference
 * on Advanced Visual Interfaces (AVI'02) by Harel and Koren.
 *
 * @author Ville Karavirta
 */
public class VisualFRGraph extends VisualKKGraph {

    public VisualFRGraph(Graph g) { super(g); }
    public VisualFRGraph(Vertex v) { super(v); }

    protected void actualAlgorithm() { frucRein(); }

    /**
     * The algorithm. 
     */
    private void frucRein() {
	double[] xDisp = new double[nodes];
	double[] yDisp = new double[nodes];
	int[] rad = new int[nodes];
	double temp = len*3;
	int steps = 0;
	Vector indexes = new Vector();
        for (int i = 0; i < nodes; i++) {
	    indexes.add(i, nodeList[i]);
	    rad[i] = Math.max(nodeList[i].lGetPreferredSize().width/2,
			      nodeList[i].lGetPreferredSize().height/2);
	}
	double maxDelta = Double.MAX_VALUE;
	while (temp > 0.05 && maxDelta > 0.5 && steps < maxSteps) {
	    for (int i = 0; i < nodes; i++) {
		xDisp[i] = 0;
		yDisp[i] = 0;
	    }
	    // calculate repulsive forces
	    double dX, dY, del, rep, delta;
	    for (int i = 0; i < nodes; i++) {
		for (int j = i + 1; j < nodes; j++) {
		    dX = xCoord[i] - xCoord[j];
		    dY = yCoord[i] - yCoord[j];
		    del = dX*dX + dY*dY;
		    rep = repulsiveForce(rad[i], rad[j], del);
		    delta = Math.sqrt(del);
		    xDisp[i] += (dX/delta)*rep;
		    yDisp[i] += (dY/delta)*rep;
		    // minus, since dx(i,j) = -dx(j,i)
		    xDisp[j] -= (dX/delta)*rep;
		    yDisp[j] -= (dY/delta)*rep;
		}
	    }
	    //calculate attractive forces
	    for (int i = 0; i < nodes; i++) {
		//		VisualGraphComponent v = nodeList[i];
		VisualGraphComponent[] suc = nodeList[i].getSuccessorComponents();
		double att;
		for (int k = 0; k < suc.length; k++) {
		    if (suc[k].equals(nodeList[i]))
			continue;
		    int j = indexes.indexOf(suc[k]);
		    dX = xCoord[i] - xCoord[j];
		    dY = yCoord[i] - yCoord[j];
		    delta = Math.sqrt(dX*dX + dY*dY);
		    att = attractiveForce(rad[i], rad[j], dX, dY);
		    xDisp[i] -= (dX/delta)*att;
		    yDisp[i] -= (dY/delta)*att;
		    // plus, since dx(i,j) = -dx(j,i)
		    xDisp[j] += (dX/delta)*att;
		    yDisp[j] += (dY/delta)*att;
		}
	    }
	    maxDelta = 0;
	    for (int i = 0; i < nodes; i++) {
		delta = Math.sqrt((xDisp[i] * xDisp[i]) +
					(yDisp[i] * yDisp[i]));
		maxDelta = Math.max(delta, maxDelta);
		if (delta > temp) {
		    xCoord[i] += xDisp[i] / (delta / temp);
		    yCoord[i] += yDisp[i] / (delta / temp);
		} else {
		    xCoord[i] += xDisp[i];
		    yCoord[i] += yDisp[i];
		}
	    }
	    steps++;
	    temp = cool(temp);
	}
    }
    
    private double cool(double val) {
	return val/1.02;
    }

    private double repulsiveForce(int ri, int rj, double delta) {
	double tmp = ri + rj + len;
	return (tmp*tmp)/(delta);
    }

    private double attractiveForce(int ri, int rj, double dX, double dY) {
	double tmp = ri + rj + len;
	return (dX*dX)/(tmp*tmp) + (dY*dY)/(tmp*tmp);
    }
}
