// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/* $Id: VisualKKGraph.java,v 1.23 2007/09/10 08:11:07 vkaravir Exp $ */

package matrix.visual;

import java.awt.*;
import java.util.*;

import matrix.structures.FDT.*;
import matrix.structures.FDT.substructures.Vertex;
import matrix.util.Note;

/**
 * A version of the Kamada-Kawai graph drawing algorithm from
 * paper "An Algorithm for Drawing General Undirected Graphs" in
 * Information Processing Letters 31 (1989) by Kamada and Kawai.
 * The algorithm has been modified for non-uniform vertices as
 * described in paper "Drawing Graphs with Non-Uniform Vertices"
 * in Proceedings of Working Conference on Advanced Visual 
 * Interfaces (AVI'02) by Harel and Koren.
 *
 * @author Ville Karavirta
 */
public class VisualKKGraph extends VisualGraph {

    private static final long serialVersionUID = -5908459017930669396L;
    /** Edge-length. */
    protected double len;
    /** Nodes of the component currently being handled. */
    protected VisualGraphComponent[] nodeList;
    /** X-coordinates of the component currently being handled. */
    protected double[] xCoord;
    /** Y-coordinates of the component currently being handled. */
    protected double[] yCoord;
    /** Number of nodes in the component currently being handled. */
    protected int nodes;
    /** Is the layout changed since last calculation. */
    private boolean isChanged = true;
    /** Maximum number of steps in the main loop of the algorithm. */
    protected int maxSteps = 5000;
    /** How small energies are tried to improve. Larger the number
	is the faster the algorithm and worse the result. */
    public static final double epsilon = 0.001;
    /** Has the user changed the edge length. */
    protected boolean lenAdjust = false;
    /** Graph-theoretic distances in the graph. */
    private Matrix2D dValues;
  
    public VisualKKGraph(Graph g) { 
	super(g); 
    }
    
    public VisualKKGraph(Vertex v) { 
	super(v); 
    }


    // These are here just for setting the isChanged value
    protected VisualGraphComponent createNewComponent(Vertex v) {
	isChanged = true;
	return super.createNewComponent(v);
    }
    
    protected VisualReference createNewReference(VisualComponent v1, 
						 VisualComponent v2) {
	isChanged = true;
	return super.createNewReference(v1, v2);
    }

    public void flipX() {
	isChanged = true;
	super.flipX();
    }

    public void flipY() {
	isChanged = true;
	super.flipY();
    }

    public void rotated() {
	isChanged = true;
	super.rotated();
    }

    protected LayoutSize doActualLayout() {
	int wid = 0, hei = 0;
	//Note.out(this, "calculating layout\n");
	Component[] com = getItems();
	VisualGraphComponent[][] coms= getConnectedComponents(com);
	//Note.out(this, "components: " + coms.length);
	for (int i = 0; i < coms.length; i++) {
	    nodeList = coms[i];
	    nodes = nodeList.length;
	    xCoord = new double[nodes];
	    yCoord = new double[nodes];
	    if (lenAdjust == false) {
                len = calculateEdgeLength();

		// Finding the length of the longest string
		// (weight). Adding this length to edge length to
		// visualize longer strings (weights) so that they
		// don't overlap with the nodes or arrow heads.
		Component[] cmp = getLinks();
		int biggestLength  = 0;
		int tempLength = 0;
		if (isWeighted) {
		    for(int k = 0; k < cmp.length; k++) {
			tempLength = getMetrics().stringWidth(((VisualReference)cmp[k]).getName()); //name of the reference is the weight
			if (tempLength > biggestLength)
			    biggestLength = tempLength;
		    }
		}
		len += biggestLength;

	    }
	    circleLayout();
	    actualAlgorithm();
	    removeOverlappingNodes();
	    LayoutSize lays = assignCoordinates(wid);
	    wid += lays.width;
	    hei = Math.max(hei, lays.height);
	    if (i != coms.length - 1)
		wid += getGapWidth();
	}
	isChanged = false;
	return new LayoutSize(wid, hei);
    }

    protected double calculateEdgeLength() {
        double len;
        double maxR = -1;
        for (int j = 0; j < nodes; j++) {
            LayoutSize nodeSize = nodeList[j].lGetPreferredSize();
            maxR = Math.max(maxR, nodeSize.width/2);
            maxR = Math.max(maxR, nodeSize.height/2);
        }       
        //lucky (or unlucky) guess for edge length
        if (isWeighted) //little bit longer edge length for weighted graphs
            len = maxR*Math.sqrt(nodes)*1.7;//*1.15; // by changing this coefficient it is possible to adjust the length of the edges
        else
            len = maxR*Math.sqrt(nodes)*1.15;
        len = Math.min(len, 200);
        return len;
    }

    /**
     * The algorithm. 
     */
    protected void actualAlgorithm() {
	kamadaKawai();
    }

    private void kamadaKawai() {
	Matrix2D wValues = calculateW();
	//Note.out(this, "using edge-length " + len);
	dValues = calculateShortestPaths(wValues);
	// not needed currently
	//double energy = calculateEnergy();
	double maxDm = 0;
	int maxDmIndex = -1;
	double dm;
	for (int i = 0; i < nodes; i++) {
	    dm = calculateDeltaM(i);
	    if (dm > maxDm) {
		maxDm = dm;
		maxDmIndex = i;
	    }
	}
	int steps = 0, subSteps = 0;
	boolean stop = false;
	double lastDm;
	double[] deltas;
	while (maxDm > epsilon && !stop) {
	    lastDm = maxDm + 2;
	    dm = maxDm;
	    //Note.out(this, "maxDm:"+maxDm+" maxDmIndex:"+maxDmIndex);
	    while (dm > epsilon && (lastDm - dm > epsilon) && !stop) {
		deltas = calculateDeltas(maxDmIndex);
		xCoord[maxDmIndex] += deltas[0];
		yCoord[maxDmIndex] += deltas[1];
		lastDm = dm;
		dm = calculateDeltaM(maxDmIndex);
		subSteps++;
		if (subSteps > maxSteps) {
		    Note.out(this, "couldn't find the best layout");
		    stop = true;
		}
	    }
	    steps++;
	    maxDmIndex = 0;
	    maxDm = calculateDeltaM(0);
	    for (int i = 1; i < nodes; i++) {
		dm = calculateDeltaM(i);
		if (dm > maxDm) {
		    maxDm = dm;
		    maxDmIndex = i;
		}
	    }
	}
	//Note.out(this, "Steps:"+steps+" SubSteps:"+subSteps + " nodes:"+nodes);
	//Note.out(this, "maxDm:"+maxDm+" index:"+maxDmIndex);
    }
    
    /**
     * Lays the nodes on a circle, or if the layout is not changed
     * after previous painting, retrieves the current coordinates.
     * In this method values in xCoord and yCoord are set correctly.
     */
    private void circleLayout() {
	int w = nodes*10;
	int h = nodes*10;
	for (int i = 0; i < nodes; i++) {
	    if (isChanged) {
	    // If you want to make this faster uncomment the previous line
	    // and comment the next line. Now the layout is always calculated
	    // from scratch, while it could be calculated completely again 
	    // only if it has changed. Note, however, that making it like 
	    // that can cause the layout to look slighty different after
	    // each time the layout is calculated.
		//if (true) {
	 	xCoord[i] = (h/2)*Math.cos(2*Math.PI*((double)i/(double)nodes) 
					   + (w/2));
		yCoord[i] = (h/2)*Math.sin(2*Math.PI*((double)i/(double)nodes)
					   + (h/2));
	    } else {
		xCoord[i] = nodeList[i].lGetBounds().x;
		yCoord[i] = nodeList[i].lGetBounds().y;
	    }
	}
    }

    /**
     * Assigns coordinates for the nodes. 
     * NOTE: This method should be called only immediately after calling
     * kamadaKawai(..).
     * @param x x-coordinate of the upper-left corner of the area
     * @return <code>LayoutSize</code> of the nodes.
     */    
    private LayoutSize assignCoordinates(int x) {
	double minX = Double.MAX_VALUE, minY = Double.MAX_VALUE;
	for (int i = 0; i < nodes; i++) {
	    minX = Math.min(minX, xCoord[i]);
	    minY = Math.min(minY, yCoord[i]);
	}
	int maxW = 0, maxH = 0;
	for (int i = 0; i < nodes; i++) {
	    LayoutSize nodeSize = nodeList[i].lGetPreferredSize();
	    int w = nodeSize.width;
	    maxW = Math.max(maxW, w);
	    int h = nodeSize.height;
	    maxH = Math.max(maxH, h);
	    int xC = (int)(xCoord[i] - minX);
	    int yC = (int)(yCoord[i] - minY);
	    nodeList[i].lSetBounds(xC + x, yC, w, h);
	    maxW = Math.max(maxW, xC + w);
	    maxH = Math.max(maxH, yC + h);
	}
	return new LayoutSize(maxW, maxH);
    }

    public void changeEdgeLength() {
	String leng = getApplication().getInput("Edge length", ""+len);
	if(leng != null) {
	    double tmp;
	    try {
		tmp = Double.parseDouble(leng);
		setInvalid();
		repaint();
	    } catch (NumberFormatException ex) {
		Note.show("Sorry", "NumberFormatException");
		return;
	    }
	    setEdgeLength(tmp);
	}
    }

    public void setEdgeLength(double newLen) {
	isChanged = true;
	lenAdjust = true;
	len = newLen;
	setInvalid();
	repaint();
    }

    public double getEdgeLength() {
	return len;
    }

    private Vector checked = null;

    /** 
     * Returns the connected components of this graph. 
     * @param com Components of the graph.
     * @return Connected components in <code>Vector</code> which
     * contains <code>Vectors</code> of the components.
     */
    private VisualGraphComponent[][] getConnectedComponents(Component[] com) {
        Vector ready = new Vector();
	checked = new Vector();
	int nod = com.length;
	VisualGraphComponent wNode;
	for (int i = 0; i < nod; i++) {
	    wNode = (VisualGraphComponent)com[i];
	    if (!checked.contains(wNode))
		ready.add(findConnectedNodes(wNode, new Vector()));
	}
	VisualGraphComponent[][] coms = new VisualGraphComponent[ready.size()][];
	Iterator it = ready.iterator();
	int i = 0;
	//Note.out(this, "components: " + coms.size());
	while (it.hasNext()) {
	    Vector nodesL = (Vector)it.next();
	    nodes = nodesL.size();
	    coms[i] = new VisualGraphComponent[nodes];
	    System.arraycopy(nodesL.toArray(), 0, coms[i], 0, nodes);
	    i++;
	}
	checked = null;
	return coms;
    }

    /**
     * Finds the components connected to the given node.
     * @param iNode The node.
     * @param comps Vector containing the other nodes the given node is
     * already known to be connected.
     * @return Vector containing all the nodes connected to the given node.
     */
    private Vector findConnectedNodes(VisualGraphComponent iNode, Vector comps) {
	comps.add(iNode);
	checked.add(iNode);
	Vector presuc = new Vector();
	VisualGraphComponent[] suc = iNode.getSuccessorComponents();
	VisualGraphComponent[] pre = iNode.getPredecessorComponents();
	for (int i = 0; i < suc.length; i++)
	    presuc.add(suc[i]);
	for (int i = 0; i < pre.length; i++)
	    if (!presuc.contains(pre[i]))
		presuc.add(pre[i]);
	Iterator edgeIter = presuc.iterator();
	VisualGraphComponent nextNode;
	while (edgeIter.hasNext()) {
	    nextNode = (VisualGraphComponent)edgeIter.next();
	    if (!checked.contains(nextNode))
		comps = findConnectedNodes(nextNode, comps);
	}
	return comps;
    }

    private void removeOverlappingNodes() {
	int total = 0;
	int nodeOverlaps = -1;
	int steps = 0;
	int moves = 0;
	while (nodeOverlaps != 0 && steps < maxSteps) {
	    nodeOverlaps = 0;
	    for (int i = 0; i < nodes - 1; i++) {
		VisualGraphComponent ic = (VisualGraphComponent)nodeList[i];
		LayoutSize nodeSize = ic.lGetPreferredSize();
		double iw = nodeSize.width/2;
		double ih = nodeSize.height/2;
		for (int j = i + 1; j < nodes; j++) {
		    VisualGraphComponent jc = (VisualGraphComponent)nodeList[j];
		    nodeSize = jc.lGetPreferredSize();
		    double jw = nodeSize.width/2;
		    double jh = nodeSize.height/2;
		    double dx = (xCoord[i] + iw) - (xCoord[j] + jw);
		    double dy = (yCoord[i] + ih) - (yCoord[j] + jh);
		    double dist = Math.sqrt(dx*dx+dy*dy);
		    //not really considering overlapping, just eliminating the
		    //possibility for nodes to overlap
		    double r = Math.max(iw, ih) + Math.max(jw, jh);
		    if (dist < r)
			nodeOverlaps++;
		    while (dist < r) {
			if (xCoord[i] + iw/2 >= xCoord[j] + jw/2) {
			    xCoord[i]++;
			    xCoord[j]--;
			} else {
			    xCoord[i]--;
			    xCoord[j]++;
			}
			if (yCoord[i] + ih/2 >= yCoord[j] + jh/2) {
			    yCoord[i]++;
			    yCoord[j]--;
			} else {
			    yCoord[i]--;
			    yCoord[j]++;
			}
			dx = (xCoord[i] + iw) - (xCoord[j] + jw);
			dy = (yCoord[i] + ih) - (yCoord[j] + jh);
			dist = Math.sqrt(dx*dx+dy*dy);
			r = Math.max(iw, ih) + Math.max(jw, jh);
			moves++;
		    }
		}
	    }
	    total += nodeOverlaps;
	    steps++;
	}
	//Note.out(this, "Nodeoverlaps:"+total+" moves:"+moves);
    }

    protected Matrix2D calculateW() {
	Matrix2D w = new Matrix2D(nodes, nodes);
	w.assign(0);
	double[] f = new double[nodes];
	LayoutSize nodeSize = nodeList[0].lGetPreferredSize();
	f[0] = Math.max(nodeSize.width/2, nodeSize.height/2);
	for (int i = 0; i < nodes - 1; i++) {
	    for (int j = i + 1; j < nodes; j++) {
		if (i == 0) {
		    nodeSize = nodeList[j].lGetPreferredSize();
		    f[j] = Math.max(nodeSize.width/2, nodeSize.height/2);
		}
		w.setValueAt(i, j, len + f[i] + f[j]);
		w.setValueAt(j, i, len + f[i] + f[j]);
	    }
	}
	return w;
    }

    /**
     * Calculates the energy of all the nodes. 
     * @param d <code>Matrix2D</code> object of the graph-theoretic distances
     * between the nodes.
     * @return The energy
     */
    private double calculateEnergy() {
	double energy = 0;
	double dx2, dy2, dVal;
	for (int i = 0; i < nodes - 1; i++) {
	    for (int j = i + 1; j < nodes; j++) {
		dx2 = (xCoord[i] - xCoord[j]) * (xCoord[i] - xCoord[j]);
		dy2 = (yCoord[i] - yCoord[j]) * (yCoord[i] - yCoord[j]);
		dVal = dValues.getValueAt(i,j);
		energy += (1 / (dVal * dVal)) * (dx2 + dy2 + dVal * dVal - 
					      2 * dVal * Math.sqrt(dx2 + dy2));
	    }
	}
	return energy;
    }

    /**
     * Calculates delta M for the given node.
     * @param m Number of the node to calculate the delta M for.
     * @param d <code>Matrix2D</code> object of the graph-theoretic distances
     * between the nodes.
     * @return Delta M for the node.
     */
    private double calculateDeltaM(int m) {
	double xPart = 0, yPart = 0;
	double dx, dy, dd, dVal, kVal;
	for (int i = 0; i < nodes; i++) {
	    if (i != m) {
		dx = xCoord[m] - xCoord[i];
		dy = yCoord[m] - yCoord[i];
		dd = Math.sqrt(dx * dx + dy * dy);
		dVal = dValues.getValueAt(m, i);
		kVal = 1 / (dVal * dVal);
		xPart += kVal * (dx - dVal * dx / dd);
		yPart += kVal * (dy - dVal * dy / dd);
	    }
	}
	return Math.sqrt(xPart * xPart + yPart * yPart);
    }
    
    /**
     * Calculates the change in x and y in one step of the algorithm for
     * the given node.
     * @param i Number of the node to calculate changes for.
     * @param d <code>Matrix2D</code> object of the graph-theoretic distances
     * between the nodes.
     * @return Array of the changes as doubles, index 0 is x and 1 is y.
     */
    private double[] calculateDeltas(int i) {
	double[] deltas = new double[2];
	double dx, dy, dd, dVal, kVal, ddCubed;
	double xPart = 0, yPart = 0;
	double xxPart = 0, xyPart = 0, yyPart = 0;
	for (int j = 0; j < nodes; j++) {
	    if (i != j) {
		dx = xCoord[i] - xCoord[j];
		dy = yCoord[i] - yCoord[j];
		dd = Math.sqrt(dx * dx + dy * dy);
		dVal = dValues.getValueAt(i, j);
		kVal = 1 / (dVal * dVal);
		ddCubed = dd * dd * dd;
		
		xPart += kVal * (dx - dVal * dx / dd);
		yPart += kVal * (dy - dVal * dy / dd);
		xxPart += kVal * (1 - dVal * dy * dy / ddCubed);
		xyPart += kVal * (dVal * dx * dy/ ddCubed);
		// xy = yx
		//yxPart += kVal * (dVal * dy * dx/ ddCubed);
		yyPart += kVal * (1 - dVal * dx * dx / ddCubed);
	    }
	}
	deltas[0] = ((-xPart) * yyPart + xyPart * yPart) /
	    (xxPart * yyPart - xyPart * xyPart);
	deltas[1] = (xxPart * (-yPart) + xPart * xyPart) /
	    (xxPart * yyPart - xyPart * xyPart);
	return deltas;
    }

    private Matrix2D calculateShortestPaths(Matrix2D w) {
	Matrix2D dist = new Matrix2D(nodes, nodes);
	dist.assign(0);
	for (int j = 0; j < nodes; j++) {
	    VisualGraphComponent origin = (VisualGraphComponent)nodeList[j];
	    Vector visited = new Vector();
	    Vector settled = new Vector();
	    Hashtable indexes = new Hashtable();
	    for (int i = 0; i < nodes; i++)
		indexes.put(nodeList[i], new Integer(i));
	    visited.add(origin);
	    BinaryHeap q = new BinaryHeap();
	    q.add(new HeapItem(origin, 0));
	    VisualGraphComponent v;
	    while (!q.isEmpty()) {
		v = ((HeapItem)q.remove()).node;
		if (!settled.contains(v)) {
		    settled.add(v);
		    Vector presuc = new Vector();
		    VisualGraphComponent[] suc = v.getSuccessorComponents();
		    for (int i = 0; i < suc.length; i++)
			presuc.add(suc[i]);
		    // handle graphs always as undirected
		    VisualGraphComponent[] pre = v.getPredecessorComponents();
		    for (int i = 0; i < pre.length; i++)
			if (!presuc.contains(pre[i]))
			    presuc.add(pre[i]);
		    Iterator it = presuc.iterator();
		    while (it.hasNext()) {
			VisualGraphComponent wn = 
			    (VisualGraphComponent)it.next();
			int index = ((Integer)indexes.get(wn)).intValue();
			int prevIndex = ((Integer)indexes.get(v)).intValue();
			double d = dist.getValueAt(j, prevIndex)
			    + w.getValueAt(prevIndex, index);
			if (!visited.contains(wn) || dist.getValueAt(j, index) > d) {
			    visited.add(wn);
			    dist.setValueAt(j, index, d);
			    q.add(new HeapItem(wn, d));
			}
		    }
		}
	    }
	}
	return dist;
    }

    private class HeapItem implements Comparable {
	VisualGraphComponent node;
	double distance;
	HeapItem(VisualGraphComponent node, double distance) {
	    this.node = node;
	    this.distance = distance;
	}
	public int compareTo(Object other) {
	    return (int)(((HeapItem)other).distance - distance);
	}
    }
    
    private class BinaryHeap {

	private Comparable[] heap;
	private int size;
	
	public BinaryHeap() {
	    heap = new Comparable[1];
	    size = 0;
	}
	
	public boolean isEmpty() {
	    return size == 0;
	}

	public void add(Comparable item) {
	    if (size == heap.length) {
		Comparable[] newHeap = new Comparable[2 * heap.length];
		System.arraycopy(heap, 0, newHeap, 0, heap.length);
		heap = newHeap;
	    }
	    int parent, child = size++;
	    while (child > 0 && heap[parent = (child - 1) / 2].compareTo(item) < 0) {
		heap[child] = heap[parent];
		child = parent;
	    }
	    heap[child] = item;
	}
	
	public Comparable remove() {
	    if (size == 0)
		throw new NoSuchElementException();
	    Comparable result = heap[0];
	    Comparable item = heap[--size];
	    int child, parent = 0; 
	    while ((child = (2 * parent) + 1) < size) {
		if (child + 1 < size && heap[child].compareTo(heap[child + 1]) < 0)
		    child++;
		if (item.compareTo(heap[child]) < 0) {
		    heap[parent] = heap[child];
		    parent = child;
		} else 
		    break;
	    }
	    heap[parent] = item;
	    return result;
	}
	
    }	

    /** Represents a two dimensional array. */
    protected class Matrix2D {
	double[][] matr;

	/**
	 * Constructs a new instance with the given number of rows 
	 * and columns.
	 * @param rows Number of rows.
	 * @param cols Number of columns.
	 */
	public Matrix2D(int rows, int cols) {
	    matr = new double[rows][cols];
	}
	/**
	 * Constructs a new instance with the given values.
	 * @param values The values.
	 */
	public Matrix2D(double[][] values) {
	    matr = values;
	}

	/**
	 * Sets the value at the given position.
	 * @param row Number of the row.
	 * @param col Number of the column.
	 * @param value New value.
	 */
	public void setValueAt(int row, int col, double value) {
	    matr[row][col] = value;
	}

	public double getValueAt(int row, int col) {
	    return matr[row][col];
	}

	public void assign(double value) {
	    for (int i = 0; i < matr.length; i++) 
		for (int j = 0; j < matr[i].length; j++) 
		    matr[i][j] = value;
	}
	
	public String toString() {
	    String str = "[";
	    for (int i = 0; i < matr.length; i++) {
		str += "[ ";
		for (int j = 0; j < matr[i].length; j++) 
		    str += matr[i][j] + " ";
		str += "] ";
	    }
	    return str;
	}
    }


    /**
     * Returns a boolean which is true if the graph layout has crossing
     * edges.
     */
    public boolean hasCrossingEdges() {
	if (isChanged)
	    doActualLayout();
	Component[] tmp = getLinks();
	VisualReference[] links = new VisualReference[tmp.length];
	for (int i = 0; i < tmp.length; i++) 
	    links[i] = (VisualReference)tmp[i];
	for (int i = 0; i < links.length; i++) {
	    Point ifrom = links[i].fromPoint;
	    Point ito = links[i].toPoint;
	    for (int j = i + 1; j < links.length; j++) {
		Point jfrom = links[j].fromPoint;
		Point jto = links[j].toPoint;
		if (crossing(ifrom, ito, jfrom, jto))
		    return true;
	    }
	}
	return false;
    }
    

    private boolean crossing(Point from1, Point to1, Point from2, Point to2) {
	double x=0, y=-1, a=0, b=0, c=0, d=0;
	if ((to1.x-from1.x) == 0)
	    return false;
	else if ((to1.y-from1.y) == 0)
	    return false;
	else
	    a = (to1.y-from1.y)/(1.0*to1.x-from1.x);
	b = -from1.x*a + from1.y;
	if (to1.equals(from2) && to2.equals(from1))
	    return false;
	if ((to2.x-from2.x) == 0)
	    return false;
	else if ((to2.y-from2.y) == 0)
	    return false;
	else
	    c = (to2.y-from2.y)/(1.0*to2.x-from2.x);
	d = -from2.x*c + from2.y;
	if (a == c && b == d)
	    return false;
	y = (-(a*d)/c + b)/(1-a/c);
	x = (y - b)/a;
	if ((from1.x < x && to1.x > x) || (from1.x > x && to1.x < x))
	    if ((from1.y < y && to1.y > y) || (from1.y > y && to1.y < y))
		if ((from2.x < x && to2.x > x) || (from2.x > x && to2.x < x))
		    if ((from2.y < y && to2.y > y) || (from2.y > y && to2.y < y))
			return true;
	return false;
    }
}
