// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/* $Id: VisualLayeredTreeComponent.java,v 1.12 2007/09/10 08:11:07 vkaravir Exp $ */

package matrix.visual;

/**
 * Intelligent component for VisualLayeredTree. Functions as a subtree
 * that can lay itself out.
 *
 * @author Riku Saikkonen
 * @version $Revision: 1.12 $
 * @see VisualLayeredTree
 * @see LayeredTreeContours
 */
public class VisualLayeredTreeComponent extends VisualTreeComponent {
    private LayeredTreeContours contours;
    private LayoutSize cachedTranslation, translation;
    private int depthOfRoot;

    public VisualLayeredTreeComponent(matrix.structures.FDT.Tree t,
                                      VisualTreeComponent father) {
        super(t, father);
    }

    /** Update and reset layout information for this subtree.
     * Does a post-order update for its children. Should be called
     * for the root before doFinalLayout when something is changed.
     * @param depth depth of this node in the tree (0=root)
     */
    public void updateLayoutInformation(int depth) {
        // Update children first
        VisualTreeComponent[] ch = getChildren();
        for (int i = 0; i < ch.length; i++) {
            VisualLayeredTreeComponent child =
                (VisualLayeredTreeComponent) ch[i];
            child.updateLayoutInformation(depth + 1);
        }

        // Now update us
        cachedTranslation = null;
        translation = new LayoutSize(0, 0);
        depthOfRoot = depth;
        calcContours();
    }

    /** Does the final layout for the tree.
     * Called for the root of the tree.
     * @param ulx x coordinate of upper left corner of drawing
     * @param uly y coordinate of upper left corner of drawing
     */
    public void doFinalLayout(int ulx, int uly) {
        //Note.out(this, "doFinalLayout("+ulx+","+uly+") *********");

        // Move whole tree to the right if a child extends that
        // far (root translation handles the leftmost children's
        // left extents, but not the others)
        if (-contours.getLeftExtent() - getXTranslation() > 0) {
            //Note.out(this, "additional translation " + (-contours.getLeftExtent() - getXTranslation()));
            translate(-contours.getLeftExtent() - getXTranslation(), 0);
        }
            
        translate(ulx, uly);
        propagateTranslations(true);
    }

    /** Moves this node and its children.
     * Caches the translation; propagateTranslations actually does it
     */
    private void translate(LayoutSize howMuch) {
        if (cachedTranslation == null)
            cachedTranslation = new LayoutSize(0, 0);
        
        cachedTranslation.width += howMuch.width;
        cachedTranslation.height += howMuch.height;
    }
    
    /** Moves this node and its children.
     * Caches the translation; propagateTranslations actually does it
     */
    private void translate(int x, int y) {
        translate(new LayoutSize(x, y));
    }
    
    /** Moves this node only (not its children). */
    private void translateThisNode(int x, int y) {
        translation.width += x;
        translation.height += y;
    }
    
    /** Gets the current X translation of this subtree */
    private int getXTranslation() {
        return translation.width +
            ((cachedTranslation == null) ? 0 : cachedTranslation.width);
    }
    
    /** Propagates translations cached by translate(LayoutSize) */
    private void propagateTranslations(boolean setB) {
        if (cachedTranslation != null) {
            // Move children
            VisualTreeComponent[] ch = getChildren();
            for (int i = 0; i < ch.length; i++) {
                VisualLayeredTreeComponent child =
                    (VisualLayeredTreeComponent) ch[i];
                child.translate(cachedTranslation);
                child.propagateTranslations(setB);
            }
                
            // Move self
            translation.width += cachedTranslation.width;
            translation.height += cachedTranslation.height;
            cachedTranslation = null;
        }
        if (setB) {
            LayoutSize vtcSize = lGetPreferredSize();

            lSetBounds(translation.width - vtcSize.width / 2,
                       translation.height,
                       vtcSize.width,
                       vtcSize.height);
        }
    }

    /** Translates nodes and calculate contours
     * Should be called after the children's contours have been calculated.
     */
    private void calcContours() {
        LayoutSize vtcSize = lGetPreferredSize();
        VisualTreeComponent[] ch = getChildren();
        int rootLeft = -vtcSize.width / 2;
        int rootRight = vtcSize.width / 2 +
            (vtcSize.width % 2 == 0 ? 0 : 1);
        int rootHeight = vtcSize.height;

        if (ch.length == 0) {
            contours = new LayeredTreeContours(rootLeft, rootRight,
                                               rootHeight);
            translateThisNode(-rootLeft, 0);
        } else {
            int transSum = 0;

            // Get initial contour from first child
            VisualLayeredTreeComponent firstChild =
                (VisualLayeredTreeComponent) ch[0];
            contours = firstChild.getContours();
            firstChild.forgetContours();
            firstChild.translate(0, getVisualContainer().getGapHeight() +
                                 rootHeight);

            // Join the other contours and translate the children
            for (int i = 1; i < ch.length; i++) {
                VisualLayeredTreeComponent child =
                    (VisualLayeredTreeComponent) ch[i];
                LayeredTreeContours childC = child.getContours();
                int trans = contours.
                    calcTranslation(childC,
                                    getVisualContainer().getGapWidth());

                transSum += trans;

                child.forgetContours();
                contours.joinWith(childC, trans);

                // Translate "trans" to the right of first child,
                // ignoring this child's previous root translation
                child.translate(firstChild.getXTranslation() +
                                trans -
                                child.getXTranslation(),
                                getVisualContainer().getGapHeight() +
                                rootHeight);
            }

            // Place the root in the horizontal center of the children
            int rootTrans = transSum / ch.length;
            contours.addOnTop(rootLeft, rootRight, rootHeight,
                              getVisualContainer().getGapHeight(),
                              rootTrans);
            translateThisNode(firstChild.getXTranslation() + rootTrans,
                              0);
            //Note.out(this, "root translated " + firstChild.getXTranslation() +"+"+ rootTrans);
        }

        //Note.out(this, "contours(" + depthOfRoot + "): w=" + contours.getWidth() + ", h=" + contours.getHeight());
        //Note.out(this, rootLeft + " " + rootRight + " " + rootHeight);
    }

    public LayeredTreeContours getContours() {
        return contours;
    }

    private void forgetContours() {
        contours = null;
    }
}
