/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: VisualMatrix.java,v 1.2 2007/11/25 20:46:56 archie Exp $ */

package matrix.visual;

import java.awt.*;

import matrix.animation.Animator;
import matrix.simulation.VisualTypeConf;
import matrix.structures.CDT.CDT;
import matrix.structures.FDT.*;
import matrix.structures.FDT.advanced.StyledArray;
import matrix.structures.FDT.probe.Key;
import matrix.structures.FDT.probe.MatrixImpl;
import matrix.structures.FDT.probe.Table;
import matrix.util.Note;
import matrix.util.export.ConfigurableGraphics;

public class VisualMatrix extends VisualArray {
	
    protected boolean indexed = true; // component show its index
	private boolean rotateComponents = false; // Rotate VisualArrayComponents
	private boolean resizeHotSpotEnabled = true;
    
	public void setRotateComponents(boolean b) {
		rotateComponents=b;
	}

	public boolean getRotateComponents() {
		return rotateComponents;
	}
	
	public VisualMatrix(Matrix m) {
		super(m);
		//this.array = a;
		this.indexed = !isNestedContainer(); //indexed; 
	}
	
	public String getRepresentationName() {
		return "Matrix";
	}
	
	public void addNullHotSpot() { /* do not add! */ }
	
	public void addComponents() {
		super.addComponents();
        if (isNestedContainer()) {
            setIndexed(false);
        }
		add(visualResizeHotSpot = new VisualResizeHotSpot(getStructure()));
		visualResizeHotSpot.setVisible(isTitled() && resizeHotSpotEnabled);
	}
    
    /** 
     * Sets the resize hot spot enabled or disabled depending on the
     * given value. If resize hot spot is not added, does nothing.
     */
    public void setResizeHotSpotEnabled(boolean value) {
        resizeHotSpotEnabled = value;
        if (visualResizeHotSpot != null)
            visualResizeHotSpot.setVisible(value);
    }
	
	/**
	 * Renames all the keys in a table.
	 *
	 * Asks user to give new keys and then inserts them into beginning of 
	 * the table and sets the size of the table so that only the given
	 * keys are visible.
	 *
	 */
	public void renameAllKeys() {
		if (getStructure() instanceof Table) {
			String keys = getApplication().getInput("Give new keys: ", "");
			if (keys.length() > 0) {
				int i = -1;
				//Component[] array = getItems();
				getAnimator().startOperation();
				while(keys.length() > 0 ){
					int index = keys.indexOf(" ");
					if (index > 0) {
						String key = keys.substring(0, index);
						i++;
						getArray().setObject(new Key(key), i);
						keys = keys.substring(index + 1);
					} 
					else if  (index == -1) {
						i++;
						getArray().setObject(new Key(keys), i);
						keys = "";
					}
					else {
						keys = keys.substring(1);
					}
				}
				getArray().setFirst(0);
				if (i == -1) i = 0;
				getArray().setLast(i);
				getAnimator().endOperation();
				setInvalid();
			}
		}
		else {
			Note.show(this, "Not a real array. Cannot rename keys.");
		}
	}
	
	
	
	public void RaiseIndex() {
		Animator.getActiveAnimator().startOperation();
		getArray().setFirst(getArray().getFirst()+1);
		getArray().setLast(getArray().getLast()+1);
		Animator.getActiveAnimator().endOperation();
		setInvalid();
	}
	
	public void LowerIndex() {
		Animator.getActiveAnimator().startOperation();
		getArray().setFirst(getArray().getFirst()-1);
		getArray().setLast(getArray().getLast()-1);
		Animator.getActiveAnimator().endOperation();
		setInvalid();
	}
	
	public Array getArray() {
		if (getStructure() instanceof Array)
			return (Array)getStructure();
		else if (getStructure() != null) {
			if (getStructure().getElement() instanceof Array)
				return (Array)getStructure().getElement();
			else
				Note.err(this, "Cannot determine array for " + getStructure());
		} else Note.err(this, "No element defined: " + getStructure());
		return null;
	}
	
	// operational methods
	public void indexed() {
		indexed = !indexed;
		setInvalid();
	}
	
	public boolean isIndexed() {
		return indexed;
	}
	
	public void setIndexed(boolean val) {
		indexed = val;
	}
	
	public void increment() {
		Animator.getActiveAnimator().startOperation();
		getArray().setObject(null, getArray().getLast()+1);
		Animator.getActiveAnimator().endOperation();
		setInvalid();
		
		if(getArray() instanceof Table) {
			if(((Table)getArray()).isRandomTable())
				debugRefresh();
		}
	}
	
	public void decrement() {
		Animator.getActiveAnimator().startOperation();
		getArray().setLast(getArray().getLast() - 1);
		Animator.getActiveAnimator().endOperation();
		setInvalid();
	}
	
	// double is a reserved word
	public void Double() {
		int newLast = 2*(getArray().getLast() - getArray().getFirst() + 1);
		Animator.getActiveAnimator().startOperation();
		getArray().setObject(null, getArray().getFirst() + newLast - 1);
		Animator.getActiveAnimator().endOperation();
		setInvalid();
		
		if(getArray() instanceof Table) {
			if(((Table)getArray()).isRandomTable())
				debugRefresh();
		}
		
	}
	
	public void halve() {
		
		int newLast = (getArray().getLast() - getArray().getFirst() + 1)/2;
		Animator.getActiveAnimator().startOperation();
		getArray().setLast(getArray().getFirst() + newLast - 1);
		Animator.getActiveAnimator().endOperation();
		setInvalid();
	}
	
	public void debugRefresh() {
		if(getArray() instanceof Table) 
			if(((Table) getArray()).isRandomTable()) {
				((Table) getArray()).randomize();
			}
		super.debugRefresh();
	}
	
	public void setTitled(boolean val) {
		super.setTitled(val);
		if (visualResizeHotSpot != null)
			visualResizeHotSpot.setVisible(val);
	}
	
	/** Return the insets for arrays.
	 * We use smaller insets for titled arrays (no extra space around the
	 * components). */
	public Insets getInsets() {
		if (isTitled() && getFont() != null) {
			int h = getMetrics().getHeight();
			int w = 1;
			return new Insets(h, w, h, w);
		} else
			return super.getInsets();
	}
	
    /*
	public void doLayout() {
		super.doLayout();
		visualResizeHotSpot.setBounds(getBounds().width-visualResizeHotSpot.getBounds().width, 
					      1, 
					      visualResizeHotSpot.getBounds().width, 
					      visualResizeHotSpot.getBounds().height);
	}
    */	
	protected LayoutSize doActualLayout() {
		Component[] cmps = getItems();
		LayoutSize sz = new LayoutSize(0, 0);
        sz.width = sz.height = 0;
        // Let's not have indexes, nested arrays do not have them anyway
        //setIndexed(false); 
        
		// compute array width (sz.width) and max height (sz.height)
		for (int i=0; i<cmps.length; i++) {
			VisualArrayComponent vc = (VisualArrayComponent) cmps[i];
			LayoutSize cs = vc.lGetPreferredSize();
			sz.width += cs.width;
			sz.height = Math.max(sz.height, cs.height);
            vc.setRotated(true);
		}
		
		// Set bounds for components. This is done separately from the
		// previous, because the height of all components should be
		// the same.
		int x = 0;
		for (int i=0; i<cmps.length; i++) {
			VisualComponent vc = (VisualComponent) cmps[i];
			LayoutSize cs = vc.lGetPreferredSize();
			vc.lSetBounds(x,0, cs.width, sz.height);
			if ((i==cmps.length-1)&&(!isRotated())&&(!isFlippedX())) {
				visualResizeHotSpot.setBounds(x, 1, cs.width, getInsets().top-1);
			}
			if ((i==0)&&(!isRotated())&&(isFlippedX())) {
				visualResizeHotSpot.setBounds(x, 1, cs.width, getInsets().top-1);
			}
			x += cs.width;
		}
		if (isRotated()) {
			int vrw=(int)visualResizeHotSpot.getPreferredSize().getWidth();
			visualResizeHotSpot.setBounds(1+x-vrw, 1, vrw, getInsets().top-1);
		}
		return sz;
	}
	
    /** Matrix is always rotated 90 degrees. */
    public boolean isRotated() {
        return true;
    }
    
    /** Matrix is always flipped from the default orientation by Y axis */
    public boolean isFlippedY() {
        return true;
    }
    
	protected boolean shouldBeRotated() {
		return getParent() instanceof VisualArrayComponent;
	}
	
	VisualArrayComponent empty = null;
	public void validateComponents() {
		super.validateComponents();
		/* if (isNestedContainer()) {
			if (shouldBeRotated()) {
				VisualArrayComponent vac = (VisualArrayComponent)getParent();
				VisualMatrix va = vac.getComponentParent();
				setRotated(!va.isRotated());
			} 
		}*/
		if (isDuplicate()) {
			Component[] cmps; 
			int j = 0;
			for (int i=getArray().getFirst(); i<=getArray().getLast(); i++) {
				cmps = getItems(); 
				
				FDT s = (j < getItemCount()) ? ((VisualType)cmps[j]).getStructure() : null;
				
				Object o = new Key("");
				
				if ((s != null) && (!s.equals(o))) {
					int tmp = 0;
					int k;
					for (k=j+1; k<getItemCount(); k++) {
						if (((VisualType)cmps[k]).getStructure().equals(o)) {
							tmp = k;
							break;
						}
					}
					if (tmp != 0) { 
						Component c = cmps[tmp];
						remove(tmp);
						add(c, j);
					} else {
						add(newComponent(i, o), j);
                        //Removal of the old element to avoid moving of empty keys
                        remove(j+1);
					}
				} else if (s == null) {
					add(newComponent(i, o), j);
				}
				VisualArrayComponent vac = 
					(VisualArrayComponent)getItem(j);
				vac.setIndex(i);    
				j++;
			}
			while (j < getItemCount())
				remove(j);
			return;
		}
		if ((getArray() != null)) {
			Component[] cmps; 
			int j = 0;
			for (int i=getArray().getFirst(); i<=getArray().getLast(); i++) {
				cmps = getItems(); //getComponents();
				Object o = getArray().getObject(i);
				
				FDT s = (j < getItemCount()) ? ((VisualType)cmps[j]).getStructure() : null;
				
				if (o == null) {
					if ((s == null) || (!(s instanceof Key)) ||	(!((Key)s).equals(Key.Empty))){
						add(newComponent(i, o), j);
                    }
				} else if ((s != null) && (!s.equals(o))) {
					int tmp = 0;
					int k;
					for (k=j+1; k<getItemCount(); k++) {
						if (((VisualType)cmps[k]).getStructure().equals(o)) {
							tmp = k;
							break;
						}
					}
					if (tmp != 0) { 
						Component c = cmps[tmp];
						remove(tmp);
						add(c, j);
					} else {
                        add(newComponent(i, o), j);
                        //Removal of the old element to avoid moving of empty keys
                        remove(j+1);
					}
				} else if (s == null) {
					add(newComponent(i, o), j);
				}
				VisualArrayComponent vac = 
					(VisualArrayComponent)getItem(j);
				vac.setIndex(i);    
				//if (getArray() instanceof StyledArray)
				//  vac.setLabeled(true);
				j++;
			}
			while (j < getItemCount()) // remove leftovers
				remove(j);
		}
	}
	
    public int getLeftMargin() {
        // getMetrics().stringWidth(getMaxIndexName()) + getInsets().left;
        VisualArrayComponent vac = 
            (VisualArrayComponent)getItem(getItemCount()-1);
        Insets in = vac.getInsets();
        Insets myIn = getInsets();
        return vac.getMetrics().stringWidth(vac.getName())+myIn.left+in.left+in.right;
        
    }
 
    private Matrix getMatrix() { 
        return (Matrix)getStructure();
    }
    
	public String getIndexName(int index) {
        return getMatrix().getIndexName(index);
    }
    
	public int getRows() {
        return getMatrix().getRows();
    }
    
    public String getRowName(int index) {
        return getIndexName(index);
    }
    
    public int getColumns() {
        return ((MatrixImpl) getStructure()).getColumns();
    }
    
    public String getColumnName(int index) {
        return getMatrix().getColumnIndexName(index);
    }
    
	/** @returns the label that has the maximum width.  A heuristics
	 is used with digits. If the array is a StyledArray, the
	 precise max width is calculated.
	 */
	public String getMaxIndexName() {
		if (getStructure() instanceof StyledArray) { // this is slow process for StyledArrays
			StyledArray sa = (StyledArray)getStructure();
			String label = "";
			int labelWidth = 0;
			FontMetrics fm = getMetrics();
			for (int i=0; i<sa.getLast(); i++) {
				String tmp = sa.getIndexName(i);
				int tmpWidth = fm.stringWidth(tmp);
				if (labelWidth < tmpWidth) {
					label = tmp;
					labelWidth = tmpWidth;
				}
			}
			return label;
		}
		// we just determine how many digits there is in the max index and multiply that by the maxWidth of a digit
		else {
			if(getStructure() instanceof Array) {
				int first = ((Array)getStructure()).getFirst();
				int last = ((Array)getStructure()).getLast();
				int maxIndex = Math.max(Math.abs(first),Math.abs(last));
				if (first < 0) maxIndex *= 10;
				String s = String.valueOf(myPow(10,discreteLog(maxIndex+1))-1); // 0,9,99,999,...
				return s;
			}
			return "";
		}
	}
	
	public static final double log10e = 0.4342944819;
	/* @returns the discrete logarithm for x. */
	private int discreteLog(int x) {
		double d = x;
		return (int)(Math.ceil(log10e * Math.log(d))); 
	}
	
	/* @returns a^b */
	private int myPow(int a, int b) {
		if (b == 0) return 1;
		if (b == 1) return a;
		if ((b % 2) == 1) return a*myPow(a, b-1);
		int tmp = myPow(a, b/2);
		return tmp*tmp;
	}
	
    /** This method draws the Matrix representation into
     ** the specified Graphics. Invoked by Java.awt.
     */
    public void paintFrame(Graphics g) {
        if (g instanceof ConfigurableGraphics) {
            if (((ConfigurableGraphics) g).isDisabled("StructureBorder")) {
                // if exporting and structure borders are not wanted then stop
                // here
                return;
            }
        }
        // Nested containers do not have a frame, thus we do this only in case
        // this is not a nested container
        if (!isNestedContainer() || hasFocus()) {
            Rectangle r = getBounds();
            int x = 0;
            int y = 0;
            int w = r.width - 1;
            int h = r.height - 1;
            // draw frame and title of this array
            setBorderColor(g);

            if (getStructure() instanceof CDT)
                g.setColor(getCDTColor());
            g.drawRoundRect(x, y, w, h, ROUNDING*2, ROUNDING*2); // bounding box
            if (isTitled()) { // calculate, e.g., places for the hot spots
                // only if the container is titled
                String name = getName();
                int nWidth = getMetrics().stringWidth(name);
                int nHeight = getMetrics().getHeight();
                int descent = getMetrics().getDescent();

                // MenuHotSpot+SearchHotSpot and nullHotSpot
                int menuHotSpotLabelWidth = 0;
                int upperRightCornerHotSpotWidth = 0;
                int nullHotSpotLabelWidth = 0;
                
                if (isHotSpotEnabled()) {
                    menuHotSpotLabelWidth = getVisualMenuHotSpot()
                            .getPreferredSize().width;
                    if (getVisualSearchHotSpot() != null)
                        upperRightCornerHotSpotWidth = getVisualSearchHotSpot()
                                .getPreferredSize().width;
                    else if (getVisualResizeHotSpot() != null)
                        upperRightCornerHotSpotWidth = getVisualResizeHotSpot()
                                .getPreferredSize().width;
                    if (getVisualNullHotSpot() != null)
                        nullHotSpotLabelWidth = getVisualNullHotSpot()
                                .getPreferredSize().width;
                }
                int roundDiam = 2 * ROUNDING;

                // top bar begins
                // left section consists of an arc on top and a rectangle below it
                g.fillArc(0, 0, roundDiam, roundDiam, 90, 90);
                g.fillRect(0, ROUNDING, ROUNDING, nHeight - ROUNDING);
                // center section is a rectangle
                g.fillRect(ROUNDING + menuHotSpotLabelWidth, y, w + 1
                        - roundDiam - menuHotSpotLabelWidth
                        - upperRightCornerHotSpotWidth, nHeight);
                // right section is again an arc and a rectangle
                g.fillArc(w - roundDiam + 1, 0, roundDiam, roundDiam, 0, 90);
                g.fillRect(w - ROUNDING + 1, ROUNDING, ROUNDING - 1, nHeight
                        - ROUNDING);
                // top bar ends

                // bottom bar begins
                // left section is a rectangle on top and an arc below it
                g.fillArc(0, y + h - roundDiam, roundDiam, roundDiam, 180, 90);
                g.fillRect(0, y + h - nHeight, ROUNDING, nHeight - ROUNDING);
               
                // center section
                g.fillRect(ROUNDING, y + h - nHeight, w + 1 - roundDiam
                        - nullHotSpotLabelWidth, nHeight);
                // right section
                g.fillArc(w - roundDiam + 1, y + h - roundDiam, roundDiam,
                          roundDiam, 0, -90);
                g.fillRect(w - ROUNDING + 1, y + h - nHeight, ROUNDING, nHeight
                        - ROUNDING);
                // bottom bar ends

                // calculate the correct position and space left for title name
                // after hot spots
                x += menuHotSpotLabelWidth;
                // cut width according to both possible hot spots
                w -= (menuHotSpotLabelWidth + upperRightCornerHotSpotWidth);

                if (w < 0)
                    return; // does not fit, just return

                while ((name.length() > 1) && (nWidth > w)) { // try to have
                    // at least some
                    // title name
                    name = name.substring(0, name.length() - 1);
                    nWidth = getFontMetrics(getFont()).stringWidth(name);
                }
                setPenColor(g); // title drawing color
                g.drawString(name, x + w / 2 - nWidth / 2, y + nHeight
                        - descent);
                // indexes
                int leftMargin = getLeftMargin();
                for (int i=0; i<getColumns(); i++) // indexes
                    g.drawString(getColumnName(i), leftMargin+i*(w-leftMargin+5)/getColumns(), y+h-descent);
            } else { // if not titled, just paint the single line borders
                g.fillRect(x, y + h - 3, w, 3); // bottom bar
                g.fillRect(x, y, w, 3); // top bar
            }
        }
    }
    
	public void paintComponent(Graphics g) {
		super.paintComponent(g);
		paintFrame(g);
	}
	
	// simulation methods
	public void write(int i) {
		getArray().setObject(read.getStructure(), i);
	}
	
	public boolean canInsert() {
		return getStructure() instanceof CDT;
	}
	
	public void insert(Object structure) {
		if (structure instanceof Key)
			if (((Key)structure).toString() == "")
				return;
		if (getStructure() instanceof CDT) {
			if (!(structure == null))
				((CDT)getStructure()).insert(structure);
		}
	}
	
	public void delete(Object structure) {
		erase(structure);
	}
	
	public void erase(Object component) {
		if(getStructure() instanceof CDT) 
			((CDT)getStructure()).delete(component);
	}
	
	protected void removeKeyHelp(Object o) {
		if(getParent() instanceof VisualType)
			((VisualType)getParent()).removeKeyHelp(o);
		else if(getStructure() instanceof CDT) {
			((CDT)getStructure()).delete(o);
			//this.setStructure(((CDT)getStructure()).delete(o));
		}
		else if(getStructure() instanceof Array) {
			Array a = (Array)getStructure();
			int i = a.getFirst();
			while(i < a.getLast()) {
				if(o.equals(a.getObject(i))) {
					a.setObject(null, i);
					break;
				}
				i++;
			}
		}
		else
			remove();	
	}	    
	
	public VisualArrayComponent newComponent(int place, Object o) {
		return new VisualArrayComponent(place, o, rotateComponents);
	}
    
    /**
     * Configures the properties related to VisualMatrix and then calls the 
     * configure method of the super class.
     */
    public void configure(VisualTypeConf vtc) {
        if (vtc != null) {
            String value = vtc.getVisualConfValue(this.getClass().getName(), VisualTypeConf.INDEXED);
            //In the following way, no need to check whether value is null or not.
            if ("true".equals(value)) { 
                setIndexed(true);
            } else if ("false".equals(value)) {
                setIndexed(false);
            }
        }
        super.configure(vtc);
    }
    
    /**
     * SerialVersionUID
     */
    private static final long serialVersionUID = 7697614843490142527L;
}
