// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/* $Id: VisualReference.java,v 1.88 2007/09/10 08:11:08 vkaravir Exp $ */

package matrix.visual;

import java.awt.*;
import java.awt.geom.CubicCurve2D;
import java.awt.geom.QuadCurve2D;
import java.io.PrintStream;
import java.util.*;

import matrix.decoration.*;
import matrix.structures.FDT.*;
import matrix.structures.FDT.probe.Key;
import matrix.structures.FDT.substructures.Source;
import matrix.util.Note;

/** This abstraction is used to represent any kind of references.
 * A subclass of the VisualContainer could have its own concrete subclass of
 * the VisualReference in order to have a special look and feel of links.
 * @see VisualContainer
 *
 * @author Ari Korhonen, Riku Saikkonen
 * @version $Revision: 1.88 $
 */
public class VisualReference extends VisualType {
    protected VisualComponent from, to, newtarget = null, newsource = null, movingTarget = null;
    protected boolean weighted = false;
    protected Point fromPoint, toPoint;

    private VisualType weight;
    private Vector midPoints;
    private boolean midPointReversal;

    private int lx;  // temporary variables for finding out the location of
    private int ly;  // new midpoints
    
    private static final double  ARROW_HALF_ANGLE = Math.PI / 6;
    private static final int     ARROW_LENGTH = 8;
    private static final double  EPSILON = 4.0;

    private static final boolean ANTIALIAS    = true;
    private static final boolean EXPERIMENTAL = true; //smooth curves

    /** Have the midpoints for self-loops been added? */
    private boolean loopPointsAdded=false;
    /** Have the midpoints for self-loops been modified externally? */
    private boolean loopPointsModified=false;

    /**
     * Returns the centre point of an AWT component converted to the
     * coordinate system of the container of this component.
     * @param c Component to get centre of
     * @return Centre point of component
     */
    private Point getCorrectedCentre(VisualComponent c) {
        if (c!=movingTarget) {
            Component cp=c.getParent();
            Component tp=getParent();	

            if ((cp!=tp)&&(cp!=null)&&(tp!=null)) {
                Point p=javax.swing.SwingUtilities.convertPoint(cp,
                        c.getCenter(),
                        tp);
                return p;
            }
        }
        return c.getCenter();
    }

    /**
     * Returns the centre point of an AWT component converted to the
     * coordinate system of the container of this component.
     * @param c Component to get centre of
     * @return Centre point of component
     */
    private Point getCorrectedBorderAtAngle(VisualComponent c, double angle) {
        Component cp=c.getParent();
        Component tp=getParent();	

        if (c!=movingTarget) {
            if ((cp!=tp)&&(cp!=null)&&(tp!=null)) {
                Point p=javax.swing.
                SwingUtilities.convertPoint(cp,
                        c.getBorderAtAngle(angle),
                        tp);
                return p;
            }	    
        }
        return c.getBorderAtAngle(angle);
    }

    // --- debug purposes only ---
    public void dump(PrintStream out) {
        super.dump(out);
        out.println("movingTarget = " + movingTarget);
        out.println("newtarget = " + newtarget + "; newsource = " + newsource);
    }

    // --- constructors ---
    public VisualReference(VisualComponent from, VisualComponent to) {
        super(null);
        this.from = from;
        this.to = to;
        fromPoint=getCorrectedCentre(from); // just to have some initial values
        toPoint=getCorrectedCentre(to);
        forgetMidPoints();

        //Note.out(this, "Checking for reference labels..");
        if (from.getStructure() instanceof LabelDecorator) {
            LabelDecorator ld = (LabelDecorator)from.getStructure();
            this.weighted = ld.isReferenceLabelEnabled();
            setName(ld.getReferenceLabel(from.getReferenceNumber(to)));
        } else if (from.getStructure() instanceof Styled) {
            StyleSheet ss = ((Styled)from.getStructure()).getStyleSheet();
            this.weighted = ss.isReferenceLabelEnabled();
            setName(ss.getReferenceLabel(from.getReferenceNumber(to)));
        } else this.weighted = false;

        if(from.getStructure() instanceof ReferenceInformationDecorator) {
            ReferenceInformationDecorator ttd = (ReferenceInformationDecorator)from.getStructure();
            if(ttd.isReferenceToolTipInfoEnabled()) {
                setToolTipText(ttd.getReferenceInformation(from.getReferenceNumber(to)));
            }
        }

        //addPopUpMenuSeparator();
        //addPopUpCheckboxMenuItem("Weighted", weighted);
    }

    // --- end of constructors ---


    // -- visualization methods: COLOR --

    public boolean hasFocus() {
        return focus;
    }

    /** Toggle weight labels. */
    public void Weighted() {
        weighted = !weighted;
        weight.setVisible(weighted);
        //setInvalid();
    }

    /** Determine the weight flag. 
     * @return true if the reference is labeled by weight; false otherwise
     */
    public boolean isWeighted() {
        return weighted; 
    }

    /** Determine the directed flag. 
     * @return true if the reference is directed; false otherwise
     */
    public boolean isDirected() {
        if (getVisualContainer() instanceof VisualList)
            return true;
        if (getVisualContainer() instanceof VisualTree)
            return ((VisualTree)getVisualContainer()).isDirected();
        if (getVisualContainer() instanceof VisualGraph)
            return ((VisualGraph)getVisualContainer()).isDirected();
        return false;
    }

    /** Forgets previously added midpoints and reversing information. */
    public void forgetMidPoints() {
        midPointReversal = false;
        midPoints = new Vector();
        loopPointsAdded=false;
    }

    public void setMidPointOrdering(boolean reversed) {
        midPointReversal = reversed;
    }

    /** Adds a midpoint to the reference.
     * The line is drawn through the point c.getCenter(). This adds a
     * midpoint after all previous midpoints (just before toPoint).
     * @param c a (possibly dummy) component placed at the midpoint
     */
    public void addMidPoint(VisualComponent c) {
        loopPointsModified=true;
        if (midPointReversal)
            midPoints.insertElementAt(c, 0);
        else
            midPoints.addElement(c);
    }

    public void addComponents() {
        if (weight == null)
            weight = new VisualLabel(new Key(getName()));
    }

    public void validateComponents() {
        if (weight != null)
            weight.setVisible(isWeighted());
        if (isWeighted()) {
            if (getComponentCount() == 0)
                add(weight);
        } else {
            removeAll();
        }
    }

    public void doLayout() {
        super.doLayout();

        // remove comments if old fashioned pointer updates are preferred
        //newtarget = null;
        //newsource = null;

        int cCount = getComponentCount();
        for (int i = 0; i < cCount; i++)
            getComponent(i).doLayout();
        if ((getSource() != null) && (getTarget() != null)) {
            fromPoint = getCorrectedCentre(getSource());
            toPoint = getCorrectedCentre(getTarget());

            doLoopPoints(); //creates midpoints for self-loop

            countFromAndToPoints(); //counts fromPoint and toPoint coordinates

            int x1, y1, x2, y2;

            if (fromPoint.x < toPoint.x) { x1 = fromPoint.x; x2 = toPoint.x; }
            else { x2 = fromPoint.x; x1 = toPoint.x; }

            if (fromPoint.y < toPoint.y) { y1 = fromPoint.y; y2 = toPoint.y; }
            else { y2 = fromPoint.y; y1 = toPoint.y; }

            Enumeration mpe = midPoints.elements();
            while (mpe.hasMoreElements()) {
                Point p = getCorrectedCentre(((VisualComponent)mpe.
                        nextElement()));
                if (p.x < x1) x1 = p.x;
                if (p.x > x2) x2 = p.x;
                if (p.y < y1) y1 = p.y;
                if (p.y > y2) y2 = p.y;
            }

            if (isDirected()) { /*Calculate arrowhead size if directed*/
                // A pessimistic estimate of arrowhead size
                int arrowWidth = (int) (Math.sin(ARROW_HALF_ANGLE) *
                        2 * ARROW_LENGTH);
                int arrowHeight = arrowWidth;
                // Add half of the arrowhead to all borders
                x1 -= arrowWidth / 2 + 1; x2 += arrowWidth / 2 + 1;
                y1 -= arrowHeight / 2 + 1; y2 += arrowHeight / 2 + 1;
            }

            // XXX check that these calculations are still correct!
            // Take the size of the weight label into account if weighted.
            if (isWeighted() && (getFont() != null) && (getName() != null)) {
                int lWidth = getMetrics().stringWidth(getName());
                int lHeight = getMetrics().getHeight();
                if (x2 - x1 < 2*lWidth) {
                    int disp = (2*lWidth - (x2 - x1)) / 2 + 1;
                    x1 -= disp; x2 += disp;
                }
                if (y2 - y1 < lHeight) {
                    int disp = (lHeight - (y2 - y1)) / 2 + 1;
                    y1 -= disp; y2 += disp;
                }
            }

            //rectangle that bounds this reference
            Rectangle r = new Rectangle(x1, y1, x2 - x1 + 1, y2 - y1 + 1);
            setBounds(r);
            width = r.width;
            height = r.height;

            int stringWidth = getMetrics().stringWidth(getName()) +6;
            int stringHeight = getMetrics().getHeight();

            // set weight - this set bounds to the weight. Have to use
            // width and height of the string to calculate right place
            // for the upper left corner of the weight. In this way,
            // different font sizes are handled so that weight should
            // be always placed in the middle of the reference.
            if (getComponentCount() > 0) {
                getComponent(0).setBounds(Math.abs((x2-x1)/2) - (stringWidth/2), //-10,
                        Math.abs((y2-y1)/2) - (stringHeight/2), //-10,
                        stringWidth, stringHeight); //20,20);
            }
        }

    }

    /**
     * Creates midpoints for self-loop.
     *
     */
    private void doLoopPoints() {
        if (loopPointsAdded&&!loopPointsModified)
            forgetMidPoints();

        if (fromPoint.equals(toPoint)&&(midPoints.size()==0)&&
                !loopPointsModified) {
            VisualReferenceTurn vrt =
                new VisualReferenceTurn(getVisualContainer(), this);
            // Add two new midpoints for a small self-loop. 
            Rectangle r = new Rectangle();
            r.x=fromPoint.x+3+getSource().getWidth();
            r.y=fromPoint.y;
            vrt.setBounds(r);
            //      addMidPoint(vrt);
            vrt = new VisualReferenceTurn(getVisualContainer(), this);
            r.y=fromPoint.y-3-getSource().getHeight();
            vrt.setBounds(r);
            //      addMidPoint(vrt);
            loopPointsModified=false;
            loopPointsAdded=true;

            //Note.err(this, 
            //       "source and target are on top of each other!");
            //                return;
        }
    }

    /**
     * Counts fromPoint and toPoint coordinates.
     * Takes into account midPoints if they exists.
     *
     */
    private void countFromAndToPoints() {
        Point secondPoint, secondToLastPoint;
        if (midPoints.size() > 0) {
            secondPoint=
                getCorrectedCentre(((VisualComponent)midPoints.
                        firstElement()));
            secondToLastPoint =
                getCorrectedCentre(((VisualComponent)midPoints.
                        lastElement()));
        } else {
            secondPoint = toPoint;
            secondToLastPoint = fromPoint;
        }

        double fromAngle =
            normalizeAngle(2*Math.PI -
                    Math.atan2(secondPoint.y - fromPoint.y,
                            secondPoint.x - fromPoint.x));
        double toAngle = 
            normalizeAngle(2*Math.PI -
                    Math.atan2(secondToLastPoint.y - toPoint.y,
                            secondToLastPoint.x - toPoint.x));

        fromPoint = getCorrectedBorderAtAngle(getSource(), fromAngle);
        toPoint = getCorrectedBorderAtAngle(getTarget(), toAngle);
    }

    /*
    public Dimension getSize() {
        return new Dimension(width, height);
    }
     */

    /** 
     * Overriding getName() -method in VisualType.  Weight is fetched
     * using VisualLabel's getKey() -method because it fetches weight
     * straight from the underlying structure.
     * @return name of the reference (weight)
     */
    public String getName() {
        if (weight != null)
            return ((VisualLabel)weight).getKey();
        else
            return super.getName();
    }

    /** Returns the source component of this reference. */
    public VisualComponent getSource() { return from; }

    /** Sets the source component of this reference. */
    public void setSource(VisualComponent t) { from = t; }

    /** Returns the component this reference is pointing to. */
    public VisualComponent getTarget() { 
        if (movingTarget != null)
            return movingTarget;
        if (newtarget != null)
            return newtarget;
        return to;
    }

    /** Sets the component this reference is pointing to. */
    public void setTarget(VisualComponent t) { 
        if ((to != t) || (newsource != getSource())) {
            newtarget = null;
            newsource = null;
            forgetMidPoints();
        }
        to = t; 
    }

    /** return the Point this reference is pointig to. */ 
    private Point getTargetLocation() {
        return toPoint;
    }

    public void assign() {
        getApplication().showMessage("Sorry", "Cannot assign "+ read +
                " into Reference.\n"+
        "Try to write (dblClick) the new label.");
    }

    public void write() {
        if ((from != null) && 
                (from.getStructure() instanceof LabelDecorator)) {
            LabelDecorator ld = (LabelDecorator)from.getStructure();
            if (ld == null)
                Note.err(this, "Illegal (null) LabelDecorator");
            if (ld.isReferenceLabelEnabled()) {
                ld.setReferenceLabel(from.getReferenceNumber(to), 
                        read.getStructure().toString());
                ((VisualLabel)weight).setKey(read.getStructure().toString());
            }
            //else Note.out(this, "Not referenceLabelEnabled "+ld);
        } else Note.warning(this, "from was null while trying to write()");
    }

    public void rename() {
        if (from.getStructure() instanceof LabelDecorator) {
            LabelDecorator ld = (LabelDecorator)from.getStructure();
            String newName = getApplication().getInput("Rename: ", ld.getReferenceLabel(from.getReferenceNumber(to)));
            if (newName == ld.getReferenceLabel(from.getReferenceNumber(to)))
                return;
            getAnimator().startOperation();
            setName(newName);
            ld.setReferenceLabel(from.getReferenceNumber(to), newName);
            ((VisualLabel)weight).setKey(newName);
            getAnimator().endOperation();
            setInvalid();
        }
    }

    public Dimension getPreferredSize() { 
        return new Dimension(10,10);
    }

    public FDT getStructure() {
        return null;
    }

    public VisualType getMovingPart() {
        //setMask(AWTEvent.MOUSE_EVENT_MASK, false);
        return movingTarget = new VisualArrow(this);
    }

    public void selectEntity(boolean isShiftDown) {
        if (getSource().getStructure() instanceof Source) {
            getAnimator().startOperation();
            ((Source)(getSource().getStructure())).referenceSelectedTo(getTarget().getStructure());
            getAnimator().endOperation();
        }	
        setInvalid();
    }

    
    public void drag(int x, int y) {
        movingTarget.setBounds(x+getBounds().x-2,y+getBounds().y-4,1,1);
        forgetMidPoints();
        doLayout();
        
        lx = x; ly = y;
        //repaint(); DO WE NEED THIS?
        //super.drag(x,y);
    }

    private VisualComponent getNode(VisualType v) {
        if (v == null)
            return null;
        if (v instanceof VisualComponent)
            return (VisualComponent)v;
        if (v.getParent() instanceof VisualType)
            return getNode((VisualType)v.getParent());
        return null;
    }

    public void released() {
        //Note.out(this, "Reference released to " + entered);
        //setMask(AWTEvent.MOUSE_EVENT_MASK, true);
        picked = false;
        if (!dragEnabled())
            return;
        if (entered instanceof VisualNullHotSpot) {
            //Note.out(this, "NullHotsSpot source = "+ getSource());
            getSource().moveReference(this, from, null);
            entered = null;
        } else {
            VisualReferenceTurn vrt = new VisualReferenceTurn(getVisualContainer(), this);

            // set the bounds (location) of the midpoint
            Rectangle r = new Rectangle(lx + getX(), ly + getY(), 1, 1);
            vrt.setBounds(r); // set ad hoc bounds for the midpoint
            //	addMidPoint(vrt); // add turn to this reference
        }
        VisualComponent vc = getNode(entered);
        if ((vc != null) && (vc instanceof VisualComponent)) {
            from.moveReference(this, to, vc); }
        newtarget = vc;
        newsource = getSource();

        movingTarget = null;
        setInvalid();
    }

    private int min(int a, int b) {
        return a < b ? a : b;
    }

    private double distance(int x, int y, int XA, int YA, int XB, int YB) {
        int YC,XC;

        XC = x + getBounds().x;
        YC = y + getBounds().y;

        double LL = Math.sqrt((XA-XB)*(XA-XB) + (YA-YB)*(YA-YB));
        return ((YB-YA)*(XC-XA)-(XB-XA)*(YC-YA))/LL;
    }

    public boolean contains(int x, int y) {
        java.util.Iterator mpe = midPoints.iterator();
        int lastX=fromPoint.x;
        int lastY=fromPoint.y;

        while (mpe.hasNext()) {
            Point p = getCorrectedCentre(((VisualComponent)mpe.next()));

            if ((super.contains(x,y) &&
                    (Math.abs(distance(x,y, lastX, lastY,
                            p.x, p.y)) < EPSILON)))
                return true;
            lastX=p.x;
            lastY=p.y;
        }
        if ((super.contains(x,y) &&
                (Math.abs(distance(x,y, lastX, lastY,
                        toPoint.x, toPoint.y)) < EPSILON)))
            return true;

        return false;
    }

    /** Normalizes angle to [0..2pi). */
    private double normalizeAngle(double angle) {
        while (angle < 0)
            angle += 2 * Math.PI;
        while (angle >= 2 * Math.PI)
            angle -= 2 * Math.PI;

        return angle;
    }

    /** Returns the point on perpendicular projection of line (x1,y1)->(x2,y2)
     * and point (x,y) that lies the specified distance away from the 
     * reference line.
     */
    private Point moveAlongPerpendicularProjection(int x1, int y1,
            int x2, int y2,
            int x, int y, 
            int d) {
        if (x1 == x2) return new Point(x+d,y);
        if (y1 == y2) return new Point(x,y-d);
        double k = (x2-x1)/(y1-y2);

        double delta_x = Math.sqrt((d*d)/(k*k+1));
        double delta_y = k*delta_x;

        return new Point(x+(int)delta_x,y+(int)delta_y);
    }

    private Point extrapolate(int x1, int y1,
            int x2, int y2,
            int d) {

        double dist = d / Math.sqrt((x1-x2)*(x1-x2) + (y1-y2)*(y1-y2));

        double delta_x = (x2 - x1) * dist;
        double delta_y = (y2 - y1) * dist;

        return new Point(x1+(int)delta_x,y1+(int)delta_y);
    }


    /*
    private void paintWeight(Graphics g, int x1, int y1, int x2, int y2) {
        if (weighted) {
            String name = getName();
            if (name == null) name = "Unknown";
            int lWidth = getMetrics().stringWidth(name);
            if (lWidth < Math.abs(x2-x1)) {
                lWidth = Math.abs(x2-x1) - lWidth;
            } else 
                lWidth = 0;
            FontMetrics fm = getMetrics();
            int lHeight = fm.getHeight() - fm.getMaxDescent();
            int x = (x1 < x2 ? x1 : x2) + lWidth/2;
            int y = y1 + (y2-y1)/2 + lHeight/2;
            //Point p = moveAlongPerpendicularProjection(x1,y1,x2,y2,x,y,10);
            //x = p.x; y = p.y;

            //g.drawString(name, x, y);
            //if (weight == null) weight = new VisualKey(new Key(name));
            //weight.setBounds(x,y,width,height);
            //weight.paint(g);
        }
    }
     */
    private void drawAngledLine(Graphics g, int x, int y,
            double angle, int length) {
        g.drawLine(x,
                y,
                x + (int) (Math.cos(angle) * length),
                y - (int) (Math.sin(angle) * length));
    }

    private void paintArrow(Graphics g, int x1, int y1, int x2, int y2) {
        if (isDirected()) {
            double toAngle = 2*Math.PI - Math.atan2(y1 - y2, x1 - x2);
            drawAngledLine(g, x2, y2, toAngle - ARROW_HALF_ANGLE, ARROW_LENGTH);
            drawAngledLine(g, x2, y2, toAngle + ARROW_HALF_ANGLE, ARROW_LENGTH);
        }
    }

    public boolean isHiding() {
        VisualComponent target = getTarget();
        if ((target == null) || (target.isHiding()))
            return true;
        return false;
    }

    public boolean isVisited() {
        if (getSource().getStructure() instanceof Source) 
            return ((Source)getSource().getStructure()).isVisitedReferenceTo(getTarget().getStructure());
        else
            return super.isVisited();
    }

    private StyleSheet defaultStyleSheet = null;
    public StyleSheet getDefaultStyleSheet() {
        if (defaultStyleSheet == null) {
            defaultStyleSheet = new StyleSheetAdapter();
            //defaultStyleSheet.setFocusedBorderColor(Color.red);
            //defaultStyleSheet.setVisitedFocusedBorderColor(Color.magenta);
        }
        return defaultStyleSheet;
    }

    void setBackgroundColor(Graphics g) {
        Color defaultColor;
        ColorDecorator cd;
        FDT fromFdt = from.getStructure();
        if (fromFdt instanceof ColorDecorator) {
            cd = (ColorDecorator)fromFdt;
        } else if (fromFdt instanceof Styled) {
            cd = ((Styled)fromFdt).getStyleSheet();
        } else 
            cd = getDefaultStyleSheet();
        if (isVisited()) {
            if (hasFocus()) 
                defaultColor = cd.getVisitedFocusedBackgroundColor();
            else
                defaultColor = cd.getVisitedBackgroundColor();
        } else {
            if (hasFocus())
                defaultColor = cd.getFocusedBackgroundColor();
            else
                defaultColor = cd.getDefaultBackgroundColor();
        }
        if (g != null)
            g.setColor(defaultColor);
    }

    void setBorderColor(Graphics g) {
        Color defaultColor;
        ColorDecorator cd;
        FDT fromFdt = from.getStructure();
        if (fromFdt instanceof ColorDecorator) {
            cd = (ColorDecorator)fromFdt;
        } else if (fromFdt instanceof Styled) {
            cd = ((Styled)fromFdt).getStyleSheet();
        } else 
            cd = getDefaultStyleSheet();
        if(hasFocus()) {
            //Note.out(this, this + " hsa focus, using focused color");
            defaultColor = cd.getFocusedReferenceColor(to.getStructure());
        } else if(isVisited()) {
            //Note.out(this, this + " hsa focus, using visited color");
            defaultColor = cd.getModifiedReferenceColor(to.getStructure());
        } else
            defaultColor = cd.getReferenceColor(to.getStructure());
        if (g != null)
            g.setColor(defaultColor);
    }

    void setPenColor(Graphics g) {
        Color defaultColor;
        ColorDecorator cd;
        FDT fromFdt = from.getStructure();
        if (fromFdt instanceof ColorDecorator) {
            cd = (ColorDecorator)fromFdt;
        } else if (fromFdt instanceof Styled) {
            cd = ((Styled)fromFdt).getStyleSheet();
        } else 
            cd = getDefaultStyleSheet();
        if (isVisited()) {
            if (hasFocus()) 
                defaultColor = cd.getVisitedFocusedPenColor();
            else
                defaultColor = cd.getVisitedPenColor();
        } else {
            if (hasFocus())
                defaultColor = cd.getFocusedPenColor();
            else
                defaultColor = cd.getDefaultPenColor();
        }
        if (g != null)
            g.setColor(defaultColor); 
    }


    public void paintComponent(Graphics g) {

        // TODO Clearing draws should all be performed before
        // the painting draws. The current painting creates
        // gaps when the clearing draw deletes a part of the
        // previous painting draw
        
        
        Object antialiasState = null;
        if (ANTIALIAS && g instanceof Graphics2D){
            antialiasState = ((Graphics2D)g).getRenderingHint(RenderingHints.KEY_ANTIALIASING);
            ((Graphics2D)g).setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        }


        if (isHiding()) return;
        setBorderColor(g);

        Point tp = getTargetLocation(); // this is because doLayout is not
        // called while dragging

        //Note.out(this, "newtarget = " + newtarget);


        // reset translation to same coordinates as components
        g.translate(-getBounds().x, -getBounds().y);        
        Point last = fromPoint;
        Enumeration mpe = midPoints.elements();


        if (EXPERIMENTAL == false || !(g instanceof Graphics2D)){
            while (mpe.hasMoreElements()) {
                Point p = getCorrectedCentre(((VisualComponent)mpe.nextElement()));

                if (!(g instanceof matrix.util.export.MatrixGraphics)) {
                    // not painting when exporting, since the result looks bad
                    // and references can't have focus when exporting
                    if (!hasFocus())
                        g.setColor(getBackground());

                    if (ANTIALIAS && g instanceof Graphics2D){
                        Graphics2D g2= (Graphics2D)g;

                        Stroke originalStroke = g2.getStroke();

                        int width = hasFocus() ? 3 : 4; //wider to paint over the focused version                    
                        g2.setStroke(new BasicStroke(width));

                        g.drawLine(last.x, last.y, p.x, p.y);

                        g2.setStroke(originalStroke);



                    }
                    else{

                        g.fillPolygon(new int[] { last.x + 2, p.x+2, p.x-2, last.x-2 },
                                new int[] { last.y, p.y, p.y, last.y }, 4);
                        g.fillPolygon(new int[] { last.x, p.x, p.x, last.x },
                                new int[] { last.y + 2, p.y+2, p.y-2,last.y-2}, 4);
                    }



                }
                setBorderColor(g);
                g.drawLine(last.x, last.y, p.x, p.y);
                last = p;
            }

            if (!(g instanceof matrix.util.export.MatrixGraphics)) {
                // not painting when exporting, since the result looks bad
                // and references can't have focus when exporting
                if (!hasFocus()) 
                    g.setColor(getBackground());

                if (ANTIALIAS && g instanceof Graphics2D){
                    Graphics2D g2= (Graphics2D)g;

                    Stroke originalStroke = g2.getStroke();

                    int width = hasFocus() ? 3 : 4; //wider to paint over the focused version                
                    g2.setStroke(new BasicStroke(width));

                    g.drawLine(last.x, last.y, tp.x, tp.y);

                    g2.setStroke(originalStroke);
                }
                else{

                    g.fillPolygon(new int[] { last.x + 2, tp.x+2, tp.x-2, last.x-2 },
                            new int[] { last.y, tp.y, tp.y, last.y }, 4);
                    g.fillPolygon(new int[] { last.x, tp.x, tp.x, last.x },
                            new int[] { last.y + 2, tp.y+2, tp.y-2,last.y-2}, 4);
                }
            }

            setBorderColor(g);
            g.drawLine(last.x, last.y, tp.x, tp.y);

        }else{

            // EXPERIMENTAL CODE - smooth curves
            
            setBorderColor(g);
            Graphics2D g2= (Graphics2D)g;
            Stroke originalStroke = g2.getStroke();

            int width = hasFocus() ? 3 : 4; //wider to paint over the focused version                
            g2.setStroke(new BasicStroke(width));

            if (!hasFocus()) 
                g.setColor(getBackground());
            
            if (midPoints.size() == 0){

                g.drawLine(last.x, last.y, tp.x, tp.y);            
                setBorderColor(g);
                g2.setStroke(originalStroke);
                g.drawLine(last.x, last.y, tp.x, tp.y);            
            
            }else if (midPoints.size() == 1){

                Point p = getCorrectedCentre(((VisualComponent)midPoints.get(0)));

                QuadCurve2D qc = new QuadCurve2D.Double(last.x, last.y, 
                        p.x,
                        p.y,
                        tp.x, tp.y);

                g2.draw(qc);
                setBorderColor(g);
                g2.setStroke(originalStroke);
                g2.draw(qc);

            }else if (midPoints.size() == 2){

                Point p = getCorrectedCentre(((VisualComponent)midPoints.get(0)));
                Point p2= getCorrectedCentre(((VisualComponent)midPoints.get(1)));

                CubicCurve2D cc = new CubicCurve2D.Double(last.x, last.y, 
                        p.x,
                        p.y,
                        p2.x,
                        p2.y,
                        tp.x, tp.y);

                g2.draw(cc);
                setBorderColor(g);
                g2.setStroke(originalStroke);
                g2.draw(cc);

            }
            
            if (!midPoints.isEmpty())
                last = getCorrectedCentre((VisualComponent)midPoints.lastElement());
            
        }

        paintArrow(g, last.x, last.y, tp.x, tp.y);
        //paintWeight(g, fromPoint.x, fromPoint.y, tp.x, tp.y);
        g.translate(getBounds().x, getBounds().y); // cancel above translation

        if (ANTIALIAS && g instanceof Graphics2D){
            ((Graphics2D)g).setRenderingHint(RenderingHints.KEY_ANTIALIASING, antialiasState);
        }

        super.paintComponent(g);
    }

    public String toString() {
        return "link " + from + "->" + to;
    }
}
