// Matrix - algorithm simulation system
// Copyright (C) 2000- 2007 Matrix group
// (See the AUTHORS file and the author tag below for more information
// on the authors.)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/* $Id: VisualTree.java,v 1.85 2007/09/10 08:11:06 vkaravir Exp $ */

package matrix.visual;

import java.awt.*;
import java.io.PrintStream;
import java.util.Stack;

import matrix.animation.StructureChangeOperation;
import matrix.decoration.LabelDecorator;
import matrix.structures.CDT.CDT;
import matrix.structures.FDT.*;
import matrix.util.Note;

/**
 * This is an abstract container class for trees. The validator is provided
 * that mirrors the actual data structure and validates all the components and
 * references within this container. The validator ensures that the component
 * hierarchy created is a valid tree even though the actual data structure
 * is not. Thus, the concrete subclass that produces the tree layout may
 * assume that there is no cycles in paths starting from the root node.
 * If the actual data structure was a graph, several flags are provided
 * to disable edges that do not exist in trees, such as back edges,
 * cross edges, etc.
 *
 * @author Ari Korhonen
 * @version $Revision: 1.85 $
 */
public abstract class VisualTree extends VisualContainer {
    // special head structure pointing to the root node of the tree
    //private Head head; 
    //private int[] forest; // store the actual tree as dad-linked "forest"
    //private VisualHeadReference vHeadRef; // visual reference for head
    private boolean emptyLeaves = true;
    private boolean backEdges = true;
    private boolean forwardEdges = true;
    private boolean crossEdges = true;
    private boolean DFS = true;
    private boolean directed = false;

    // debug
    public void dump(PrintStream out) {
        super.dump(out);
        out.println("tree=" + getStructure());
    }

    /** Returns the horizontal gap between nodes in layout coordinates. */
    public int getGapWidth() { return super.getGapWidth(); }
    /** Returns the vertical gap between nodes in layout coordinates. */
    public int getGapHeight() {
        if (getStructure() instanceof LabelDecorator && 
                ((LabelDecorator)getStructure()).isReferenceLabelEnabled())
            return 25; //make the gap little bigger if reference label is enabled
        return super.getGapHeight();
    }
    
    /** Construct a new representation for the specified tree.
     * @param t the tree
     */
    public VisualTree(Tree t) {
        super(t);
    }

    /*
    public FDT getStructure() {
	FDT currentFDT = super.getStructure();
	if (currentFDT instanceof Tree)
	    return currentFDT;
	if (currentFDT instanceof CDT)
	    return (FDT)currentFDT.getElement();
	Note.err(this, "Illegal FDT structure");
	return null;
    }
    */
    /*
    public void setStructure(FDT fdt) {
      root.setObject(fdt,0);
    }
    */

    /** Subclasses should override this method in order to 
     ** return a new concrete VisualTreeComponent for the specified Tree.
     * By default this method returns a new VisualTreeComponent.
     * @see VisualTreeComponent
     * @param f will be the father of the new component
     */
    protected VisualTreeComponent createNewComponent(Tree t, 
                                                     VisualTreeComponent f) {
	/* Code removed because of bug fix repoted in flyspray Task #45 
	boolean h=hasRepresentation(t);
	if (h) {
	    Note.out(this, "create visualization for null tree node.");
	    return new VisualTreeComponent(null,f);
	}
	else */
	return new VisualTreeComponent(t,f);
    }

    /** Subclasses should override this method to return a new concrete
     ** VisualReference between the specified Vertices.
     * By default this method returns a new VisualReference
     * @see VisualReference
     */
    protected VisualReference createNewReference(VisualComponent from,
                                                 VisualComponent to) {
        return new VisualReference(from, to);
    }

    // popUpCheckboxMenuItems

    /** Toggle directed flag. */
    public void Directed() {
        directed = !directed;
        setInvalid();
    }

    /** Toggle emptyLeaves flag. */
    public void EmptyLeaves() {
        emptyLeaves = !emptyLeaves;
        setInvalid();
    }

    /** Toggle DFS flag. */
    public void DFSvalidate() {
        DFS = !DFS;
        setInvalid();
    }

    /** Toggle backEdges flag. */
    public void BackEdges() {
        backEdges = !backEdges;
        setInvalid();
    }

    /** Toggle forwardEdges flag. */
    public void ForwardEdges() {
        forwardEdges = !forwardEdges;
        setInvalid();
    }
    
    /** Toggle crossEdges flag. */
    public void CrossEdges() {
        crossEdges = !crossEdges;
        setInvalid();
    }

    /** Return true if the edges are directed. */
    public boolean isDirected() {
        return directed;
    }

    /** Return true if empty subtrees are visible. */
    public boolean isEmptySubTreesEnabled() { 
        return emptyLeaves; 
    }

    /** Return true if back edges are enabled. */
    public boolean isBackEdgesEnabled() { 
        return backEdges; 
    }

    /** Return true if forward edges are enabled. */
    public boolean isForwardEdgesEnabled() { 
        return forwardEdges; 
    }

    /** Return true if cross edges are enabled. */
    public boolean isCrossEdgesEnabled() { 
        return crossEdges; 
    }

    public boolean isDFSValidated() {
	return DFS;
    }

    private int getIndexOfComponent(VisualType vt) {
        for (int i=0; i<getComponentCount(); i++)
            if (vt.equals(getComponent(i))) return i;
        return -1;
    }
    private VisualTreeComponent getVisualTreeComponent(int index) {
        return (VisualTreeComponent)getItem(index);
    }

    private VisualTreeComponent setComponents(Tree t, int position, 
                                              VisualTreeComponent f) {
        if (position == 0) return setRootComponent(t);
        
        int k = lookUpVisualItem(t, position);
        if (k > position) { // found, but not in correct place -> move to correct place
            VisualReference vr = getLink(k);
            removeLink(k);
            add(vr,position);
            VisualComponent vc = getItem(k);
            removeItem(k);
            add(vc,position);
        } else if (k == -1) { 
            VisualComponent vc = createNewComponent(t, f);
            add(createNewReference(f, vc), position);
            add(vc, position);
        } 
        VisualTreeComponent vtc = getVisualTreeComponent(position);
        vtc.setFather(f);
        VisualReference vr = getLink(position);
        vr.setSource(f);
        vr.setTarget(vtc);
        return vtc;
    }

    private VisualTreeComponent setRootComponent(Tree t) {
    	VisualTreeComponent oldRootNode = getRootNode();
    	int k = lookUpVisualItem(t,0);
    	
    	if (k < 0) { // representation not found 
    		add(createNewComponent(t,null),0);
    	}
    	else if (k > 0) { // found but at wrong position
            removeLink(k);
    		VisualComponent vc = getItem(k);
    		removeItem(k);
    		add(vc,0);
    		VisualTreeComponent vtc = getVisualTreeComponent(0);
    		vtc.setFather(null);
    	} // else representation found at correct position (== first)
    	
    	if (getLinkCount() == 0) // reference not found
    		add(new VisualHeadReference(getRootNode()),0);
    	else if (getLink(0) instanceof VisualHeadReference) // ref found
    		((VisualHeadReference)getLink(0)).setTarget(getRootNode());
    	else Note.err(this, "setRoot():Cannot find VisualHeadReference: "+
    			getLink(0));
    	
        /* We have possibly created a new component (the root is changed
         * while VisualHeadReference already exists) but not any
         * references, thus in order to prevent the number of items
         * to be different than the number of references we might have
         * to create one (some) dummy reference(s)
         */
        while (getItemCount() > getLinkCount()) {
            add(createNewReference(getRootNode(),oldRootNode));
        }
      
	if(getItemCount() > getLinkCount()) {
	    Note.err(this, "Too many items");
	}
        
        return (VisualTreeComponent)getItem(0);
    }
    /*
    private boolean isForwardEdge(Tree r, Tree c, int[] forest, int i) {
	if (!isDuplicate(c,i)) return false;
	return hasPredecessor(r,c,forest,i-1);
    }
    */
    // public section 

    private VisualTreeComponent getDuplicate(Tree t) {
        int k = lookUpVisualItem(t,0);
        if (k<0) Note.err(this, "getDuplicate(): index out of bounds");
        return (VisualTreeComponent)getItem(k);
    }
    
    private boolean hasPredecessor(Tree t, VisualTreeComponent f) {
        return getDuplicate(t).hasPredecessor((Tree)f.getStructure());
    }

    private boolean hasAncestor(Tree t, VisualTreeComponent f) {
        return hasPredecessor((Tree)f.getStructure(), getDuplicate(t));
    }

    // a class for storing father and node during validateComponents
    class Entry {
        public Tree tree;
        public VisualTreeComponent father;
        public Entry(Tree tree, VisualTreeComponent father) {
            this.tree = tree;
            this.father = father;
        }
        public String toString() {
            return "tree="+tree+" (father="+father+")";
        }
    }

    class Deque {
	Stack s1 = new Stack();
	Stack s2 = new Stack();
	public boolean empty() {
	    return (s1.empty() && s2.empty());
	}
	public void push(Entry e) { s1.push(e); }
	public Entry pop() { return (Entry)s1.pop(); }
	public Entry get() { 
	    if (s2.empty())
		while (!s1.empty())
		    s2.push(s1.pop());
	    return (Entry)s2.pop();
	}
    }
    
    // validation should prevent recursive visualizations to exist in which a structure is it's own key
    private boolean parentHasThisStructureAsKey() {
	VisualComponent vc = (getParent() instanceof VisualComponent) ? (VisualComponent)getParent() : null;
	if (vc == null) return false;
	VisualTree vtree = (vc.getParent() instanceof VisualTree) ? (VisualTree)vc.getParent() : null;
	//if (vtree == null) return false;
	while (vtree != null) {
	    //Note.out(this, "parent = " + vc.getStructure());
	    //Note.out(this, "parent' = " + vc.getStructure().getElement());
	    //Note.out(this, "this = " + this.getStructure());	
	    // Let us check the equality od the parent structure and it's element in case the parent structure is a CDT
	    //if ((vtree.getStructure() == this.getStructure()) || (vtree.getStructure().getElement() == this.getStructure())) {
	    if ((vtree.getStructure() == this.getStructure()) && ((vtree.getRootNode() == vc))) {
		//if (vtree.getRootNode() == vc) {
		Note.warning(this, "Recurring structure encountered: " + this.getStructure());
		return true;
	    }
	    vc = (vtree.getParent() instanceof VisualComponent) ? (VisualComponent)vtree.getParent() : null;
	    if (vc == null) return false;
	    vtree = (vc.getParent() instanceof VisualTree) ? (VisualTree)vc.getParent() : null;
	}
	return false;
    }

    public void validateComponents() {
    	super.validateComponents();
        
    	VisualTreeComponent visualComponent;
    	
    	if ((getRoot() != null) && (!parentHasThisStructureAsKey())) { // non empty tree
    		Deque que = new Deque();
    		// We use class Entry to keep track of all information 
    		// about one node; start by pushing the root into the stack
    		que.push(new Entry(getRoot(), null)); 
    		int i = 0; 
    		// valid VisualTreeComponents (VTC) should appear in DFS/BFS order
    		while (!que.empty()) { 
    			Entry entry = DFS ? que.pop() : que.get(); // root of (sub)tree

    			// determine tree, forward, back, and cross edges
    			boolean isTreeEdge = !hasRepresentation(entry.tree, i);
    			boolean isForwardEdge = 
    				(!isTreeEdge && hasPredecessor(entry.tree, entry.father));
    			if (isForwardEdge && !isForwardEdgesEnabled())
    				continue;
    			boolean isBackEdge =
    				(!isTreeEdge && hasAncestor(entry.tree, entry.father));
    			if (isBackEdge && !isBackEdgesEnabled())
    				continue;
    			boolean isCrossEdge =
    				(!isTreeEdge && !isForwardEdge && !isBackEdge);
    			if (isCrossEdge && !isCrossEdgesEnabled())
    				continue;
    			
    			// put the next component and its father reference at pos i, and update i++
    			visualComponent = setComponents(entry.tree,i,entry.father);
    			if ((entry.tree != null) && hasRepresentation(entry.tree, i)) 
    				visualComponent.setMinimized(true); // minimize if seen
    			i++;
    			
    			// push subtrees into the deque
    			
    			if ((entry.tree != null) && 
    					(entry.tree.getSubTreeCount() > 0)) {
    				Tree[] subTree = entry.tree.getSubTrees();
    				if ((subTree != null) && 
    						(subTree.length > 0) &&
							(!visualComponent.isDuplicate()) &&
							(!visualComponent.isMinimized())) {
    					if (DFS) // push subtrees for DFS
    						for (int j=subTree.length-1; j>=0; j--) {
    							if ((subTree[j] != null) || // empty subtrees
    									(isEmptySubTreesEnabled())) 
    								que.push(new Entry(subTree[j], 
    										visualComponent)); 
    						}
    					else // push subtrees for BFS
    						for (int j=0; j<subTree.length; j++) {
    							if ((subTree[j] != null) || // empty subtrees
    									(isEmptySubTreesEnabled())) 
    								que.push(new Entry(subTree[j], 
    										visualComponent)); 
    						}
    				}
    			}
    		}
    		
    		// remove unused components
    		while (i < getItemCount()) { // too many VTC's (could be deleted)
    			visualComponent = getVisualTreeComponent(i);
    			visualComponent.setFather(null);
    			removeItem(i);
    		}
    		while (i < getLinkCount())
    			removeLink(i);
    	} else { // Root deleted or tree is empty
    		// delete all components if there is any
    		removeAll();
    		// create new empty root
    		setRootComponent(null);
    	}
    }

    /** Return the VisualTreeComponent of the root of the tree. */
    protected VisualTreeComponent getRootNode() {
        if (getItemCount() > 0)
            return (VisualTreeComponent) getItem(0);
        else
            return null; // no root node found
    }
    
    /** Return the root node of the tree. */
    protected Tree getRoot() {
	FDT currentFDT = getStructure();
	if (currentFDT instanceof CDT && currentFDT.getElement() instanceof Tree)
	    return (Tree)currentFDT.getElement();
	if (currentFDT instanceof Tree)
	    return (Tree)currentFDT;
	Note.err(this, "Illegal FDT structure");
	return null;
    }
    
    protected void setRoot(Tree newRoot) {
	Note.out(this, "setRoot " + newRoot);
	    FDT str = getStructure();
	    getAnimator().put(new StructureChangeOperation(str, newRoot));
            setStructure(newRoot);
    }

    public void write() {       
        getApplication().showMessage("Sorry", "write invoked: not impemented yet");
    }

    public boolean canInsert() {
	return getStructure() instanceof CDT;
    }
    
    public void insert(Object structure) {
        if (getStructure() instanceof CDT) {
            CDT cdt = (CDT)getStructure();
            FDT newRoot = cdt.insert(structure);
            if (newRoot != getStructure()) {
                if (newRoot instanceof Tree)
                    setRoot((Tree)newRoot);
                else
                    Note.err(this, 
                             "tree "+ newRoot +" not a dynamic tree anymore!");
            }
        } else 
            getApplication().showMessage("Sorry", 
                                   "Cannot insert into non CDT " + this);
    }

    public void delete(Object structure) {
        if (getStructure() instanceof CDT) {
            CDT cdt = (CDT)getStructure();

            FDT newRoot = cdt.delete(structure);
            if (newRoot != getStructure()) {
                if (newRoot instanceof Tree)
                    setRoot((Tree)newRoot);
                else
                    Note.err(this, 
                             "tree "+ newRoot +" not a dynamic tree anymore!");
            }
        } else 
            getApplication().showMessage("Sorry", 
                                   "Cannot delete from non CDT " + this);
    }

    /*
    public Point getCursorPoint() {
        Note.out(this, "getCursorPoint for "+this);
        VisualHeadReference vHeadRef = (VisualHeadReference)getLink(0);
        if (vHeadRef != null) {
            VisualType v = vHeadRef.getSource();
            if (v != null) {
                Rectangle r = v.getBounds();
                Note.out(this, "r="+r);
                return new Point(r.x, r.y);
            } else Note.warning(this, "getCursorPoint(): missing head source");
        } else Note.warning(this, 
			    "getCursorPoint(): missing head reference");
        return null;
    }
    */
    public Point getCursorPoint() {
        VisualComponent vtc = getItem(0);
        if (vtc != null) {
            return vtc.getLocation();
        } else Note.warning(this, "getCursorPoint(): missing root node");
        return null;
    }

    public Polygon getMinimizedPolygon(int x, int y, int width, int height) {
        int xpoints[] = new int[4];
        int ypoints[] = new int[4];
        
        xpoints[0] = width/2-1; ypoints[0] = 2;
        xpoints[1] = width/2+1; ypoints[1] = 2;
        xpoints[2] = width; ypoints[2] = height-2;
        xpoints[3] = 1; ypoints[3] = height-2;
        
        return new Polygon(xpoints, ypoints, 4);
    }
    
    public void paintComponent(Graphics g) {
        super.paintComponent(g);
        if (isMinimized()) { paint_minimized(g); return; }
        if (isPicked()) return; // no frames
        paintFrame(g);
    }
    
}
