// MatrixPro - algorithm simulation system
// Copyright (C) 2000-2007 See the AUTHORS file and the author tag
// below for more information on the authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package matrixpro.animation;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Stack;
import java.util.StringTokenizer;

import matrix.animation.AnimatedOperation;
import matrix.animation.Animator;
import matrix.animation.endOperation;
import matrix.animation.startOperation;

/**
 * Animator that has a number for each of the steps in the animation.
 * 
 * @version $Id: CountingAnimator.java,v 1.13 2007/09/10 08:32:32 vkaravir Exp $
 */
public class CountingAnimator extends Animator implements Serializable {
    /**
     * States and their numbers. Operation number as key (Integer) and step
     * number as value (String)
     */
    private HashMap states;

    /** All the steps numbers. */
    private ArrayList steps;

    /**
     * Default constructor for animator named as defaultAnimator. Currently
     * animator's names are not used by matrix framework.
     */
    public CountingAnimator() {
        super();
        init();
    }

    public CountingAnimator(Animator a) {
        this();
        a.rewind();
        Stack s = a.disposeRedo();
        Stack tmp = new Stack();
        while (!s.empty())
            tmp.push(s.pop());
        while (!tmp.empty())
            put((AnimatedOperation) tmp.pop());
    }

    /** Initializes the values for this animator. */
    private void init() {
        steps = new ArrayList();
        steps.add(0, new Integer(1));
        states = new HashMap();
        states.put(new Integer(0), "1");
    }

    /**
     * Clear redo-stack. Because redo - stack is cleared <code>redo</code> is
     * disposed.
     * 
     * @return stack Containing removed elements from redo - stack.
     */
    public Stack disposeRedo() {
        Stack removed = super.disposeRedo();
        int maxStep = getMaximum();
        for (int i = getState(); i < maxStep; i++)
            states.remove(states.get(new Integer(i)));
        String cur = getStep();
        Integer lastStep;
        if (cur.indexOf(".") != -1)
            lastStep = new Integer(new Integer(cur.substring(0, cur
                    .indexOf("."))).intValue() + 1);
        else
            lastStep = new Integer(cur.substring(0, cur.length()));
        steps.set(0, lastStep);
        if (getState() == 1) {
            steps = new ArrayList();
            steps.add(0, new Integer(1));
            states = new HashMap();
            states.put(new Integer(0), "1");
        }
        return removed;
    }

    public void disposeUndo() {
        super.disposeUndo();
        // clear the steps
        steps = new ArrayList();
        steps.add(0, new Integer(1));
        states = new HashMap();
        states.put(new Integer(0), "1");
        int startsBefore = super.getStartsBefore();
        // call super.disposeRedo since the numbering is handled here
        Stack removed = super.disposeRedo();
        Stack tmp = new Stack();
        // rewind here to be able to insert again the startOperations
        // inserted in Animator (to get the numbering updated)
        rewind();
        for (int i = 0; i < startsBefore; i++)
            put(new startOperation());
        while (!removed.empty())
            tmp.push(removed.pop());
        while (!tmp.empty())
            put((AnimatedOperation) tmp.pop());
        rewind();
    }

    /**
     * Put a new AnimatedOperation in Animator.
     * 
     * @param ao
     *            Operation to put in Animator
     * @see matrix.animation.AnimatedOperation
     */
    public void put(AnimatedOperation ao) {
        String stepBeforePut = getStep();
        int parenthesesBefore = parentheses;
        AnimatedOperation peek = null;
        if (!getUndoStack().empty())
            peek = (AnimatedOperation) getUndoStack().peek();
        super.put(ao);
        if (peek instanceof startOperation && ao instanceof endOperation) {
            // the operation already removed in Animator and
            // here only the numbering must be updated
            if (((Integer) steps.get(parenthesesBefore)).intValue() != 1)
                steps.set(parenthesesBefore, new Integer(((Integer) steps
                        .get(parenthesesBefore)).intValue() - 1));
            return;
        }
        int parentheses = super.getParentheses();
        StringTokenizer tokenizer = new StringTokenizer(stepBeforePut, ".");
        if (parentheses <= 0 && !(ao instanceof endOperation))
            return;
        if (steps.size() <= parentheses)
            steps.add(parentheses, new Integer(0));
        for (int i = 0; i < tokenizer.countTokens(); i++)
            steps.set(i, new Integer(tokenizer.nextToken()));
        steps.set(parentheses, new Integer(((Integer) steps.get(parentheses))
                .intValue() + 1));
        String ste = steps.get(0) + "";
        for (int i = 1; i < steps.size() && i <= parentheses; i++)
            ste += "." + steps.get(i);
        if (ao instanceof startOperation) {
        } else if (ao instanceof endOperation) {
            steps.set(parentheses + 1, new Integer(1));
        } else {

        }
        states.put(new Integer(getState()), ste);
    }

    /** Returns the number of the current step as String. */
    public String getStep() {
        return (String) states.get(new Integer(getState()));
    }

    /** Returns the number of the last step as int. */
    public int getMaxStep() {
        return ((Integer) steps.get(0)).intValue();
    }

    static final long serialVersionUID = 1075844722174565390L;

    public void endOperation() {
        super.endOperation();
    }
}
