// MatrixPro - algorithm simulation system
// Copyright (C) 2000-2007 See the AUTHORS file and the author tag
// below for more information on the authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/* $Id: MatrixProRBTree.java,v 1.9 2007/09/10 08:32:32 vkaravir Exp $ */

package matrixpro.structures;

import matrix.structures.CDT.probe.RBTree;
import matrix.structures.CDT.probe.RedBlackNode;
import matrix.structures.FDT.BinaryTree;
import matrix.structures.FDT.FDT;
import matrix.structures.FDT.Tree;
import matrix.structures.FDT.probe.Table;
import matrix.structures.other.Rotations;
import matrix.structures.simulationextensions.CallableMethods;
import matrix.structures.util.MatrixComparable;
import matrix.util.Note;

/**
 * A Red-Black Tree for MatrixPro. Automatic balancing can be turned off to make
 * it easier to simulate the rotations.
 * 
 * @author Ville Karavirta
 */
public class MatrixProRBTree extends RBTree implements CallableMethods {

    /** Constructs a new MatrixProRBTree */
    public MatrixProRBTree() {
        this(null);
    }

    /**
     * Constructs a new MatrixProRBTree
     * 
     * @param o
     *            the key to the root node
     */
    public MatrixProRBTree(Object o) {
        super(o);
    }

    public FDT getNewStructure() {
        return new MatrixProRedBlackNode();
    }

    public Tree getNewNode(Object key) {
        return new MatrixProRedBlackNode((MatrixComparable) key);
    }

    public boolean equals(Object o) {
        if (o instanceof MatrixProRBTree) {
            if (getElement() != null)
                return getElement().equals(((MatrixProRBTree) o).getElement());
            else
                return o == null;
        } else
            return super.equals(o);
    }

    public class MatrixProRedBlackNode extends RedBlackNode implements
            CallableMethods {

        /** default constructor for empty tree */
        public MatrixProRedBlackNode() {
            super();
        }

        /** Constructor for tree with one element */
        public MatrixProRedBlackNode(MatrixComparable e) {
            super(e);
        }

        public FDT insert(Object o) {
            Note.out(this, "INSERT " + o + " to " + this);

            if (o == null)
                return this;
            if (o instanceof Table) { // recursive insertion to Tables
                Table t = (Table) o;
                MatrixProRedBlackNode r = this; // to store new roots
                for (int i = 0; i < t.size(); i++)
                    r = (MatrixProRedBlackNode) r.insert(t.getObject(i));
                return r;
            } else if (o instanceof MatrixComparable) {
                if (empty()) { // inserted element is only element
                    setElement((MatrixComparable) o);
                    this.setRed(false);
                    return this;
                } else {
                    BinaryTree node = this, prev = this;
                    do {
                        prev = node;
                        if (((MatrixComparable) node.getElement())
                                .gt((MatrixComparable) o))
                            node = node.getLeft();
                        else
                            node = node.getRight();
                    } while (node != null); // new leaf will now be prev's child
                    node = new MatrixProRedBlackNode((MatrixComparable) o);
                    ((MatrixProRedBlackNode) node)
                            .setParent((MatrixProRedBlackNode) prev);
                    if (((MatrixComparable) prev.getElement())
                            .gt((MatrixComparable) o)) {
                        prev.setLeft(node);
                    } else {
                        prev.setRight(node);
                    }
                    if (balance())
                        return rotateUp((MatrixProRedBlackNode) node);
                    else
                        return this;
                }
            } else
                Note.show(this, "Cannot insert object of type " + o.getClass());
            return this;

        }

        // Interface CallableMethods
        /**
         * Returns an array of strings describing the names of methods that
         * could be invoked during the simulation. The methods should not take
         * any argument.
         */
        public String[] getCallableMethodNames() {
            return new String[] { "toggleColor", "rotateLeft", "rotateRight" };
        }

        // Methods allowed for CallableMethods interface
        /** If the color of the node is red - set it black and vice versa */
        public void toggleColor() {
            setRed(!isRed());
        }

        public void rotateRight() {
            Rotations.rotateWithLeft_(this);
        }

        public void rotateLeft() {
            Rotations.rotateWithRight_(this);
        }

    }

    /**
     * Returns an array of strings describing the names of methods that could be
     * invoked during the simulation. The methods should not take any argument.
     */
    public String[] getCallableMethodNames() {
        return new String[] { "toggleBalance" };
    }

    /**
     * Method used to toggle on or off the automatic balancing of this AVL tree.
     * To get the current value of the balancing, use the balance() method.
     * 
     * @see #balance()
     */
    public void toggleBalance() {
        balance = !balance;
    }

    private boolean balance = true;

    /**
     * Returns the current value of the automatic balancing.
     */
    public boolean balance() {
        return balance;
    }
}
