// MatrixPro - algorithm simulation system
// Copyright (C) 2000-2007 See the AUTHORS file and the author tag
// below for more information on the authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package matrixpro.ui;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.ArrayList;
import java.util.Stack;

import javax.swing.ImageIcon;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.KeyStroke;
import javax.swing.ToolTipManager;
import javax.swing.UIManager;

import prototype.ui.ExerciseFrame;
import content.interfaces.SimulationExercise;

import matrix.animation.AnimatedOperation;
import matrix.animation.Animator;
import matrix.structures.FDT.FDT;
import matrix.uitools.MatrixFrame;
import matrix.uitools.MenuFrame;
import matrix.uitools.StructurePanel;
import matrix.util.ApplicationAdapter;
import matrix.util.Note;
import matrix.util.configuration.MatrixConfiguration;
import matrix.visual.RepresentationFactory;
import matrix.visual.VisualContainer;
import matrix.visual.VisualType;

import matrixpro.animation.CountingAnimator;
import matrixpro.animation.SmallVisualAnimator;
import matrixpro.toolbar.Toolbar;
import matrixpro.toolbar.ToolbarFactory;
import matrixpro.toolbar.ToolbarPanel;
import matrixpro.util.FileHelper;
import matrixpro.util.PropertiesHandler;
import matrixpro.util.Property;
import matrixpro.util.UndoRedoHandler;
import matrixpro.util.export.MPExporter;

/**
 * The Frame used in <b>MatrixPro</b>. Contains the <code>VisualAnimator,
 * Toolbar</code>
 * and <code>StructurePanel.</code>
 * 
 * @author Ville Karavirta
 * @version $Id: MainFrame.java,v 1.45 2009/07/16 13:07:04 tred Exp $
 */
public class MainFrame extends MenuFrame implements ActionListener,
        PropertyChangeListener {
    private static final long serialVersionUID = 600449572314208632L;

    public static final String UNTITLED_FILE_NAME = "untitled";

    /** Number of recent files in the recent menu. */
    private static final int RECENT_FILES_VISIBLE = 5;

    private static final String VERSION = "v1.4";

    private static final String VERSION_DATE = "(16 July 2009)";

    /** ToolbarPanel of this frame. */
    private ToolbarPanel tb;

    /** Menu of the recently accessed files. */
    private JMenu recent;

    /** ActionListener for the recently accessed files. */
    private ActionListener recentListener;

    private static VisualType clip;

    private JMenuItem cutMenu;

    private JMenuItem copyMenu;

    private JMenuItem pasteMenu;

    private JMenuItem pasteNewMenu;

    private JMenuItem deleteMenu;

    private JMenuItem undoMenu;

    private JMenuItem redoMenu;

    private JPanel vanim;

    /**
     * Is the toolbar visible or invisible when a new MainFrame is created.
     */
    private static boolean toolbarVisibleWhenCreated = true;

    private String fileName;

    private JMenuItem newMenu;

    private JMenuItem openMenu;

    private JMenuItem saveMenu;

    private JMenuItem exportMenu;

    private JMenuItem pageSetupMenu;

    private JMenuItem printMenu;

    /**
     * Constructs a new instance of this class.
     */
    public MainFrame() {
        this(new CountingAnimator());
    }

    /**
     * Constructs a new instance of this class with the given Animator.
     * 
     * @param a
     *            The Animator to be used with this Frame
     */
    public MainFrame(Animator a) {
        this(new CountingAnimator(a));
    }

    public MainFrame(CountingAnimator ca) {
        this(new SmallVisualAnimator(ca), ca, new StructurePanel(ca));
    }

    /**
     * Constructs a new instance of this class with the given StructurePanel.
     * 
     * @param p
     *            The SturcturePanel to be used in this Frame
     */
    public MainFrame(StructurePanel p) {
        this(p.getAnimator(), p);
    }

    /**
     * Constructs a new instance of this class with the given Animator and
     * StructurePanel.
     * 
     * @param a
     *            The Animator to be used int this Frame
     * @param p
     *            The SturcturePanel to be used in this Frame
     */
    protected MainFrame(Animator a, StructurePanel p) {
        this(new CountingAnimator(a), p);
    }

    public MainFrame(CountingAnimator a, StructurePanel p) {
        this(new SmallVisualAnimator(a), a, p);
    }

    /**
     * Constructs a new instance of this class with the given VisualAnimator,
     * Animator and StructurePanel.
     * 
     * @param va
     *            The VisualAnimator to be used in this Frame
     * @param ca
     *            The CountingAnimator to be used in this Frame
     * @param p
     *            The SturcturePanel to be used in this Frame
     */
    public MainFrame(SmallVisualAnimator va, CountingAnimator ca,
            StructurePanel p) {
        super(va, ca, p);
        ToolTipManager.sharedInstance().setDismissDelay(1500);
    }

    /**
     * Constructs a new instance of this class with the given structures and
     * Animator.
     * 
     * @param structures
     *            The structures to be added to this Frame
     * @param a
     *            The Animator to be used in this Frame
     */
    public MainFrame(FDT[] structures, CountingAnimator a) {
        this(a);
        addStructures(structures);
    }

    /** Sets the position and size of the window. */
    protected void setPosition() {
        Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
        if (lastXOpen == -1)
            lastXOpen = screen.width / 5;

        else
            lastXOpen += X_SPACING;
        if (lastYOpen == -1)
            lastYOpen = screen.height / 5;
        else
            lastYOpen += Y_SPACING;
        if (lastXOpen + 2 * screen.width / 3 > screen.width
                || lastYOpen + 2 * screen.height / 3 > screen.height) {
            lastXOpen = 0;
            lastYOpen = 0;
        }
        setBounds(lastXOpen, lastYOpen, 2 * screen.width / 3,
                2 * screen.height / 3);
    }

    /** Shows the optionsdialog. Sets all the open windows unenabled. */
    public void showOptionsDialog() {
        Frame[] fr = getFrames();
        for (int i = 0; i < fr.length; i++)
            fr[i].setEnabled(false);
        new OptionsDialog(this, true);
        for (int i = 0; i < fr.length; i++)
            fr[i].setEnabled(true);
    }

    /** Adds the components. */
    protected void layoutComponents() {
        ImageIcon icon = FileHelper.getImageIcon("/pics/matrixpro-icon.png");
        if (icon != null)
            setIconImage(icon.getImage());
        tb = (ToolbarPanel) ToolbarFactory.getToolbarFromProperties(this);
        if (!toolbarVisibleWhenCreated)
            setVisualAnimatorVisible();
        addPropertyChangeListener(this);
        setClipboardEnabled();
        fileName = UNTITLED_FILE_NAME;
        JScrollPane sp = getScrollPane();
        sp.setViewportView(getStructurePanel());
        getContentPane().setLayout(new BorderLayout());
        setPosition();
        setTitle(fileName + " - MatrixPro " + getVersionNumber());
        JSplitPane split = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, tb, sp);
        split.setOneTouchExpandable(true);
        split.setDividerLocation(tb.getMinimumSize().width + 20);
        getContentPane().add(split, BorderLayout.CENTER);
    }

    public JFrame getNewFrame(StructurePanel p) {
        return new MainFrame(p);
    }

    public JFrame getNewFrame() {
        return new MainFrame();
    }

    public JFrame getNewFrame(FDT[] structures, Animator a) {
        return new MainFrame(structures, new CountingAnimator(a));
    }

    public void actionPerformed(ActionEvent e) {
        tb.update();
    }

    /**
     * Sets the Toolbar of this Frame to the given ToolbarPanel.
     * 
     * @param tbp
     *            The ToolbarPanel to set as the toolbar.
     */
    public void setToolbar(ToolbarPanel tbp) {
        if (tb.getParent() != null) {
            remove(tb);
            tb = tbp;
            getContentPane().add(tb, BorderLayout.WEST);
            vanim = null;
        } else {
            tb = tbp;
            remove(vanim);
            vanim = new JPanel();
            vanim.add(getVisualAnimator());
            getContentPane().add(vanim, BorderLayout.NORTH);
        }
        validate();
    }

    /** Exports the contents of this frame. */
    public void export() {
        MPExporter.getInstance().export(this);
    }

    /**
     * Opens a dialog to open a previously saved animation or structure.
     */
    public void open() {
        String file = FileHelper.open(this);
        if (!file.equals(""))
            addRecent(file);
    }

    /**
     * Opens a dialog to save the animation in the <code>Frame</code> this
     * menu is in.
     */
    public void save() {
        String file = FileHelper.save(this);
        if (!file.equals(""))
            addRecent(file);
    }

    /** Creates a MenuItem for the recently opened/saved files. */
    public JMenuItem createRecentMenu() {
        recent = new JMenu("Open recent");
        recent.setMnemonic(KeyEvent.VK_R);
        setRecent(recent);
        return recent;
    }

    /**
     * Adds an entry to the recently opened files menu.
     * 
     * @param fileName
     *            Name of the recently opened file
     */
    protected void addRecent(String fileName) {
        ArrayList recentFiles = new ArrayList();
        recentFiles.add(fileName);
        Property property = PropertiesHandler.getProperty("recent-file-1");
        int i = 2;
        while (property.getValue() != null) {
            if (!recentFiles.contains(property.getValue()))
                recentFiles.add(property.getValue());
            property = PropertiesHandler.getProperty("recent-file-" + i++);
        }
        for (i = 0; i < RECENT_FILES_VISIBLE && i < recentFiles.size(); i++)
            PropertiesHandler.setProperty("recent-file-" + (i + 1), recentFiles
                    .get(i).toString());
        PropertiesHandler.store();
        MainFrame.firePropertyChangeForAllFrames("recentFiles", null,
                new Object());
    }

    public void propertyChange(PropertyChangeEvent e) {
        String property = e.getPropertyName();
        // System.out.println(property+":"+e);
        if (property.equals("recentFiles"))
            setRecent(recent);
        if (property.equals("clipboard"))
            setClipboardEnabled();
    }

    /** Sets the menu of the recently accessed files to the given menu. */
    protected void setRecent(JMenu menu) {
        if (recentListener == null)
            recentListener = new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    // System.out.println("action");
                    openFile(((JMenuItem) e.getSource()).getText());
                }
            };
        Property property = PropertiesHandler.getProperty("recent-file-1");
        recent.removeAll();
        int i = 2;
        JMenuItem m;
        while (property.getValue() != null) {
            m = new JMenuItem(property.getValue());
            m.addActionListener(recentListener);
            menu.add(m);
            property = PropertiesHandler.getProperty("recent-file-" + i++);
        }
    }

    /** Opens the file given as parameter. */
    protected void openFile(String fileName) {
        try {
            FileHelper.structureFromFile(fileName, this);
        } catch (Exception e) {
            Note.err(this, "Couldn't open file " + fileName + ". Original "
                    + "exception was " + e);
        }
        addRecent(fileName);
    }

    /**
     * Opens a new Exercise.
     * 
     * @param e
     *            The exercise to open.
     */
    public void newExercise(SimulationExercise e) {
        e.setSeed(System.currentTimeMillis());
        ExerciseFrame ef = new MPExerciseFrame(e);
        ef.setVisible(true);
    }

    public static void firePropertyChangeForAllFrames(String propertyName,
            Object oldValue, Object newValue) {
        Frame[] f = getFrames();
        for (int i = 0; i < f.length; i++)
            if (f[i] instanceof MainFrame)
                ((MainFrame) f[i]).firePropertyChange(propertyName, oldValue,
                        newValue);
    }

    public void firePropertyChangeEvent(String propertyName, Object oldValue,
            Object newValue) {
        firePropertyChange(propertyName, oldValue, newValue);
    }

    public void aboutDialog() {
        showDialog(
                getVersionInfo()
                        + "\n"
                        + "Copyright (C) 2000-2009 Matrix group\n"
                        + "(See the AUTHORS file for more information on the authors.)\n"
                        + "\n"
                        + "This program is free software; you can redistribute it and/or modify\n"
                        + "it under the terms of the GNU General Public License as published by\n"
                        + "the Free Software Foundation; either version 2 of the License, or\n"
                        + "(at your option) any later version."
                        + "\n\n"
                        + "This program is distributed in the hope that it will be useful,\n"
                        + "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
                        + "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the\n"
                        + "GNU General Public License for more details."
                        + "\n\n"
                        + "You should have received a copy of the GNU General Public License\n"
                        + "along with this program; if not, write to the Free Software\n"
                        + "Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA\n"
                        + "\n\n"
                        + "For more information see http://www.cs.hut.fi/Research/MatrixPro/\n\n"
                        + "Please, send all comments and suggestions to matrix@cs.hut.fi.",
                "About");
    }

    public void helpDialog() {
        new HelpFrame().setVisible(true);
    }

    private void showDialog(String message, String title) {
        ImageIcon icon = FileHelper.getImageIcon("/pics/MatrixLogo_128.gif");
        JOptionPane.showMessageDialog(this, message, title,
                JOptionPane.PLAIN_MESSAGE, icon);
    }

    public void close() {
        if ((getFileName().equals(UNTITLED_FILE_NAME) && getStructurePanel()
                .getStructures().length == 0)
                || getFrameCount() == 1)
            super.close();
        else {
            super.clear();
            tb.resetSelectedVisualType();
            tb.update();
            setFileName(UNTITLED_FILE_NAME);
        }
    }

    public void clear() {
        setUndoPoint("clear");
        super.clear();
        tb.resetSelectedVisualType();
        tb.update();
    }

    public void setBegin() {
        setUndoPoint("set begin");
        super.setBegin();
    }

    public void setEnd() {
        setUndoPoint("set end");
        super.setEnd();
    }

    public void update() {
        getStructurePanel().invalidateComponents();
        getStructurePanel().validate();
        // setClipboardEnabled();
        validate();
        tb.update();
        deleteMenu.setEnabled(tb.visualTypeSelected());
    }

    private static String getVersionInfo() {
        return "MatrixPro version " + getVersionNumber() + " "
                + getVersionDate();
        // + " based on Matrix build 11";
    }

    public Toolbar getToolbar() {
        return tb;
    }

    public JMenuItem createCopyItem() {
        ImageIcon ii = FileHelper.getImageIcon("/pics/Copy16.gif");
        copyMenu = new JMenuItem("Copy", ii);
        copyMenu.setAccelerator(KeyStroke.getKeyStroke('C', Toolkit
                .getDefaultToolkit().getMenuShortcutKeyMask()));
        copyMenu.setMnemonic(KeyEvent.VK_C);
        copyMenu.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                copy();
                MainFrame.firePropertyChangeForAllFrames("clipboard", null, "");
            }
        });
        copyMenu.setEnabled(false);
        return copyMenu;
    }

    public JMenuItem createCutItem() {
        ImageIcon ii = FileHelper.getImageIcon("/pics/Cut16.gif");
        cutMenu = new JMenuItem("Cut", ii);
        cutMenu.setAccelerator(KeyStroke.getKeyStroke('X', Toolkit
                .getDefaultToolkit().getMenuShortcutKeyMask()));
        cutMenu.setMnemonic(KeyEvent.VK_T);
        cutMenu.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                cut();
                MainFrame.firePropertyChangeForAllFrames("clipboard", null, "");
            }
        });
        cutMenu.setEnabled(false);
        return cutMenu;
    }

    public JMenuItem createPasteItem() {
        ImageIcon ii = FileHelper.getImageIcon("/pics/Paste16.gif");
        pasteMenu = new JMenuItem("Paste", ii);
        pasteMenu.setAccelerator(KeyStroke.getKeyStroke('V', Toolkit
                .getDefaultToolkit().getMenuShortcutKeyMask()));
        pasteMenu.setMnemonic(KeyEvent.VK_P);
        pasteMenu.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                paste();
                MainFrame.firePropertyChangeForAllFrames("clipboard", null, "");
            }
        });
        pasteMenu.setEnabled(false);
        return pasteMenu;
    }

    public JMenuItem createPasteAsNewItem() {
        pasteNewMenu = new JMenuItem("Paste as duplicate");
        pasteNewMenu.setMnemonic(KeyEvent.VK_A);
        pasteNewMenu.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                pasteAsNew();
                MainFrame.firePropertyChangeForAllFrames("clipboard", null, "");
            }
        });
        pasteNewMenu.setEnabled(false);
        return pasteNewMenu;
    }

    public JMenuItem createDeleteItem() {
        ImageIcon ii = FileHelper.getImageIcon("/pics/Delete16.gif");
        deleteMenu = new JMenuItem("Delete", ii);
        deleteMenu.setMnemonic(KeyEvent.VK_D);
        deleteMenu.setAccelerator(KeyStroke.getKeyStroke("DELETE"));
        deleteMenu.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                delete();
                MainFrame.firePropertyChangeForAllFrames("clipboard", null, "");
            }
        });
        deleteMenu.setEnabled(false);
        return deleteMenu;
    }

    public void copy() {
        if (tb.visualTypeSelected())
            clip = tb.getSelectedVisualType();
        firePropertyChangeForAllFrames("clipboard", null, "");
    }

    public void cut() {
        if (tb.visualTypeSelected()) {
            setUndoPoint("cut");
            VisualType selection = tb.getSelectedVisualType();
            clip = selection;
            selection.delete();
            tb.resetSelectedVisualType();
            update();
            firePropertyChangeForAllFrames("clipboard", null, "");
        }
    }

    public void paste() {
        if (clip != null && tb.visualTypeSelected()) {
            setUndoPoint("paste");
            VisualType newVt = RepresentationFactory.getRepresentation(clip
                    .getStructure());
            VisualType.read = newVt;
            tb.getSelectedVisualType().assign();
            // getStructurePanel().validate();
            update();
        }
    }

    public void pasteAsNew() {
        if (clip != null && clip instanceof VisualContainer) {
            setUndoPoint("paste as new");
            String name = clip.getName();
            VisualType[] newVt = getStructurePanel().addStructures(
                    new FDT[] { clip.getStructure() },
                    new String[] { clip.getNameOfRepresentation() });
            newVt[0].rename(name);
            update();
        }
    }

    public void delete() {
        if (tb.visualTypeSelected()) {
            setUndoPoint("delete");
            tb.getSelectedVisualType().delete();
            tb.resetSelectedVisualType();
            update();
        }
    }

    public JMenuItem createNewItem() {
        ImageIcon ii = FileHelper.getImageIcon("/pics/New16.gif");
        newMenu = new JMenuItem("New Window", ii);
        newMenu.setMnemonic(KeyEvent.VK_N);
        newMenu.setAccelerator(KeyStroke.getKeyStroke('N', Toolkit
                .getDefaultToolkit().getMenuShortcutKeyMask()));
        newMenu.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                getNewFrame();
            }
        });
        return newMenu;
    }

    public JMenuItem createOpenItem() {
        ImageIcon ii = FileHelper.getImageIcon("/pics/Open16.gif");
        openMenu = new JMenuItem("Open", ii);
        openMenu.setMnemonic(KeyEvent.VK_O);
        openMenu.setAccelerator(KeyStroke.getKeyStroke('O', Toolkit
                .getDefaultToolkit().getMenuShortcutKeyMask()));
        openMenu.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                open();
            }
        });
        return openMenu;
    }

    public JMenuItem createSaveItem() {
        ImageIcon ii = FileHelper.getImageIcon("/pics/Save16.gif");
        saveMenu = new JMenuItem("Save As...", ii);
        saveMenu.setMnemonic(KeyEvent.VK_S);
        saveMenu.setAccelerator(KeyStroke.getKeyStroke('S', Toolkit
                .getDefaultToolkit().getMenuShortcutKeyMask()));
        saveMenu.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                save();
            }
        });
        return saveMenu;
    }

    public JMenuItem createExportItem() {
        ImageIcon ii = FileHelper.getImageIcon("/pics/Export16.gif");
        exportMenu = new JMenuItem("Export...", ii);
        exportMenu.setMnemonic(KeyEvent.VK_E);
        exportMenu.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                export();
            }
        });
        return exportMenu;
    }

    public JMenuItem createPageSetupItem() {
        ImageIcon ii = FileHelper.getImageIcon("/pics/PageSetup16.gif");
        pageSetupMenu = new JMenuItem("Page Setup...", ii);
        pageSetupMenu.setMnemonic(KeyEvent.VK_U);
        pageSetupMenu.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                pageSetup();
            }
        });
        return pageSetupMenu;
    }

    public JMenuItem createPrintItem() {
        ImageIcon ii = FileHelper.getImageIcon("/pics/Print16.gif");
        printMenu = new JMenuItem("Print...", ii);
        printMenu.setMnemonic(KeyEvent.VK_P);
        printMenu.setAccelerator(KeyStroke.getKeyStroke('P', Toolkit
                .getDefaultToolkit().getMenuShortcutKeyMask()));
        printMenu.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                print(false);
            }
        });
        return printMenu;
    }

    public JMenuItem createUndoItem() {
        ImageIcon ii = FileHelper.getImageIcon("/pics/Undo16.gif");
        undoMenu = new JMenuItem("Undo", ii);
        undoMenu.setMnemonic(KeyEvent.VK_U);
        undoMenu.setAccelerator(KeyStroke.getKeyStroke('Z', Toolkit
                .getDefaultToolkit().getMenuShortcutKeyMask()));
        undoMenu.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                undo();
            }
        });
        undoMenu.setEnabled(false);
        return undoMenu;
    }

    public JMenuItem createRedoItem() {
        ImageIcon ii = FileHelper.getImageIcon("/pics/Redo16.gif");
        redoMenu = new JMenuItem("Redo", ii);
        redoMenu.setMnemonic(KeyEvent.VK_R);
        redoMenu.setAccelerator(KeyStroke.getKeyStroke('Y', Toolkit
                .getDefaultToolkit().getMenuShortcutKeyMask()));
        redoMenu.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                redo();
            }
        });
        redoMenu.setEnabled(false);
        return redoMenu;
    }

    public void undo() {
        if (undo != null) {
            UndoRedoHandler.UndoRedoWrapper wrap = undo.undo(
                    (CountingAnimator) getAnimator(), getStructurePanel()
                            .getStructures());
            CountingAnimator anim = wrap.getAnimator();
            FDT[] s = wrap.getStructures();
            getStructurePanel().setStructures(s);
            tb.updateApplication();
            CountingAnimator a = (CountingAnimator) getAnimator();
            a.rewind();
            a.disposeRedo();
            int pos = anim.getState();
            anim.rewind();
            Stack st = anim.disposeRedo();
            Stack tmp = new Stack();
            while (!st.empty())
                tmp.push(st.pop());
            while (!tmp.empty())
                a.put((AnimatedOperation) tmp.pop());
            a.rewind();
            a.end();
            a.rewind();
            a.GoTo(pos, a.getMaximum());
            redoMenu.setEnabled(true);
            undoMenu.setEnabled(false);
            update();
        }
    }

    private UndoRedoHandler undo;

    public void setUndoPoint() {
        undo = new UndoRedoHandler((CountingAnimator) getAnimator(),
                getStructurePanel().getStructures());
        undoMenu.setEnabled(true);
        undoMenu.setText("Undo");
        redoMenu.setEnabled(false);
        redoMenu.setText("Redo");
    }

    public void setUndoPoint(String commandName) {
        undo = new UndoRedoHandler((CountingAnimator) getAnimator(),
                getStructurePanel().getStructures());
        undoMenu.setEnabled(true);
        undoMenu.setText("Undo - " + commandName);
        redoMenu.setEnabled(false);
        redoMenu.setText("Redo - " + commandName);
    }

    public void redo() {
        if (undo != null) {
            UndoRedoHandler.UndoRedoWrapper wrap = undo.redo(
                    (CountingAnimator) getAnimator(), getStructurePanel()
                            .getStructures());
            CountingAnimator anim = wrap.getAnimator();
            FDT[] s = wrap.getStructures();
            getStructurePanel().setStructures(s);
            tb.updateApplication();
            CountingAnimator a = (CountingAnimator) getAnimator();
            a.rewind();
            a.disposeRedo();
            int pos = anim.getState();
            anim.rewind();
            Stack st = anim.disposeRedo();
            Stack tmp = new Stack();
            while (!st.empty())
                tmp.push(st.pop());
            while (!tmp.empty())
                a.put((AnimatedOperation) tmp.pop());
            a.rewind();
            a.end();
            a.rewind();
            a.GoTo(pos, a.getMaximum());
            undoMenu.setEnabled(true);
            redoMenu.setEnabled(false);
            update();
        }
    }

    public boolean isUndoEnabled() {
        if (undoMenu == null)
            return false;
        return (undoMenu.isEnabled());
    }

    public boolean isRedoEnabled() {
        if (redoMenu == null)
            return false;
        return (redoMenu.isEnabled());
    }

    public boolean isPasteEnabled() {
        return clip != null;// && VisualType.read.getApplication() ==
                            // getStructurePanel());// &&
                            // tb.visualTypeSelected());
    }

    public void setClipboardEnabled() {
        boolean selected = tb.visualTypeSelected()
                && VisualType.read.getApplication() == getStructurePanel();
        if (copyMenu != null)
            copyMenu.setEnabled(selected);
        if (cutMenu != null)
            cutMenu.setEnabled(selected);
        if (pasteMenu != null)
            pasteMenu.setEnabled(isPasteEnabled());
        if (pasteNewMenu != null)
            pasteNewMenu.setEnabled(clip != null
                    && clip instanceof VisualContainer);
    }

    public void setToolbarVisible() {
        remove(vanim);
        getContentPane().add(tb, BorderLayout.WEST);
        tb.add(vanim, 0);
        tb.setVisible(true);
        validate();
    }

    public void setVisualAnimatorVisible() {
        remove(tb);
        if (vanim == null) {
            vanim = new JPanel();
            vanim.add(getVisualAnimator());
        }
        getContentPane().add(vanim, BorderLayout.NORTH);
        vanim.setVisible(true);
        validate();
        getStructurePanel().requestFocus();
    }

    public void validate() {
        if (tb != null)
            tb.updateStepLabel();
        super.validate();
    }

    public static String getVersionNumber() {
        return VERSION;
    }

    public static String getVersionDate() {
        return VERSION_DATE;
    }

    /**
     * Set the toolbar visible or invisible when a new MainFrame is created.
     * 
     * @param value
     *            True, if the toolbar should be visible, false otherwise.
     */
    public static void setToolbarVisibleWhenCreated(boolean value) {
        toolbarVisibleWhenCreated = value;
    }

    /**
     * Returns the name of the file in this frame.
     * 
     * @return the name of the file in this frame.
     */
    public String getFileName() {
        return fileName;
    }

    /**
     * Sets the name of the file in this frame
     * 
     * @param string
     *            the new name of the file in this frame
     */
    public void setFileName(String string) {
        fileName = string;
        setTitle(fileName + " - MatrixPro " + getVersionNumber());
    }

    public static void main(String[] args) {
        if (System.getProperty("mrj.version") != null) {
            /*
             * If we are running on MacOS, we set some Mac specific settings.
             * mrj.version system property is set only in MacOS, so it can be
             * used to check whether we are running on MacOS or not.
             */
            try {
                /* Set used Look and Feel to MacOS's own Look and Feel. */
                UIManager.setLookAndFeel(UIManager
                        .getSystemLookAndFeelClassName());
                /*
                 * Next system property sets the menu bar to be displayed at the
                 * top of the screen in Mac OS. In Swing, menu bar is by default
                 * displayed at the top of the application window (JFrame) but
                 * this is not a natural location in Mac OS.
                 */
                System.setProperty("apple.laf.useScreenMenuBar", "true");
            } catch (Exception e) {
                System.err
                        .println("Problems initializing Look and feel for MacOs.");
                e.printStackTrace();
            }
        }

        if (args.length > 0 && args[0].equals("-version")) {
            System.out.println(getVersionInfo());
            System.exit(0);
        }
        MatrixConfiguration.setInputStream(new FileHelper().getClass()
                .getResourceAsStream("/matrixconf.xml"));
        if (args.length > 0 && !(args[0].equals(""))) {
            MatrixConfiguration.setConfPath(args[0]);
        }

        if (args.length > 1) {
            String arg = args[1].toLowerCase();
            if (arg.equals("t") || arg.equals("true"))
                MatrixConfiguration.setValidatingParser(true);
        }

        MatrixConfiguration.getInstance();
        MatrixFrame mf = null;
        // print everything from note to a file
        // Note.setPrintStream(new java.io.PrintStream(new
        // java.io.FileOutputStream(new
        // java.io.File("errors"+System.currentTimeMillis()))));
        mf = new MainFrame(new CountingAnimator());
        ApplicationAdapter.setApplication(mf.getStructurePanel());
    }

}
