// MatrixPro - algorithm simulation system
// Copyright (C) 2000-2007 See the AUTHORS file and the author tag
// below for more information on the authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package matrixpro.util;

import java.awt.Component;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.JarURLConnection;
import java.net.URL;
import java.util.Stack;
import java.util.jar.JarFile;

import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JOptionPane;

import matrix.animation.AnimatedOperation;
import matrix.animation.Animator;
import matrix.uitools.StructurePanel;
import matrix.util.Application;
import matrix.util.ApplicationAdapter;
import matrix.util.Note;
import matrix.util.configuration.MatrixConfiguration;
import matrix.util.io.ASCIISaver;
import matrix.util.io.FileUtil;
import matrix.util.io.MatrixSerialization;

import matrixpro.animation.CountingAnimator;
import matrixpro.ui.MainFrame;

/**
 * Utility class that contains static methods to open, save and export files.
 * 
 * @version $Id: FileHelper.java,v 1.29 2007/09/10 08:35:53 vkaravir Exp $
 */
public class FileHelper {
    public static final String DEFAULT_JAR_FILE = "matrixpro-full.jar";

    /**
     * Saves the contents of the given <code>MatrixFrame</code> to the file
     * selected by the user from the opened file dialog.
     * 
     * @param f
     *            The frame that is to be saved.
     */
    public static String save(MainFrame f) {
        boolean ok = true;
        JFileChooser fileChooser = new JFileChooser(System.getProperties()
                .getProperty("user.dir"));
        fileChooser.removeChoosableFileFilter(fileChooser.getFileFilter());
        SimpleFileFilter ff = new SimpleFileFilter("Serialization");
        fileChooser.addChoosableFileFilter(ff);
        // fileChooser.addChoosableFileFilter(new SimpleFileFilter("MatrixPro
        // animation"));
        fileChooser.addChoosableFileFilter(new SimpleFileFilter("ASCII"));
        fileChooser.setFileFilter(ff);
        File saveFile;
        do {
            int state = fileChooser.showSaveDialog(f);
            if (state == JFileChooser.CANCEL_OPTION)
                return "";
            String fileName = fileChooser.getSelectedFile().getAbsolutePath();
            saveFile = new File(fileName);
            if (saveFile.exists()) {
                int overwrite = JOptionPane.showConfirmDialog(f, "File "
                        + fileName + " exists! Overwrite?", "File exists",
                        JOptionPane.YES_NO_OPTION);
                if (overwrite == JOptionPane.YES_OPTION)
                    ok = true;
                else
                    ok = false;
            } else {
                ok = true;
            }
        } while (!ok);
        Application a = ApplicationAdapter.getApplication();
        a.getAnimator().end();
        f.getStructurePanel().validate();
        f.validate();
        try {
            if (fileChooser.getFileFilter().getDescription().equals(
                    "Serialization")) {
                // create a normal Matrix Animator so the files can be loaded
                // with any Matrix application
                Animator an = a.getAnimator();
                an.rewind();
                Stack s = an.disposeRedo();
                Animator na = new Animator();
                Stack temp = new Stack();
                while (!s.empty())
                    temp.push(s.pop());
                while (!temp.empty()) {
                    AnimatedOperation ao = (AnimatedOperation) temp.pop();
                    an.put(ao);
                    na.put(ao);
                }
                na.end();
                an.rewind();
                an.end();
                // MatrixSerialization.save(na, a.getStructures(), saveFile);
                MatrixSerialization.save(na, a.getStructures(), saveFile, true,
                        f.getStructurePanel()); // added by Kimmo. Saves also
                                                // information about
                                                // visualizations of structures.
                f.setFileName(saveFile.getAbsoluteFile().getName());
            } else if (fileChooser.getFileFilter().getDescription().equals(
                    "ASCII")) {
                StructurePanel sp = f.getStructurePanel();
                Component[] cs = sp.getComponents();
                ASCIISaver as = new ASCIISaver(cs, saveFile);
                as.saveStructuresToFile();
            }
            // f.setTitle(saveFile.getName() + " - " + f.getTitle());
        } catch (SecurityException e) {
            Note.show(f, "Applet cannot write local filesystem. " +
            // "See http://www.cs.hut.fi/Research/Matrix/applet/security.html."
            // +
                    "Original exception was:" + e.getMessage());

        }
        return saveFile.getAbsolutePath();
    }

    /**
     * Opens the user-selected file to the given <code>MatrixFrame</code>.
     * Only if the selected file is a saved animation, the animation is opened
     * in a new window.
     */
    public static String open(MainFrame f) {
        JFileChooser fileChooser = new JFileChooser(System.getProperties()
                .getProperty("user.dir"));
        fileChooser.setBackground(f.getBackground());
        fileChooser.setForeground(f.getForeground());
        int state = fileChooser.showOpenDialog(f);
        if (state == JFileChooser.CANCEL_OPTION)
            return "";
        String fileName = fileChooser.getSelectedFile().getAbsolutePath();
        File tempFile = new File(fileName);
        if (!tempFile.exists()) { // checking that the file exists
            Note.show(f, "File " + fileName + " not found.");
            return "";
        }
        try {
            structureFromFile(fileName, f);
        } catch (SecurityException e) {
            Note.show(f, "Applet cannot read local filesystem. " +
            // "See http://www.cs.hut.fi/Research/Matrix/applet/security.html."
            // +
                    "Original exception was:" + e.getMessage());
        }
        return fileName;
    }

    /**
     * Opens the given file to the given frame. Only if the selected file is a
     * saved animation, the animation is opened in a new window.
     * 
     * @param fileName
     *            Name of the file.
     * @param f
     *            The frame to open the file to.
     */
    public static void structureFromFile(String fileName, MainFrame f) {
        int type = FileUtil.sniffType(fileName);
        String fName = new java.io.File(fileName).getName();
        MainFrame mf = null;
        switch (type) {
        case FileUtil.FDT_CLASS:
            matrix.structures.FDT.FDT tempFDT = FileUtil
                    .instantiateFDTClass(fileName);
            if (tempFDT != null) { // we don't want to add null structure to
                                    // application
                mf = (MainFrame) f.getNewFrame();
                // setFileName(mf, fName);
                ApplicationAdapter.getApplication().addStructure(tempFDT);
                mf.setVisible(true);
            }
            break;
        case FileUtil.GRAPH_DESCRIPTION:
            FileUtil.loadVisualStructuresFromFile(fileName, f);
            // setFileName(f, fileName);
            f.setVisible(true);
            break;
        case FileUtil.SERIALIZED_ANIMATION:
            MatrixSerialization ms = FileUtil.openAnimation(fileName);
            Animator animator = ms.getAnimator();
            animator.rewind();
            Stack s = animator.disposeRedo();
            Stack tmp = new Stack();
            while (!s.empty())
                tmp.push(s.pop());
            CountingAnimator countingAnim = new CountingAnimator();
            while (!tmp.empty())
                countingAnim.put((AnimatedOperation) tmp.pop());
            countingAnim.rewind();
            mf = (MainFrame) f.getNewFrame(ms.getStructures(), countingAnim);
            mf.setFileName(fName);
            ApplicationAdapter.getApplication().getAnimator().rewind();
            if (ms.getLayout() != null) // added by Kimmo
                ms.getLayout().restoreProperties(mf.getStructurePanel()); // added
                                                                            // by
                                                                            // Kimmo.
                                                                            // Restores
                                                                            // the
                                                                            // saved
                                                                            // visualization.
            mf.getStructurePanel().validate();
            mf.validate();
            mf.setVisible(true);
            break;
        case FileUtil.TABLE_DESCRIPTION:
            FileUtil.loadVisualStructuresFromFile(fileName, f);
            // setFileName(f, fileName);
            f.setVisible(true);
            break;
        case FileUtil.STRUCTURES_DESCRIPTION:
            FileUtil.loadVisualStructuresFromFile(fileName, f);
            // setFileName(f, fileName);
            f.setVisible(true);
            break;
        case -1: // the file is not in the supported format (sniffType()
                    // returned -1)
            Note.show(null, "The file " + fileName
                    + " is not in a file format supported by MatrixPro.");
        }
    }

    /**
     * Returns a JButton that has the picture from the file fileName. If the
     * file can't be found, then the given text buttonText is used.
     */
    public static JButton getJButton(String fileName, String buttonText) {
        ImageIcon i = getImageIcon(fileName);
        if (i != null)
            return new JButton(i);
        else
            return new JButton(buttonText);
    }

    public static ImageIcon getImageIcon(String fileName) {
        return new ImageIcon(getBytesFromFile(fileName));
    }

    private static byte[] getBytesFromFile(String fileName) {
        try {
            byte[] b = getBytesFromResource(fileName);
            if (b != null && b.length > 0) {
                return b;
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
        File f = new File("." + fileName);
        try {
            if (f.exists()) {
                InputStream in = new FileInputStream(f);
                byte[] b = readFromStream(in);
                return b;
            } else {
                String jarFile = MatrixConfiguration.getJarFile();
                if (jarFile == null)
                    jarFile = System.getProperty("java.class.path",
                            DEFAULT_JAR_FILE);
                JarFile jar = new JarFile(MatrixConfiguration.getJarFile());
                InputStream in = jar.getInputStream(jar.getEntry(fileName));
                byte[] b = readFromStream(in);
                return b;
            }
        } catch (Exception e) {
            // e.printStackTrace();
        }
        try {
            String jarURL = "jar:" + MatrixConfiguration.getCodebaseURL()
                    + DEFAULT_JAR_FILE + "!/";
            URL u = new URL(jarURL);
            JarURLConnection conn = (JarURLConnection) u.openConnection();
            JarFile jar = conn.getJarFile();
            InputStream in = jar.getInputStream(jar.getEntry(fileName));
            byte[] b = readFromStream(in);
            return b;
        } catch (Exception ex) {
            // ex.printStackTrace();
        }
        return null;
    }

    private static byte[] readFromStream(InputStream in) throws IOException {
        int cursor = 0;
        int readBytes = 0;
        byte[] b = new byte[65536];
        while ((readBytes = in.read(b, cursor, 10000)) != -1)
            cursor += readBytes;
        in.close();

        byte[] c = new byte[cursor];
        System.arraycopy(b, 0, c, 0, cursor);
        return c;
    }

    public static String getDocumentFile(String fileName) {
        return new String(getBytesFromFile(fileName));
    }

    private static byte[] getBytesFromResource(String fileName) {
        if (fileName == null)
            return null;
        InputStream in = FileHelper.class.getResourceAsStream(fileName);
        try {
            byte[] b = readFromStream(in);
            return b;
        } catch (Exception exc) {
            // exc.printStackTrace();
            return null;
        }
    }
}
