// MatrixPro - algorithm simulation system
// Copyright (C) 2000-2007 See the AUTHORS file and the author tag
// below for more information on the authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package matrixpro.util.export;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Vector;

import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.RepaintManager;

import matrix.util.Note;

import matrixpro.ui.MainFrame;

/**
 * Class used in MatrixPro to export the visualizations and animations.
 * 
 * @version $Id: MPExporter.java,v 1.7 2007/09/10 08:35:56 vkaravir Exp $
 */
public class MPExporter implements PropertyChangeListener {

    private static Collection exportClasses = new Vector();

    private HashMap filters;

    private JFileChooser fileChooser;

    private MPExporter() {
    }

    /**
     * Returns an instance of MPExporter.
     * 
     * @return an instance of MPExporter
     */
    public static MPExporter getInstance() {
        return new MPExporter();
    }

    /**
     * Exports the contents of the given MainFrame in the format the user
     * selects.
     * 
     * @param mf
     *            The MainFrame the contents of which are exported.
     */
    public void export(MainFrame mf) {
        if (exportClasses.size() < 1) {
            Note.show(null, "No export formats available!");
            return;
        }
        filters = new HashMap();
        fileChooser = new JFileChooser(System.getProperties().getProperty(
                "user.dir"));
        // remove the default FileFilter and set values
        fileChooser.removeChoosableFileFilter(fileChooser.getFileFilter());
        fileChooser.setDialogTitle("Export");
        fileChooser.setApproveButtonToolTipText("Export visualization");
        fileChooser.setApproveButtonMnemonic('e');
        Iterator iter = exportClasses.iterator();
        MPExportFormat exp = null;
        while (iter.hasNext()) {
            Class c = (Class) iter.next();
            try {
                exp = (MPExportFormat) c.newInstance();
            } catch (InstantiationException e) {
                // discard exception, can't happen
                e.printStackTrace();
                continue;
            } catch (IllegalAccessException e) {
                // discard exception, can't happen
                e.printStackTrace();
                continue;
            }
            fileChooser.addChoosableFileFilter(exp.getFileFilter());
            filters.put(exp.getFileFilter().getClass().getName(), exp);
        }
        fileChooser.setAccessory(exp.getAccessory());
        fileChooser.addPropertyChangeListener(this);
        File saveFile;
        boolean ok = false;
        do {
            int state = fileChooser.showDialog(mf, "Export");
            if (state == JFileChooser.CANCEL_OPTION)
                return;
            String fileName = fileChooser.getSelectedFile().getAbsolutePath();
            // automatically add the file extension
            String extension = ((MPExportFormat) filters.get(fileChooser
                    .getFileFilter().getClass().getName())).getFileExtension();
            if (!fileName.toLowerCase().endsWith(extension.toLowerCase()))
                fileName += "." + extension;
            saveFile = new File(fileName);
            ok = checkFile(saveFile, mf);
        } while (!ok);
        RepaintManager.currentManager(mf.getStructurePanel())
                .setDoubleBufferingEnabled(false);
        ((MPExportFormat) filters.get(fileChooser.getFileFilter().getClass()
                .getName())).export(mf, saveFile);
        RepaintManager.currentManager(mf.getStructurePanel())
                .setDoubleBufferingEnabled(true);
    }

    private static boolean checkFile(File file, MainFrame mf) {
        if (file.exists()) {
            int overwrite = JOptionPane.showConfirmDialog(mf, "File "
                    + file.getName() + " exists! Overwrite?", "File exists",
                    JOptionPane.YES_NO_OPTION);
            if (overwrite == JOptionPane.YES_OPTION)
                return true;
            else
                return false;
        } else
            return true;
    }

    /**
     * Adds the given classname to the list of available export formats.
     * 
     * @param className
     *            Name of the class of the export format. The value should be a
     *            full class name. The class given should be a subclass of
     *            matrixpro.util.export.MPExportFormat.
     * @see matrixpro.util.export.MPExportFormat
     */
    public static void addExportFormat(String className) {
        try {
            Class cl = Class.forName(className);
            MPExportFormat ef = (MPExportFormat) cl.newInstance();
            exportClasses.add(cl);
        } catch (ClassNotFoundException e) {
            Note.err(null, "Invalid export format class: " + className
                    + "\nClass not found!");
            e.printStackTrace();
        } catch (InstantiationException e) {
            Note.err(null, "Invalid export format class: " + className
                    + "\nCoudn't create new instance!");
            e.printStackTrace();
        } catch (IllegalAccessException e) {
            Note.err(null, "Invalid export format class: " + className
                    + "\nCoudn't create new instance!");
            e.printStackTrace();
        } catch (ClassCastException e) {
            Note
                    .err(
                            null,
                            "Invalid export format class: "
                                    + className
                                    + "\nNot an instance of matrixpro.util.export.MPExportFormat!");
            e.printStackTrace();
        }
    }

    /**
     * @see java.beans.PropertyChangeListener#propertyChange(java.beans.PropertyChangeEvent)
     */
    public void propertyChange(PropertyChangeEvent evt) {
        if (evt.getPropertyName().equals(
                JFileChooser.FILE_FILTER_CHANGED_PROPERTY)) {
            fileChooser.setAccessory(((MPExportFormat) filters.get(fileChooser
                    .getFileFilter().getClass().getName())).getAccessory());
        } else if (evt.getPropertyName().equals(
                JFileChooser.ACCESSORY_CHANGED_PROPERTY))
            fileChooser.validate();
    }
}
